#include <winsock2.h>	// this needs to be before #include <windows.h> (which happens in "zoomsearch.h")
#include "zoomsearch.h"

// The rest of this file is excluded from compilation if FastCGI is disabled.

#ifdef FASTCGI_ENABLED

#include <stdio.h>
#include <stdlib.h>
//#include <windows.h>
#include <msxml2.h>
#include <atlbase.h>

#include "fastcgi.h"

#pragma comment(lib, "ws2_32.lib")	// needed to link to ws2 lib for winsock commands

enum FASTCGI_PROTOCOL {
	FcgiConfigProtocolNamedPipe,
	FcgiConfigProtocolTcp,
	FcgiConfigProtocolInvalid
}; 

FASTCGI_PROTOCOL protocol;
HANDLE hStdin;
SOCKET acceptSocket = INVALID_SOCKET;

class TRANSPORT
{
public: 

	TRANSPORT()
	{}
	~TRANSPORT()
	{} 

	HRESULT Send( PBYTE pBuffer, DWORD dwSize)
	{
		return SendReceive( TRUE, pBuffer, dwSize );
	}
	HRESULT Receive( PBYTE pBuffer, DWORD dwSize)
	{
		return SendReceive( FALSE, pBuffer, dwSize );
	} 

protected: 

	virtual HRESULT SendReceive( bool fSend, PBYTE pBuffer, DWORD dwSize ) = 0;
}; 

class TRANSPORT_NAMEDPIPE : public TRANSPORT
{
public: 

	TRANSPORT_NAMEDPIPE()
	{}
	~TRANSPORT_NAMEDPIPE()
	{}
protected:            

	HRESULT SendReceive( bool fSend, PBYTE pBuffer, DWORD dwSize )
	{
		BOOL fRet  = FALSE;
		DWORD dwLastError = 0;
		DWORD nNumberOfBytesRead = 0; 

		if( fSend == TRUE )
		{        
			fRet = WriteFile( hStdin, pBuffer, dwSize, &nNumberOfBytesRead, NULL );
		}
		else
		{
			fRet = ReadFile( hStdin, pBuffer, dwSize, &nNumberOfBytesRead, NULL );
		} 

		if(!fRet)
		{
			dwLastError = GetLastError();
		} 

		return HRESULT_FROM_WIN32( dwLastError );
	}
}; 

class TRANSPORT_TCP : public TRANSPORT
{
public: 

	TRANSPORT_TCP()
	{}
	~TRANSPORT_TCP()
	{} 

protected: 

	HRESULT SendReceive( bool fSend, PBYTE pBuffer, DWORD dwSize )
	{
		int iRet = 0;
		WSABUF winsockBuffer;
		DWORD   dwBytesTransferred  = 0;
		DWORD   dwFlags             = 0;
		DWORD   dwLastError            = 0; 

		winsockBuffer.len = dwSize;
		winsockBuffer.buf = (char *)pBuffer; 

		if( fSend == TRUE )
		{
			iRet = WSASend( acceptSocket, &winsockBuffer, 1, &dwBytesTransferred, 0, NULL, NULL ); 
		}
		else
		{
			iRet = WSARecv( acceptSocket, &winsockBuffer, 1, &dwBytesTransferred, &dwFlags, NULL, NULL );
		} 

		if( SOCKET_ERROR == iRet )
		{
			dwLastError = WSAGetLastError();
		} 

		return HRESULT_FROM_WIN32( dwLastError );
	}
};

bool InitializeWinsock( )
{
	WORD wVersionRequested;
	WSADATA wsaData;
	int errt;

	wVersionRequested = MAKEWORD( 2, 2 );		

	if ( ( errt = WSAStartup( wVersionRequested, &wsaData ) ) != 0 ) 
	{
		return FALSE;
	}

	return TRUE;
}

bool IsApplicationInvokedUsingFastCGI( HANDLE s )
{
	// Determine if this application was invoked using CGI or FastCGI
	struct sockaddr nameOfConnectedSocket;
	int nameLenOfConnectedSocket;
	HRESULT hr = S_OK;
	int errs; 

	errs = getpeername( (SOCKET)s, &nameOfConnectedSocket, &nameLenOfConnectedSocket );        
	if ( errs == SOCKET_ERROR )
	{
		hr = HRESULT_FROM_WIN32( WSAGetLastError() );
		if(  hr == HRESULT_FROM_WIN32( WSAENOTCONN ) )
		{
			//LogMessage( TRACE_MESSAGE, "Application was invoked using FastCGI over TCP, got a WSAENOTCONN on a call to getpeername()" );                    
			return TRUE;
		}
		else            
			if( hr == HRESULT_FROM_WIN32( WSAENOTSOCK ) ) 
			{                    
				//LogMessage( TRACE_MESSAGE, "Application was invoked using FastCGI over Named Pipe, got a WSAENOTSOCK on a call to getpeername()" );                    
				return TRUE;
			}            
			else            
			{                    				    
				return FALSE;            
			}
	}        
	else
	{
		// Quit, this app is not compatible with FastCGI
		//LogMessage( TRACE_MESSAGE, "Application was invoked using CGI, getpeername() successful" );
		return FALSE;
	}        
}

bool IsNamedPipeOrSocket( HANDLE hStdin, FASTCGI_PROTOCOL * useNamedPipeOrSocket )
{
	HRESULT hr = S_OK; 

	try
	{
		acceptSocket = accept( (SOCKET)hStdin, NULL, NULL ); 

		if( acceptSocket == INVALID_SOCKET )
		{
			(*useNamedPipeOrSocket) = FcgiConfigProtocolNamedPipe;                
		}
		else
		{
			(*useNamedPipeOrSocket) = FcgiConfigProtocolTcp;
		}            
	}
	catch(...)
	{
		//LogMessage( TRACE_EXCEPTION, "Exception is IsNamedPipeOrSocket() function" );
		return FALSE;
	}
	return TRUE;
}

extern bool InitializeCGI();
extern bool CloseCGI();

TRANSPORT* pTransport;	
unsigned char CurrentRequestIdB1;
unsigned char CurrentRequestIdB0;


int BeginFastCGI(zoomSearch* zs, const zoomSearchSetup *zsSetup)
{
	// ----------------------------------------------------------------------------
	// This is the FastCGI main loop
	// ----------------------------------------------------------------------------

	
	PrintUserDebug("BeginFastCGI was called");

	// Initialize Winsock. Only then socket calls will succeed, otherwise they will return a WSAENOTINITIALIZED error.
	if( !InitializeWinsock() )
	{
		//LogMessage( TRACE_ERROR, "InitializeWinsock() failed with error message" );            
	}

	// Obtain the Standard Input Handle. The ServerSocket was duplicated onto this handle by FastCGI for TCP.
	// The initial connection is made by the FastCGI handler by hijacking the STD_INPUT_HANDLE handle 
	// for the FastCGI application process        
	hStdin = GetStdHandle( STD_INPUT_HANDLE );

	if( !IsApplicationInvokedUsingFastCGI( hStdin ) ) 
	{
		//LogMessage( TRACE_ERROR, "IsApplicationInvokedUsingFastCGI() failed with error message" );
		return 1;
	}

	// Determine whether the communication should happen over TCP or Named Pipes
	// Make a socket call on the STD_INPUT_HANDLE handle and if it fails, use Named Pipes, else use sockets.            
	FASTCGI_PROTOCOL namePipeOrSocket;
	if ( !IsNamedPipeOrSocket( hStdin, &namePipeOrSocket ) )
	{
		//LogMessage( TRACE_ERROR, "IsNamedPipeOrSocket() failed with error message" );
		return 1;
	}

	FCGI_Header headerOfMessage;
	FCGI_EndRequestRecord message_EndRequest;        
	//FCGI_StdOutErrorRecord message_StdOut;
	//FCGI_StdOutErrorRecord message_StdErr;		

	DWORD dwBytesWritten = 0;
	DWORD dwError = 0;
	DWORD dwResponseSize  = 0;
	PBYTE pBuffer = NULL;
	HRESULT hr = S_OK;
	pTransport = NULL;	

	int MyTestInstanceCounter = 0;

	char * content = new char[10];

	// Initialize the transport connection based on whether the Server is using Named Pipes or TCP.
	if( namePipeOrSocket == FcgiConfigProtocolNamedPipe )
	{
		pTransport = new TRANSPORT_NAMEDPIPE();
	}
	else
	{
		pTransport = new TRANSPORT_TCP();
	}

	CurrentRequestIdB1 = 0;
	CurrentRequestIdB0 = 0;

	// We do this stuff only once per instance of the FastCGI
	int statusOK = true;

	if (zs->setup1((zoomSearchSetup*)zsSetup) != SUCCESS)
	{
		statusOK = false;		
	}

	if (statusOK)
	{
		statusOK = zs->OpenIndexFiles(zsSetup);
		if (statusOK)
			statusOK = zs->LoadDictionary(zsSetup);
	}

	if (statusOK == false)
	{
		// error report here
		//FastCGIPrintf("FastCGI Error: Initial loading of index data failed.");
		goto FinishResponse;
	}

#ifdef DEBUG_FASTCGI
	int mainLoopCount = 0;
#endif

	//FCGI_NameValuePair11 tmpValuePair;
	FCGI_NameValuePair tmpValuePair;
	FCGI_NameValuePair44 tmpParseBytes;
	DWORD pbi = 0;

	char* namePtr = NULL;
	char* valuePtr = NULL;
	char* bigNameBuffer = NULL;
	char* bigValueBuffer = NULL;
	
	while ( TRUE )
	{        
		ZeroMemory( &headerOfMessage, sizeof( headerOfMessage ) );
#ifdef DEBUG_FASTCGI
		if (mainLoopCount == 1)
		{
			headerOfMessage.type = FCGI_PARAMS;
			dwResponseSize = 100;
		}
		else if (mainLoopCount < 10)
		{
			headerOfMessage.type = FCGI_STDIN;
			dwResponseSize = 0;
			headerOfMessage.requestIdB0 = 0;
			headerOfMessage.requestIdB1 = mainLoopCount;
		}
		else
		{
			headerOfMessage.type = FCGI_END_REQUEST;
		}
		mainLoopCount++;
#else
		hr = pTransport->Receive( (PBYTE)&headerOfMessage, sizeof(headerOfMessage) );            
		if( FAILED(hr) )
		{
			goto FinishResponse;
			//LogMessage( TRACE_ERROR, "Receive failed, could not read message header" );
		}

		// Calculate size of response based on information in the message header
		dwResponseSize = ( headerOfMessage.contentLengthB1 << 8 )
			+ headerOfMessage.contentLengthB0
			+ headerOfMessage.paddingLength;
#endif


		if( dwResponseSize > 0 )
		{                
			if (pBuffer)
			{
				delete(pBuffer);
				pBuffer = NULL;
			}

			// Allocate memory for response                
			pBuffer = new BYTE[dwResponseSize + 1];
			if ( NULL == pBuffer )
			{
				dwError = ERROR_OUTOFMEMORY;
				//LogMessage( TRACE_ERROR, "Could not allocate memory" );
				goto FinishResponse;
			}
			ZeroMemory( pBuffer, dwResponseSize );                // Receive the rest of the message
#ifdef DEBUG_FASTCGI
			pBuffer[0] = strlen("QUERY_STRING");
			pBuffer[1] = strlen("zoom_query=test");
			strcpy((char*)(pBuffer+2), "QUERY_STRING");
			strcpy((char*)(pBuffer+2+pBuffer[0]), "zoom_query=test");
#else
			hr = pTransport->Receive( pBuffer, dwResponseSize );
			if( FAILED(hr) )
			{
				goto FinishResponse;
				//LogMessage( TRACE_ERROR, "Receive failed, could not read message body" );
			}
#endif
		}        
		// Switch based on header type and define your own response to each of the messages from the FastCGI Server
		switch( headerOfMessage.type )
		{
		case FCGI_BEGIN_REQUEST : 
			//LogMessage( TRACE_MESSAGE, "Got an FCGI_BEGIN_REQUEST" );                                                                                                                            
			break;

		case FCGI_PARAMS :
			// LogMessage( TRACE_MESSAGE, "Got an FCGI_PARAMS" );
			pbi = 0;
			while (pbi < dwResponseSize)
			{								
				// check high order bit of first byte to determine format
				if (pBuffer[pbi] >> 7 == 1)
				{			
					// first four bytes is nameLength
					tmpParseBytes.nameLengthB3 = pBuffer[pbi++];
					tmpParseBytes.nameLengthB2 = pBuffer[pbi++];
					tmpParseBytes.nameLengthB1 = pBuffer[pbi++];
					tmpParseBytes.nameLengthB0 = pBuffer[pbi++];

					tmpValuePair.nameLength = ((tmpParseBytes.nameLengthB3 & 0x7f) << 24) + (tmpParseBytes.nameLengthB2 << 16) + (tmpParseBytes.nameLengthB1 << 8) + tmpParseBytes.nameLengthB0;
				}
				else
				{
					// first byte is nameLength
					tmpValuePair.nameLength = pBuffer[pbi];
					pbi++;
				}
				
				// check high order bit of first byte for format
				if (pBuffer[pbi] >> 7 == 1)
				{
					tmpParseBytes.valueLengthB3 = pBuffer[pbi++];
					tmpParseBytes.valueLengthB2 = pBuffer[pbi++];
					tmpParseBytes.valueLengthB1 = pBuffer[pbi++];
					tmpParseBytes.valueLengthB0 = pBuffer[pbi++];

					tmpValuePair.valueLength = ((tmpParseBytes.valueLengthB3 & 0x7f) << 24) + (tmpParseBytes.valueLengthB2 << 16) + (tmpParseBytes.valueLengthB1 << 8) + tmpParseBytes.valueLengthB0;
				}
				else
				{
					tmpValuePair.valueLength = pBuffer[pbi];
					pbi++;
				}

				namePtr = NULL;
				valuePtr = NULL;
				bigNameBuffer = NULL;
				bigValueBuffer = NULL;

				if (tmpValuePair.nameLength > FCGI_NAMEVALUEPAIR11_MAXDATALEN)
				{
					bigNameBuffer = (char*) malloc(sizeof(char) * tmpValuePair.nameLength+1);
					namePtr = bigNameBuffer;
				}
				else
				{
					memset(tmpValuePair.nameData, 0, sizeof(tmpValuePair.nameData));				
					memcpy(tmpValuePair.nameData, pBuffer+pbi, tmpValuePair.nameLength);					
					namePtr = (char*) &(tmpValuePair.nameData);
				}
				pbi += tmpValuePair.nameLength;

				if (tmpValuePair.valueLength > FCGI_NAMEVALUEPAIR11_MAXDATALEN)
				{
					bigValueBuffer = (char*) malloc(sizeof(char) * tmpValuePair.valueLength+1);
					valuePtr = bigValueBuffer;
				}
				else
				{
					memset(tmpValuePair.valueData, 0, sizeof(tmpValuePair.valueData));
					memcpy(tmpValuePair.valueData, pBuffer+pbi, tmpValuePair.valueLength);					
					valuePtr = (char*) &(tmpValuePair.valueData);
				}
				pbi += tmpValuePair.valueLength;
																
				if (tmpValuePair.nameLength > 0)
				{
#ifndef NON_WINDOWS
					//_putenv_s("REQUEST_METHOD", "GET");	
					_putenv_s(namePtr, valuePtr);
#else
					//putenv("REQUEST_METHOD=GET");	
					char tmpstr[1024];
					sprintf(tmpstr, "%s=%s", namePtr, valuePtr);
					putenv(tmpstr);
#endif
				}

				if (bigNameBuffer)
					free(bigNameBuffer);
				if (bigValueBuffer)
					free(bigValueBuffer);
			}
			break;

		case FCGI_STDIN :
			//Call Search Class					
			
			CurrentRequestIdB1 = headerOfMessage.requestIdB1;
			CurrentRequestIdB0 = headerOfMessage.requestIdB0;
			
			statusOK = InitializeCGI();

			zs->InitQuery();			
			if (zs->LoadQuery(zsSetup))
			{
				// only perform search if LoadQuery returned 1
				zs->search(zsSetup);
			}

			// Send FCGI_END_REQUEST
			ZeroMemory( &message_EndRequest, sizeof( message_EndRequest ) );   
			message_EndRequest.header.version = FCGI_VERSION_1;
			message_EndRequest.header.type = FCGI_END_REQUEST;
			message_EndRequest.header.requestIdB0 = headerOfMessage.requestIdB0;
			message_EndRequest.header.requestIdB1 = headerOfMessage.requestIdB1;
			message_EndRequest.header.contentLengthB0 = FCGI_END_REQUEST_BODY_LEN;
			pTransport->Send( (PBYTE)&message_EndRequest, sizeof( message_EndRequest ) );
			
			zs->cleanupQuery();
			zs->InitBuffers();
			CloseCGI();
			zs->ResetStartTime();
			
			break;
			
		case FCGI_NULL_REQUEST_ID :
			//LogMessage( TRACE_MESSAGE, "FCGI_NULL_REQUEST_ID obtained" );
			break;

#ifdef DEBUG_FASTCGI
		case FCGI_END_REQUEST:
			goto FinishResponse;
			break;
#endif

		default :
			//LogMessage( TRACE_MESSAGE, "Unhandled type obtained, quitting" );
			break;

		}
	} // end while

FinishResponse:  

	zs->cleanup();
	if (pTransport)
		delete(pTransport);
	if (content)
		delete(content);
	if (pBuffer)
		delete(pBuffer);

	return 1;
}


int FastCGIPrintf(char* strData)
{
	if (pTransport == NULL || (CurrentRequestIdB0 == 0 && CurrentRequestIdB1 == 0))
		return 0;

	int strDataLen = strlen(strData);

	char* strTruncatedPtr = NULL;
	char tmpSavedChar = 0;

	FCGI_StdOutErrorRecord message_StdOut;
	ZeroMemory( &message_StdOut, sizeof( message_StdOut ) );                                            
	message_StdOut.header.version = FCGI_VERSION_1;
	message_StdOut.header.type = FCGI_STDOUT;											
	message_StdOut.header.requestIdB0 = CurrentRequestIdB0;
	message_StdOut.header.requestIdB1 = CurrentRequestIdB1;											
	if (strDataLen > sizeof(message_StdOut.contentData))
	{
		// This string is too large to fit in a single FCGI output
		// since the header only allows 2 bytes for the contentLength 
		// (thus, 0xFFFF is FCGI_MAX_LENGTH)
		strTruncatedPtr = strData + sizeof(message_StdOut.contentData) - 1;	// -1 for null terminator
		if (strTruncatedPtr != NULL)
		{
			// remember the char we replace with the NULL terminator
			tmpSavedChar = *strTruncatedPtr;
			*strTruncatedPtr = '\0';
			// adjust string length
			strDataLen = strlen(strData);
		}
		else
		{
			// This should never happen, but in case it does, let's ignore the print, rather than corrupt mem.
			return 0;
		}
	}

	strcpy(message_StdOut.contentData, strData);	

	message_StdOut.header.contentLengthB0 = strDataLen & 0xff;
	message_StdOut.header.contentLengthB1 = (strDataLen >> 8) & 0xff;
	
	pTransport->Send( (PBYTE)&message_StdOut, sizeof( message_StdOut.header ) + strDataLen );

	if (strTruncatedPtr != NULL)
	{
		// restore the overwritten character
		*strTruncatedPtr = tmpSavedChar;
		// Recurse and print the remaining string
		strDataLen += FastCGIPrintf(strTruncatedPtr);
	}

	return strDataLen;
}

#endif
