#ifndef FASTCGI_H
#define FASTCGI_H
/*		
 * Listening socket file number
 */

#define FCGI_LISTENSOCK_FILENO 0

typedef struct {
	unsigned char version;
	unsigned char type;
	unsigned char requestIdB1;
	unsigned char requestIdB0;
	unsigned char contentLengthB1;
	unsigned char contentLengthB0;
	unsigned char paddingLength;
	unsigned char reserved;
} FCGI_Header;

/*
 * Number of bytes in the FCGI_Header. Future versions of the protocol
 * will not reduce this number
 */
#define FCGI_HEADER_LEN		8
#define FCGI_END_REQUEST_BODY_LEN 8

/*
 * Value for version component of FCGI_Header
 */
#define FCGI_VERSION_1		1

/*
 * Values for type component of FCGI_Header
 */
#define FCGI_BEGIN_REQUEST		1
#define FCGI_ABORT_REQUEST		2
#define FCGI_END_REQUEST		3
#define FCGI_PARAMS				4
#define FCGI_STDIN				5
#define FCGI_STDOUT				6
#define FCGI_STDERR				7
#define FCGI_DATA				8
#define FCGI_GET_VALUES			9
#define FCGI_GET_VALUES_RESULT	10
#define FCGI_UNKNOWN_TYPE		11
#define FCGI_MAXTYPE (FCGI_UNKNOWN_TYPE)

#define FCGI_MAX_LENGTH 0xffff

/*
 * Value for requestId component of FCGI_Header
 */
#define FCGI_NULL_REQUEST_ID    0

/*
 * Mask for flags component of FCGI_BeginRequestBody
 */
#define FCGI_KEEP_CONN	1

/*
 * Values for role component of FCGI_BeginRequestBody
 */
#define FCGI_RESPONDER	1
#define FCGI_AUTHORIZER 2
#define FCGI_FILTER		3

typedef struct {
	unsigned char appStatusB3;
	unsigned char appStatusB2;
	unsigned char appStatusB1;
	unsigned char appStatusB0;
	unsigned char protocolStatus;
	unsigned char reserved[3];
} FCGI_EndRequestBody;

typedef struct {
	FCGI_Header header;
	FCGI_EndRequestBody body;
} FCGI_EndRequestRecord;


typedef struct {
	FCGI_Header header;
	char contentData[FCGI_MAX_LENGTH];
} FCGI_StdOutErrorRecord;

/*
 * Values for protocolStatus component of FCGI_EndRequestBody
 */

#define FCGI_REQUEST_COMPLETE	0
#define FCGI_CANT_MPX_CONN		1
#define FCGI_OVERLOADED			2
#define FCGI_UNKNOWN_ROLE		3


/* Name-value pairs added from 3.4 of FastCGI specifications by RL @ Wrensoft
 */
#define FCGI_NAMEVALUEPAIR11_MAXDATALEN 255
typedef struct {
	unsigned char nameLengthB0;  /* nameLengthB0  >> 7 == 0 */
	unsigned char valueLengthB0; /* valueLengthB0 >> 7 == 0 */
	unsigned char nameData[FCGI_NAMEVALUEPAIR11_MAXDATALEN];
	unsigned char valueData[FCGI_NAMEVALUEPAIR11_MAXDATALEN];
} FCGI_NameValuePair11;

typedef struct {
	unsigned long nameLength;  
	unsigned long valueLength; 
	char nameData[FCGI_NAMEVALUEPAIR11_MAXDATALEN];
	char valueData[FCGI_NAMEVALUEPAIR11_MAXDATALEN];
} FCGI_NameValuePair;

typedef struct {
	unsigned char nameLengthB3;  /* nameLengthB3  >> 7 == 1 */
	unsigned char nameLengthB2;
	unsigned char nameLengthB1;
	unsigned char nameLengthB0;
	unsigned char valueLengthB3; /* valueLengthB3 >> 7 == 1 */
	unsigned char valueLengthB2;
	unsigned char valueLengthB1;
	unsigned char valueLengthB0;
	/*
	unsigned char nameData[nameLength
		((B3 & 0x7f) << 24) + (B2 << 16) + (B1 << 8) + B0];
	unsigned char valueData[valueLength
		((B3 & 0x7f) << 24) + (B2 << 16) + (B1 << 8) + B0];
	*/
} FCGI_NameValuePair44;


#endif