#include "zoomsearch.h"
#ifdef CGI_MODE

#include "yacgi.h" //include the YACGI library header for CGI helper functions

#ifdef FASTCGI_ENABLED
#include "fastcgi.h"
extern enum FASTCGI_PROTOCOL;
bool InitializeWinsock();
bool IsApplicationInvokedUsingFastCGI(HANDLE s);
bool IsNamedPipeOrSocket( HANDLE hStdin, FASTCGI_PROTOCOL * useNamedPipeOrSocket );
int BeginFastCGI(zoomSearch* zs, const zoomSearchSetup *zsSetup);
int FastCGIPrintf(char* str);
#endif

bool InitializeCGI();
bool CloseCGI();
int cgiValueIntegerArray(CGI *cgi, char *name, int* resultArray, int* num_results, int maxValue=0);

// ----------------------------------------------------------------------------
// Global variables
// ----------------------------------------------------------------------------
zoomSearch zs;
zoomSearchSetup zsSetup;
CGI* cgi;
TIMER_VALUE starttime;
TIMER_VALUE endtime;
char TempPrintfBuffer[MAX_PRINTF_BUFFERLEN+1];	// for zoomprintf()

#ifdef DEBUG_TIMER
USERDEBUGSTATS UserDebugStats;
#endif

// ----------------------------------------------------------------------------
// main()
// ----------------------------------------------------------------------------

int main(int argc, char* argv[],  char *envp[]) //added  char *envp[] to print environment, MAY BREAK UNDER LINUX
{
#ifdef DEBUG_FASTCGI
	_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );	
#endif

	bool statusOK = true;
			
	char* TmpScriptName = (strrchr(argv[0], '\\')); //grab after last / or just take argv[0]
	if(TmpScriptName == NULL) //try looking for other /
		TmpScriptName = (strrchr(argv[0], '/'));

	if(TmpScriptName == NULL)
	{
		//probably linux/bsd where only filename is passed
		strcpy(zsSetup.ScriptName, argv[0]);
		
		#ifdef NON_WINDOWS
			char tmpBuf[512];
			getcwd(tmpBuf, 512);
			strcpy(zsSetup.ExecutePath, tmpBuf); //set to empty string
		#else
			strcpy(zsSetup.ExecutePath, ".");
		#endif		
	}
	else
	{
		strcpy(zsSetup.ScriptName, TmpScriptName+1);
		
		//save path
		strcpy(zsSetup.ExecutePath, argv[0]);

		char* namePtr = strrchr(zsSetup.ExecutePath, '\\');
		if(namePtr == NULL) //look for /
			namePtr = strrchr(zsSetup.ExecutePath, '/');

		if(namePtr != NULL)
			*namePtr = '\0'; //remove filename
	}

	//Check for command line options
	for(int a=1; a< argc; a++)
	{		
		// Specify alternate path of index files (when index files not in same folder as CGI)
		if (argv[a][0] == '/' && toupper(argv[a][1]) == 'I' && argc > a) 
		{
			if (strlen(argv[a+1]) < MAXPATH)
			{
				strcpy(zsSetup.ExecutePath, argv[a+1]);
			}
		}
	}


#ifdef FASTCGI_ENABLED

	BeginFastCGI(&zs, &zsSetup);

#else

	// ----------------------------------------------------------------------------
	// This is the ordinary CGI main code
	// ----------------------------------------------------------------------------

	if ( zs.setup1(&zsSetup) != SUCCESS)
	{
		statusOK = false;		
	}

	if (statusOK)
	{
		statusOK = InitializeCGI();
	}

	//Call Search Class
	if (statusOK)
	{		
		zs.InitQuery();		
		if (zsSetup.IsZoomAC)
		{
			if (zs.LoadQuery(&zsSetup))
			{
				zs.searchAC(&zsSetup);
			}
		}
		else
		{
			if (zs.LoadQuery(&zsSetup))
			{
				if (zs.OpenIndexFiles(&zsSetup) && zs.LoadDictionary(&zsSetup))
					zs.search(&zsSetup);		
			}
		}
	}	

#endif
	
	//cleanup
	CloseCGI();

	//zs.cleanup();

	return 0;
}

bool CloseCGI()
{
	if (zs.UseCats)
	{
		if (zsSetup.Cat)
		{
			free(zsSetup.Cat);
			zsSetup.Cat = NULL;
		}
	}

	if (zs.UseMetaFields)
	{
		if (zsSetup.meta_query)
		{
			for (int fieldnum = 0; fieldnum < zs.NumMetaFields; fieldnum++)
			{
				free(zsSetup.meta_query[fieldnum]);
				zsSetup.meta_query[fieldnum] = NULL;
			}
			free(zsSetup.meta_query);
			zsSetup.meta_query = NULL;
		}
	}

	if (cgi)
	{
		cgiClose(cgi);	
		cgi = NULL;
	}
	return true;
}

bool InitializeCGI()
{
	bool statusOK = true;	
	
	
	// *Environment variable dump used to happen here, but its since been moved further down
	// so that we can write the content-type header AFTER we read the settings file.
	

#ifdef VC_DEBUGGER_GETQUERY
#ifndef NON_WINDOWS
	_putenv_s("REQUEST_METHOD", "GET");	
	_putenv_s("QUERY_STRING", VC_DEBUGGER_GETQUERY);
#else
	putenv("REQUEST_METHOD=GET");	
	char tmpstr[1024];
	sprintf(tmpstr, "QUERY_STRING=%s", VC_DEBUGGER_GETQUERY);
	putenv(tmpstr);
#endif	
#endif

    /*-----------------------------------------------------------
	*              Opening Relation
	*-----------------------------------------------------------*/	
    cgi = cgiOpen(); 
    if(!cgi)
    {
		printf("Content-type: text/html\n\n");
		printf("<H3>%s</H3>%c",cgiStateMsg(),10);		
#ifndef DEBUG_MODE
		exit(1);
#endif
    }


	if (cgiValueInteger(cgi, "zoom_xml", &zsSetup.UseXML, 1) != CGI_OK)
		zsSetup.UseXML = INPUT_RETRIEVE_ERROR;


	//get the IP of the client running the search
	char* tmpIP = cgiEnvGet("REMOTE_ADDR");
	if(tmpIP == NULL)
		strcpy(zsSetup.clientIP, "unknown");
	else
		strcpy(zsSetup.clientIP, tmpIP);


	// Vales from GET
	//Query search string
	zsSetup.IsZoomQuery = false;
	char* temp = cgiValueFirst(cgi, "zoom_query"); 
	if(temp != NULL)
	{
		memset(zsSetup.Query, 0, sizeof(char)*(MAX_QUERY_LENGTH+1));
		strncpy((char*)zsSetup.Query, temp, MAX_QUERY_LENGTH);
		zsSetup.IsZoomQuery = true;
	}
	else
	{
		// to cater for Flash's search form having a quirk of not handling underscore characters:
		temp = cgiValueFirst(cgi, "zoom%5Fquery");
		if (temp != NULL)
		{
			strcpy((char*)zsSetup.Query, temp);
			zsSetup.IsZoomQuery = true;
		}
		else
			strcpy((char*)zsSetup.Query, "");	
	}



	// number of results per page, defaults to 10 if not specified
	if(cgiValueInteger(cgi, "zoom_per_page", &zsSetup.Per_Page, 1) != CGI_OK)
		zsSetup.Per_Page=10;


	// current result page number, defaults to the first page if not specified
	zsSetup.NewSearch = 0;
	if(cgiValueInteger(cgi, "zoom_page", (long*)&zsSetup.Page, 1) != CGI_OK)
	{
		zsSetup.Page = 1;
		zsSetup.NewSearch = 1;
	}
		
	if(cgiValueInteger(cgi, "zoom_ac", (long*)&zsSetup.IsZoomAC, 0) != CGI_OK)
		zsSetup.IsZoomAC = 0;				
	
	if(cgiValueInteger(cgi, "zoom_and", &zsSetup.And, 1) != CGI_OK)
		zsSetup.And = INPUT_RETRIEVE_ERROR;

	if (cgiValueInteger(cgi, "zoom_sort", &zsSetup.Sort, 1) != CGI_OK)
		zsSetup.Sort = zsSetup.DefaultSort;

	// read in dateFromStr and dateToStr
	if (cgiValueString(cgi, "zoom_datefrom", zsSetup.DateFromStr, 10) != CGI_OK)
		strcpy(zsSetup.DateFromStr, "");
	
	if (cgiValueString(cgi, "zoom_dateto", zsSetup.DateToStr, 10) != CGI_OK)
		strcpy(zsSetup.DateToStr, "");

	if(cgiValueInteger(cgi, "zoom_ltimefrom", (long*)&zsSetup.LinuxTimeFrom, 1) != CGI_OK)
		zsSetup.LinuxTimeFrom = 0;

	if(cgiValueInteger(cgi, "zoom_ltimeto", (long*)&zsSetup.LinuxTimeTo, 1) != CGI_OK)
		zsSetup.LinuxTimeTo = 0;

	if(cgiValueInteger(cgi, "masternode", &zsSetup.ShowXMLSettings, 1) != CGI_OK)
			zsSetup.ShowXMLSettings = 0;


	char* scriptName = cgiEnvGet("SCRIPT_NAME");
	if (scriptName != NULL && strlen(scriptName) < TEMP_STR_SIZE)
		strcpy(zsSetup.ExeName, scriptName);					

	//categories selection
	if (zs.UseCats)
	{
		zsSetup.Cat = (int*) malloc(zs.NumCats * sizeof(int));
		if (zsSetup.Cat == NULL)
		{
			printf("Content-type: text/html\n\n");
			printf("Error: Memory Allocation Error");
			statusOK = false;
		}	
		else
		{
			zsSetup.num_zoom_cats = zs.NumCats;		
			if (cgiValueIntegerArray(cgi, "zoom_cat%5B%5D", zsSetup.Cat, &zsSetup.num_zoom_cats, zs.NumCats) != CGI_OK)
			{
				if (cgiValueIntegerArray(cgi, "zoom_cat[]", zsSetup.Cat, &zsSetup.num_zoom_cats, zs.NumCats) != CGI_OK)
				{
					long tmpCat = 0;
					if(cgiValueInteger(cgi, "zoom_cat", &tmpCat, 1) == CGI_OK)	// old single cat param
						zsSetup.Cat[0] = tmpCat;
					else				
						zsSetup.Cat[0] = -1;		// neither parameters are present, default option to all
					zsSetup.num_zoom_cats = 1;
				}				
			}		
		}
	}

	//custom metafields 
	if (zs.UseMetaFields)
	{
		zsSetup.meta_query = (char**) malloc(zs.NumMetaFields * sizeof(char*));		
		if (zsSetup.meta_query == NULL)
		{
			printf("Content-type: text/html\n\n");
			printf("Error: Memory Allocation Error");
			statusOK = false;
		}
		else
		{
			// initialized to zeros
			memset(zsSetup.meta_query, 0, zs.NumMetaFields * sizeof(char*));
			zsSetup.NumMetaFields = zs.NumMetaFields;

			char metaMultiParamStr[METAFIELD_NAME_LEN+1];
			int* tmpIntArray = NULL;
			tmpIntArray = (int*) malloc((METAFIELD_MULTI_MAX+1) * sizeof(int));
			int num_multi_values = 0;

			for (int fieldnum = 0; fieldnum < zs.NumMetaFields; fieldnum++)
			{
				if (zs.metafields[fieldnum].type == METAFIELD_TYPE_MULTI)
				{
					// meta_query will store numeric values up to 255 per char
					// +1 for count at [0] index
					zsSetup.meta_query[fieldnum] = (char*) malloc((METAFIELD_MULTI_MAX+1) * sizeof(char));	
					if (zsSetup.meta_query[fieldnum] == NULL)
					{
						printf("Content-type: text/html\n\n");
						printf("Error: Memory Allocation Error");
						statusOK = false;
						break;
					}

					sprintf(metaMultiParamStr, "%s%%5B%%5D", zs.metafields[fieldnum].name);
					// set num_multi_values to maximum number of results we can receive
					num_multi_values = METAFIELD_MULTI_MAX;
					if (cgiValueIntegerArray(cgi, metaMultiParamStr, tmpIntArray, &num_multi_values, METAFIELD_MULTI_MAX) != CGI_OK)
					{
						sprintf(metaMultiParamStr, "%s[]", zs.metafields[fieldnum].name);
						if (cgiValueIntegerArray(cgi, metaMultiParamStr, tmpIntArray, &num_multi_values, METAFIELD_MULTI_MAX) != CGI_OK)
						{
							long tmpMulti = 0;
							if(cgiValueInteger(cgi, zs.metafields[fieldnum].name, &tmpMulti, 1) == CGI_OK)	// old single cat param
								tmpIntArray[0] = tmpMulti;
							else				
								tmpIntArray[0] = -1;		// neither parameters are present, default option to all							
							num_multi_values = 1;
						}
					}

					zsSetup.meta_query[fieldnum][0] = num_multi_values;
					if (num_multi_values > 0)
					{
						int j = 1;	// start at 1 since 0 holds the count
						for (int i = 0; i < num_multi_values; i++)
						{
							if (tmpIntArray[i] < METAFIELD_MULTI_MAX)
							{
								zsSetup.meta_query[fieldnum][j] = (char)tmpIntArray[i];
								j++;
							}
						}
					}
				}
				else
				{
					zsSetup.meta_query[fieldnum] = (char*) malloc((METAFIELD_TEXT_LEN+1) * sizeof(char));
					if (zsSetup.meta_query[fieldnum] == NULL)
					{
						printf("Content-type: text/html\n\n");
						printf("Error: Memory Allocation Error");
						statusOK = false;
						break;
					}				
					if (cgiValueString(cgi, zs.metafields[fieldnum].name, zsSetup.meta_query[fieldnum], METAFIELD_TEXT_LEN) != CGI_OK)
						strcpy(zsSetup.meta_query[fieldnum], "");
				}
			}	// end for
			if (tmpIntArray)
				free(tmpIntArray);
		}
	}
	return statusOK;
}

// Custom cgi value function (based on cgiValueInteger() from yacgi.c)
// to retrieve an array of CGI values with the same name.
//
// Note that I've omitted the extra code for parsing out of long strings, signs, etc.
// (originally in the yacgi code) and assume that the value will be an int as specified. 
//
// [in]		resultArray		a pointer to an allocated array of int's where the results are stored
// [in/out]	num_results		pointer to an integer storing the maximum number of results that can be returned 
//							returns the actual number of results added to the array.
// [in]		maxValue		Optional. If this is greater than zero, the function will ignore values
//							which are greater than maxValue.
int cgiValueIntegerArray(CGI *cgi, char *name, int* resultArray, int* num_results, int maxValue)
{
    char  *string;
	int	  result_count = 0;

    string = cgiValueFirst(cgi, name);
    if (!string)
        return cgiStateSet(CGI_VAL_NOTFOUND);
    if (!strlen(string))
        return cgiStateSet(CGI_VAL_EMPTY);

	resultArray[result_count] = atol(string);
	if (maxValue == 0 || resultArray[result_count] <= maxValue)
		result_count++;
	else
		return cgiStateSet(CGI_VAL_EMPTY);

	string = cgiValueNext(cgi, name);
	while (string != NULL && result_count < *num_results)
	{
		if (strlen(string) > 0)
		{
			resultArray[result_count] = atol(string);
			if (maxValue == 0 || resultArray[result_count] <= maxValue)
				result_count++;	// only keep this value (and not allow it to be overwritten) if under maxValue
		}
		string = cgiValueNext(cgi, name);
	}
	*num_results = result_count;
    return cgiStateSet(CGI_OK);
}

#endif //CGI
