#ifndef ZOOMSEARCH_H
#define ZOOMSEARCH_H

#ifdef _DEBUG
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>
//#define DEBUG_NEW new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

#define _CRT_SECURE_NO_WARNINGS

#ifndef O_LARGEFILE
#define O_LARGEFILE	0100000	// large file support is a Linux/BSD only thing.
#endif

// ----------------------------------------------------------------------------
// Zoom Search Engine CGI source code
// CGI search function to be used with Zoom Search Engine
//
// Copyright (C) 2004-2016 PassMark Software Pty Ltd, trading as Wrensoft
// ----------------------------------------------------------------------------

// Uncomment when compiling for CGI (if not already specified in makefile or .vcproj file)
//#define CGI_MODE

// Uncomment this define if compiling on non-windows platform (if not already specified in makefile or project file)
//#define NON_WINDOWS

// Uncomment when compiling for ASPX (if not already specified in project file)
//#define ASPX

//	Uncomment this for Sun Sparc build (if not already specified in makefile)
//  Sparc uses Big Endian (as opposed to Intel's Little Endian)
//	Required to reverse the byte ordering of binary data stored in zoom_wordmap.zdat
//	Remember to uncomment #define NON_WINDOWS (i.e. NON_WINDOWS must be definded)
//#define	SUN_SPARC_BUILT

//#define FASTCGI_ENABLED

//#define DEBUG_MODE	// This is for running in a debugger with a query specified below
//#define DEBUG_FILE	// Uncomment this to print out debugging information to file ./logs/debug_log.txt
//#define DEBUG_TIMER		// This enables printing and tracking of time spent.
//#define DEBUG_FASTCGI

// Uncomment this line to use the VC++ debugger to step through the CGI
//#define VC_DEBUGGER_GETQUERY "zoom_sort=0&zoom_xml=0&zoom_query=&zoom_per_page=10&zoom_and=1&zoom_cat%5B%5D=5&ContributedFrom=petit&ContributedBy=&ContributedWhen="
//#define VC_DEBUGGER_GETQUERY "zoom_sort=0&zoom_xml=0&zoom_query=&zoom_per_page=10&zoom_and=1&zoom_cat%5B%5D=-1&ContributedFrom=bnfl&ContributedBy=&ContributedWhen=2004"
//#define VC_DEBUGGER_GETQUERY "zoom_query=page&ZOOM_PST_FROM=&ZOOM_PST_TO=&ZOOM_PST_CC=&zoom_per_page=5000&zoom_and=0&zoom_xml=1"
//#define VC_DEBUGGER_GETQUERY "zoom_query=stiinta+spirituala\""
//#define VC_DEBUGGER_GETQUERY "zoom_query=test"

/*
The 2GB (O_LARGEFILE) problem:
------------------------------
This flag when used with the open() calls throughout the CGI will allow the
CGI to use .ZDAT files greater than 2 GB in size. However, due to older
versions of Linux and BSD not supporting this flag, we are unable to provide
this option and maintain the compatibility with old *nix web servers.

At the moment, the CGI is limited to ZDAT files under 2 GB in size for the
Linux and BSD builds for this reason. The Win32 build will work for files
larger than 2 GB. In the future, we can re-enable this to allow *nix CGIs
to read larger than 2 GB, when the newer versions of libc/*nix servers 
become more common (we are using pair.com as the baseline for comparison).
*/

#if !defined(_WINDOWS)
// Define the following to support .zdat files > 2GB on Linux.
// Will need to compile on more modern linux distributions. Currently tested on Ubuntu 12 x64.
#if _FILE_OFFSET_BITS == 64
#define LINUX_LARGEFILE_USEFSEEKO
#endif
#endif


#define ZOOM_AUTOCOMPLETE_JS	"zoom_autocomplete.js"
#define ZOOM_AUTOCOMPLETE_CSS	"zoom_autocomplete.css"
#define ZOOM_DATEPICKER_JS		"zoom_datepicker.js"
#define ZOOM_DATEPICKER_CSS		"zoom_datepicker.css"

#include <fcntl.h>      /* Needed only for _O_RDWR definition */

#ifndef NON_WINDOWS
#include <io.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <stdarg.h>
//#include <locale.h>


//#include "pcre.h" //include the PCRE library for regular expressions

#include "ucdata.h"

#ifndef NON_WINDOWS
// Windows exclusive includes
#include <windows.h>
#include <mbstring.h>
#else
// Linux/BSD exclusive operatives
#define _LARGEFILE_SOURCE 1
#include <sys/types.h>
#include <sys/time.h>
#endif
#include "strtable.h"

#include "libstemmer_c/include/libstemmer.h"

#define NUM_STEMMING_ALGORITHMS	16	// hand counted from algorithm_names in modules.h

#ifdef ASPX
#using <System.dll>
#using <System.Web.dll>
#include <vcclr.h>
using namespace System;
using namespace System::Text;
using namespace System::Web::UI;
using namespace System::Collections;
//using namespace System::Runtime::InteropServices;

//extern gcroot<HtmlTextWriter^> aspxoutput;

#endif


// This is a fix for Unix systems which do not have EDEADLOCK defined, only EDEADLK
#ifndef EDEADLOCK
#define EDEADLOCK EDEADLK
#endif

#define HISTOGRAM_NUM		10000


//#define CULL_PAGES_LIMIT	1000

#define		TITLELEN		1024		//_MAX_FNAME+_MAX_EXT		
#define		URLLENGTH		32+3+2048	//INTERNET_MAX_URL_LENGTH
#define		MAXPATH			260			//MAX_PATH
#define		MAXDESCLEN		2000

// This is the size of the global buffer TempPrintfBuffer that we use for sprintf() calls 
// (used whenever we call ResultsBufferPrintf() etc.) It needs to be AT LEAST as 
// large as the URLLENGTH or whatever the longest thing is that we will potentially
// printf() out. Since context description size can be user specified, we'll make this
// a healthy 10k in size (and since we only have one global buffer now)
#define		MAX_PRINTF_BUFFERLEN	10240	

//Functions


// WORDLENGTH is now as long as maximum word length (ZOOM_MAXWORDLENGTH_LIMIT) in the Indexer (100).
// But this is in UTF-8. So potentially 4 byte per word, and we are using this for single-byte char arrays.
// so we'll use 400. It's only used for a few temp buffers here and there and nowhere too damaging.
#define WORDLENGTH 400
#define MAXDICTIDLEN 5

#ifdef NON_WINDOWS 
#define	_open		open
#define _close		close
#define _read		read
#define _O_RDONLY	O_RDONLY
#define _O_BINARY	0x0000	// binary mode does not exist outside of windows
#define O_BINARY	0x0000
//#define __time32_t	signed long
#define __int64		long long
#endif

// -------------------------------------------------------------------------
// Defines for Win32 types needed for the CGI
// -------------------------------------------------------------------------
#ifndef BYTE
typedef unsigned char BYTE;
#endif

#ifndef BOOL
typedef int BOOL;
#define TRUE	1
#define FALSE	0
#endif

#ifndef ULONG32
typedef unsigned int ULONG32;
#endif

#ifndef LONG32
typedef signed int LONG32;
#endif

#ifndef _TIME32_T_DEFINED
typedef LONG32 __time32_t;
#define _TIME32_T_DEFINED
#endif

#ifndef _TIME64_T_DEFINED
typedef __int64 __time64_t;
#define _TIME64_T_DEFINED
#endif


static char* MonthNames[] =
{
	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
static char* DayOfWeekNames[] = 
{
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

// Final context seeks define the number of pages before MaxContextSeeks is reached
// in which we change the behaviour of context seeks in an attempt to pickup 
// more obvious phrase matches where the stored offset pointers are next to each other.
#define FINAL_CONTEXTSEEKS	50		

// We allow this many punctuation characters within a considered exact phrase
#define PHRASE_NUM_PUNCT_TOLERANCE 3

#define VARIANT_BUFFER_AVGWORDLEN	10

//struct to hold information from dictionary file (word and pointer)
struct dictstruct
{
	char* wordPtr;
	__int64 ptrID;
	unsigned int variantCount;
	char* variantPtr;
#ifdef OSF
	int numResults;			// Added for OSF ListView specifically (not in Zoom)
#endif
};

#define REC_KEYWORD_LEN 500
struct recstruct
{
	char word[REC_KEYWORD_LEN];
	__int64 index;
};

#define SPCODE_LEN	6

struct spellstruct
{
	char* spcode;
	unsigned long id;
	unsigned long id2;
	unsigned long id3;
};

// Note: the score in res_table_item is a different size to the Indexer. It is
// 2 bytes here, but only 1 byte in the Indexer. The reason is that we need to
// store a larger value when we run through the histogram and recalculate values
// based on number of terms matched
struct res_table_item
{
	unsigned long score;
	unsigned char proximity;
	unsigned long terms;
	__int64 txtptr1;
	__int64 txtptr2;
	__int64 txtptr3;
	unsigned long and_counter;
};
struct output_table_item
{
	unsigned long pageindex;
	unsigned long score;	
	unsigned long terms;
	__int64 txtptr1;
	__int64 txtptr2;
	__int64 txtptr3;	
};

//Structure to hold score, page number and offset
/*
#pragma pack(2)
struct word_data
{
	unsigned char score : 8;
	unsigned char proximity : 8;
	unsigned long pagenum : 32;
	unsigned long textoffset : 32;		
	
	//BYTE	score;		// 2 bytes
	//LONG	pagenum;	// 4 bytes
	//LONG	textoffset;	// 4 bytes
};	
#pragma pack()
*/

#pragma pack(2)
struct word_data
{
	unsigned char score : 8;
	unsigned char proximity : 8;
	//unsigned long pagenum : 32;
	//unsigned long textoffset : 32;		
	unsigned long long pagenum_offset : 64;
	
	/*
	BYTE	score;		// 2 bytes
	LONG	pagenum;	// 4 bytes
	LONG	textoffset;	// 4 bytes
	*/
};	
#pragma pack()

#define MAKE_PAGENUM_AND_TEXTOFFSET(a,b) ((unsigned long long)(((long)(a)& 0xFFFFFFLL) | (((long long)((__int64) (b) & 0xFFFFFFFFFFLL)) << 24)))
#define GET_PAGENUM(dwl) ((dwl) & 0x0000000000FFFFFFLL)
#define GET_TEXTOFFSET(dwl) ((__int64)(((long long)(dwl) >> 24) & 0xFFFFFFFFFFLL))

//Error codes
enum ERRORS
{
	SUCCESS,				//general success
	FAILURE,				//general failure
	MISSING_CAT,			//categories files are missing
	EMPTY_DESCRIPTIONS,		//descripton file is empty
	OUT_OF_MEMORY,			//malloc call failed due to lack of memory
	UNSUPPORTED_FORMAT,		// for when we load in an index from a newer version (or older) that this CGI does not support
	FILE_NOT_FOUND
};


#define OVECCOUNT				33	// must be a multiple of 3.  
#define MAX_QUERY_LENGTH		2085
#define	DATESTRING_LENGTH		14
#define TEMP_STR_SIZE			2048
//#define PAGEINFO_MAX_LINELEN	1024	// no longer used, dynamically allocated
#define PAGEDATA_MAX_LINELEN	7261 //5178
// url|title|desc|imageurl|actualurl
// TITLELEN = 256 + 256, URLLENGTH = 2048 + 3 + 32, MAXDESCLEN = 500, 

// Index format information
#define PAGEDATA_URL		0
#define PAGEDATA_TITLE		1
#define PAGEDATA_DESC		2
#define PAGEDATA_IMG		3

#define PAGEINFO_OFFSET		0
#define PAGEINFO_DATETIME	1	
#define PAGEINFO_FILESIZE	2
#define PAGEINFO_BOOST		3
#define PAGEINFO_LINKACTION 4
#define PAGEINFO_CAT		5
#define PAGEINFO_METAFIRST	6


// This PageInfoData format allows us to support 64-bit sized indexes
// without wasting too much space or needing to switch formats. The
// dataoffset is 5 bytes rather than 8 bytes (int64). This means we need to
// strip an __int64 value to fit in the 5 bytes.
#define PAGEINFODATA_DATAOFFSET_SIZE 5
#define OLDPAGEINFODATA_MAXRECSIZE			65535	// max value for 2 byte (unsigned short)
#define PAGEINFODATA_MAXRECSIZE			4294967295	// max value for 4 byte (unsigned int)
#pragma pack(1)
struct OldPageInfoData		// this format is still used for PHP and ASP and smaller capacity platforms as of V7 build 1014
{
	unsigned short recsize;		// 2 byte format from V7 build 1013 or earlier	
	BYTE dataoffset[PAGEINFODATA_DATAOFFSET_SIZE];
	ULONG32 filesize;
	__time32_t datetime;
	unsigned char boost;
	unsigned char filetype;	
	// This is a variable number of bytes which depends on settings. The first couple of bytes
	// may be used to store the category number(s), while the meta data may follow.
	BYTE data;
};
struct PageInfoData
{	
	unsigned int recsize;			// 4 byte format from V7 build 1014 and later
	BYTE dataoffset[PAGEINFODATA_DATAOFFSET_SIZE];	
	ULONG32 filesize;
	__time32_t filetime;
	unsigned char boost;
	unsigned char filetype;
	// This is a variable number of bytes which depends on settings. The first couple of bytes
	// may be used to store the category number(s), while the meta data may follow.
	BYTE data;	
};
#pragma pack()

static bool CopyOldPageInfoDataToPageInfoData(OldPageInfoData* opid, PageInfoData* pid)
{	
	pid->recsize = (unsigned int) opid->recsize;
	memcpy(pid->dataoffset, opid->dataoffset, sizeof(BYTE) * PAGEINFODATA_DATAOFFSET_SIZE);
	pid->filesize = opid->filesize;
	pid->filetime = opid->datetime;
	pid->boost = opid->boost;
	pid->filetype = opid->filetype;
	return true;
}

// Assistant access functions to switch between OldPageInfoData and PageInfoData 

static unsigned int GetPageInfo_Recsize(PageInfoData* pid, int UseBigPageInfoData=1)
{
	if (UseBigPageInfoData) return pid->recsize;
	OldPageInfoData* opid = (OldPageInfoData*)pid;
	return opid->recsize;
}
static ULONG32 GetPageInfo_Filesize(PageInfoData* pid, int UseBigPageInfoData=1)
{
	if (UseBigPageInfoData)	return pid->filesize;
	OldPageInfoData* opid = (OldPageInfoData*)pid;
	return opid->filesize;
}
static __time32_t GetPageInfo_Filetime(PageInfoData* pid, int UseBigPageInfoData=1)
{
	if (UseBigPageInfoData)	return pid->filetime;
	OldPageInfoData* opid = (OldPageInfoData*)pid;
	return opid->datetime;
}
static unsigned char GetPageInfo_Boost(PageInfoData* pid, int UseBigPageInfoData=1)
{
	if (UseBigPageInfoData) return pid->boost;
	OldPageInfoData* opid = (OldPageInfoData*)pid;
	return opid->boost;
}
static unsigned char GetPageInfo_Filetype(PageInfoData* pid, int UseBigPageInfoData=1)
{
	if (UseBigPageInfoData)	return pid->filetype;
	OldPageInfoData* opid = (OldPageInfoData*)pid;
	return opid->filetype;
}
static BYTE* GetPageInfo_DataOffset(PageInfoData* pid, int UseBigPageInfoData=1)
{
	if (UseBigPageInfoData)	return pid->dataoffset;
	OldPageInfoData* opid = (OldPageInfoData*)pid;
	return opid->dataoffset;
}




// Highlight types used in highlightArray[]
#define HIGHLIGHT_NONE		0
#define HIGHLIGHT_SINGLE	1
#define HIGHLIGHT_START		2
#define HIGHLIGHT_END		3

//---------------------------------------------------------------------
// Custom meta field related structs and defines
//---------------------------------------------------------------------
#define METAFIELD_TYPE		0
#define METAFIELD_NAME		1
#define METAFIELD_SHOW		2
#define METAFIELD_FORM		3
#define METAFIELD_METHOD	4	
#define METAFIELD_DROPDOWN	5

#define METAFIELD_TYPE_NUMERIC	0
#define METAFIELD_TYPE_TEXT		1
#define METAFIELD_TYPE_DROPDOWN 2
#define METAFIELD_TYPE_MULTI	3
#define METAFIELD_TYPE_MONEY	4

#define METAFIELD_METHOD_EXACT			0
#define METAFIELD_METHOD_LESSTHAN		1
#define METAFIELD_METHOD_LESSTHANORE	2	// less than or equal to
#define METAFIELD_METHOD_GREATERTHAN	3
#define METAFIELD_METHOD_GREATERTHANORE 4	// greater than or equal to
#define METAFIELD_METHOD_SUBSTRING		5

#define METAFIELD_NAME_LEN	100
#define METAFIELD_SHOW_LEN	100
#define METAFIELD_TEXT_LEN			65535		// 2-byte limit for length of meta text field
#define METAFIELD_TEXT_SIZE_BYTES	2			// number of bytes used to represent the metafield text length
#define METAFIELD_OLD_TEXT_LEN		255		// 1-byte limit for length of meta text field
#define METAFIELD_OLD_TEXT_SIZE_BYTES 1			// number of bytes used to rep metafield text length in old format


#define METAFIELD_MULTI_MAX	255

#define METAFIELD_DROPDOWN_LEN	64

#define METAFIELD_NOVALUE_MARKER 4294967295UL
#define METAFIELD_DROPDOWN_NOVALUE	255


struct metafield_elem
{
	char name[METAFIELD_NAME_LEN+1];
	int type;
	char showname[METAFIELD_SHOW_LEN+1];	
	char formname[METAFIELD_SHOW_LEN+1];

	int method;
	
	int DropdownCount;
	char** DropdownValues;// we need to allocate this array dynamically
};


// ----------------------------------------------------------------------------
// Template buffers
// ----------------------------------------------------------------------------
// defines for output elements
#define OUTPUT_FORM_START			0
#define OUTPUT_FORM_END				1
#define OUTPUT_FORM_SEARCHBOX		2
#define OUTPUT_FORM_SEARCHBUTTON	3
#define OUTPUT_FORM_RESULTSPERPAGE	4
#define OUTPUT_FORM_MATCH			5
#define OUTPUT_FORM_CATEGORIES		6
#define OUTPUT_FORM_CUSTOMMETA		7

#define OUTPUT_HEADING		8
#define	OUTPUT_SUMMARY		9
#define OUTPUT_SUGGESTION	10
#define OUTPUT_PAGESCOUNT	11
#define OUTPUT_SORTING		12
#define OUTPUT_SEARCHTIME	13
#define OUTPUT_RECOMMENDED	14
#define OUTPUT_PAGENUMBERS	15
#define OUTPUT_CATSUMMARY	16
#define OUTPUT_JAVASCRIPTS	17
#define OUTPUT_FORM_DATERANGE 18

#define OUTPUT_TAG_COUNT	19

// This is the initial size of the OutputBuffers. It will be realloc'ed as necessary.
// Make it bigger for faster/less alloc, but more wasted space in memory.
#define OUTPUTBUFFERS_INITSIZE 100

static char* TemplateShowTags[] =
{
	"<!--ZOOM_SHOW_FORMSTART-->", 
	"<!--ZOOM_SHOW_FORMEND-->", 
	"<!--ZOOM_SHOW_SEARCHBOX-->", 
	"<!--ZOOM_SHOW_SEARCHBUTTON-->",
	"<!--ZOOM_SHOW_RESULTSPERPAGE-->",
	"<!--ZOOM_SHOW_MATCHOPTIONS-->",
	"<!--ZOOM_SHOW_CATEGORIES-->",
	"<!--ZOOM_SHOW_CUSTOMMETAOPTIONS-->",
	"<!--ZOOM_SHOW_HEADING-->",
	"<!--ZOOM_SHOW_SUMMARY-->",
	"<!--ZOOM_SHOW_SUGGESTION-->",
	"<!--ZOOM_SHOW_PAGESCOUNT-->",
	"<!--ZOOM_SHOW_SORTING-->",
	"<!--ZOOM_SHOW_SEARCHTIME-->",
	"<!--ZOOM_SHOW_RECOMMENDED-->",
	"<!--ZOOM_SHOW_PAGENUMBERS-->",
	"<!--ZOOM_SHOW_CATSUMMARY-->",
	"<!--ZOOM_SHOW_JAVASCRIPTS-->",
	"<!--ZOOM_SHOW_DATERANGE-->"
};

// we pre-count the length of these strings to prevent
// having to do a strlen() repeatedly later on
#define TEMPLATE_DEFAULT_TAG		"<!--ZOOMSEARCH-->"
#define TEMPLATE_DEFAULT_TAG_LEN	17
#define TEMPLATE_SEARCHFORM_TAG		"<!--ZOOM_SHOW_SEARCHFORM-->"
#define TEMPLATE_SEARCHFORM_TAG_LEN	27
#define TEMPLATE_RESULTS_TAG		"<!--ZOOM_SHOW_RESULTS-->"
#define TEMPLATE_RESULTS_TAG_LEN	24
#define TEMPLATE_QUERY_TAG			"<!--ZOOM_SHOW_QUERY-->"
#define TEMPLATE_QUERY_TAG_LEN		22

#define INPUT_RETRIEVE_ERROR -1001

#ifdef NON_WINDOWS
#include <unistd.h>                         
#include <stdlib.h>
#endif

//#define OUT_BUFFER_SIZE 65536 //64kb currently
#define OUT_BUFFER_SIZE 1048576	// 640kb


#ifndef NON_WINDOWS
#define TIMER_VALUE LARGE_INTEGER
#else
#define TIMER_VALUE double
#endif

//struct to group together all the data that needs to be passed in to search
struct zoomSearchSetup
{
	//Fill these two then call setup1
	char ScriptName[TEMP_STR_SIZE];
	char ExecutePath[512];
	char ExeName[TEMP_STR_SIZE];

	//Fill the rest and then call search
	long UseXML;
	char clientIP[32];
	bool IsZoomQuery;
	unsigned char Query[MAX_QUERY_LENGTH+1];	
	long Per_Page;
	int NewSearch;
	unsigned long Page;
	long And;
	long IsUserDebugMode;
	long ShowXMLSettings;
	long Sort;
	time_t LinuxTimeFrom;	// this will override DateFromStr and DateToStr. For use with OSF or any cases where time is important
	time_t LinuxTimeTo;
	char DateFromStr[DATESTRING_LENGTH+1];	// "dd/mm/yyyy" + null terminator 
	char DateToStr[DATESTRING_LENGTH+1];
	int* Cat;
	int num_zoom_cats;	

	int DefaultSort;

	int IsZoomAC;
	
	int NumMetaFields;
	char** meta_query;
};

struct context_elem
{
	unsigned long dict_id;
	unsigned int variant_index;
};

struct zoomMetaInfo {
	char szName[METAFIELD_NAME_LEN+1];
	char szValue[METAFIELD_TEXT_LEN+1];
};

struct zoomResultInfo {
	char szTitle[TITLELEN+1];
	char szLink[1024+MAX_QUERY_LENGTH];
	unsigned char filetype;
	char szContext[2048];
	char szDescription[MAXDESCLEN+1];
	char szPubDate[128];
	int iTermsMatched;
	int iScore;

	int iNumMetaFields;
	zoomMetaInfo *metaList;
};

#define DEBUGTIMER_DICTLOOKUP		0
#define DEBUGTIMER_WORDMAPLOOKUP	1
#define DEBUGTIMER_CONTEXTSEEK		2
#define DEBUGTIMER_MEMALLOC			3
#define DEBUGTIMER_PREPQUERY		4
#define DEBUGTIMER_FINDRECLINKS		5
#define DEBUGTIMER_STEMMING			6
#define DEBUGTIMER_PHRASE			7
#define DEBUGTIMER_LOADMETA			8
#define DEBUGTIMER_LOADCATS			8
#define DEBUGTIMER_LOADSTRTABLE		10
#define DEBUGTIMER_COUNT			11
#ifdef DEBUG_TIMER
struct USERDEBUGSTATS
{
	TIMER_VALUE StartTime[DEBUGTIMER_COUNT];
	double TotalTime[DEBUGTIMER_COUNT];	// this is stored in MS for accuracy with small operations
	int SaveCounter[DEBUGTIMER_COUNT];
};
#endif

//possible need to find a way to move these inside the class
int sw_compare(const void *arg1, const void *arg2 );


bool DecodeEntities(char* origStr, char* destStr, char* entitychars, int destLen);
bool EncodeEntities(char* origStr, char* destStr, char* entitychars, int destLen);

#ifdef FASTCGI_ENABLED
int FastCGIPrintf(char* str);
#endif

#ifdef OSF
int OSFPrintf(char* str);
void OSFResultsCallback(zoomResultInfo *resultItem);
#endif

TIMER_VALUE GetTimeNow();
double GetTimeDifference(TIMER_VALUE starttime, TIMER_VALUE endtime);
double GetTimeDifferenceInMilliseconds(TIMER_VALUE starttime, TIMER_VALUE endtime);

#ifdef DEBUG_TIMER

void PrintUserDebug(const char *fmt, ...);
void UserDebugSaveTime(int debugTimerIndex);
void UserDebugAddTime(int debugTimerIndex);
extern USERDEBUGSTATS UserDebugStats;

#else

#define UserDebugSaveTime(int) ((void)0)
#define UserDebugAddTime(int) ((void)0)

#ifdef NON_WINDOWS
#define PrintUserDebug(string, args...) ((void)0)
#else
#define PrintUserDebug(string, ...) ((void)0)
#endif

#endif

extern TIMER_VALUE starttime;
extern TIMER_VALUE endtime;

//extern char TempPrintfBuffer[MAX_PRINTF_BUFFERLEN+1];

//#pragma managed(push, off)
//int zoomprintf(const char *fmt, ...);
//#pragma managed(pop)



class zoomSearch
{
	public:
		zoomSearch();
		~zoomSearch();


		int InitBuffers();
		int	setup1(zoomSearchSetup *zsSetup);

		int InitQuery();
		int LoadQuery(const zoomSearchSetup *zsSetup);
		int OpenIndexFiles(const zoomSearchSetup *zsSetup);
		int LoadDictionary(const zoomSearchSetup *zsSetup, BOOL bLoadNumResults=FALSE);

		int getDictionary(dictstruct **dict);
		unsigned long getNumKeywords();
		unsigned long getDictReservedLimit();

		int	search(const zoomSearchSetup *zsSetup);
		int searchAC(const zoomSearchSetup *zsSetup);

		bool	GetStem(char* word, unsigned int wordlen);
		void ResetStartTime();
		
		int UseCats;
		int	NumCats;								//number of categories CatNames in array
		int NumCatBytes;
		int DisplayCatSummary;

		int UseUTF8;
		char Charset[1024];
		int Codepage;
		int NoCharset;
		int DefaultToXML;
		void cleanup();
		void cleanupQuery();

		int UseMetaFields;
		int NumMetaFields;
		int MetaMoneyShowDec;
		char MetaMoneyCurrency[10];
		metafield_elem* metafields;

		int DotNetUseFormTags;
		int DotNetUsePostBacks;				

		BOOL LoadNumResultsFromDict;

#ifdef ASPX
		gcroot<HtmlTextWriter^> aspxoutput;
		void SetASPXOutput(HtmlTextWriter^ outputPtr);
#endif

	private:

		int zoomprintf(const char *fmt, ...);
		int PrintStr(const char* str);
		int ResultsBufferPrintf(const char *fmt, ...);
		int OutputBufferPrintf(int bufferIndex, const char *fmt, ...);
		void OutputBufferClear(int bufferIndex);


		//functions
		int SortByDate(const void *arg1, const void *arg2 );
		int SortCompare(const void *arg1, const void *arg2 );
		int SortByDateAsc(const void *arg1, const void *arg2 );

		//void	PrintEndOfTemplate();
		void	IncludeAdditionalJS();
		void	ShowTemplate();
		void	ShowDefaultForm();
		bool	IncludeFile(char* filepath);

		int readSettings();
		//int readCategories();
		void PrintNumResults(int num, char* text);
		long myfseek(FILE* stream, __int64 pos);
		FILE* myfopen(const char* filename, const char* mode);

		unsigned long GetNextDictWord(FILE* fp_pagetext, unsigned long* dict_id, unsigned int* variant_index=NULL);
		char* GetDictionaryWord(unsigned long dict_id, unsigned int variant_index);
		char* GetSpellingWord(unsigned long dict_id);

		char* FindVariant(char* variantPtr, int variant_index);

		//unsigned long myfseek(int fd, unsigned long pos);
		//int myfread(int fd, void *buffer, unsigned int count);
		int readDictionary();
		
		
		int GetBytes(FILE* fp, int numbytes);

		int readPageinfo();
		int readRecommended();
		
		
		bool addDictionaryEntry(char* line, unsigned long lineNo);
		char* addVariantEntry(char* line);
		bool strReplace(char* search, char* replace, char* subject, bool char4char, int MAXSIZE);
		void SkipSearchWord(int word);
		int GetDictID(char* word);
		void PrintHighlightDescription(char* line, int bufferIndex=-1);
		bool RecLinkWordMatch(char* rec_word, __int64 rec_idx);
		void ToLowerAndStem(char* destBuffer);

		void AddParamToURL(char* destURL, char* srcURL, char* paramStr, char* paramValue);
		void RecLinkAddParamToURL(char* destURL, char* srcURL, char* paramStr, char* paramValue);
		//unsigned long GetNextDictWord(int fp_pagetext, unsigned long* dict_id);		

		char* pattern2regexp(char* pattern);
		void QueryEntities(char* str, int MAXSIZE);
		void htmlspecialchars(char* str, int MAXSIZE);
		void urlencode(char* src, int MAXSIZE);
		bool isWordChar(char test);
		bool ConvertAccentedWord(char* word, int MAXSIZE);
		int Trim(char* buffer); //remove leading and trailing spaces from buffer
		char* iStrStr(const char *string, const char *strCharSet, int nStrLen=0); //case insensitive strstr 
		//void PrintDebugLine(char* string);

		bool GetSPCode(char* word, char* soundex);
		bool GetMetaphone(char* word, char* soundex, int len);
		bool GetPageData(__int64 index, char* url, char* title, char* desc, char* image);
		bool GetUrlFromPageData(__int64 index, char* urlBuffer);

		int ASCII_to_Integer(char *ASCIIString);
		__int64 ASCII_to_Int64(char *ASCIIString);

		//int WildCardMatch (char* haystack, char* needle, int CaseInsensitive, int SubstringMatch);
		bool WildCardCompare(
						char * pTameText,             // A string without wildcards
						char * pWildText,             // A (potentially) corresponding string with wildcards
						bool bCaseSensitive = false,  // By default, match on 'X' vs 'x'
						char cAltTerminator = '\0'    // For function names, for example, you can stop at the first '('
						);

		void DisplayXMLSettings();
		void utf8_strlwr(unsigned char* string);
		int mystrcmp_rightcase(const char *string1, const char *string2);
		int mystricmp(const char *string1, const char *string2);
		int mystrnicmp(const char *string1, const char *string2, int len);
		char* preg_quote(char c, char delim);			

		void HighlightContextArray();		

		void TruncateInt64ToBytes(__int64 srcValue, BYTE* destBytesArray, int numBytes);
		__int64 ExpandBytesToInt64(BYTE* byteArray, int numBytes);

		int CheckBitInByteArray(int bitnum, BYTE* byteArray);

		void endianAware_memcpy(void *dst, void *src, size_t size, size_t size_dst = 0);




		//-------------------------------------------------------------
		//-------------------------------------------------------------
		//variables
		//-------------------------------------------------------------
		//-------------------------------------------------------------
		char tmpPath[1024];
		char target[TEMP_STR_SIZE];

		unsigned long histogram_max;
		unsigned long histogram_min;
		unsigned long histogram_threshold;
		unsigned long cullpages_count;
		bool IsCulled;
		char** StrTable;
		int NewSearch;

		struct sb_stemmer* stemmer;
		const sb_symbol* stemmed;


		unsigned int context_word_count;
		//unsigned long* contextArray;
		context_elem* contextArray;
		BYTE* highlightArray;

		char*** SearchPhrases;
		char*** SearchPhrasesForHL;
		
		char**	OutputBuffers;
		int*	OutputBufferSizes;
		int*	OutputBufferLens;

		char* OutputResultsBuffer;
		int OutputResultsSize;
		int OutputResultsLen;

		char tmpGetParamBuffer[URLLENGTH+1+20];	//+20 for strcat operation for zoom_sort= parameter
		char tmpLinkBuffer[URLLENGTH+1];

		//Flags - set after reading the settings.dat file
		// ----------------------------------------------------------------------------
		// Settings (change if necessary)
		// ----------------------------------------------------------------------------

		int Timing;         // 0 = Do not display and calculate search time
		// 1 = Display and calculate search time

		int Logging;        // 0 = No logging of words that a user enter.
		// 1 = Words are logged to a file for later analysis. (See
		//     documentation for file permission issues)

		int ZoomInfo;       // 0 = Don't display Zoom info line at bottom of search
		// 1 = Display Zoom info line at bottom of search

		
		char* GetNextDictWordBuff;
		int GetNextDictWordBuffRead;
		int GetNextDictWordBuffCount;

		int SearchMultiCats;

		int FormFormat;

		char ContentType[1024];
		long UseXML;
		
		long ShowXMLSettings;
		char XMLTitle[TITLELEN+1];
		char XMLDescription[MAXDESCLEN+1];
		char XMLLink[URLLENGTH+1];
		char XMLOpenSearchDescURL[URLLENGTH+1];
		char XMLStyleSheetURL[URLLENGTH+1];
		int XMLHighlight;
		
		int UseStemming;
		int StemmingAlgo;		
		
		int MapAccents;
		unsigned char** AccentChars;
		char** NormalChars;
		unsigned int MinWordLen;
		int Highlighting;
		int HighlightLimit;
		int GotoHighlight;
		int PdfHighlight;
		int PluginOpenNewWindow;

		//int ResultFormat;

		int UseDateTime;
		int DisplayDate;
		int DefaultSort;
		int DateRangeSearch;
		int DateRangeFormat;
		char DatePickerPath[URLLENGTH+1];

		int UseAutoComplete;


		int DisplayMetaDesc;
		int	DisplayContext;
		int DisplayNumber;
		int DisplayTitle;
		int DisplayTerms;
		int DisplayScore;
		int DisplayURL;
		int DisplayFilesize;
		int DisplayMetaFields;
		int UseZoomImage;

		int TruncateShowURL;

		int Spelling;
		int SpellingWhenLessThan;

		int MaxKeyWordLineLen;
		int OutputBasewordBufferSize;
		int OutputVariantBufferSize;

		int DictIDLen;
		unsigned long DictReservedLimit;
		unsigned long DictReservedPrefixes;
		unsigned long DictReservedSuffixes;
		unsigned long DictReservedNoSpaces;
		int UseBigPageInfoData;
		int FileTypePluginStart;
		int WordSplit;
		int SearchAsSubstring;
		int ToLowerSearchWords;
		int DefaultToAnd;
		int ContextSize;
		unsigned long MaxContextKeywords;
		int AllowExactPhrase;
		int MaxResultsPerQuery;
		
		unsigned long MaxMatches;
		int MaxContextSeeks;
		int MaxSearchTime;
		
		int UseLinkTarget;
		char WordJoinChars[1024];
		char LinkTarget[2048];
		int ExactPhrase;
		int NumAccents;
		unsigned long NumKeywords;
		long NumPages;
		long NumVariants;
		unsigned int PageInfoSize;
		unsigned int PageDataSize;

		signed int WeightProximity;

		char VersionInfo[1024];
		unsigned long NumSpellings;
		int Recommended;
		unsigned int NumRecommended;
		int RecommendedMax;


		//Global Vars, that are no longer global.
		char ExecutePath[512];
		char SettingsFilename[256];				//settings file filename
		char LogFilename[256];					//Path and File name of search word log file
		char TemplateFilename[256];				//Set this to your template HTML page (for the formatting and location of the search form)
		//char CategoryFilename[256];				//categories filename
		char ZoomPagedataFilename[256];
		char ZoomPageTextFilename[256];			//zoom pagetext filename
		char ZoomPageinfoFilename[256];			// zoom pageinfo filename
		char ZoomRecommendedFilename[256];		// zoom recommended filename
		char ZoomAutocompleteFilename[256];		// zoom autocomplete filename


		unsigned char Query[MAX_QUERY_LENGTH+1];				//query search string
		unsigned char QueryForSearch[MAX_QUERY_LENGTH+50];	//query encoded for searching (some special chars encoded to match indexer behaviour)
		unsigned char QueryForHTML[MAX_QUERY_LENGTH+50];	//query encoded with HTML entities, ready for HTML output
		unsigned char QueryForURL[MAX_QUERY_LENGTH+50];		//query encoded for inclusion as part of a URL (eg. GET parameter)
		char metaParams[MAX_QUERY_LENGTH+1];
		char dateRangeParams[100];	// "&amp;zoom_datefrom=dd%2Fmm%2Fyyyy&amp;zoom_dateto=dd%2Fmm%2Fyyyy" + null
		
		char DateFromStr[DATESTRING_LENGTH+1];	// "dd/mm/yyyy" + null terminator (but encoded is "dd%2Fmm%2Fyyyy")
		char DateToStr[DATESTRING_LENGTH+1];


		char SkippedOutputStr[MAX_QUERY_LENGTH+50]; //skipped words string
		char** CatNames;							//arry of category names
		
		int* Cat;									//selected category from search (now an array)
		int num_zoom_cats;							// the number of categories that has been queried

		int num_recs_found;
		char zoom_target[TEMP_STR_SIZE];


		char** meta_query;
		//char*** metavalues;
		
		__int64 LastRecLinkIdx;

		//unsigned long* pgdataoffset;		

		BYTE* pageinfobin;
		PageInfoData** pageinfo;
		BYTE** catpages;
		BYTE** metapages;

		spellstruct* Spell;

		FILE* templateFp;						//file pointer to template file, opened in readTemplate, closed in PrintEndOfTemplate
		FILE* fp_pagetext;
		FILE* fp_wordmap;
		//int fp_pagetext;
		//int fp_wordmap;

		//FILE* fp_dict;
		FILE* fp_spell;
		FILE* fp_pagedata;

		bool IsPageDataPreloaded;
		BYTE* pagedatabin;

		char clientIP[32];
		char** SearchWords;
		char** RegExpSearchWords;
		unsigned long NumSearchWords;
		int SkippedWords;
		dictstruct* Dict;

		int* search_terms_ids;
		int** phrase_terms_ids;

		char* baseword_buffer;
		char* baseword_endPtr;
		unsigned long baseword_size;
		unsigned long baseword_len;

		char* variant_buffer;
		char* variant_endPtr;
		unsigned long variant_size;
		unsigned long variant_len;

		int exclude_count;

		recstruct* Rec;
		int rec_count;

		//unsigned long NumWords;
		BYTE* UseWildCards;

		int PerPageOptions[5];		// The options available in the dropdown menu for number of results per page
		long Per_Page;				// The selected opton
		unsigned long Page;
		long And;					// AND operator 1 if we are searching for ALL terms 0 if we are searching for ANY terms (default)
		long Sort;					// sorting operator. zero is default, 1 is sort by date if date/time is available.

		long IsUserDebugMode;

		char ScriptName[TEMP_STR_SIZE];	//name of the running executable (may be overwritten by LinkBackURL)		
		char LinkBackJoinChar[10];	// this is for "?" or "&amp;"
		char ExeName[TEMP_STR_SIZE];
		

		bool IsMaxLimitExceeded;

		bool IsZoomQuery;
		bool IsNoKeywordQuery;

		bool IsPhraseStarted;	
		bool IsPhraseFound;		


		//Debug mode functions and vars
#ifdef DEBUG_FILE
		char debugFilename[256];		
#endif
		

#ifndef NON_WINDOWS 
		//LARGE_INTEGER starttime2;
		char* GetLastErrorString();
#else
		char* _strdup(const char *strSource);
		int _stricmp(const char *string1, const char *string2);
		int _strnicmp(const char *string1, const char *string2, int len);
		void _strlwr(char* string);
		void _mbslwr(unsigned char* string);
		int _eof(int filedes);
#endif


};

#endif
