// ----------------------------------------------------------------------------
// Zoom Search Engine CGI
//
// CGI search function to be used with Zoom Search Engine
//
// A fast custom website search engine using pre-indexed data files.
// Copyright (C) 2004-2016 PassMark Software Pty Ltd, trading as Wrensoft
//
//	This software is the confidential and proprietary information of Wrensoft
//  You shall not disclose such Confidential Information and shall use it
//  only in accordance with the terms of the agreement you entered into
//  with Wrensoft.
//
//
// WRENSOFT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
// NO EVENT WILL WRENSOFT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
// CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
// OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
// OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
// USE OR PERFORMANCE OF THIS SOFTWARE.
//
// email: zoom@wrensoft.com
// www: http://www.wrensoft.com
// ----------------------------------------------------------------------------
//
// main.cpp - main file for zoom cgi app
// Created: 16/03/04 
#include "zoomsearch.h"
#include "qsort.h"
#include "libstemmer_c/libstemmer/modules.h"

#ifdef ASPX
#include <windows.h>
#include <msclr/marshal.h>
using namespace msclr::interop;
#endif

#ifndef snprintf
#ifdef _WINDOWS
#define snprintf _snprintf
#endif
#endif

/*
// For Debian 2.2 or other old versions of Linux without largefile support
#include <stdio.h>
#ifndef fseeko
// if we don't have fseeko then we can't use it. force old fseek() behaviour
#ifdef LINUX_LARGEFILE_USEFSEEKO
#undef LINUX_LARGEFILE_USEFSEEKO
#endif
#endif
*/

void zoomSearch::PrintNumResults(int num, char* text)
{		
    if (num == 0)
		OutputBufferPrintf(OUTPUT_SUMMARY, "%s %s ", StrTable[STR_NO_RESULTS], text);        
    else if (num == 1)
		OutputBufferPrintf(OUTPUT_SUMMARY, "%d %s %s ", num, StrTable[STR_RESULT], text);
    else
	{
		if (IsMaxLimitExceeded)
			OutputBufferPrintf(OUTPUT_SUMMARY, "%s ", StrTable[STR_MORETHAN]);
		OutputBufferPrintf(OUTPUT_SUMMARY, "%d %s %s ", num, StrTable[STR_RESULTS], text);
	}
}

zoomSearch::zoomSearch()
{
	StrTable = NULL;
	pageinfo = NULL;
	pageinfobin = NULL;
	pagedatabin = NULL;
	catpages = NULL;
	metapages = NULL;
	Spell = NULL;	

	//Global Flags - set after reading the settings.dat file
	// ----------------------------------------------------------------------------
	// Settings (change if necessary)
	// ----------------------------------------------------------------------------

	Timing = 0;         // 0 = Do not display and calculate search time
	// 1 = Display and calculate search time

	Logging = 0;        // 0 = No logging of words that a user enter.
	// 1 = Words are logged to a file for later analysis. (See
	//     documentation for file permission issues)

	ZoomInfo = 1;       // 0 = Don't display Zoom info line at bottom of search
	// 1 = Display Zoom info line at bottom of search
	

	UseCats = 0;		
	SearchMultiCats = 0;
	UseWildCards = NULL;

	FormFormat = 0;

	ContentType[0] = '\0';
	UseXML = 0;
	DefaultToXML = 0;
	ShowXMLSettings = 0;
	XMLTitle[0] = '\0';
	XMLDescription[0] = '\0';
	XMLLink[0] = '\0';
	XMLOpenSearchDescURL[0] = '\0';
	XMLStyleSheetURL[0] = '\0';
	XMLHighlight = 0;

	UseUTF8 = 0;
	Charset[0] = '\0';
	Codepage = 0;	// 0 is CP_ACP (default ANSI codepage) in WinNls.h 	

	UseStemming = 0;
	StemmingAlgo = 2;	// default to english

	NoCharset = 0;
	MapAccents = 0;
	AccentChars = NULL;
	NormalChars = NULL;
	MinWordLen = 0;
	Highlighting = 0;
	HighlightLimit = 30;
	GotoHighlight = 0;
	PdfHighlight = 0;
	PluginOpenNewWindow = 0;

	//int ResultFormat = 2;	

	DotNetUseFormTags = 1;	// default to on for backwards compatibility
	DotNetUsePostBacks = 0;

	UseDateTime = 0;
	DisplayDate = 0;
	DefaultSort = 0;

	DateRangeSearch = 0;
	DateRangeFormat = 0;
	DatePickerPath[0] = '\0';

	UseAutoComplete = 0;

	DisplayCatSummary = 0;

	DisplayMetaDesc = 0;
	DisplayContext = 0;
	DisplayNumber = 0;
	DisplayTitle = 0;
	DisplayTerms = 0;
	DisplayScore = 0;
	DisplayURL = 0;
	DisplayFilesize = 0;
	DisplayMetaFields = 0;
	UseZoomImage = 0;

	TruncateShowURL = 0;

	Spelling = 0;
	SpellingWhenLessThan = 0;

	MaxKeyWordLineLen = 32;
	OutputBasewordBufferSize = 0;
	OutputVariantBufferSize = 0;

	DictIDLen = 0;
	DictReservedLimit = 0;
	DictReservedPrefixes = 0;
	DictReservedSuffixes = 0;
	DictReservedNoSpaces = 0;
	UseBigPageInfoData = 0;
	FileTypePluginStart = 4;
	WordSplit=0;
	SearchAsSubstring = 0;
	ToLowerSearchWords = 0;
	DefaultToAnd = 0;
	ContextSize = 0;
	MaxContextKeywords = 0;
	AllowExactPhrase = 0;
	MaxContextSeeks = 0;
	MaxMatches = 0;
	MaxSearchTime = 30;
	MaxResultsPerQuery = 1000;

	UseLinkTarget = 0;
	WordJoinChars[0] = '\0';
	LinkTarget[0] = '\0';
	ExactPhrase=0;
	NumAccents=0;
	NumKeywords = 0;
	NumPages = 0;
	NumVariants = 0;
	VersionInfo[0] = '\0';
	NumSpellings = 0;
	Recommended = 0;
	NumRecommended = 0;
	RecommendedMax = 0;
	
	//Global Vars, that are no longer global.

	strcpy(SettingsFilename, "settings.zdat");				//settings file filename
	strcpy(LogFilename, ".//logs//searchwords.log");	//Path and File name of search word log file
	strcpy(TemplateFilename,"search_template.html");		//Set this to your template HTML page (for the formatting and location of the search form)
	//strcpy(CategoryFilename, "zoom_cats.zdat");					//categories filename
	strcpy(ZoomPagedataFilename, "zoom_pagedata.zdat");
	strcpy(ZoomPageTextFilename, "zoom_pagetext.zdat");			//zoom pagetext filename
	strcpy(ZoomPageinfoFilename, "zoom_pageinfo.zdat");			// zoom pageinfo filename
	strcpy(ZoomRecommendedFilename, "zoom_recommended.zdat");	// zoom recommended filename
	strcpy(ZoomAutocompleteFilename, "zoom_autocomplete.zdat");

	tmpLinkBuffer[0] = '\0';
	tmpGetParamBuffer[0] = '\0';


	SkippedOutputStr[0]	= '\0';		//skipped words string
	CatNames			= NULL;		//arry of category names
	NumCats				= 0;		//number of categories CatNames in array
	//long Cat			= -1;		//selected category from search
	Cat					= NULL;		//selected category from search (now an array)
	num_zoom_cats		= 0;		// the number of categories that has been queried	
	

	//pgdataoffset		= NULL;

	//datetime			= NULL;		// array of datetimes

	UseMetaFields		= 0;
	NumMetaFields		= 0;
	MetaMoneyShowDec	= 0;
	strcpy(MetaMoneyCurrency, "&#36;");	// default is dollar sign
	metafields			= NULL;
	meta_query			= NULL;
	//metavalues			= NULL;

	templateFp			= NULL;		//file pointer to template file, opened in readTemplate, closed in PrintEndOfTemplate
	fp_pagetext			= NULL;
	fp_wordmap			= NULL;

	//fp_dict			= NULL;
	fp_spell			= NULL;
	fp_pagedata			= NULL;

	SearchWords			= NULL;
	RegExpSearchWords	= NULL;
	Dict				= NULL;
	variant_buffer		= NULL;
	baseword_buffer		= NULL;

	Rec					= NULL;
	rec_count			= 0;

	// The options available in the dropdown menu for number of results per page
	PerPageOptions[0]	= 10; 
	PerPageOptions[1]	= 20; 
	PerPageOptions[2]	= 50; 
	PerPageOptions[3]	= 100; 
	PerPageOptions[4]	= -1; 

	OutputBuffers = NULL;
	OutputBufferSizes = NULL;	
	OutputBufferLens = NULL;

	OutputResultsBuffer = NULL;
	OutputResultsSize = 0;	
	OutputResultsLen = 0;

	zoom_target[0] = '\0';

	search_terms_ids = NULL;
	phrase_terms_ids = NULL;

	GetNextDictWordBuff = NULL;

	PageInfoSize = 0;
	PageDataSize = 0;
	IsPageDataPreloaded = false;

	LoadNumResultsFromDict = FALSE;

	InitQuery();

#ifdef DEBUG_TIMER
	memset(&UserDebugStats, 0, sizeof(USERDEBUGSTATS));
#endif

#ifndef FASTCGI_ENABLED
	//get current time 	
	starttime = GetTimeNow();
#endif

#ifdef DEBUG_FILE
	strcpy(debugFilename, "error_log.txt");
#endif

}

zoomSearch::~zoomSearch()
{
	//cleanup();
}

int zoomSearch::InitBuffers()
{
	// initialize output buffers	
	OutputBuffers = (char**) malloc(sizeof(char*) * OUTPUT_TAG_COUNT);
	if (OutputBuffers == NULL)
		return FAILURE;

	// make an array of int's to keep track of each of the sizes for the above buffers
	OutputBufferSizes = (int*) malloc(sizeof(int) * OUTPUT_TAG_COUNT);
	if (OutputBufferSizes == NULL)
		return FAILURE;

	OutputBufferLens = (int*) malloc(sizeof(int) * OUTPUT_TAG_COUNT);
	if (OutputBufferLens == NULL)
		return FAILURE;

	for (int bufnum = 0; bufnum < OUTPUT_TAG_COUNT; bufnum++)
	{	
		OutputBuffers[bufnum] = (char*) malloc(sizeof(char) * OUTPUTBUFFERS_INITSIZE);
		if (OutputBuffers[bufnum] == NULL)
			return FAILURE;
		OutputBuffers[bufnum][0] = '\0';
		OutputBufferSizes[bufnum] = OUTPUTBUFFERS_INITSIZE;		
		OutputBufferLens[bufnum] = 0;
	}
#ifdef OSF
	OutputResultsSize = 10 * PAGEDATA_MAX_LINELEN;
#else
	OutputResultsSize = Per_Page * PAGEDATA_MAX_LINELEN;
#endif
	OutputResultsBuffer = (char*) malloc(sizeof(char) * OutputResultsSize);
	OutputResultsBuffer[0] = '\0';
	OutputResultsLen = 0;

	return SUCCESS;
}

#ifdef ASPX
void zoomSearch::SetASPXOutput(HtmlTextWriter^ outputPtr)
{
	aspxoutput = outputPtr;
}
#endif


int	zoomSearch::setup1(zoomSearchSetup *zsSetup)
{
	strcpy(ExecutePath, zsSetup->ExecutePath);
	strcpy(ScriptName, zsSetup->ScriptName);
	zsSetup->IsZoomAC = 0;	
	UseBigPageInfoData = 0; // need to be reset here since OSF/FastCGI won't re-initialize the zoomsearch class per query
	
	// default link back join char (this is for joining selfURL to "zoom_query=blah&etc.")
	strcpy(LinkBackJoinChar, "?");

	// Initialize output buffers
	InitBuffers();

	// initialize string table
	StrTable = (char**) malloc((STR_TABLE_COUNT)  * sizeof(char*));
	memset(StrTable, 0, STR_TABLE_COUNT * sizeof(char*));
	
	int nRet = readSettings();
	if (nRet != SUCCESS)
	{		
#ifndef ASPX
		zoomprintf("Content-type: text/html\n\n");
#endif
		
		switch(nRet)		
		{
		case OUT_OF_MEMORY:
			zoomprintf("Unable to load settings.zdat file due to lack of memory.<br />Check that the file is valid and not corrupted and that server load is normal.");
			break;
		case FAILURE:
			zoomprintf("Unable to load settings.zdat file.<br />The file appears to be invalid and may have been corrupted.");
			break;
		case UNSUPPORTED_FORMAT:
			zoomprintf("Unable to load settings.zdat file.<br />The file appears to be invalid and may have been corrupted.");
			break;
		case FILE_NOT_FOUND:
#if !defined(NON_WINDOWS) && !defined(ASPX)
			char* msg = GetLastErrorString();
			zoomprintf("Unable to open settings.zdat or file is invalid.<br />%s<br />Check file permissions and that file exists", msg);
			LocalFree(msg);
#else
			zoomprintf("Unable to open %s/%s.<br />Check file permissions and that file exists", ExecutePath, SettingsFilename);
#endif
			break;
		}
			
		return FAILURE;				
	}	

	// Settings we need to remember in main.cpp
	zsSetup->DefaultSort = DefaultSort;

	return SUCCESS;
}

TIMER_VALUE GetTimeNow()
{	
	TIMER_VALUE timerVal;
#ifndef NON_WINDOWS	
	QueryPerformanceCounter(&timerVal);			
#else	
	struct timeval timeval_start;
	gettimeofday(&timeval_start, NULL);
	timerVal = timeval_start.tv_sec + (double) timeval_start.tv_usec/1000000;		
#endif
	return timerVal;
}

double GetTimeDifference(TIMER_VALUE starttime, TIMER_VALUE endtime)
{
	double timeDiff;
#ifndef NON_WINDOWS	
	LARGE_INTEGER frequency;//todo: move global & init once
	QueryPerformanceFrequency(&frequency);	
	//double hres = (double)( ( endtime.QuadPart - starttime.QuadPart) /	frequency.QuadPart);
	timeDiff = (double) ( endtime.QuadPart - starttime.QuadPart ) / (double) frequency.QuadPart;
#else	
	timeDiff = endtime - starttime;
#endif
	return timeDiff;
}

double GetTimeDifferenceInMilliseconds(TIMER_VALUE starttime, TIMER_VALUE endtime)
{
	double timeDiff;
#ifndef NON_WINDOWS	
	LARGE_INTEGER frequency;//todo: move global & init once
	QueryPerformanceFrequency(&frequency);	
	frequency.QuadPart = frequency.QuadPart / 1000;	// get it in milliseconds	
	timeDiff = (double) ( endtime.QuadPart - starttime.QuadPart ) / (double) frequency.QuadPart;
#else	
	// TODO: millisecond conversion
	timeDiff = endtime - starttime;
#endif
	return timeDiff;
}


#ifdef DEBUG_TIMER
void UserDebugSaveTime(int debugTimerIndex)
{
	UserDebugStats.StartTime[debugTimerIndex] = GetTimeNow();	
}

void UserDebugAddTime(int debugTimerIndex)
{
	TIMER_VALUE endtime = GetTimeNow();
	double timeDifference = GetTimeDifferenceInMilliseconds(UserDebugStats.StartTime[debugTimerIndex], endtime);
	UserDebugStats.TotalTime[debugTimerIndex] += timeDifference;
	UserDebugStats.SaveCounter[debugTimerIndex]++;
}

void PrintUserDebug(const char *fmt, ...) 
{ 
	//if (IsUserDebugMode == 0)
	//	return;

	va_list	ap;		// variable list
	if (fmt == NULL) 
		return;

#ifdef DEBUG_FILE
	FILE* fp = fopen(debugFilename, "a");
	if (fp == 0)
		return;
#endif

	TIMER_VALUE endtime = GetTimeNow();
	double timeDifference = GetTimeDifference(starttime, endtime);

	char userDebugBuffer[MAX_PRINTF_BUFFERLEN+1] = {0};
		
	va_start(ap, fmt);
		//vfprintf(fp, fmt, ap);
		//ResultsBufferPrintf(fmt, ap);
		vsnprintf(userDebugBuffer, MAX_PRINTF_BUFFERLEN, fmt, ap);
#ifdef DEBUG_FILE
		vfprintf(fp, fmt, ap);
#endif
	va_end(ap);

//	ResultsBufferPrintf("<div class=\"debug\">(%.6f) %s</div>\n", timeDifference, userDebugBuffer);	
#ifdef FASTCGI_ENABLED
	char tmpstr[1024];
	sprintf(tmpstr, "<div class=\"debug\">(%.6f) %s</div>\n", timeDifference, userDebugBuffer);	
	FastCGIPrintf(tmpstr);
#else
	zoomprintf("<div class=\"debug\">(%.6f) %s</div>\n", timeDifference, userDebugBuffer);	
#endif

#ifdef DEBUG_FILE
	fclose(fp);
#endif
}
#endif

int zoomSearch::InitQuery()
{
	histogram_max = 0;
	histogram_min = 0;
	histogram_threshold = 0;
	cullpages_count = 0;
	IsCulled = false;

	NumSearchWords		= 0;
	SkippedWords		= 0;

	Per_Page			= 10;	//The selected opton
	Page				= 1;
	And					= 0;	//AND operator 1 if we are searching for ALL terms 0 if we are searching for ANY terms (default)
	Sort				= 0;	// sorting operator. zero is default, 1 is sort by date if date/time is available.
	LastRecLinkIdx = 0;

	DateFromStr[0] = '\0';
	DateToStr[0] = '\0';

	IsMaxLimitExceeded	= false;
	IsZoomQuery			= false;
	IsNoKeywordQuery	= false;

	num_recs_found		= 0;
	exclude_count		= 0;

	stemmer = NULL;

	Query[0]			= '\0';		//query search string
	QueryForSearch[0]	= '\0';		//query encoded for searching (some special chars encoded to match indexer behaviour)
	QueryForHTML[0]		= '\0';		//query encoded with HTML entities, ready for HTML output
	QueryForURL[0]		= '\0';		//query encoded for inclusion as part of a URL (eg. GET parameter)

	return 1;
}

int zoomSearch::LoadQuery(const zoomSearchSetup *zsSetup) //added  char *envp[] to print environment, MAY BREAK UNDER LINUX
{
	//copy all the input vars to class vars, might want to redo the rest of the code later so this isn't needed.
#ifndef NON_WINDOWS
	if (starttime.HighPart == 0 && starttime.LowPart == 0 && starttime.QuadPart == 0)
#else
	if (starttime == 0)
#endif
	{
		//get current time 	
		starttime = GetTimeNow();
	}
	
	UseXML = zsSetup->UseXML;
	strcpy(clientIP, zsSetup->clientIP);
	IsZoomQuery = zsSetup->IsZoomQuery;
	strcpy((char*)Query, (char*)zsSetup->Query);	
	Per_Page = zsSetup->Per_Page;
	NewSearch = zsSetup->NewSearch;
	Page = zsSetup->Page;
	ShowXMLSettings = zsSetup->ShowXMLSettings;	
	Sort = zsSetup->Sort;
	And = zsSetup->And;
	IsUserDebugMode = zsSetup->IsUserDebugMode;	

	strcpy(ExeName, zsSetup->ExeName);

	strcpy((char*)DateFromStr, (char*)zsSetup->DateFromStr);
	strcpy((char*)DateToStr, (char*)zsSetup->DateToStr);

	//NumCats = zsSetup->NumCats;				
	Cat	= zsSetup->Cat;							
	num_zoom_cats = zsSetup->num_zoom_cats;					

	meta_query = zsSetup->meta_query;	
		
	if (zsSetup->IsZoomAC)
		SearchAsSubstring = 1;
			 
	// ----------------------------------------------------------------------------
	// Starts here
	// ----------------------------------------------------------------------------
	
	
	
#ifndef ASPX

	if(UseXML == INPUT_RETRIEVE_ERROR)
	{
		if (DefaultToXML == 1)
			UseXML = 1;
		else
			UseXML = 0;
	}		


	if (UseXML)
		sprintf(ContentType, "text/xml");
	else
		sprintf(ContentType, "text/html");
	
	// Write HTTP header including charset if specified in the settings file.
	// Note that this has to be the first thing outputted for it to take effect
	if (NoCharset == 0)
		zoomprintf("Content-type: %s; charset=%s\n\n", ContentType, Charset);
	else
		zoomprintf("Content-type: %s;\n\n", ContentType);	

#endif //ASPX

	if (STEMMER_VERSION != 1)
	{
		// This source tree is expecting a particular version of the libstemmer library
		// that we have customized to handle a few bugs/scenarios. If the libstemmer 
		// components are being updated, please verify the changes necessary. (see libstemmer.h)
		zoomprintf("ERROR: Incorrect stemmer version.");
		exit(1);
	}


#ifdef VC_DEBUGGER_TESTQUERY
	strcpy((char*)Query, VC_DEBUGGER_TESTQUERY);
#endif	

	// Can't ask for 1000+ results on a page, when we will only have 1000 results in total!
	if (Per_Page > MaxResultsPerQuery)
		Per_Page = MaxResultsPerQuery;
			 
	
			 
	//AND operator. 
	if(And == INPUT_RETRIEVE_ERROR)
	{
		if (DefaultToAnd == 1)
			And = 1;
		else
			And = 0;
	}	
			 
			 
	// init. link target string
	strcpy(target, "");
	if (UseLinkTarget == 1)
		sprintf(zoom_target, "target=\"%s\"",LinkTarget);
		
	if (UseXML == 1)
	{
		zoomprintf("<?xml version=\"1.0\" encoding=\"%s\"?>\n", Charset);
		if (strlen(XMLStyleSheetURL) > 0)
			zoomprintf("<?xml-stylesheet type=\"text/xsl\" href=\"%s\"?>\n", XMLStyleSheetURL);
		zoomprintf("<!--Zoom Search Engine %s-->\n", VersionInfo);
		zoomprintf("<rss version=\"2.0\" xmlns:opensearch=\"http://a9.com/-/spec/opensearch/1.1/\" xmlns:atom=\"http://www.w3.org/2005/Atom\" xmlns:zoom=\"http://www.wrensoft.com/zoom/response/5.0/schema/\">\n");
		zoomprintf("<channel>\n");
		zoomprintf("<title>%s</title>\n", XMLTitle);
		zoomprintf("<description>%s</description>\n", XMLDescription);
		zoomprintf("<link>%s</link>\n", XMLLink);
		if (strlen(XMLOpenSearchDescURL) > 0)
			zoomprintf("<opensearch:link rel=\"search\" href=\"%s\" type=\"application/opensearchdescription+xml\" />\n", XMLOpenSearchDescURL);

		if (ShowXMLSettings == 1)
			DisplayXMLSettings();	

		if (XMLHighlight == 1)
			Highlighting = 1;	//force highlighting on if XML highlighting enabled
	}
	else
	{
		ResultsBufferPrintf("<!--Zoom Search Engine %s-->\n", VersionInfo);
	}
		
	strcpy((char*)QueryForHTML, (char*)Query);
	htmlspecialchars((char*)QueryForHTML, MAX_QUERY_LENGTH+50);
		
	// This is moved to AFTER the readSettings() call because the ScriptName may have been overwritten by the Linkback URL setting
	if (FormFormat > 0 && UseXML == 0)
	{
#if defined(ASPX)
		if (DotNetUseFormTags == 1)
		{
			OutputBufferPrintf(OUTPUT_FORM_START, "<form method=\"get\" action=\"%s\" class=\"zoom_searchform\">\n",ScriptName);
			OutputBufferPrintf(OUTPUT_FORM_END, "</form>\n");
		}
#else
		OutputBufferPrintf(OUTPUT_FORM_START, "<form method=\"get\" action=\"%s\" class=\"zoom_searchform\">\n",ScriptName);
		OutputBufferPrintf(OUTPUT_FORM_END, "</form>\n");
#endif 
	}

	// Replace the key text <!--ZOOMSEARCH--> with the following
	if (FormFormat > 0 && UseXML == 0)
	{
		// Insert the form		
		OutputBufferPrintf(OUTPUT_FORM_SEARCHBOX, "<label for=\"zoom_searchbox\">%s</label><input type=\"text\" name=\"zoom_query\" size=\"20\" value=\"%s\" id=\"zoom_searchbox\" class=\"zoom_searchbox\" />\n", StrTable[STR_FORM_SEARCHFOR], QueryForHTML);		
		OutputBufferPrintf(OUTPUT_FORM_SEARCHBUTTON, "<input type=\"submit\" value=\"%s\" class=\"zoom_button\" />\n", StrTable[STR_FORM_SUBMIT_BUTTON]);
		
		if (FormFormat == 2)
		{
			//print the results per page part of form
			OutputBufferPrintf(OUTPUT_FORM_RESULTSPERPAGE, "<span class=\"zoom_results_per_page\"><label for=\"zoom_per_page\">%s</label>\n", StrTable[STR_FORM_RESULTS_PER_PAGE]);
			OutputBufferPrintf(OUTPUT_FORM_RESULTSPERPAGE, "<select name=\"zoom_per_page\" id=\"zoom_per_page\">\n");
			for(int a=0; PerPageOptions[a] != -1; a++)
			{
				OutputBufferPrintf(OUTPUT_FORM_RESULTSPERPAGE, "<option");
				if (PerPageOptions[a] == Per_Page)
					OutputBufferPrintf(OUTPUT_FORM_RESULTSPERPAGE, " selected=\"selected\"");
				OutputBufferPrintf(OUTPUT_FORM_RESULTSPERPAGE, ">%d</option>\n", PerPageOptions[a]);
			}
			OutputBufferPrintf(OUTPUT_FORM_RESULTSPERPAGE, "</select><br /><br /></span>\n");
			
			//print categroies part of form
			if (UseCats) 
			{
				OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "<span class=\"zoom_categories\"><label for=\"zoom_cat\">%s</label>", StrTable[STR_FORM_CATEGORY]);				
				if (SearchMultiCats)
				{
					OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "<ul>\n");
        			OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "<li><input type=\"checkbox\" name=\"zoom_cat[]\" id=\"zoom_cat-1\" value=\"-1\"");
        			if (Cat[0] == -1)
	        			OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, " checked=\"checked\"");
	        		OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "><label for=\"zoom_cat-1\">%s</label></input></li>\n", StrTable[STR_FORM_CATEGORY_ALL]);
					for (int i = 0; i < NumCats; i++)
					{
        				OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "<li><input type=\"checkbox\" name=\"zoom_cat[]\" id=\"zoom_cat%d\" value=\"%d\"", i, i);
        				if (Cat[0] != -1)
        				{
	        				for (int catit = 0; catit < num_zoom_cats; catit++)
	        				{								
	        					if (i == Cat[catit])
	        					{
	        						OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, " checked=\"checked\"");
	        						break;
	        					}
	        				}
	        			}
        				OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, " /><label for=\"zoom_cat%d\">%s</label></li>\n", i, CatNames[i]);        		
        			}
        			OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "</ul><br /><br /></span>\n");				
				}
				else
				{
					OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "<select name=\"zoom_cat[]\" id=\"zoom_cat\">");
					// 'all cats option
					OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "<option value=\"-1\">%s</option>", StrTable[STR_FORM_CATEGORY_ALL]);
					for(int a = 0; a < NumCats; a++) 
					{
						OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "<option value=\"%d\"", a);
						if (a == Cat[0])
							OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, " selected=\"selected\"");
						OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, ">%s</option>", CatNames[a]);
					}
					OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "</select>&nbsp;&nbsp;\n");
				}
				OutputBufferPrintf(OUTPUT_FORM_CATEGORIES, "</span>\n");
			}
			if (UseMetaFields)
			{
				OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<span class=\"zoom_metaform\">\n");				
				char labelIDStr[METAFIELD_NAME_LEN+14+1] = {0};
				char tmpQueryValueForHTML[METAFIELD_TEXT_LEN+50+1] = {0};
				for (int fieldnum = 0; fieldnum < NumMetaFields; fieldnum++)
				{
					strcpy(tmpQueryValueForHTML, meta_query[fieldnum]);
					htmlspecialchars((char*)tmpQueryValueForHTML, METAFIELD_TEXT_LEN+50);	// prepare for HTML output
					sprintf(labelIDStr, "zoom_metaform_%s", metafields[fieldnum].name);

					if (metafields[fieldnum].type == METAFIELD_TYPE_NUMERIC)
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<label for=\"%s\">%s: </label><input type=\"text\" id=\"%s\" name=\"%s\" size=\"20\" value=\"%s\" class=\"zoom_metaform_numeric\" />\n", labelIDStr, metafields[fieldnum].formname, labelIDStr, metafields[fieldnum].name, tmpQueryValueForHTML);
					else if (metafields[fieldnum].type == METAFIELD_TYPE_DROPDOWN)
					{
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<label for=\"%s\">%s: </label><select name=\"%s\" id=\"%s\" class=\"zoom_metaform_dropdown\">\n", labelIDStr, metafields[fieldnum].formname, metafields[fieldnum].name, labelIDStr);
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<option value=\"-1\">%s</option>", StrTable[STR_FORM_CATEGORY_ALL]);
						int metaqueryVal = -1;
						if (strlen(meta_query[fieldnum]) > 0)
							metaqueryVal = atoi(meta_query[fieldnum]);
						for (int ddi = 0; ddi < metafields[fieldnum].DropdownCount; ddi++)
						{
							OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<option value=\"%d\"", ddi);
							if (ddi == metaqueryVal)
								OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, " selected=\"selected\"");
							OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, ">%s</option>", metafields[fieldnum].DropdownValues[ddi]);
						}
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "</select>\n");
					}
					else if (metafields[fieldnum].type == METAFIELD_TYPE_MULTI)
					{
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<label for=\"%s\">%s: </label><select multiple id=\"%s\" name=\"%s[]\" class=\"zoom_metaform_multi\">\n", labelIDStr, metafields[fieldnum].formname, labelIDStr, metafields[fieldnum].name);
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<option value=\"-1\">%s</option>", StrTable[STR_FORM_CATEGORY_ALL]);						
						int num_multi_query = (int)meta_query[fieldnum][0];
						for (int ddi = 0; ddi < metafields[fieldnum].DropdownCount; ddi++)
						{
							OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<option value=\"%d\"", ddi);
							for (int mqi = 1; mqi <= num_multi_query; mqi++)
							{
								if (ddi == (int)(meta_query[fieldnum][mqi]))
									OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, " selected=\"selected\"");
							}
							OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, ">%s</option>", metafields[fieldnum].DropdownValues[ddi]);
						}
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "</select>&nbsp;&nbsp;\n");
					}
					else if (metafields[fieldnum].type == METAFIELD_TYPE_MONEY)
					{
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<label for=\"%s\">%s: </label>%s<input type=\"text\" id=\"%s\" name=\"%s\" size=\"7\" value=\"%s\" class=\"zoom_metaform_money\" />\n", labelIDStr, metafields[fieldnum].formname, MetaMoneyCurrency, labelIDStr, metafields[fieldnum].name, tmpQueryValueForHTML);
					}
					else
						OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "<label for=\"%s\">%s: </label><input type=\"text\" id=\"%s\" name=\"%s\" size=\"20\" value=\"%s\" class=\"zoom_metaform_text\" />\n", labelIDStr, metafields[fieldnum].formname, labelIDStr, metafields[fieldnum].name, tmpQueryValueForHTML);
				}
				OutputBufferPrintf(OUTPUT_FORM_CUSTOMMETA, "</span>\n");
			}
			OutputBufferPrintf(OUTPUT_FORM_MATCH, "<span class=\"zoom_match\"><label for=\"zoom_match\">%s</label>\n", StrTable[STR_FORM_MATCH]);
			if (And == 0) 
			{
				OutputBufferPrintf(OUTPUT_FORM_MATCH, "<label for=\"zoom_match_any\"><input type=\"radio\" id=\"zoom_match_any\" name=\"zoom_and\" value=\"0\" checked=\"checked\" />%s</label>\n", StrTable[STR_FORM_ANY_SEARCH_WORDS]);
				OutputBufferPrintf(OUTPUT_FORM_MATCH, "<label for=\"zoom_match_all\"><input type=\"radio\" id=\"zoom_match_all\" name=\"zoom_and\" value=\"1\" />%s</label>\n", StrTable[STR_FORM_ALL_SEARCH_WORDS]);
			} 
			else 
			{
				OutputBufferPrintf(OUTPUT_FORM_MATCH, "<label for=\"zoom_match_any\"><input type=\"radio\" id=\"zoom_match_any\" name=\"zoom_and\" value=\"0\" />%s</label>\n", StrTable[STR_FORM_ANY_SEARCH_WORDS]);
				OutputBufferPrintf(OUTPUT_FORM_MATCH, "<label for=\"zoom_match_all\"><input type=\"radio\" id=\"zoom_match_all\" name=\"zoom_and\" value=\"1\" checked=\"checked\" />%s</label>\n", StrTable[STR_FORM_ALL_SEARCH_WORDS]);
			}			
			OutputBufferPrintf(OUTPUT_FORM_START, "<input type=\"hidden\" name=\"zoom_sort\" value=\"%d\" />\n", Sort);
			OutputBufferPrintf(OUTPUT_FORM_START, "<input type=\"hidden\" name=\"zoom_xml\" value=\"%d\" />\n", UseXML);

			OutputBufferPrintf(OUTPUT_FORM_MATCH, "<br /><br /></span>\n");

			if (DateRangeSearch == 1)
			{
				OutputBufferPrintf(OUTPUT_FORM_DATERANGE, "<span class=\"zoom_daterange\">\n");
				char dateformatStr[4];
				if (DateRangeFormat == 1)
					strcpy(dateformatStr, "mdy");
				else
					strcpy(dateformatStr, "dmy");				 
				OutputBufferPrintf(OUTPUT_FORM_DATERANGE, "<label for=\"zoom_datefrom\">%s </label><input id=\"zoom_datefrom\" name=\"zoom_datefrom\" value=\"%s\"><input type=\"button\" value=\"%s\" onclick=\"displayDatePicker('zoom_datefrom', false, '%s');\" />", StrTable[STR_FORM_DATE_FROM], DateFromStr, StrTable[STR_FORM_DATE_BUTTON], dateformatStr);
				OutputBufferPrintf(OUTPUT_FORM_DATERANGE, "<label for=\"zoom_dateto\">%s </label><input id=\"zoom_dateto\" name=\"zoom_dateto\" value=\"%s\"><input type=\"button\" value=\"%s\" onclick=\"displayDatePicker('zoom_dateto', false, '%s');\" />", StrTable[STR_FORM_DATE_TO], DateToStr, StrTable[STR_FORM_DATE_BUTTON], dateformatStr);				
				OutputBufferPrintf(OUTPUT_FORM_DATERANGE, "</span>\n");
				/*
				if (strlen(DatePickerPath) > 0)
					OutputBufferPrintf(OUTPUT_JAVASCRIPTS, "<script language=\"JavaScript\" src=\"%s/zoom_datepicker.js\"></script>\n", DatePickerPath);
				else
					OutputBufferPrintf(OUTPUT_JAVASCRIPTS, "<script language=\"JavaScript\" src=\"zoom_datepicker.js\"></script>\n");
				*/
			}

		}
		else
		{
			OutputBufferPrintf(OUTPUT_FORM_START, "<input type=\"hidden\" name=\"zoom_per_page\" value=\"%d\" />\n", Per_Page);
			OutputBufferPrintf(OUTPUT_FORM_START, "<input type=\"hidden\" name=\"zoom_and\" value=\"%d\" />\n", And);
			OutputBufferPrintf(OUTPUT_FORM_START, "<input type=\"hidden\" name=\"zoom_sort\" value=\"%d\" />\n", Sort);
			OutputBufferPrintf(OUTPUT_FORM_START, "<input type=\"hidden\" name=\"zoom_xml\" value=\"%d\" />\n", UseXML);
		}		
	}
				
	// Give up early if no search words provided	
	if (strlen((char*)Query) < 1 && zsSetup->IsZoomAC == 0)
	{
		bool NoSearch = false;

		if (UseMetaFields)
		{
			if (IsZoomQuery)
				IsNoKeywordQuery = true;
			else
				NoSearch = true;
		}
		else
		{
			// only display 'no query' line if zoom_query parameter was not specified at all
			if (IsZoomQuery == true)
			{
				OutputBufferPrintf(OUTPUT_SUMMARY, "<div class=\"summary\">%s</div>\n", StrTable[STR_NO_QUERY]);
			}

			NoSearch = true;
		}

		if (NoSearch)
		{
			//Let others know about Zoom.
			if (ZoomInfo == 1)
				OutputBufferPrintf(OUTPUT_PAGENUMBERS, "\n<center><p><small>%s <a href=\"http://www.wrensoft.com/zoom/\" target=\"_blank\"><b>Zoom Search Engine</b></a></small></p></center>\n", StrTable[STR_POWEREDBY]);

			ShowTemplate();
#if !defined(FASTCGI_ENABLED) && !defined(OSF)
			cleanup(); 	//clean up
#endif
			return 0;
		}
	}

	// Give up early if we are asking for search results beyond our cull limit
	if (((Page-1)*Per_Page) >= (unsigned int)MaxResultsPerQuery)
	{
		if (UseXML)
			zoomprintf("<zoom:error>%s</zoom:error>", StrTable[STR_MAX_RESULTS]);
		else
			ResultsBufferPrintf("%s<br />", StrTable[STR_MAX_RESULTS]);
		ShowTemplate();
		cleanup();
		return 0;
	}
	
	// Initialize UTF8 data - we are only UTF8 aware from here on ------------------

	if (UseUTF8 == 1)
	{
		if (ucdata_load(".", UCDATA_UTF8)) 
		{
			if (UseXML) zoomprintf("<zoom:error>");
			zoomprintf("Error loading in UTF8 data.");
			if (UseXML) zoomprintf("</zoom:error>");
			ShowTemplate();
			cleanup();
			return 0;
		}			
	}

	PrintUserDebug("Settings loaded.<br>"\
				   "Loading string table took: %.6f<br>"\
				   "Loading meta names took: %.6f<br>"\
				   "Loading categories took: %.6f<br>",
				   UserDebugStats.TotalTime[DEBUGTIMER_LOADSTRTABLE]/1000,
				   UserDebugStats.TotalTime[DEBUGTIMER_LOADMETA]/1000,
				   UserDebugStats.TotalTime[DEBUGTIMER_LOADCATS]/1000);

	///////////////////////////////////////////////////////////////////////////
	// Prepare query for search

	if (MapAccents == 1) 
	{
		PrintUserDebug("Converting accent mapped characters...");
		ConvertAccentedWord((char*)Query, MAX_QUERY_LENGTH);
		PrintUserDebug("Finished converting accent mapped characters.");
	}
		  
	if (ToLowerSearchWords == 1)
	{
		if (UseUTF8 == 1)
		{
			//_mbslwr(Query);						
			utf8_strlwr(Query);
		}
		else
			_strlwr((char*)Query); 
	}	

	// prepare search query, strip quotes, trim whitespace
	if (AllowExactPhrase == 0)
	{
		strReplace("\"", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	}
	
	if (strchr(WordJoinChars, '.') == NULL)
		strReplace(".", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	
	if (strchr( WordJoinChars, '-') == NULL)	
	{
		//strReplace("-", " ", (char*)Query, true, MAX_QUERY_LENGTH);	

		// Remove hyphens only if it is NOT the first character of a word
		// (for the sake of negative searches)
		int tmplen = strlen((char*)Query);
		bool IsStartOfWord;		
		for (int i = 0; i < tmplen; i++)
		{
			if (i == 0 || (i > 0 && Query[i-1] == ' '))
				IsStartOfWord = true;
			else
				IsStartOfWord = false;

			if (!IsStartOfWord && Query[i] == '-')
				Query[i] = ' ';
		}				
	}

	/*
	if (strchr(WordJoinChars, '#') == NULL)	
		strReplace("#", " ", (char*)Query, true, MAX_QUERY_LENGTH);	
	*/

	// Remove hash/sharp sign only if it is NOT the last character of a word
	// (for "C#" etc.)
	if (strchr(WordJoinChars, '#') == NULL)	
	{
		int tmplen = strlen((char*)Query);
		bool IsEndOfWord;		
		for (int i = tmplen-1; i >= 0; i--)
		{
			if (i == tmplen-1 || (i < tmplen-1 && Query[i+1] == ' '))
				IsEndOfWord = true;
			else
				IsEndOfWord = false;

			if (!IsEndOfWord && Query[i] == '#')
				Query[i] = ' ';
		}				
	}

	// Replace + sign only if it is not the last character of a word
	// and repeated (eg. "C++")
	if (strchr(WordJoinChars, '+') == NULL)	
	{
		int tmplen = strlen((char*)Query);
		bool IsEndOfWord;		
		for (int i = tmplen-1; i >= 0; i--)
		{
			if (i == tmplen-1 || (i < tmplen-1 && Query[i+1] == ' '))
				IsEndOfWord = true;
			else
				IsEndOfWord = false;

			if (Query[i] == '+')
			{
				// if end of word and char before it is NOT '+', we remove
				// and if NOT end of word, but char after it is NOT '+', we remove
				if (IsEndOfWord && i > 0 && Query[i-1] != '+')
					Query[i] = ' ';
				else if (!IsEndOfWord && i < tmplen-1 && Query[i+1] != '+')
					Query[i] = ' ';
			}			
		}				
	}

	if (strchr( WordJoinChars, '_') == NULL)	
		strReplace("_", " ", (char*)Query, true, MAX_QUERY_LENGTH);	
	
	if (strchr(WordJoinChars, '\'') == NULL)
		strReplace("'", " ", (char*)Query, true, MAX_QUERY_LENGTH);	
	
	if (strchr(WordJoinChars, '$') == NULL)	
		strReplace("$", " ", (char*)Query, true, MAX_QUERY_LENGTH);	
	
	if (strchr(WordJoinChars, ',') == NULL)	
		strReplace(",", " ", (char*)Query, true, MAX_QUERY_LENGTH);		  

	if (strchr(WordJoinChars, ':') == NULL)	
	{
		// We temporarily replace the search syntax for "site:" and "url:"
		// to avoid stripping it out when colon is disabled for word join
		strReplace("site:", "[#@!]", (char*)Query, false, MAX_QUERY_LENGTH);
		strReplace("url:", "[#@]", (char*)Query, false, MAX_QUERY_LENGTH);
		strReplace(":", " ", (char*)Query, true, MAX_QUERY_LENGTH);

		// We then restore it
		strReplace("[#@!]", "site:", (char*)Query, false, MAX_QUERY_LENGTH);
		strReplace("[#@]", "url:", (char*)Query, false, MAX_QUERY_LENGTH);
	}

	if (strchr(WordJoinChars, '&') == NULL)	
		strReplace("&", " ", (char*)Query, true, MAX_QUERY_LENGTH);		  

	if (strchr(WordJoinChars, '/') == NULL)	
		strReplace("/", " ", (char*)Query, true, MAX_QUERY_LENGTH);		  

	if (strchr(WordJoinChars, '\\') == NULL)	
		strReplace("\\", " ", (char*)Query, true, MAX_QUERY_LENGTH);		  

	if (strchr(WordJoinChars, '@') == NULL)	
		strReplace("@", " ", (char*)Query, true, MAX_QUERY_LENGTH);		  
	
	// strip slashes, sloshes, paretheses, and other regexp elements
	//strReplace("+", " ", (char*)Query, true, MAX_QUERY_LENGTH);		  
	strReplace("(", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	strReplace(")", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	strReplace("^", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	strReplace("[", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	strReplace("]", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	strReplace("|", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	strReplace("{", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	strReplace("}", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	strReplace("%", " ", (char*)Query, true, MAX_QUERY_LENGTH);
	//strReplace("", " ", (char*)Query, true, MAX_QUERY_LENGTH);	// This would break some Japanese and Chinese characters. MUST NOT STRIP.
	strReplace("!", " ", (char*)Query, true, MAX_QUERY_LENGTH);
		
	// strip any of the wordjoinchars if followed immediately by a space
	strReplace("_ ", " ", (char*)Query, false, MAX_QUERY_LENGTH);	
	strReplace("' ", " ", (char*)Query, false, MAX_QUERY_LENGTH);
	//strReplace("# ", " ", (char*)Query, false, MAX_QUERY_LENGTH);	
	strReplace("$ ", " ", (char*)Query, false, MAX_QUERY_LENGTH);		
	strReplace(". ", " ", (char*)Query, false, MAX_QUERY_LENGTH);
	strReplace(", ", " ", (char*)Query, false, MAX_QUERY_LENGTH);
	strReplace(": ", " ", (char*)Query, false, MAX_QUERY_LENGTH);		  
	strReplace("& ", " ", (char*)Query, false, MAX_QUERY_LENGTH);	
	strReplace("/ ", " ", (char*)Query, false, MAX_QUERY_LENGTH);	
	strReplace("\\ ", " ", (char*)Query, false, MAX_QUERY_LENGTH);	
	
	int queryLen = strlen((char*)Query);
	if (Query[queryLen-1] == '_' ||
		Query[queryLen-1] == '\'' ||
		//Query[queryLen-1] == '#' ||
		Query[queryLen-1] == '$' ||
		Query[queryLen-1] == '.' ||
		Query[queryLen-1] == ',' ||
		Query[queryLen-1] == ':' ||
		Query[queryLen-1] == '&' ||
		Query[queryLen-1] == '/' ||
		Query[queryLen-1] == '\\')
		Query[queryLen-1] = '\0';

	Trim((char*)Query);	
	
	// encode the query entities remaining
	strcpy((char*)QueryForSearch, (char*)Query);
	//QueryEntities((char*)QueryForSearch, MAX_QUERY_LENGTH+50);
	queryLen = strlen((char*)QueryForSearch);

	// Re-prepare a copy of the query for printing out for HTML (since it may have been modified by the stripping)
	strcpy((char*)QueryForHTML, (char*)Query);
	htmlspecialchars((char*)QueryForHTML, MAX_QUERY_LENGTH+50);

	// Split search phrase into words	
	
	int max_number_terms = 1;	// include first one (no space before it)
	// count number of spaces to determine possible number of search terms.
	for (int i = 0; i < queryLen; i++)
	{
		if (QueryForSearch[i] == ' ')
			max_number_terms++;		
	}
	
	IsPhraseStarted = false;	
	IsPhraseFound = false;

	// we allocate the number of pointers for these terms (not necessarily use them all)
	SearchWords = (char**) calloc(max_number_terms, sizeof(char*));	
	int j = 0;

	// allocate memory for first word in advance
	SearchWords[NumSearchWords] = (char*) malloc((MAX_QUERY_LENGTH+1) * sizeof(char));
	
	//unsigned long utf8ch = 0;
	for (int i = 0; i < queryLen; i++)
	{	
		/*
		if (UseUTF8 && i+2 < queryLen)
		{
			utf8ch = uctoutf32(QueryForSearch+i, 3);
			if (utf8ch > 0)
				i += 2;
		}
		printf("%d\n", utf8ch);
		*/
		//if (QueryForSearch[i] == '\"' || (UseUTF8 && (utf8ch == 8222 || utf8ch == 12319)))
		if (QueryForSearch[i] == '\"')
		{
			if (IsPhraseStarted)
			{
				// End of phrase
				IsPhraseStarted = false;
			}
			else
			{
				// New phrase
				if (i > 0 && QueryForSearch[i-1] == '-')
				{
					// if the previous character was a negative sign, then we don't start a new word
					IsPhraseStarted = true;
				}
				else
				{
					IsPhraseStarted = true;				
					j = 0;
				}
			}
		}
		else if (!IsPhraseStarted && QueryForSearch[i] == ' ')
		{			
			if (j > 0)	// leave out extraneous spaces
			{
				SearchWords[NumSearchWords][j++] = '\0';
				NumSearchWords++;
				SearchWords[NumSearchWords] = (char*) malloc((MAX_QUERY_LENGTH+1) * sizeof(char));
				j = 0;
			}
		}
		else 
		{
			SearchWords[NumSearchWords][j++] = QueryForSearch[i];
		}

		if (IsPhraseStarted)
			IsPhraseFound = true;
	}

	if (queryLen > 0)	// added 5/feb/09 to avoid a NumSearchWords==1 when blank query
	{
		SearchWords[NumSearchWords][j++] = '\0';
		NumSearchWords++;
	}
	
	// Prepare an array for the regexp friendly version of search terms for highlighting
	if (NumSearchWords > 0)
		RegExpSearchWords = (char**) malloc(NumSearchWords * sizeof(char*));
	for (unsigned long a = 0; a < NumSearchWords; a++)
	{
		RegExpSearchWords[a] = (char*) malloc((MAX_QUERY_LENGTH+1) * sizeof(char));
	}


	if (strchr((char*)QueryForSearch, '-') != NULL)
	{
		//move the negative searches to the end of the array
		qsort((void*)SearchWords, NumSearchWords, sizeof(char*), sw_compare);
		
	}

	return 1;
}

int zoomSearch::OpenIndexFiles(const zoomSearchSetup *zsSetup) //added  char *envp[] to print environment, MAY BREAK UNDER LINUX
{
	// Load index data files (*.zdat) ---------------------------------------------

	PrintUserDebug("Opening pagedata file...");

	sprintf(tmpPath, "%s/%s", ExecutePath, ZoomPagedataFilename);

#if defined(FASTCGI_ENABLED) || defined(OSF)
	IsPageDataPreloaded = true;
	HANDLE hFile = CreateFileA(tmpPath, GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		zoomprintf("Unable to open pagedata file<br />");
		cleanup(); 
		return 0;
	}
	PageDataSize = GetFileSize(hFile, NULL);
	CloseHandle(hFile);
#else
	IsPageDataPreloaded = false;
#endif
		
	if (IsPageDataPreloaded)
		fp_pagedata = myfopen(tmpPath, "rb"); 	
	else
		fp_pagedata = myfopen(tmpPath, "r"); 	

	if (fp_pagedata == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
#if !defined(NON_WINDOWS) && !defined(ASPX)
		char* msg = GetLastErrorString();
		zoomprintf("Unable to open %s<br />%s<br />Check file permissions and that file exists", ZoomPagedataFilename, msg);
		LocalFree(msg);
#else
		zoomprintf("Unable to open %s<br />Check file permissions and that file exists", ZoomPagedataFilename);
#endif
		if (UseXML) zoomprintf("</zoom:error>\n");
		ShowTemplate();
		cleanup(); 
		return 0;			
	}

#if defined(FASTCGI_ENABLED) || defined(OSF)
	// We were going to make this available in the normal CGI as well. But we can't use 
	// win32 call to determine filesizes and we really need the Indexer to write it out
	// in settings.zdat. So for now, we'll make it FCGI/Win32 only.
	if (IsPageDataPreloaded)
	{
		// For FastCGI, we will pre-load the pagedata file

		// First we determine the size of the file by jumping to the end and using ftell
		/*
		fseek(fp_pagedata, SEEK_END, 0);
		PageDataSize = ftell(fp_pagedata);
		fseek(fp_pagedata, SEEK_SET, 0);	// return to beginning of file
		*/		

		pagedatabin = (BYTE*) malloc(PageDataSize);
		if (pagedatabin == NULL)
		{
			zoomprintf("Unable to allocate memory for pagedata file for FastCGI mode.<br />");
			fclose(fp_pagedata);
			cleanup(); 
			return 0;
		}

		unsigned int bytesread = fread(pagedatabin, sizeof(BYTE), PageDataSize, fp_pagedata);
		if (bytesread != PageDataSize)
		{
			zoomprintf("Pagedata file does not match PageDataSize specified.");
			fclose(fp_pagedata);
			cleanup(); 
			return 0;
		}		
		fclose(fp_pagedata);
		fp_pagedata = NULL;
	}
#endif
		
	if (Recommended == 1)
	{
		PrintUserDebug("Loading recommended links...");
		readRecommended();
		PrintUserDebug("Recommended links loaded.");
	}
	
	PrintUserDebug("Loading pageinfo file...");

	// Open pageinfo file
	int res = readPageinfo();
	if (res == FAILURE)
	{
		if (UseXML) zoomprintf("<zoom:error>");
#if !defined(NON_WINDOWS) && !defined(ASPX)
		char* msg = GetLastErrorString();
		zoomprintf("Unable to open %s<br />%s<br />Check file permissions and that file exists", ZoomPageinfoFilename, msg);
		LocalFree(msg);
#else
		zoomprintf("Unable to open %s<br />Check file permissions and that file exists", ZoomPageinfoFilename);
#endif	
		if (UseXML) zoomprintf("</zoom:error>\n");
		ShowTemplate();
		cleanup(); 
		return 0;			
	}
	
	//PrintDebugLine("pageinfo loaded. before opening pagetext file\n");
	PrintUserDebug("Pageinfo loaded.");
	
	// Open pagetext file
	if (DisplayContext == 1 || AllowExactPhrase == 1)
	{				
		PrintUserDebug("Opening pagetext file...");

		sprintf(tmpPath, "%s/%s", ExecutePath, ZoomPageTextFilename);
		fp_pagetext = myfopen(tmpPath, "rb");
		//fp_pagetext = _open(tmpPath, _O_RDONLY|_O_BINARY);		
		
		//if(fp_pagetext == -1)
		if (fp_pagetext == NULL)
		{
			if (UseXML) zoomprintf("<zoom:error>");
			#if !defined(NON_WINDOWS) && !defined(ASPX)
				char* msg = GetLastErrorString();
				zoomprintf("Unable to open zoom_pagetext.zdat<br />%s<br />Check file permissions and that file exists", msg);
				LocalFree(msg);
			#else
				zoomprintf("Unable to open zoom_pagetext.zdat<br />Check file permissions and that file exists");
			#endif
			if (UseXML) zoomprintf("</zoom:error>\n");
			ShowTemplate();
			cleanup(); 	
			return 0;			
		}
		
		char teststr[8] = {0};

		fgets(teststr, 7, fp_pagetext);
		//_read(fp_pagetext, &teststr, 7);			

		if (teststr[0] == 'T' && teststr[2] == 'h' && teststr[4] == 'i' && teststr[6] == 's')
		{
			if (UseXML) zoomprintf("<zoom:error>");
			zoomprintf("<b>Zoom config error:</b> The \"zoom_pagetext.zdat\" file is not properly created for the search settings specified in \"setting.php\"<br />");
			zoomprintf("Please check that you have re-indexed and try again.");
			if (UseXML) zoomprintf("</zoom:error>\n");
			fclose(fp_pagetext);
			ShowTemplate();					
			cleanup();
			return 0;
		}
	}
		
	PrintUserDebug("Opening wordmap file...");
	
	//Open keywords file
	//tmpPath[1024];
	sprintf(tmpPath, "%s/zoom_wordmap.zdat", ExecutePath);
	fp_wordmap = myfopen(tmpPath, "rb");
	//fp_wordmap = _open(tmpPath, _O_RDONLY|_O_BINARY);
	
	//if (fp_wordmap == -1)
	if (fp_wordmap == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Can not find one or more of the Zoom index files.<br />Please make sure the generated index files are uploaded to the same path as this search script.<br />Also check file permissions have been set to allow reading.<br />\n");
		if (UseXML) zoomprintf("</zoom:error>\n");
		ShowTemplate();
		cleanup();
		return 0;
	}
	
	if (Spelling == 1)
	{
		PrintUserDebug("Loading spelling suggestions...");

		// load in spellings file
		sprintf(tmpPath, "%s/zoom_spelling.zdat", ExecutePath);
		fp_spell = fopen(tmpPath, "rt");		
		if(fp_spell == NULL)
		{
			if (UseXML) zoomprintf("<zoom:error>");
			zoomprintf("Can not find one or more of the Zoom index files. Please make sure the generated index files are uploaded to the same path as this search script. Also check file permissions have been set to allow reading.");
			if (UseXML) zoomprintf("</zoom:error>\n");
			ShowTemplate();
			cleanup();
			return 0;
		}					
		unsigned int i = 0;
		char* tempLine = (char*) malloc(MaxKeyWordLineLen * sizeof(char));		
		if(tempLine == NULL)
		{
			if (UseXML) zoomprintf("<zoom:error>");
			zoomprintf("Error: Out of memory");
			if (UseXML) zoomprintf("</zoom:error>\n");
			ShowTemplate();
			cleanup();
			return 0;
		}		
				
		Spell = (spellstruct*) malloc((NumSpellings + 1) * sizeof(spellstruct));
			
		i = 0;	
			
		while (!feof(fp_spell))
		{			
			if(fgets(tempLine, MaxKeyWordLineLen, fp_spell) == NULL)
				break;

			char* tok = strtok(tempLine, " \n\r");
			int j = 0;
			bool gotSPCode = false;
			bool gotSPid = false;
			while (tok != NULL && j < 4)
			{
				switch (j)
				{
				case 0:
					Spell[i].spcode = (char*) calloc(SPCODE_LEN+1, sizeof(char));					
					strncpy(Spell[i].spcode, tok, SPCODE_LEN);									
					gotSPCode = true;					
					break;
				case 1:
					Spell[i].id = ASCII_to_Integer(tok);
					Spell[i].id2 = 0;
					Spell[i].id3 = 0;
					gotSPid = true;
					break;
				case 2:
					Spell[i].id2 = ASCII_to_Integer(tok);	
					break;
				case 3:
					Spell[i].id3 = ASCII_to_Integer(tok);	
					break;
				}
				j++;
				tok = strtok(NULL, " \n\r");
			}	

			if (gotSPCode && gotSPid)
			{				
				//increment spelling array element
				i++;
			}
		}
		fclose(fp_spell);
		free(tempLine);
		PrintUserDebug("Spelling suggestions loaded.");		
	}

	return 1;
}


int zoomSearch::LoadDictionary(const zoomSearchSetup *zsSetup, BOOL bLoadNumResults) //added  char *envp[] to print environment, MAY BREAK UNDER LINUX
{
	PrintUserDebug("Loading dictionary file...");
	
#ifdef OSF
	// This will populate the Dict.NumResults field. Note that this is only available if building for OSF
	LoadNumResultsFromDict = bLoadNumResults;
#endif

	if (readDictionary() == FAILURE)
	{
		if (UseXML) printf("<zoom:error>");
		printf("Failed to load dictionary file.<br />\n");
		if (UseXML) printf("</zoom:error>\n");
		ShowTemplate();
		cleanup();
		return 0;
	}

	PrintUserDebug("Dictionary file loaded.");
	return 1;
}

int zoomSearch::getDictionary(dictstruct **dict)
{
	*dict = (dictstruct*) calloc((NumKeywords + 1), sizeof(dictstruct));
	memcpy(*dict, Dict, (NumKeywords + 1) * sizeof(dictstruct));
	return 1;
}

unsigned long zoomSearch::getNumKeywords()
{
	return NumKeywords;
}

unsigned long zoomSearch::getDictReservedLimit()
{
	return DictReservedLimit;
}

int zoomSearch::searchAC(const zoomSearchSetup *zsSetup) //added  char *envp[] to print environment, MAY BREAK UNDER LINUX
{	
	char tmpPath[1024];
	sprintf(tmpPath, "%s/%s", ExecutePath, ZoomAutocompleteFilename);
	FILE* fp_ac = myfopen(tmpPath, "r"); 
	char buffer[TEMP_STR_SIZE]="";

	if (fp_ac == NULL)
	{
		zoomprintf("Can not find the zoom_autocomplete.zdat index file.<br />Please make sure the generated index files are uploaded to the same path as this search script.<br />Also check file permissions have been set to allow reading.<br />\n");
		return 0;
	}
		
	unsigned int i = 0;
	int sep = 0;
	char* sepPtr = NULL;
	char* tempLine = (char*) malloc(MaxKeyWordLineLen * sizeof(char));
	
	int lineLen = 0;
	
	if (tempLine == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Error: out of memory");
		if (UseXML) zoomprintf("</zoom:error>");
		ShowTemplate();
		cleanup();
		return 0;
	}			  
			  
	char* tok = 0;
	while (!feof(fp_ac))
	{
		if (fgets(tempLine, MaxKeyWordLineLen, fp_ac) == NULL)
			break;
		
		lineLen = strlen(tempLine);
		tempLine[lineLen-1] = '\0';
		if (tempLine[0] != '\0')
		{
			tok = strtok(tempLine, "\n\r");
			for (unsigned long sw = 0; tok != NULL && sw < NumSearchWords; sw++)
			{
				if (_strnicmp(tok, SearchWords[sw], strlen(SearchWords[sw])) == 0)
				{
					//htmlspecialchars((char*)tok, MaxKeyWordLineLen);
					//strReplace("'", " ", (char*)tok, true, MaxKeyWordLineLen);
					zoomprintf("<div class=\"zoom_ac_item\" onclick=\"ZoomAutoComplete_UseSuggestion(&quot;%s&quot;);\"><a class=\"zoom_ac_item\" href=\"#\" target=\"_self\">%s</a></div>\n", tok, tok);
				}
			}
		}
	}	

	fclose(fp_ac);
	free(tempLine);	
	return 1;
}


int zoomSearch::search(const zoomSearchSetup *zsSetup) //added  char *envp[] to print environment, MAY BREAK UNDER LINUX
{	
	unsigned int i = 0;
			
	GetNextDictWordBuff = (char*)malloc(sizeof(char)*128*DictIDLen);
	GetNextDictWordBuffRead = 0;
	GetNextDictWordBuffCount = 0;

		  
	// Prepare query for search ---------------------------------------------------

	if (UseStemming == 1)
	{
		if (StemmingAlgo < 0 || StemmingAlgo >= NUM_STEMMING_ALGORITHMS)
		{
			if (UseXML) printf("<zoom:error>");
			printf("Invalid stemming algorithm specified in settings file.<br />\n");
			if (UseXML) printf("</zoom:error>\n");
			ShowTemplate();
			cleanup();
			return 0;
		}

		if (UseUTF8)
		{
			stemmer = sb_stemmer_new(algorithm_names[StemmingAlgo], "UTF_8");
		}
		else
		{
			if (StemmingAlgo == 11)	// only romanians use ISO_8859_2 according to modules.h
				stemmer = sb_stemmer_new(algorithm_names[StemmingAlgo], "ISO_8859_2");
			else if (StemmingAlgo == 12)	// only russians need KOI8_R according to modules.h
				stemmer = sb_stemmer_new(algorithm_names[StemmingAlgo], "KOI8_R");
			else
				stemmer = sb_stemmer_new(algorithm_names[StemmingAlgo], "ISO_8859_1");
		}

		if (stemmer == NULL)
		{
			if (UseXML) zoomprintf("<zoom:error>");
			zoomprintf("Failed to initialize stemmer.<br />\n");
			if (UseXML) zoomprintf("</zoom:error>\n");
			ShowTemplate();
			cleanup();
			return 0;	
		}
	}
		  
		
	char query_zoom_cats[MAX_QUERY_LENGTH+1];
	strcpy(query_zoom_cats, "");
	
	// Print the heading of the search results
	if (UseXML) 
	{		
		zoomprintf("<zoom:searchquery>%s</zoom:searchquery>\n", QueryForHTML);
		if (UseCats)
		{
			if (Cat[0] == -1)
				zoomprintf("<zoom:searchcategory>All</zoom:searchcategory>\n");
			else
			{
				char tmpstr[100];
				for (int catit = 0; catit < num_zoom_cats; catit++)
				{
					zoomprintf("<zoom:searchcategory>%s</zoom:searchcategory>\n", CatNames[Cat[catit]]);					
					sprintf((char*)&tmpstr, "&amp;zoom_cat%%5B%%5D=%d", Cat[catit]);
					strcat(query_zoom_cats, tmpstr);
				}
			}
		}
	}
	else 
	{
		OutputBufferPrintf(OUTPUT_HEADING, "<div class=\"searchheading\">");
		OutputBufferPrintf(OUTPUT_HEADING, "%s %s", StrTable[STR_RESULTS_FOR], QueryForHTML);
		//OutputBufferPrintf(OUTPUT_HEADING, "[%d %d(%c) %d %d]", QueryForHTML[0], QueryForHTML[1], QueryForHTML[1], QueryForHTML[2], QueryForHTML[strlen((char*)QueryForHTML)-1]);
		if (UseCats) 
		{
			if (Cat[0] == -1)
				OutputBufferPrintf(OUTPUT_HEADING, " %s", StrTable[STR_RESULTS_IN_ALL_CATEGORIES]);
			else
			{
				char tmpstr[100];				
				OutputBufferPrintf(OUTPUT_HEADING, " %s ", StrTable[STR_RESULTS_IN_CATEGORY]);
				for (int catit = 0; catit < num_zoom_cats; catit++)
				{
					if (Cat[catit] < NumCats)
					{
						if (catit > 0)
							OutputBufferPrintf(OUTPUT_HEADING, ", ");
						OutputBufferPrintf(OUTPUT_HEADING, "\"%s\"", CatNames[Cat[catit]]);
						sprintf((char*)&tmpstr, "&amp;zoom_cat%%5B%%5D=%d", Cat[catit]);					
						strcat(query_zoom_cats, tmpstr);
					}
				}				
			}
		}	
		OutputBufferPrintf(OUTPUT_HEADING, "<br /><br /></div>\n");
	}

	if (UseXML == 0) 
		ResultsBufferPrintf("<div class=\"results\">\n");
	
	PrintUserDebug("Begin main search loop...");
			 

	// Begin main search loop -----------------------------------------------------

	//NumWords = NumSearchWords;
	unsigned long PagesCount = NumPages;
	int OutputLine = 0;	
	int ExcludeTerm = 0;
	
	UserDebugSaveTime(DEBUGTIMER_MEMALLOC);
	res_table_item* res_table = (res_table_item*) calloc(PagesCount, sizeof(res_table_item));
	if (res_table == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Out of memory");
		if (UseXML) zoomprintf("</zoom:error>");
		ShowTemplate();
		cleanup();
		return 0;
	}
	UserDebugAddTime(DEBUGTIMER_MEMALLOC);
		
	dictstruct dictline; 
	int result=0;
	unsigned long data_count=0;		
	word_data* data = NULL;

	int num_phrase_terms=0;
	word_data** phrase_terms_data = NULL;
	int *phrase_data_count = NULL;
	phrase_terms_ids = NULL;

	//char **phrase_terms = NULL;

	SearchPhrases = NULL;	
	SearchPhrasesForHL = NULL;

	int ContextSeeks=0;
	unsigned long matches=0;	
	unsigned long fullmatches = 0;		

	unsigned long wordsmatched = 0;
	unsigned long totalwordsmatched = 0;

	unsigned long bytesread = 0;	
	unsigned char bytes_buffer[MAXDICTIDLEN]={0};
	unsigned long dict_id=0;

	/*
	char pattern[TEMP_STR_SIZE];
	pcre *re; //compiled regular expression
	pcre_extra* extra = NULL;
	*/
	
	
	unsigned long maxscore = 0;
	unsigned long maxtermsmatched = 0;
			 
	int context_maxgoback = 1;	
	int SkippedExactPhrase = 0;

	unsigned long tmp_histogram_value = 0;
			
	UserDebugSaveTime(DEBUGTIMER_MEMALLOC);

	int* sw_results = (int*) calloc(NumSearchWords, sizeof(int));

	search_terms_ids = (int*) calloc(NumSearchWords, sizeof(int));

	if (AllowExactPhrase == 1)
	{
		phrase_terms_ids = (int**) malloc(sizeof(int*) * NumSearchWords);
		memset(phrase_terms_ids, 0, sizeof(int*) * NumSearchWords);
	}

	UseWildCards = (BYTE*) calloc(NumSearchWords, sizeof(BYTE));

	UserDebugAddTime(DEBUGTIMER_MEMALLOC);

	UserDebugSaveTime(DEBUGTIMER_PREPQUERY);
	//char* tmpPattern = NULL;
	char tmpWordBuffer[WORDLENGTH+1];
	char searchUrlBuffer[URLLENGTH+1] = {0};	// this is for "site:blah" searches
	char* colonPtr;

	for (unsigned long sw = 0; sw < NumSearchWords; sw++)
	{	
		// Prepare search terms to be used in regexp for highlighting in descriptions
		if (strchr(SearchWords[sw], '*') != NULL || strchr(SearchWords[sw], '?') != NULL)
		{
			UseWildCards[sw] = 1;
		}

		if (SearchAsSubstring == 1)
		{
			UseWildCards[sw] = 1;
			// We need to re-prepare the search term, and add "*" before and after 
			// the word (if necessary). This was needed after dropping the use of 
			// WildCatMatch() for the more accurate WildCardCompare() which does not
			// support substring matching.
			tmpWordBuffer[0] = '\0';
			if (SearchWords[sw][0] != '*')
				strcat(tmpWordBuffer, "*");
			strcat(tmpWordBuffer, SearchWords[sw]);
			if (SearchWords[sw][strlen(SearchWords[sw])-1] != '*')
				strcat(tmpWordBuffer, "*");
			strcpy(SearchWords[sw], tmpWordBuffer);
		}


		if (Highlighting == 1)
		{
			// back up of the search words before we stem them!

			//tmpPattern = pattern2regexp(SearchWords[sw]);	
			strcpy(RegExpSearchWords[sw], SearchWords[sw]);					
			//free(tmpPattern);									
		}
	}
	UserDebugAddTime(DEBUGTIMER_PREPQUERY);

	// make a copy of the encoded query, for use as URL parameter
	strcpy((char*)QueryForURL, (char*)Query);
	urlencode((char*)QueryForURL, MAX_QUERY_LENGTH+50);
	
	// Find recommended links if any (before stemming, but after substring)
	UserDebugSaveTime(DEBUGTIMER_FINDRECLINKS);
	char* rec_multiwords = NULL;
	num_recs_found = 0;
	if (Recommended == 1)
	{		
		char* rec_word;
		__int64 rec_idx;
		for (int rl = 0; rl < rec_count && num_recs_found < RecommendedMax; rl++)
		{
			rec_word = Rec[rl].word;
			rec_idx = Rec[rl].index;
			if (strchr(rec_word, ',') != NULL)
			{
				rec_multiwords = NULL;
				rec_multiwords = strtok(rec_word, ",");
				while (rec_multiwords != NULL)
				{
					if (RecLinkWordMatch(rec_multiwords, rec_idx))
						break;
					rec_multiwords = strtok(NULL, ",");
				}
			}
			else
				RecLinkWordMatch(rec_word, rec_idx);
		}
		if (num_recs_found > 0)
		{
			if (UseXML == 0)
				OutputBufferPrintf(OUTPUT_RECOMMENDED, "</div>\n");
		}
	}
	UserDebugAddTime(DEBUGTIMER_FINDRECLINKS);

	stemmed = NULL;	

	// Update MaxMatches to reflex the number of search terms.
	// We want to allow each search term to have a good number of results (and not let early
	// search terms max out the limit eg. "*e* dog"

	unsigned long MaxMatchesPerWord = MaxMatches;
	if (NumSearchWords > 0)
		MaxMatchesPerWord = (long)ceil( (double)MaxMatches / (double)NumSearchWords );	

	for (unsigned long sw = 0; IsNoKeywordQuery == false && sw < NumSearchWords; sw++) 
	{
		if (SearchWords[sw][0] =='\0')
			continue;

		// PrintUserDebug("searchword is: %s<br>", SearchWords[sw]);
		
		// check min length
		if (strlen(SearchWords[sw]) < MinWordLen) 
		{
			SkipSearchWord(sw);
			continue;
		}
		
		ExactPhrase = 0;
		ExcludeTerm = 0;
		wordsmatched = 0;
		
		if (SearchWords[sw][0] == '-')
		{
			strcpy(SearchWords[sw], SearchWords[sw]+1);
			ExcludeTerm = 1;
			exclude_count++;
		}

		// check if this is a search option (e.g. "site:mysite.com")
		if ((colonPtr = strchr(SearchWords[sw], ':')) != NULL)
		{
			if (strncmp(SearchWords[sw], "site:", 5) == 0 || strncmp(SearchWords[sw], "url:", 4) == 0)
			{
				strcpy(searchUrlBuffer, colonPtr+1);
				exclude_count++;
				// if this site: term is the only search term in the query, then we enable 
				// empty query behaviour (so we return all results)
				if (NumSearchWords == 1)
				{
					IsNoKeywordQuery = true;
					// we need to break out of the loop here and not continue searching through
					// the dictionary (and waste time) when there's no word to look up!
					// we'll continue to the top of the loop (where the while condition will break
					// since IsNoKeywordQuery is flagged)
					continue;	
				}
			}
		}

		if (UseStemming == 1)
		{
			// Stem the words if necessary (NOTE: only do it AFTER stripping out the exclusion char)
			// don't do this for phrases, we will have to do it later when we split the phrase terms
			if (!IsPhraseFound || strchr(SearchWords[sw], ' ') == NULL)
			{				
				GetStem(SearchWords[sw], strlen(SearchWords[sw]));
			}
		}

		if (AllowExactPhrase == 1 && strstr(SearchWords[sw], " ") != NULL)
		{
			UserDebugSaveTime(DEBUGTIMER_PREPQUERY);
			if (SearchPhrases == NULL)
			{
				// This is an array indexed the same as the SearchWords array 
				// (an entry per search term in the query) but containing either
				// NULL (for no phrase) or a pointer to another array which contains
				// the strings of that phrase, split up by spaces.
				SearchPhrases = (char***) malloc(NumSearchWords * sizeof(char*));
				memset(SearchPhrases, 0, NumSearchWords * sizeof(char*));

				if (Highlighting == 1)
				{
					SearchPhrasesForHL = (char***) malloc(NumSearchWords * sizeof(char*));
					memset(SearchPhrasesForHL, 0, NumSearchWords * sizeof(char*));
				}
			}

			// Initialise exact phrase matching for this search term
			ExactPhrase = 1;
						
			//all this just to replace the php function "split"
			char temp_phrase_terms[MAX_QUERY_LENGTH];
			strcpy(temp_phrase_terms, SearchWords[sw]);
			
			//at the moment, just split word based on spaces
			char* nextword = strtok(temp_phrase_terms, " ");
			
			//count tokens
			int count=0;
			while( nextword != NULL )
			{
				nextword = strtok( NULL, " ");
				count++;
			}
			
			num_phrase_terms = count;
			
			//for max context matching
			if (num_phrase_terms > context_maxgoback)
			{
				context_maxgoback =num_phrase_terms;				
			}									
			
			//save tokens in array
			SearchPhrases[sw] = (char**) malloc((num_phrase_terms+1) * sizeof(char*));
			if (SearchPhrases[sw] == NULL)
			{
				if (UseXML) zoomprintf("<zoom:error>");
				zoomprintf("Out of memory");
				if (UseXML) zoomprintf("</zoom:error>");
				if (sw_results != NULL) 
					free(sw_results);
				ShowTemplate();
				cleanup();
				return 0;
			}
			// initialize to zeros
			memset(SearchPhrases[sw], 0, (num_phrase_terms+1) * sizeof(char*));

			if (Highlighting == 1)
			{
				// backup buffer of search phrases for highlighting (without stemmed word)
				SearchPhrasesForHL[sw] = (char**) malloc((num_phrase_terms+1) * sizeof(char*));
				if (SearchPhrasesForHL[sw] == NULL)
				{
					if (UseXML) zoomprintf("<zoom:error>");
					zoomprintf("Out of memory");
					if (UseXML) zoomprintf("</zoom:error>");
					if (sw_results != NULL) 
						free(sw_results);
					ShowTemplate();
					cleanup();
					return 0;
				}
				// initialize to zeros
				memset(SearchPhrasesForHL[sw], 0, (num_phrase_terms+1) * sizeof(char*));
			}

			strcpy(temp_phrase_terms, SearchWords[sw]);
			nextword = strtok(temp_phrase_terms, " ");
			
			for(int a = 0; a < num_phrase_terms; a++)
			{
				if(nextword == NULL)
					break;
				
				SearchPhrases[sw][a] = (char*) malloc((strlen(nextword)+1) * sizeof(char));
				strcpy(SearchPhrases[sw][a], nextword);					
				if (Highlighting == 1)
				{
					// we'll back up the search phrases only if we need to overwrite them for stemming and highlighting
					SearchPhrasesForHL[sw][a] = (char*) malloc((strlen(nextword)+1) * sizeof(char));
					strcpy(SearchPhrasesForHL[sw][a], nextword);
				}
				if (UseStemming)
				{
					GetStem(SearchPhrases[sw][a], strlen(SearchPhrases[sw][a]));
				}

				// we allocate +1 space because we need it to be null terminated
				phrase_terms_ids[sw] = (int*) malloc(sizeof(int) * (num_phrase_terms+1));
				memset(phrase_terms_ids[sw], 0, sizeof(int) * (num_phrase_terms+1));

				nextword = strtok(NULL, " ");				
			}		
						
			int tmpid = 0;
			int WordNotFound = 0;
			
			//create phrase data count array
			phrase_data_count = (int*) malloc(sizeof(int) * num_phrase_terms);
			
			phrase_terms_data = (word_data**) malloc(sizeof(word_data*) * num_phrase_terms);
			// set to zero, so that we can check if pointer is NULL before deleting later
			memset(phrase_terms_data, 0, sizeof(word_data*) * num_phrase_terms);
															
			for (int j = 0; j < num_phrase_terms; j++)
			{    									
				tmpid = GetDictID(SearchPhrases[sw][j]);
				if (tmpid == -1)	// word is not in dictionary
				{					
					WordNotFound = 1;    			
					break;
				}
				phrase_terms_ids[sw][j] = tmpid;

				__int64 wordmap_row = Dict[tmpid].ptrID;
				if (wordmap_row != -1)
				{
					// seek to position in wordmap file
					UserDebugSaveTime(DEBUGTIMER_WORDMAPLOOKUP);
					if (myfseek(fp_wordmap, wordmap_row) == -1)
					{
						UserDebugAddTime(DEBUGTIMER_WORDMAPLOOKUP);
						return 0;
					}
					
					// first 4 bytes is data count
					fread(&(phrase_data_count[j]), 4, 1, fp_wordmap);
					
					if (phrase_data_count[j] > 0)					
						phrase_terms_data[j] = (word_data*)malloc(sizeof(word_data)*phrase_data_count[j]);
							
					//read in infrom from wordmapfile (score, pagenum, offset)
					if (phrase_terms_data[j] != NULL)
					{
#if defined(NON_WINDOWS) && defined(SUN_SPARC_BUILT)
						//	Reverse the byte order for Sparc
						unsigned char *LE_bytes = (unsigned char*) malloc(sizeof(word_data) * phrase_data_count[j]);						
						if(LE_bytes!=NULL)
						{
							if (fread(LE_bytes, sizeof(word_data), phrase_data_count[j], fp_wordmap) == 0)			
							{
								if (UseXML) zoomprintf("<zoom:error>");
								zoomprintf("Error in wordmap file: expected data not found.");							
								if (UseXML) zoomprintf("</zoom:error>\n");
								ShowTemplate();
								cleanup();
								return 1;
							}
							else
							{
								unsigned long offset;
								for (unsigned long i = 0; i < phrase_data_count[j]; i++)
								{
									offset = i*sizeof(word_data);
									phrase_terms_data[j][i].score = LE_bytes[offset];
									phrase_terms_data[j][i].proximity = LE_bytes[offset+1];
									/*phrase_terms_data[j][i].pagenum = LE_bytes[offset+2] | LE_bytes[offset+3]<<8 | LE_bytes[offset+4]<<16 | LE_bytes[offset+5]<<24;
									phrase_terms_data[j][i].textoffset = LE_bytes[offset+6] | LE_bytes[offset+7]<<8 | LE_bytes[offset+8]<<16 | LE_bytes[offset+9]<<24;*/
									
									phrase_terms_data[j][i].pagenum_offset = 0;
									for (int pnj = 0; pnj < 8; pnj++)
									{
										phrase_terms_data[j][i].pagenum_offset <<= 8;
										phrase_terms_data[j][i].pagenum_offset |= LE_bytes[offset+(9-pnj)];
									}
								}
							}
							free(LE_bytes);
						}
#else
						//if (_read(fp_wordmap, phrase_terms_data[j], sizeof(word_data) * phrase_data_count[j]) == 0)						
						if (fread(phrase_terms_data[j], sizeof(word_data), phrase_data_count[j], fp_wordmap) == 0)						
						{
							if (UseXML) zoomprintf("<zoom:error>");
							zoomprintf("Error in wordmap file: expected data not found.");							
							if (UseXML) zoomprintf("</zoom:error>\n");
							UserDebugAddTime(DEBUGTIMER_WORDMAPLOOKUP);
							ShowTemplate();
							cleanup();
							return 1;
						}
#endif					
					}			
					UserDebugAddTime(DEBUGTIMER_WORDMAPLOOKUP);
				}
				else
				{
					phrase_data_count[j] = 0;
					phrase_terms_data[j] = 0;
				}
			}
			
			if (WordNotFound == 1)
				continue;			

			UserDebugAddTime(DEBUGTIMER_PREPQUERY);
		}
		/*
		else if (strchr(SearchWords[sw], '*') != NULL || strchr(SearchWords[sw], '?') != NULL)		
		{
			UseWildCards = 1;
		}
		*/
		
		// start looking through dictionary, starting from DictReservedLimit (after the punctuation marks)
		for (unsigned long i = DictReservedLimit+1; i < NumKeywords && (!IsMaxLimitExceeded||ExcludeTerm); i++)
		{		
			// ------------- Begin dictionary lookup section ---------------------
			UserDebugSaveTime(DEBUGTIMER_DICTLOOKUP);

			dictline = Dict[i];
		
			// if we're not using wildcards, direct match
			if (ExactPhrase == 1)
			{
				// todo: move to next phrase term if first phrase term is skipped?
				// compare first term in exact phrase
				//result = mystrcmp_rightcase(SearchPhrases[sw][0], Dict[i].word);				
				if (i == phrase_terms_ids[sw][0])
					result = 0;
				else
					result = 1;
			}
			else if (UseWildCards[sw] == 0)
			{
				result = strcmp(SearchWords[sw], Dict[i].wordPtr);
				/*
				if (SearchAsSubstring == 0)
				{
					result = strcmp(SearchWords[sw], Dict[i].wordPtr);					
				}
				else
				{  
					if(strlen(SearchWords[sw]) <= strlen(Dict[i].wordPtr))					
					{	
						char* tmp = NULL;
						if (ToLowerSearchWords)
							tmp = iStrStr(Dict[i].wordPtr, SearchWords[sw]);
						else
							tmp = strstr(Dict[i].wordPtr, SearchWords[sw]);
						
						if (tmp == NULL)
							result = 1;
						else
							result=0;
					}
					else
						result=1; //searchword is bigger than word, not found
				}			
				*/
			} 
			else 
			{				
				// both wildcards and substring search will use this:
				result = 1;	// set to not found before loop
				char* variantPtr = NULL;
				for (unsigned int variant_index = 0; result != 0 && variant_index <= Dict[i].variantCount; variant_index++)
				{
					variantPtr = GetDictionaryWord(i, variant_index);
					if (variantPtr != NULL)
					{					
						if (WildCardCompare(variantPtr, SearchWords[sw], !ToLowerSearchWords))
							result = 0;
						else 
							result = 1;
					}
				}
			}

			UserDebugAddTime(DEBUGTIMER_DICTLOOKUP);
			// ------------- End dictionary lookup section ---------------------

			// word found but indicated to be not indexed or skipped
			if (result == 0 && dictline.ptrID == -1)
			{				
				if (UseWildCards[sw] == 0 && SearchAsSubstring == 0) 
				{
					//PrintDebugLine("skipping word\n"); 		

					if (ExactPhrase == 1)
						SkippedExactPhrase = 1;
					
					SkipSearchWord(sw);
					break;
				}
				else
				{
					continue;				
				}
			}
			
			if (result == 0)
			{
				wordsmatched++;
				totalwordsmatched++;

				if (!ExcludeTerm && totalwordsmatched > MaxMatches)
				{							
					IsMaxLimitExceeded = true;
					break;
				}
				if (!ExcludeTerm && wordsmatched > MaxMatchesPerWord)
				{
					break;
				}

				// remember the dictionary ID for this matched search term.
				if (search_terms_ids[sw] == 0)
					search_terms_ids[sw] = i;

				// keyword found in the dictionary
				// PrintUserDebug("keyword found\n"); 					
				
				if (ExactPhrase == 1)
				{					
					// we'll use the wordmap data for the first term that we have worked out earlier
					data_count = phrase_data_count[0];
					
					if(data_count > 0)
					{
						UserDebugSaveTime(DEBUGTIMER_MEMALLOC);

						data = (word_data *) malloc(sizeof(word_data) * data_count);
						
						for(unsigned int a=0; a< data_count; a++)
							data[a] = phrase_terms_data[0][a];

						UserDebugAddTime(DEBUGTIMER_MEMALLOC);
					}					
					//data = phrase_terms_data[0];					
					ContextSeeks = 0;
				}
				else
				{		
					// keyword found in the dictionary
					// PrintUserDebug("keyword found - not exact phrase\n"); 
										
					// seek to position in wordmap file					
					UserDebugSaveTime(DEBUGTIMER_WORDMAPLOOKUP);
					if(myfseek(fp_wordmap, dictline.ptrID) != 0)
					{
						UserDebugAddTime(DEBUGTIMER_WORDMAPLOOKUP);
						return 0;					
					}
					
					// first 4 bytes is data count
					fread(&data_count, 4, 1, fp_wordmap);
					
					//allocate memory
					if (data_count > 0)
					{
						data = (word_data *) malloc(sizeof(word_data) * data_count);
						//read in from wordmapfile (score, pagenum, offset)
						if (data != NULL)
						{
#if defined(NON_WINDOWS) && defined(SUN_SPARC_BUILT)
							//	Reverse the byte order for Sparc
							unsigned char *LE_bytes = (unsigned char*) malloc(sizeof(word_data) * data_count);						
							if(LE_bytes!=NULL)
							{
								if (fread(LE_bytes, sizeof(word_data), data_count, fp_wordmap) == 0)
								{
									if (UseXML) zoomprintf("<zoom:error>");
									zoomprintf("Error in wordmap file: expected data not found.");
									if (UseXML) zoomprintf("</zoom:error>\n");							
									ShowTemplate();
									cleanup();
									exit(1);
								}
								else
								{
									unsigned long offset;
									for (unsigned long i = 0; i < data_count; i++)
									{
										/*
										if(i==0)
										{
											for(j=0; j<sizeof(word_data); j++)
												zoomprintf("LE_bytes[%d] = %u\n", j, LE_bytes[j]);
										}
										*/
										offset = i*sizeof(word_data);
										/*
										data[i].score = LE_bytes[offset] + LE_bytes[offset+1]*256;
										data[i].pagenum = LE_bytes[offset+2] + LE_bytes[offset+3]*256 + LE_bytes[offset+4]*256*256 + LE_bytes[offset+5]*256*256*256;
										data[i].textoffset = LE_bytes[offset+6] + LE_bytes[offset+7]*256 + LE_bytes[offset+8]*256*256 + LE_bytes[offset+9]*256*256*256;
										*/
										
										data[i].score = LE_bytes[offset];
										data[i].proximity = LE_bytes[offset+1];
										/*phrase_terms_data[j][i].pagenum = LE_bytes[offset+2] | LE_bytes[offset+3]<<8 | LE_bytes[offset+4]<<16 | LE_bytes[offset+5]<<24;
										phrase_terms_data[j][i].textoffset = LE_bytes[offset+6] | LE_bytes[offset+7]<<8 | LE_bytes[offset+8]<<16 | LE_bytes[offset+9]<<24;*/
									
										data[i].pagenum_offset = 0;
										for (int j = 0; j<8; j++)
										{
											data[i].pagenum_offset <<= 8;
											data[i].pagenum_offset |= LE_bytes[offset+(9-j)];
										}
										
										
									}
									//zoomprintf("2	data[0].score = %u\n", data[0].score);
									//zoomprintf("2	data[0].pagenum = %u\n", data[0].pagenum);
									//zoomprintf("2	data[0].textoffset = %u\n", data[0].textoffset);
								}
								free(LE_bytes);
							}

#else
							//if (_read(fp_wordmap, data, sizeof(word_data)*data_count) == 0)
							if (fread(data, sizeof(word_data), data_count, fp_wordmap) == 0)
							{
								if (UseXML) zoomprintf("<zoom:error>");
								zoomprintf("Error in wordmap file: expected data not found.");
								if (UseXML) zoomprintf("</zoom:error>\n");							
								UserDebugAddTime(DEBUGTIMER_WORDMAPLOOKUP);
								ShowTemplate();
								cleanup();
								return 0;
							}
#endif
						}				
					}
					UserDebugAddTime(DEBUGTIMER_WORDMAPLOOKUP);
				}

				sw_results[sw] += data_count;

				// PrintUserDebug("go through wordmap for each page this word appears on\n"); 
									
				// Go through wordmap for each page this word appears on
				for (unsigned long j = 0; j < data_count; j++)
				{    				
					UserDebugSaveTime(DEBUGTIMER_WORDMAPLOOKUP);

					unsigned short score = data[j].score;
					unsigned char proximity = data[j].proximity;
					/*
					unsigned long txtptr = data[j].textoffset;        
					int ipage = data[j].pagenum;
					*/
					__int64 txtptr = GET_TEXTOFFSET(data[j].pagenum_offset);

					unsigned long ipage = GET_PAGENUM(data[j].pagenum_offset);

					unsigned char boost = GetPageInfo_Boost(pageinfo[ipage], UseBigPageInfoData);										
					if (boost != 0)				
					{
						float tmpscore = score * (boost / 10.0f);
						score = (unsigned short) (tmpscore + 0.5f);
						//score = (unsigned short) ((float)score * ((float)pageinfo[ipage]->boost / 10.0f)) + 0.5f; // +0.5 for rounding
					}

					UserDebugAddTime(DEBUGTIMER_WORDMAPLOOKUP);

					if (score == 0)
						continue;	// skip this entry if score is zero'ed
					
					if (ExactPhrase == 1) 
					{
						UserDebugSaveTime(DEBUGTIMER_PHRASE);

						//unsigned long maxptr = data[j].textoffset;
						//unsigned long maxptr_term = 0;
						__int64 maxptr = GET_TEXTOFFSET(data[j].pagenum_offset);
						__int64 maxptr_term = 0;
						int GotoNextPage = 0;	
						unsigned char overlapProx = 0;

						unsigned long xbi_pagenum;
						__int64 xbi_offset;

						unsigned long currword_pagenum;
						currword_pagenum = GET_PAGENUM(data[j].pagenum_offset);
						
						// Check if all of the other words in the phrase appears on this page.
						int xi=1;
						int xi_data_count;
						for (xi = 1; xi < num_phrase_terms && GotoNextPage == 0; xi++)
						{					
							xi_data_count = phrase_data_count[xi];
							// see if this word appears at all on this page, if not, we stop scanning page.
							// do not check for skipped words (data count value of zero)
							if (xi_data_count != 0)
							{
								int xbi = 0;
								// check wordmap for this search phrase to see if it appears on the current page.
								for (xbi = 0; xbi < xi_data_count; xbi++) 
								{
									xbi_pagenum = GET_PAGENUM(phrase_terms_data[xi][xbi].pagenum_offset);									
									if (xbi_pagenum == currword_pagenum)
									{										
										overlapProx = phrase_terms_data[xi][xbi].proximity << 1;
										if ((proximity & phrase_terms_data[xi][xbi].proximity) == 0 &&
											(proximity & overlapProx) == 0)
										{
											GotoNextPage = 1;
										}
										else
										{
											xbi_offset = GET_TEXTOFFSET(phrase_terms_data[xi][xbi].pagenum_offset);

											// intersection, this term appears on both pages, goto next term
											// remember biggest pointer.
											//if (phrase_terms_data[xi][xbi].textoffset > maxptr)
											if (xbi_offset > maxptr)
											{
												//maxptr = phrase_terms_data[xi][xbi].textoffset;
												maxptr = xbi_offset;
												maxptr_term = xi;			    	 					
											}											
											score += phrase_terms_data[xi][xbi].score;										
										}
										break;
									}
								}
								if (xbi == xi_data_count) // if not found
								{									
									GotoNextPage = 1;			    	   			
									break;	// goto next page
								}
							}							
						}	// end phrase term for loop

						UserDebugAddTime(DEBUGTIMER_PHRASE);
																													
						// Check how many context seeks we have made.

						if (ContextSeeks > (MaxContextSeeks+FINAL_CONTEXTSEEKS))
						{
							// When we are up to the final number of context seeks before reaching
							// MaxContextSeeks, we change our behaviour here in an attempt to find
							// more obvious phrases which have pointers occuring next to each other.

							// At this point, we skip the page (and don't count it as a context seek) 
							// unless the stored offsets for the words of the phrase are all within
							// close proximity of each other.

							// txtptr contains the offset for the first word in the phrase
							// maxptr will need to be the last word in the phrase, and within proximity
							
							if (maxptr > txtptr + ((num_phrase_terms+PHRASE_NUM_PUNCT_TOLERANCE) * DictIDLen))
							{
								GotoNextPage = 1;
							}
						}

						if (GotoNextPage == 1)
							continue;

						ContextSeeks++;					
						if (ContextSeeks > MaxContextSeeks)
						{							
							IsMaxLimitExceeded = true;								
							break;
						}

						// Check if we are taking too long here
						TIMER_VALUE checkTime = GetTimeNow();
						double checkTimeDiff = GetTimeDifference(starttime, checkTime);
						if (checkTimeDiff > MaxSearchTime)
						{							
							IsMaxLimitExceeded = true;								
							break;
						}
						
						// ok, so this page contains all of the words in the phrase
						int FoundPhrase = 0;
						int FoundFirstWord = 0;					
						unsigned long xword_id=0;
						unsigned int variant_index=0;						
						
						// we goto the first occurance of the first word in pagetext   					
						__int64 pos = 0;

						__int64 data_j_offset = GET_TEXTOFFSET(data[j].pagenum_offset);

						// work-around to test for a temporary negative value while using unsigned longs						
						__int64 tmp_maxptr_goback = (maxptr_term+PHRASE_NUM_PUNCT_TOLERANCE) * DictIDLen;	// assume 3 possible punctuations.						
						if (maxptr < tmp_maxptr_goback)
							pos = data_j_offset;
						else
							pos = maxptr - tmp_maxptr_goback;

						// do not seek further back than the occurance of the first word (avoid wrong page)						
						if (pos < data_j_offset)
							pos = data_j_offset;
						
						UserDebugSaveTime(DEBUGTIMER_CONTEXTSEEK);
						if (myfseek(fp_pagetext, pos) == -1)
						{
							UserDebugAddTime(DEBUGTIMER_CONTEXTSEEK);
							return 0;						
						}						

						GetNextDictWordBuffCount = 0;
						GetNextDictWordBuffRead = 0;
					
						// now we look for the phrase within the context of this page
						do
						{							
							for (xi = 0; xi < num_phrase_terms; xi++)
							{										
								do
								{
									xword_id = 0;
									bytesread = GetNextDictWord(fp_pagetext, &xword_id, &variant_index);
									/*
									//bytesread = _read(fp_pagetext, bytes_buffer, DictIDLen);	// grab 2 bytes
									bytesread = fread(bytes_buffer, 1, DictIDLen, fp_pagetext);	// grab next dictionary word
									if ((int)bytesread < DictIDLen)
									{
										if (UseXML) zoomprintf("<zoom:error>");
										zoomprintf("Error in pagetext file: Unexpected dictionary ID length.");
										if (UseXML) zoomprintf("</zoom:error>\n");
										return 0;
									}									
									for (int i = 0; i < DictIDLen; i++)
										xword_id = xword_id | (bytes_buffer[i] << (8*i));
									*/
									pos += bytesread;
									
									// check if we are at the end of page (wordid = 0)
									if (xword_id == 0 || xword_id == 1 || xword_id >= NumKeywords)
										break;

								} while (xword_id <= DictReservedLimit && !feof(fp_pagetext));

								if (xword_id == 0 || xword_id == 1 || xword_id >= NumKeywords)
									break;
								
								// if the words are NOT the same, we break out								
								if (xword_id != phrase_terms_ids[sw][xi])
								{
									// also check against the first word									
									if (xi != 0 && xword_id == phrase_terms_ids[sw][0])
										xi = 0;	// matched first word
									else
										break;
								}
								
								// remember how many times we find the first word on this page
								if (xi == 0)
								{
									FoundFirstWord++;
									// remember the position of the 'start' of this phrase that was found									
									txtptr = pos - bytesread;									
								}
								
							}	
							if (xi == num_phrase_terms)
							{
								// exact phrase found!
								FoundPhrase = 1;
							}														
						} 
						while (xword_id != 0 && FoundPhrase == 0 && FoundFirstWord <= data[j].score);

						UserDebugAddTime(DEBUGTIMER_CONTEXTSEEK);
						
						if (FoundPhrase != 1)
							continue;	// goto next page.
					}		
					
					// PrintUserDebug("Check if page is already in output list\n");			
					//Check if page is already in output list
					int pageexists = 0;					

					if (ipage > (int)PagesCount)
					{
						if (UseXML) zoomprintf("<zoom:error>");
						zoomprintf("Error in wordmap: page number is invalid.");
						if (UseXML) zoomprintf("</zoom:error>\n");
						ShowTemplate();
						cleanup();
						return 0;
					}
					
					if (ExcludeTerm == 1)
					{
						// we clear out the score entry so that it'll be excluded at the filtering stage						
						res_table[ipage].score = 0;
						matches--;
					}
					else if (res_table[ipage].score == 0) 
					{
						// PrintUserDebug("not in list, count this page as a unique match\n");				
						// not in list, count this page as a unique match	
						matches++;
						res_table[ipage].score = score;
						res_table[ipage].txtptr1 = txtptr;
						res_table[ipage].proximity = proximity;
					}
					else 
					{						
						// already in list
						if (res_table[ipage].score > 10000) 
						{
							// PrintUserDebug("already in list\n");								
							// take it easy if its too big (to prevent huge scores)
							res_table[ipage].score += 1;
						}
						else
						{
							res_table[ipage].score += score;    //Add in score
							// This is no longer necessary with our new scoring system:
							//res_table[ipage].score *= 2;        //Double Score as we have two words matching
						}
												
						// store the next two searchword matches
						if (res_table[ipage].terms > 0 && res_table[ipage].terms < MaxContextKeywords) 
						{							
							if (res_table[ipage].txtptr2 == 0)
								res_table[ipage].txtptr2 = txtptr;
							else if (res_table[ipage].txtptr3 == 0)
								res_table[ipage].txtptr3 = txtptr;
						}	
						res_table[ipage].proximity &= proximity;
					}
					res_table[ipage].terms += 1;	

					if (Sort > 0)
					{
						// sorting by date						
						//tmp_histogram_value = (unsigned long) pageinfo[ipage]->filetime;
						tmp_histogram_value = (unsigned long) GetPageInfo_Filetime(pageinfo[ipage], UseBigPageInfoData);

						// remember the max. and min. score
						if (matches == 1)
							histogram_min = tmp_histogram_value;

						if (tmp_histogram_value > histogram_max)
							histogram_max = tmp_histogram_value;

						if (tmp_histogram_value < histogram_min)
							histogram_min = tmp_histogram_value;
					}
					else
					{	
						// sorting by relevance
						tmp_histogram_value = res_table[ipage].score; // score

						if (res_table[ipage].terms > maxtermsmatched)
							maxtermsmatched = res_table[ipage].terms;

						if (res_table[ipage].score > maxscore)
							maxscore = res_table[ipage].score;
					}
										
					// store the 'and' user search terms matched' value
					if (res_table[ipage].and_counter == sw ||  res_table[ipage].and_counter == sw-SkippedWords-exclude_count)
						res_table[ipage].and_counter += 1;  

				}
							
				if(data != NULL)
				{											
					free(data);
					data = NULL;
				}
				
				if (UseWildCards[sw] == 0 && SearchAsSubstring == 0)
					break;  //This search word was found, so skip to next				
			}			
		}				
	}
	
	PrintUserDebug("End main search loop.<br>"\
		"Total time spent looking in dictionary: %.6f (%d iterations)<br>"\
		"Total time spent looking in wordmap: %.6f (%d iterations)<br>"\
		"Total time spent seeking in context: %.6f (%d iterations)<br>"\
		"Total time spent searching phrases: %.6f (%d iterations)<br>"\
		"Total time spent allocating memory: %.6f (%d iterations)<br>"\
		"Total time spent looking for rec links: %.6f (%d iterations)<br>"\
		"Total time spent stemming: %.6f (%d iterations)<br>"\
		"Total time spent preparing query: %.6f (%d iterations)<br>", 
		UserDebugStats.TotalTime[DEBUGTIMER_DICTLOOKUP]/1000, UserDebugStats.SaveCounter[DEBUGTIMER_DICTLOOKUP], 
		UserDebugStats.TotalTime[DEBUGTIMER_WORDMAPLOOKUP]/1000, UserDebugStats.SaveCounter[DEBUGTIMER_WORDMAPLOOKUP],
		UserDebugStats.TotalTime[DEBUGTIMER_CONTEXTSEEK]/1000, UserDebugStats.SaveCounter[DEBUGTIMER_CONTEXTSEEK],
		UserDebugStats.TotalTime[DEBUGTIMER_PHRASE]/1000, UserDebugStats.SaveCounter[DEBUGTIMER_PHRASE],
		UserDebugStats.TotalTime[DEBUGTIMER_MEMALLOC]/1000, UserDebugStats.SaveCounter[DEBUGTIMER_MEMALLOC],
		UserDebugStats.TotalTime[DEBUGTIMER_FINDRECLINKS]/1000, UserDebugStats.SaveCounter[DEBUGTIMER_FINDRECLINKS],
		UserDebugStats.TotalTime[DEBUGTIMER_STEMMING]/1000, UserDebugStats.SaveCounter[DEBUGTIMER_STEMMING],
		UserDebugStats.TotalTime[DEBUGTIMER_PREPQUERY]/1000, UserDebugStats.SaveCounter[DEBUGTIMER_PREPQUERY]);
	


	if (Sort == 0)
	{
		// sort by relevance, need to initialize histogram_max and histogram_min
		histogram_max = maxscore + (maxscore * (maxtermsmatched-1));			
		histogram_min = 0;
	}
	
	//free phrase_terms_data	
	if (phrase_terms_data != NULL)
	{
		for(int a=0; a<num_phrase_terms; a++)
		{
			if (phrase_terms_data[a] != NULL)
			{
				free(phrase_terms_data[a]);
				phrase_terms_data[a] = NULL;
			}
		}
		free(phrase_terms_data);
		phrase_terms_data = NULL;
	}

	if (phrase_data_count != NULL)
	{
		free(phrase_data_count);
		phrase_data_count = NULL;
	}
		
#if !defined(FASTCGI_ENABLED) && !defined(OSF)
	//Close the files	
	fclose(fp_wordmap);
	fp_wordmap = NULL;
#endif
	
	if (SkippedWords > 0)
	{
		if (UseXML)
		{
			zoomprintf("<zoom:summary>%s %s", StrTable[STR_SKIPPED_FOLLOWING_WORDS], SkippedOutputStr);
			if (SkippedExactPhrase == 1)
				zoomprintf(" %s", StrTable[STR_SKIPPED_PHRASE]);
			zoomprintf("</zoom:summary>\n");
		}
		else
		{
			OutputBufferPrintf(OUTPUT_SUMMARY, "<div class=\"summary\">%s %s<br />\n", StrTable[STR_SKIPPED_FOLLOWING_WORDS], SkippedOutputStr);
			if (SkippedExactPhrase == 1)
				OutputBufferPrintf(OUTPUT_SUMMARY, "%s<br />\n", StrTable[STR_SKIPPED_PHRASE]);
			OutputBufferPrintf(OUTPUT_SUMMARY, "<br /></div>\n");
		}
	}

	unsigned int full_numwords = NumSearchWords - SkippedWords - exclude_count;

	BYTE* metaDataPtr = NULL;
	int IsAnyDropdown;	

	int* CatCounter = NULL;
	int CatCounterFilled = 0;
	if (UseCats && DisplayCatSummary == 1)
	{
		if (Cat[0] == -1 || num_zoom_cats > 1)
			CatCounter = (int*)calloc(NumCats, sizeof(int));
		else
			DisplayCatSummary = 0;
	}

	strcpy(dateRangeParams, "");
	bool UseDateRange = false;
	time_t from_datetime = 0;
	time_t to_datetime = 0;
	if (DateRangeSearch == 1)
	{
		if (zsSetup->LinuxTimeFrom > 0 && zsSetup->LinuxTimeTo > 0)
		{
			from_datetime = zsSetup->LinuxTimeFrom;
			to_datetime = zsSetup->LinuxTimeTo;
		}
		else
		{
			int tmpFromMonth, tmpFromDay, tmpFromYear, tmpToMonth, tmpToDay, tmpToYear;
			if (DateRangeFormat == 1)
			{
				sscanf(DateFromStr, "%d/%d/%d", &tmpFromMonth, &tmpFromDay, &tmpFromYear);
				sscanf(DateToStr, "%d/%d/%d", &tmpToMonth, &tmpToDay, &tmpToYear);
			}
			else
			{
				sscanf(DateFromStr, "%d/%d/%d", &tmpFromDay, &tmpFromMonth, &tmpFromYear);
				sscanf(DateToStr, "%d/%d/%d", &tmpToDay, &tmpToMonth, &tmpToYear);
			}
			if (tmpFromDay > 0 && tmpFromMonth > 0 && tmpFromYear > 0 && tmpToDay > 0 && tmpToMonth > 0 && tmpToYear > 0)
			{
				tm tt;
				memset(&tt, 0, sizeof(tm));
				tt.tm_hour = 0;
				tt.tm_min = 0;
				tt.tm_sec = 0;
				tt.tm_mon = tmpFromMonth-1;
				tt.tm_mday = tmpFromDay;
				tt.tm_year = tmpFromYear - 1900;
				from_datetime = mktime(&tt);

				memset(&tt, 0, sizeof(tm));
				tt.tm_hour = 23;
				tt.tm_min = 59;
				tt.tm_sec = 59;
				tt.tm_mon = tmpToMonth-1;
				tt.tm_mday = tmpToDay;
				tt.tm_year = tmpToYear - 1900;
				to_datetime = mktime(&tt);
				if (from_datetime != -1 && to_datetime != -1)
				{
					UseDateRange = true;
					urlencode((char*)DateFromStr, DATESTRING_LENGTH);
					urlencode((char*)DateToStr, DATESTRING_LENGTH);
					sprintf(dateRangeParams, "&amp;zoom_datefrom=%s&amp;zoom_dateto=%s", DateFromStr, DateToStr);
				}
			}
		}
	}

	if (matches > 0 || IsNoKeywordQuery)
	{
		PrintUserDebug("Begin histogram building (allocating)...");

		unsigned long* histogram = NULL;
		unsigned long histogram_bucket_size;

		if ((histogram_max-histogram_min) < HISTOGRAM_NUM)
			histogram_bucket_size = 1;
		else
			histogram_bucket_size = (unsigned long) ceil(((float)(histogram_max - histogram_min) / (float)HISTOGRAM_NUM));

		histogram = (unsigned long*) calloc(HISTOGRAM_NUM, sizeof(unsigned long));
		if (histogram == NULL)
		{
			if (UseXML) zoomprintf("<zoom:error>");
			zoomprintf("Error: Out of memory");
			if (UseXML) zoomprintf("</zoom:error>\n");
			if (res_table) free(res_table);			
			ShowTemplate();
			cleanup();
			return 0;
		}

		unsigned long histogram_pos;	
		PrintUserDebug("First pass filtering (and populating histogram)...");	

		bool IsFiltered;

		char pageUrlBuffer[PAGEDATA_MAX_LINELEN+1];

		// reset matches to zero, might filter some out
		matches = 0;
		for (unsigned int i = 0; i < PagesCount; i++)
		{
			IsFiltered = false;

			if ((res_table[i].score == 0 || res_table[i].terms == 0) &&	IsNoKeywordQuery == false)
			{
				IsFiltered = true;
			}

			if (IsFiltered == false && searchUrlBuffer[0] != '\0')
			{
				// filter by URL	
				//char tmpURL[URLLENGTH+1];
				//if (GetPageData(i, (char*)&tmpURL, NULL, NULL, NULL))
				if (GetUrlFromPageData(i, (char*)&pageUrlBuffer))
				{
					//if (iStrStr(tmpURL, searchUrlBuffer) == NULL)
					if (iStrStr(pageUrlBuffer, searchUrlBuffer) == NULL)
					{
						IsFiltered = true;
					}
					else
					{
						// the site/URL matched with the site: parameter
						if (IsNoKeywordQuery)
						{
							// increment score/terms if we're allowing this
							// through on the site: match alone (no keyword search)
							res_table[i].score++;
							res_table[i].terms++;
						}
					}
				}
			}

			if (UseMetaFields && IsFiltered == false)
			{				
				metaDataPtr = metapages[i];

				for (int fieldnum = 0; fieldnum < NumMetaFields && !IsFiltered; fieldnum++)
				{
					IsAnyDropdown = false;					

					if (metafields[fieldnum].type == METAFIELD_TYPE_DROPDOWN)
					{
						if (atoi(meta_query[fieldnum]) == -1)
							IsAnyDropdown = true;
					}
					else if (metafields[fieldnum].type == METAFIELD_TYPE_MULTI)
					{
						if (meta_query[fieldnum][0] > 0)
						{
							int num_multi_query = (int) meta_query[fieldnum][0];
							for (int mqi = 0; mqi < num_multi_query && !IsAnyDropdown; mqi++)
								if (meta_query[fieldnum][mqi+1] == -1)
									IsAnyDropdown = true;
						}
					}

					if (strlen(meta_query[fieldnum]) > 0 && IsAnyDropdown == false)
					{
						if (metafields[fieldnum].type == METAFIELD_TYPE_TEXT)
						{								
							//BYTE metaStrLen = metaDataPtr[0];
							unsigned short metaStrLen = 0;
							if (UseBigPageInfoData)
							{
								memcpy(&metaStrLen, metaDataPtr, METAFIELD_TEXT_SIZE_BYTES);
								metaDataPtr += METAFIELD_TEXT_SIZE_BYTES;// move pointer to after length
							}
							else
							{
								memcpy(&metaStrLen, metaDataPtr, METAFIELD_OLD_TEXT_SIZE_BYTES);
								metaDataPtr += METAFIELD_OLD_TEXT_SIZE_BYTES;// move pointer to after length
							}														
							
							if (metaStrLen == 0)
							{
								IsFiltered = true;
							}
							else 
							{
								int metaQueryLen = strlen(meta_query[fieldnum]);
								if (metafields[fieldnum].method == METAFIELD_METHOD_SUBSTRING)
								{
									if (metaQueryLen <= (int)metaStrLen)
									{	
										if (iStrStr((char*)metaDataPtr, meta_query[fieldnum], (int)metaStrLen) == NULL)
											IsFiltered = true;
									}
									else
										IsFiltered = true;
								}
								else
								{
									// this is an exact match. should only proceed if they're the same length
									if ((int)metaStrLen == metaQueryLen)
									{
										if (mystrnicmp(meta_query[fieldnum], (char*)metaDataPtr, metaQueryLen) != 0)
											IsFiltered = true;
									}
									else
										IsFiltered = true;
								}
							}
							// move pointer to (possibly) next record
							metaDataPtr += sizeof(char) * metaStrLen;
						}						
						else if (metafields[fieldnum].type == METAFIELD_TYPE_DROPDOWN)
						{
							BYTE metaNumValue = 0;
							metaNumValue = metaDataPtr[0];
							metaDataPtr += sizeof(BYTE);
							if (metaNumValue == METAFIELD_DROPDOWN_NOVALUE)
								IsFiltered = true;
							else 
							{
								unsigned int queryValue = 0;
								sscanf(meta_query[fieldnum], "%u", &queryValue);
								if (metaNumValue != queryValue)
									IsFiltered = true;
							}							
						}
						else if (metafields[fieldnum].type == METAFIELD_TYPE_MULTI)
						{							
							IsFiltered = true;
							BYTE metaValCount = 0;
							int metaNumValue = 0;
							metaValCount = metaDataPtr[0];							
							metaDataPtr += sizeof(BYTE);
							if (metaValCount > 0)
							{
								unsigned int num_multi_query = 0;
								unsigned int queryValue = 0;		// this is important to be unsigned. otherwise we have trouble at 128 (which becomes -128)
								num_multi_query = (unsigned int) meta_query[fieldnum][0];
								for (unsigned int mqi = 0; mqi < num_multi_query && IsFiltered; mqi++)
								{
									queryValue = (unsigned int)(unsigned char) meta_query[fieldnum][mqi+1];// important to cast to unsigned char first to make sense of the value stored in signed char
									for (unsigned int mvi = 0; mvi < (unsigned int)metaValCount; mvi++)
									{
										metaNumValue = (unsigned int) metaDataPtr[mvi];																			
										if (metaNumValue == queryValue)
										{
											IsFiltered = false;
											break;
										}
									}
								}
								metaDataPtr += sizeof(BYTE) * metaValCount;
							}
						}
						else
						{
							// numeric comparison here
							unsigned int metaNumValue = 0;
							endianAware_memcpy(&metaNumValue, metaDataPtr, sizeof(unsigned int));

							bool bRet;
							
							if (metaNumValue == METAFIELD_NOVALUE_MARKER)
								bRet = false;
							else 
							{
								unsigned int queryValue = 0;								

								if (UseMetaFields == 1 && MetaMoneyShowDec == 1 && 
									metafields[fieldnum].type == METAFIELD_TYPE_MONEY)
								{
									float tmpFloatVal = 0;
									sscanf(meta_query[fieldnum], "%f", &tmpFloatVal);	
									tmpFloatVal = tmpFloatVal + 0.005f; // rounding to address float inaccuracy
									queryValue = (int) (tmpFloatVal * 100);
								}
								else
									sscanf(meta_query[fieldnum], "%u", &queryValue);	

								if (metafields[fieldnum].method == METAFIELD_METHOD_LESSTHAN)
									bRet = metaNumValue < queryValue;
								else if (metafields[fieldnum].method == METAFIELD_METHOD_LESSTHANORE)
									bRet = metaNumValue <= queryValue;
								else if (metafields[fieldnum].method == METAFIELD_METHOD_GREATERTHAN)
									bRet = metaNumValue > queryValue;
								else if (metafields[fieldnum].method == METAFIELD_METHOD_GREATERTHANORE)
									bRet = metaNumValue >= queryValue;
								else
									bRet = metaNumValue == queryValue;
							}
							
							if (!bRet)
								IsFiltered = true;

							metaDataPtr += sizeof(unsigned int);
						}
					}
					else
					{
						// we need to skip the pointer forwards through the metadata
						if (metafields[fieldnum].type == METAFIELD_TYPE_TEXT)
						{							
							//BYTE metaStrLen = 0;
							//memcpy(&metaStrLen, metaDataPtr, sizeof(BYTE));							
							unsigned short metaStrLen = 0;
							if (UseBigPageInfoData)
							{
								memcpy(&metaStrLen, metaDataPtr, METAFIELD_TEXT_SIZE_BYTES);								
								metaDataPtr += METAFIELD_TEXT_SIZE_BYTES + (sizeof(char) * metaStrLen);
							}
							else
							{
								memcpy(&metaStrLen, metaDataPtr, METAFIELD_OLD_TEXT_SIZE_BYTES);								
								metaDataPtr += METAFIELD_OLD_TEXT_SIZE_BYTES + (sizeof(char) * metaStrLen);
							}
						}
						else if (metafields[fieldnum].type == METAFIELD_TYPE_DROPDOWN)
						{
							metaDataPtr += sizeof(BYTE);
						}
						else if (metafields[fieldnum].type == METAFIELD_TYPE_MULTI)
						{
							BYTE metaValCount = 0;							
							metaValCount = metaDataPtr[0];							
							metaDataPtr += sizeof(BYTE);
							metaDataPtr += sizeof(BYTE) * metaValCount;
						}
						else
						{
							metaDataPtr += sizeof(int);
						}
					}

					if (IsNoKeywordQuery && !IsFiltered)
					{
						res_table[i].score++;
						res_table[i].terms++;
					}
				}
			}

			if (IsFiltered == false) 
			{
				if (res_table[i].and_counter < full_numwords && And == 1)
				{
					// if AND search, only copy AND results
					IsFiltered = true;						
				}
			}

			// Repositioned to AFTER boolean AND and AFTER meta fields
			// so category summary reflect these options enabled
			if (UseCats && Cat[0] != -1 && IsFiltered == false) 
			{
				// Using cats and not doing an "all cats" search				
				bool bFoundCat = false;
				int cati;
				for (cati = 0; cati < num_zoom_cats; cati++)
				{
					if (CheckBitInByteArray(Cat[cati], catpages[i]) != 0)
					{
						if (DisplayCatSummary == 1)
						{
							CatCounter[Cat[cati]]++;
							CatCounterFilled = 1;
						}
						bFoundCat = true;
					}
				}				
				//if (cati == num_zoom_cats)
				if (bFoundCat == false)
					IsFiltered = true;
			}

			if (IsFiltered == false && DateRangeSearch == 1 && UseDateRange == true)
			{
				__time32_t filetime = GetPageInfo_Filetime(pageinfo[i], UseBigPageInfoData);
				if (filetime < from_datetime)
					IsFiltered = true;
				else if (filetime > to_datetime)
					IsFiltered = true;
			}

			if (IsFiltered)
			{
				// we want to remove this from the list later, so we mark them as score=0
				res_table[i].score = 0;
			}
			else
			{
				// we can only count our AND total here AFTER we've filtered out the cats
				if (res_table[i].and_counter >= full_numwords)
				{
					fullmatches++;
				}

				if (Sort > 0)
				{
					// sorting by date, we will build a histogram with the date value
					tmp_histogram_value = (unsigned long) GetPageInfo_Filetime(pageinfo[i], UseBigPageInfoData);
				}
				else
				{
					// sorting by relevance, we will build a histogram with the score value.
					// since we need to involve the "terms matched", we will work out a new
					// score value based on:
					// NewScore = CurrentScore + (MaxScore * (TermsMatched-1))
					// PS. we could modify this in the future to allow for overlap and less prioritise
					// terms matched over score.
					tmp_histogram_value = res_table[i].score + (maxscore * (res_table[i].terms - 1));
					// update the score value with the new value
					res_table[i].score = tmp_histogram_value;
				}
				histogram_pos = (unsigned long) ceil(((float)(tmp_histogram_value - histogram_min) / (float)histogram_bucket_size));
				if (histogram_pos >= HISTOGRAM_NUM)
					histogram_pos = HISTOGRAM_NUM-1;
				histogram[histogram_pos]++;		
				matches++;				
			}
		}
				
		PrintUserDebug("Determining histogram threshold...");	

		if (Sort == 2)
		{
			// Determine threshold for num. pages required but for ascending sort
			int numpagesfound = 0;
			int hi = 0;
			for (hi = 0; hi < HISTOGRAM_NUM && numpagesfound < MaxResultsPerQuery; hi++)
			{
				numpagesfound += histogram[hi];
			}
			cullpages_count = numpagesfound;
			
			// We need to make sure hi is not -1 (which it would be if it went through all results)
			// This was why we had hi+1 before! (see below)
			if (hi >= HISTOGRAM_NUM)
			{
				// NULL the histogram threshold so it doesn't get applied
				histogram_threshold = 0;
			}
			else
				histogram_threshold = histogram_min + (histogram_bucket_size * hi);
		}
		else
		{
			// Determine threshold for num. pages required
			int numpagesfound = 0;
			int hi = 0;
			for (hi = HISTOGRAM_NUM-1; hi >= 0 && numpagesfound < MaxResultsPerQuery; hi--)
			{
				numpagesfound += histogram[hi];
			}
			cullpages_count = numpagesfound;
			
			// We need to make sure hi is not -1 (which it would be if it went through all results)
			// This was why we had hi+1 before! (see below)
			if (hi < 0)
				hi = 0;

			// this caused the last bucket to be filtered out, so we dont get quite 1000 results (16/1/2007)		
			//histogram_threshold = histogram_min + (histogram_bucket_size * (hi+1));

			// changed to the following (16/1/2007) which in tests, returned the number of pages in cullpages_count
			// uncertain why it was hi+1 before		
			histogram_threshold = histogram_min + (histogram_bucket_size * hi);
		}
		
		//zoomprintf("Histogram threshold: %d<BR>", histogram_threshold);		
		if (histogram)
			free(histogram);
	}


	PrintUserDebug("Second pass result filtering");	

	// Second pass, we cull the results to exclude:
	// - results filtered out by boolean AND
	// - results filtered out by category selection
	// - results outside of the NumResultsReqd based on sorting factor (score or date)	

	// allocate output to this size and not pagescount. we only want that many..
	output_table_item* output = (output_table_item*) calloc(cullpages_count, sizeof(output_table_item));
	if (output == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Out of memory\n");
		if (UseXML) zoomprintf("</zoom:error>\n");
		ShowTemplate();
		cleanup();
		return 0;
	}

	//Count number of output lines that match ALL search terms
	unsigned long oline = 0;
		
	float baseScale, proxScale;	
	float finalScale = 0;
	baseScale = 1.3f;
	proxScale = 1.7f;
	if (WeightProximity != 0)
		proxScale += (WeightProximity/10.0f);
						
	for (unsigned long culli = 0; culli < PagesCount /*&& oline < cullpages_count*/; culli++)
	{
		int IsFiltered = false;

		if (res_table[culli].score > 0)
		{		
			if (histogram_threshold > 0)
			{
				// Cull according to histogram thresholds
				if (Sort == 0)
				{
					// sorting by relevance
					if (res_table[culli].score < histogram_threshold)
					{
						IsFiltered = true;
						IsCulled = true;
					}
				}
				else if (Sort == 1)
				{
					// sorting by date
					__time32_t filetime = GetPageInfo_Filetime(pageinfo[culli], UseBigPageInfoData);
					if (filetime < (__time32_t) histogram_threshold)
					{
						IsFiltered = true;
						IsCulled = true;
					}
				}
				else
				{
					// sorting by date
					__time32_t filetime = GetPageInfo_Filetime(pageinfo[culli], UseBigPageInfoData);
					if (filetime > (__time32_t) histogram_threshold)
					{
						IsFiltered = true;
						IsCulled = true;
					}
				}
			}
			
			// We need to count them regardless of culling!! This was moved out from the 
			// if (isFiltered) block below because it meant we didn't count the items
			// that were culled out.
			if (UseCats == 1 && DisplayCatSummary == 1 && Cat[0] == -1)
			{
				// if we are doing an All category search AND we're showing cat summary
				for (int catit = 0; catit < NumCats; catit++)
				{					
					if (CheckBitInByteArray(catit, catpages[culli]) != 0)
					{
						CatCounter[catit]++;
						CatCounterFilled = 1;
					}
				}					
			}

			if (IsFiltered == false && oline < cullpages_count)
			{
				// copy if not filtered out
				output[oline].pageindex = culli;
												
				finalScale = ((res_table[culli].proximity / 255.0f) * proxScale) + baseScale;
				if (res_table[culli].terms > 1)	// multiword search
				{
					if (res_table[culli].terms <= 10)
						finalScale = pow(finalScale, (float)(res_table[culli].terms-1));
					else
					{
						finalScale = pow(finalScale, 10);
            			finalScale += res_table[culli].terms - 10;
					}
				}

				output[oline].score = (unsigned long) (res_table[culli].score * finalScale + 0.5f);	// +0.5 for rounding			
				output[oline].terms = res_table[culli].terms;
				output[oline].txtptr1 = res_table[culli].txtptr1;
				output[oline].txtptr2 = res_table[culli].txtptr2;
				output[oline].txtptr3 = res_table[culli].txtptr3;				
				oline++;
			}		

		}
	}

	// Note: 
	// matches is the total number of results found (including the ones that were culled out with the histogram)
	// output_count is just the number of entries in the output array
	unsigned long output_count = oline;	
	
	PrintUserDebug("Sorting results...");
	
	//Sort results in order of score, use the "SortCompare" function		
	if (matches > 1)
	{				
		if (UseDateTime == 1 && Sort == 1)
			myqsort((void*)output, output_count, sizeof(output_table_item), &zoomSearch::SortByDate, this);	// sort by date
		else if (UseDateTime == 1 && Sort == 2)
			myqsort((void*)output, output_count, sizeof(output_table_item), &zoomSearch::SortByDateAsc, this);	// sort by date ascending
		else
			myqsort((void*)output, output_count, sizeof(output_table_item), &zoomSearch::SortCompare, this);	// default sort by relevance					
	}	

	PrintUserDebug("Sorted results.");

	char tmpParams[MAX_QUERY_LENGTH];
	strcpy(metaParams, "");
	if (UseMetaFields == 1)
	{
		for (int fieldnum = 0; fieldnum < NumMetaFields; fieldnum++)
		{
			if (metafields[fieldnum].type == METAFIELD_TYPE_MULTI)
			{
				int num_multi_query = (int)meta_query[fieldnum][0];
				for (int mqi = 1; mqi <= num_multi_query; mqi++)
				{					
					sprintf(tmpParams, "&amp;%s[]=%d", metafields[fieldnum].name, meta_query[fieldnum][mqi]);
					strcat(metaParams, tmpParams);
				}
			}
			else
			{
				if (strlen(meta_query[fieldnum]) > 0)
				{
					sprintf(tmpParams, "&amp;%s=%s", metafields[fieldnum].name, meta_query[fieldnum]);
					strcat(metaParams, tmpParams);
				}
			}						
		}
	}

	PrintUserDebug("Display search results ...");

	//Display search result information	
	if (UseXML)
	{
		zoomprintf("<opensearch:totalResults>%d</opensearch:totalResults>\n", matches);
		int startIndex = 1;
		if (Page > 1 && matches > 0)
			startIndex = ((Page - 1) * Per_Page) + 1;
		zoomprintf("<opensearch:startIndex>%d</opensearch:startIndex>\n", startIndex);

		/*
		int itemsOnThisPage = Per_Page;
		if ((unsigned long)(startIndex+Per_Page) > matches)
			itemsOnThisPage = matches - startIndex;
		*/
		zoomprintf("<opensearch:itemsPerPage>%d</opensearch:itemsPerPage>\n", Per_Page);
	}
	else
	{
		OutputBufferPrintf(OUTPUT_SUMMARY, "<div class=\"summary\">\n");
		if (IsMaxLimitExceeded)
			OutputBufferPrintf(OUTPUT_SUMMARY, "%s<br /><br />", StrTable[STR_PHRASE_CONTAINS_COMMON_WORDS]);

		if (matches == 0)
			OutputBufferPrintf(OUTPUT_SUMMARY, "%s", StrTable[STR_SUMMARY_NO_RESULTS_FOUND]);
		else if (NumSearchWords > 1 && And == 0)
		{
			//OR
			int SomeTermMatches = matches - fullmatches;
			PrintNumResults(fullmatches, StrTable[STR_SUMMARY_FOUND_CONTAINING_ALL_TERMS]);		
			if (SomeTermMatches > 0)
				PrintNumResults(SomeTermMatches, StrTable[STR_SUMMARY_FOUND_CONTAINING_SOME_TERMS]);
		}
		else if (NumSearchWords > 1 && And == 1) //AND
			PrintNumResults(fullmatches, StrTable[STR_SUMMARY_FOUND_CONTAINING_ALL_TERMS]);
		else
			PrintNumResults(matches, StrTable[STR_SUMMARY_FOUND]);

		OutputBufferPrintf(OUTPUT_SUMMARY, "<br />\n</div>\n");		
	}

	if (matches < 3) 
	{
		if (And == 1 && NumSearchWords > 1)
		{			
#if defined(ASPX)			
			if (DotNetUsePostBacks == 1)
			{
				snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_per_page=%d%s%s&amp;zoom_xml=%d&amp;zoom_and=0&amp;zoom_sort=%d", QueryForHTML, metaParams, Per_Page, query_zoom_cats, dateRangeParams, UseXML, Sort);
				snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
			}
			else
#endif
			{
				snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_per_page=%d%s%s&amp;zoom_xml=%d&amp;zoom_and=0&amp;zoom_sort=%d", QueryForURL, metaParams, Per_Page, query_zoom_cats, dateRangeParams, UseXML, Sort);
				snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
			}

			OutputBufferPrintf(OUTPUT_SUGGESTION, "<div class=\"suggestion\"><br />%s <a href=\"%s\">%s</a>.</div>", StrTable[STR_POSSIBLY_GET_MORE_RESULTS], tmpLinkBuffer, StrTable[STR_ANY_OF_TERMS]);
		}
		else if (UseCats && Cat[0] != -1)
		{			
#if defined(ASPX)
			if (DotNetUsePostBacks == 1)
			{
				snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_per_page=%d&amp;zoom_cat=-1%s&amp;zoom_xml=%d&amp;zoom_and=%d&amp;zoom_sort=%d", QueryForHTML, metaParams, Per_Page, dateRangeParams, UseXML, And, Sort);
				snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
			}
			else
#endif
			{
				snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_per_page=%d&amp;zoom_cat=-1%s&amp;zoom_xml=%d&amp;zoom_and=%d&amp;zoom_sort=%d", QueryForURL, metaParams, Per_Page, dateRangeParams, UseXML, And, Sort);
				snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
			}

			OutputBufferPrintf(OUTPUT_SUGGESTION, "<div class=\"suggestion\"><br />%s <a href=\"%s\">%s</a>.</div>", StrTable[STR_POSSIBLY_GET_MORE_RESULTS], tmpLinkBuffer, StrTable[STR_ALL_CATS]);
		}
	}

	// Show category summary (no XML output)
	if (UseCats == 1 && DisplayCatSummary == 1 && CatCounterFilled == 1)
	{
		if (UseXML == 0)
			OutputBufferPrintf(OUTPUT_CATSUMMARY, "<div class=\"cat_summary\"><br />%s<ul>",StrTable[STR_CAT_SUMMARY]);

		int catSummaryItemCount = 0;

		for (int catit = 0; catit < NumCats; catit++)
		{
			if (CatCounter[catit] > 0)
			{
				if (CatCounter[catit] != matches)
				{
					catSummaryItemCount++;
					if (UseXML == 1)
					{
						zoomprintf("<zoom:categorySummary>\n");
						zoomprintf("<zoom:categoryName>%s</zoom:categoryName>\n", CatNames[catit]);
						zoomprintf("<zoom:categoryCount>%d</zoom:categoryCount>\n", CatCounter[catit]);
						zoomprintf("</zoom:categorySummary>\n");
					}
					else
					{						
#if defined(ASPX)
						if (DotNetUsePostBacks == 1)
						{
							snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_cat=%d%s&amp;zoom_per_page=%d&amp;zoom_xml=%d&amp;zoom_and=%d&amp;zoom_sort=%d", QueryForHTML, metaParams, catit, dateRangeParams, Per_Page, UseXML, And, Sort);						
							snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
						}
						else
#endif
						{
							snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_cat=%d%s&amp;zoom_per_page=%d&amp;zoom_xml=%d&amp;zoom_and=%d&amp;zoom_sort=%d", QueryForURL, metaParams, catit, dateRangeParams, Per_Page, UseXML, And, Sort);						
							snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
						}

						OutputBufferPrintf(OUTPUT_CATSUMMARY, "<li><a href=\"%s\">%s</a>", tmpLinkBuffer, CatNames[catit]);
						OutputBufferPrintf(OUTPUT_CATSUMMARY, " (%d)</li>", CatCounter[catit]);						
					}
				}
			}
		}

		if (catSummaryItemCount == 0)
		{
			// Clear cat summary if everything belongs in one category afterall			
			OutputBufferClear(OUTPUT_CATSUMMARY);
		}
		else if (UseXML == 0)
			OutputBufferPrintf(OUTPUT_CATSUMMARY, "</ul></div>");
	}

	if (CatCounter)
	{
		free(CatCounter);
		CatCounter = NULL;
	}
	
	// Prevent loading in entire suggestions file even when we dont need it.
	if (Spelling == 1)
	{
		Spelling = 0;
		for (unsigned long sw = 0; sw < NumSearchWords; sw++)
		{						
			if (sw_results[sw] < SpellingWhenLessThan)
			{				
				// this word needs suggestions
				Spelling = 1;
				break;
			}
		}
	}

	if (Spelling == 1 && Spell != NULL)
	{
		char* sw_spcode;		
		sw_spcode = (char*)calloc(SPCODE_LEN+1, sizeof(char));		
		
		char SuggestStr[MAX_QUERY_LENGTH*2];	// double'd because the suggested word could be longer for each word.
		strcpy(SuggestStr, "");

		bool SuggestionFound = false;
		int SuggestionCount = 0;

		char* word = NULL;
		char* word2 = NULL;
		char* word3 = NULL;

		char tmpWordStr[WORDLENGTH+1];

		for (unsigned long sw = 0; sw < NumSearchWords; sw++)
		{						
			if (sw_results[sw] >= SpellingWhenLessThan)
			{				
				// this word has enough results, and don't need suggestions
				if (sw > 0)
					strcat(SuggestStr, " ");
				strcat(SuggestStr, SearchWords[sw]);
			}
			else
			{		
				PrintUserDebug("Getting metaphone...");
				//if (GetSPCode(SearchWords[sw], sw_spcode) == true)

				// We are now storing SOUNDEX_LEN size buffers for the metaphone
				// However, we still only ask for 4 character keys. This is because the metaphone
				// function can return "KS" for certain characters, causing it to overflow. At the moment,
				// we are truncating the metaphone at SOUNDEX_LEN-1 character long
				if (GetMetaphone(SearchWords[sw], sw_spcode, SPCODE_LEN) == true)
				{				  
					PrintUserDebug("Got metaphone.");
					//zoomprintf("spcode is: %s<br>", sw_spcode);
					SuggestionFound = false;
					
					for (i = 0; i < NumSpellings && !SuggestionFound; i++)
					{						
						if (strcmp(Spell[i].spcode, sw_spcode) == 0)
						{				
							int j = 0;	// count the dictid's
							while (SuggestionFound == 0 && j < 3)
							{								
								word = NULL;
								switch (j)
								{
								case 0:
									if (Spell[i].id > 0)
										word = GetSpellingWord(Spell[i].id);
										//word = (char*)&Dict[Spell[i].id].word;
									break;
								case 1:
									if (Spell[i].id2 > 0)
										word = GetSpellingWord(Spell[i].id2);
										//word = (char*)&Dict[Spell[i].id2].word;
									break;
								case 2:
									if (Spell[i].id3 > 0)
										word = GetSpellingWord(Spell[i].id3);
										//word = (char*)&Dict[Spell[i].id3].word;
									break;
								}		
								
								if (word != NULL)
								{
									strcpy(tmpWordStr, word);
									ToLowerAndStem(tmpWordStr);
								}

								if (word == NULL || strcmp(SearchWords[sw], tmpWordStr) == 0)
								{																
									SuggestionFound = false;									
								}
								else
								{			
									PrintUserDebug("Suggestion found.");
									SuggestionFound = true;
									SuggestionCount++;									
									if (NumSearchWords == 1) // if single word search
									{
										if (j < 1 && Spell[i].id2 != 0)
										{		
											//word2 = (char*)&Dict[Spell[i].id2].word;
											word2 = GetSpellingWord(Spell[i].id2);
											if (word2 != NULL)
											{
												strcpy(tmpWordStr, word2);
												ToLowerAndStem(tmpWordStr);
												if (strcmp(SearchWords[sw], tmpWordStr) == 0)
													word2 = NULL;
											}
										}
										if (j < 2 && Spell[i].id3 != 0)
										{	
											//word3 = (char*)Dict[Spell[i].id3].word;
											word3 = GetSpellingWord(Spell[i].id3);
											if (word3 != NULL)
											{
												strcpy(tmpWordStr, word3);
												ToLowerAndStem(tmpWordStr);
												if (strcmp(SearchWords[sw], tmpWordStr) == 0)
													word3 = NULL;
											}
										}
									}
								}
								j++;
							}							
						}
						else if (strcmp(Spell[i].spcode, sw_spcode) > 0)
						{						
							break;				
						}
					}					

					if (SuggestionFound)				
					{											
						if (sw > 0)
							strcat(SuggestStr, " ");
						strcat(SuggestStr, word);					
					}
				}
			}
		}
		
		if (SuggestionCount > 0)
		{			
			PrintUserDebug("Displaying suggestions...");
			unsigned char suggest_out[MAX_QUERY_LENGTH*2];
			strcpy((char*)suggest_out, (char*)SuggestStr);
			strReplace(" ", "+", (char*)suggest_out, true, MAX_QUERY_LENGTH*2);				
			htmlspecialchars((char*)suggest_out, MAX_QUERY_LENGTH*2);	

			if (UseXML)
			{
				zoomprintf("<zoom:spelling>%s</zoom:spelling>\n", SuggestStr);
				if (word2 != NULL)
					zoomprintf("<zoom:spelling>%s</zoom:spelling>\n", word2);				
				if (word3 != NULL)
					zoomprintf("<zoom:spelling>%s</zoom:spelling>\n", word3);								
			}
			else
			{
				snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_per_page=%d%s%s&amp;zoom_and=0&amp;zoom_sort=%d", suggest_out, metaParams, Per_Page, query_zoom_cats, dateRangeParams, Sort);
				snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
#if defined(ASPX)
				if (DotNetUsePostBacks == 1)
					snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
#endif
				OutputBufferPrintf(OUTPUT_SUGGESTION, "<div class=\"suggestion\"><br />%s <a href=\"%s\">%s</a>", StrTable[STR_DIDYOUMEAN], tmpLinkBuffer, SuggestStr);
				if (word2 != NULL)
				{
					strcpy((char*)suggest_out, (char*)word2);
					strReplace(" ", "+", (char*)suggest_out, true, MAX_QUERY_LENGTH*2);				
					htmlspecialchars((char*)suggest_out, MAX_QUERY_LENGTH*2);

					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_sort=%d", suggest_out, metaParams, Per_Page, query_zoom_cats,dateRangeParams, And, Sort);
					snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
#if defined(ASPX)
					if (DotNetUsePostBacks == 1)
						snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
#endif
					OutputBufferPrintf(OUTPUT_SUGGESTION, " %s <a href=\"%s\">%s</a>", StrTable[STR_OR], tmpLinkBuffer, word2);	
				}
				if (word3 != NULL)
				{
					strcpy((char*)suggest_out, (char*)word3);
					strReplace(" ", "+", (char*)suggest_out, true, MAX_QUERY_LENGTH*2);				
					htmlspecialchars((char*)suggest_out, MAX_QUERY_LENGTH*2);				

					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_sort=%d", suggest_out, metaParams, Per_Page, query_zoom_cats, dateRangeParams, And, Sort);
					snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
#if defined(ASPX)
					if (DotNetUsePostBacks == 1)
						snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
#endif
					OutputBufferPrintf(OUTPUT_SUGGESTION, " %s <a href=\"%s\">%s</a>", StrTable[STR_OR], tmpLinkBuffer, word3);
				}						
				OutputBufferPrintf(OUTPUT_SUGGESTION, "?</div>");
			}
		}		
	
		if (Spell != NULL)
		{
			for (i = 0; i < NumSpellings; i++)
			{
				if (Spell[i].spcode)
					free(Spell[i].spcode);
			}
			free(Spell);
		}

		if (sw_spcode != NULL)
			free(sw_spcode);

		PrintUserDebug("Suggestions completed.");
	}

	if (sw_results != NULL)
		free(sw_results);

	unsigned long arrayline=0;
	
	// Number of pages of results
	unsigned long NumPagesResults = (long)ceil((double)matches / (double)Per_Page);	
	if (NumPagesResults > 1)		
	{
		if (!UseXML)
			OutputBufferPrintf(OUTPUT_PAGESCOUNT, "<div class=\"result_pagescount\"><br />%ld %s</div>\n", NumPagesResults, StrTable[STR_PAGES_OF_RESULTS]);
	}	
	
	PrintUserDebug("Calculated %d number of pages of results.", NumPagesResults);

	char urlLink[1024+MAX_QUERY_LENGTH] = {0};
	char url[URLLENGTH+1] = {0};
	char title[TITLELEN+1] = {0};
	char description[MAXDESCLEN+1] = {0};
	char image[URLLENGTH+1] = {0};

	// Show sorting options
	if (matches > 1 && UseXML == 0)
	{
		if (UseDateTime == 1)
		{
			OutputBufferPrintf(OUTPUT_SORTING, "<div class=\"sorting\">");
			if (Sort > 0)
			{				
#if defined(ASPX)
				if (DotNetUsePostBacks == 1)
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_xml=%d&amp;zoom_sort=0", QueryForHTML, metaParams,Page,Per_Page,query_zoom_cats, dateRangeParams, And, UseXML);
					snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
				}
				else
#endif
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_xml=%d&amp;zoom_sort=0", QueryForURL, metaParams,Page,Per_Page,query_zoom_cats, dateRangeParams, And, UseXML);
					snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
				}

				OutputBufferPrintf(OUTPUT_SORTING, "<a href=\"%s\">%s</a> / ", tmpLinkBuffer, StrTable[STR_SORTBY_RELEVANCE]);
								
#if defined(ASPX)
				if (DotNetUsePostBacks == 1)
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_xml=%d&amp;zoom_sort=%d", QueryForHTML, metaParams,Page,Per_Page,query_zoom_cats, dateRangeParams, And, UseXML, Sort+1);					
					snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
				}
				else
#endif
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_xml=%d&amp;zoom_sort=%d", QueryForURL, metaParams,Page,Per_Page,query_zoom_cats, dateRangeParams, And, UseXML, Sort+1);					
					snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
				}

				if (Sort == 1)
					OutputBufferPrintf(OUTPUT_SORTING, "<a href=\"%s\">%s</a>", tmpLinkBuffer, StrTable[STR_SORTEDBY_DATE]);
				else
					OutputBufferPrintf(OUTPUT_SORTING, "<a href=\"%s\">%s</a>", tmpLinkBuffer, StrTable[STR_SORTEDBY_DATE_ASC]);
			}
			else
			{				
#if defined(ASPX)
				if (DotNetUsePostBacks == 1)
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_xml=%d&amp;zoom_sort=1", QueryForHTML,metaParams,Page,Per_Page,query_zoom_cats,dateRangeParams,And, UseXML);
					//snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack('zoom_postback', '%s')", tmpGetParamBuffer);
					snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
				}
				else
#endif
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_xml=%d&amp;zoom_sort=1", QueryForURL,metaParams,Page,Per_Page,query_zoom_cats,dateRangeParams,And, UseXML);
					snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
				}
				OutputBufferPrintf(OUTPUT_SORTING, "<b>%s</b> / <a href=\"%s\">%s</a>", StrTable[STR_SORTEDBY_RELEVANCE], tmpLinkBuffer, StrTable[STR_SORTBY_DATE]);
			}
			OutputBufferPrintf(OUTPUT_SORTING, "</div>\n");
		}		
	}
	
	// Determine current line of result from the $output array
	if (Page == 1)
	{
		arrayline = 0;
	} 
	else 
	{
		arrayline = (Page - 1) * Per_Page;
	}
	
	// The last result to show on this page
	unsigned long result_limit = arrayline + Per_Page;
	
	PrintUserDebug("Rendering results...");		
	
	char cssFieldName[METAFIELD_NAME_LEN+50];
	char cssValueName[METAFIELD_NAME_LEN+50];


	
	zoomMetaInfo *metaInfoList = NULL;
	if (NumMetaFields > 0)
		metaInfoList = (zoomMetaInfo *)malloc(NumMetaFields * sizeof(zoomMetaInfo));
	
	// Display the results	
	while (arrayline < output_count && arrayline < result_limit) 
	{	
		zoomResultInfo resultInfo;
		memset(&resultInfo, 0, sizeof(resultInfo));

		unsigned long ipage = output[arrayline].pageindex;
		unsigned long score = output[arrayline].score;		
		unsigned char filetype = GetPageInfo_Filetype(pageinfo[ipage], UseBigPageInfoData);
		__time32_t filetime = GetPageInfo_Filetime(pageinfo[ipage], UseBigPageInfoData);
		ULONG32 filesize = GetPageInfo_Filesize(pageinfo[ipage], UseBigPageInfoData);
		
		if (GetPageData(ipage, (char*)&url, (char*)&title, (char*)&description, (char*)&image) == false)
		{
			if (metaInfoList) free(metaInfoList);
			if (UseXML) zoomprintf("<zoom:error>");
			zoomprintf("Error: Corrupted or invalid index files. Please re-index and try again.");
			if (UseXML) zoomprintf("</zoom:error>\n");			
			ShowTemplate();
			cleanup();
			return 0;
		}

		strcpy(urlLink, url);

		// Print the result link
		if (GotoHighlight == 1)
		{
			if (SearchAsSubstring == 1)
				AddParamToURL((char*)&urlLink, url, "zoom_highlightsub=", (char*)QueryForURL);				
			else
				AddParamToURL((char*)&urlLink, url, "zoom_highlight=", (char*)QueryForURL);				
		}
		else
			strcpy(urlLink, url);

		if (PdfHighlight == 1)
		{
			if (iStrStr(urlLink, ".pdf") != NULL)
			{
				char tmpQuery[MAX_QUERY_LENGTH];
				strcpy(tmpQuery, (char*)Query);
				strReplace("\"", "", (char*)tmpQuery, false, MAX_QUERY_LENGTH);
				sprintf(urlLink, "%s#search=%%22%s%%22", url, (char*)tmpQuery);
			}
		}

		if (UseXML)
		{
			zoomprintf("<item>\n");
		}
		else
		{
			if (arrayline % 2 == 0)
				ResultsBufferPrintf("<div class=\"result_block\">\n");
			else
				ResultsBufferPrintf("<div class=\"result_altblock\">\n");		
		}
		
		if (PluginOpenNewWindow == 1 && filetype >= FileTypePluginStart)
			strcpy(target, "target=\"_blank\"");
		else
			strcpy(target, zoom_target);

		if (UseZoomImage == 1)
		{
			if (strlen(image) > 0)
			{
				if (UseXML)
					zoomprintf("<zoom:imageURL>%s</zoom:imageURL>", image);
				else
				{
					ResultsBufferPrintf("<div class=\"result_image\">");
					ResultsBufferPrintf("<a href=\"%s\" %s>", urlLink, target);
					ResultsBufferPrintf("<img src=\"%s\" alt=\"\" class=\"result_image\" />", image);
					ResultsBufferPrintf("</a></div>");
				}
			}
		}
		
		strcpy(resultInfo.szTitle,title);
		strcpy(resultInfo.szLink,urlLink);
		resultInfo.filetype = filetype;

		if (UseXML)
		{
			zoomprintf("<title>%s</title>\n", title);
			zoomprintf("<link>%s</link>\n", urlLink);			
			if (UseCats)
			{			
				for (int cati = 0; cati < NumCats; cati++)
				{										
					//if ((pageinfo[ipage]->catnumber & (1 << cati)) != 0)
					if (CheckBitInByteArray(cati, catpages[ipage]) != 0)
						zoomprintf("<category>%s</category>\n", CatNames[cati]);						
				}
			}
			zoomprintf("<zoom:filetype>%d</zoom:filetype>\n", filetype);
		}
		else
		{
			ResultsBufferPrintf("<div class=\"result_title\">");
			if (DisplayNumber == 1)
				ResultsBufferPrintf("<b>%d.</b>&nbsp;", arrayline+1);		
			ResultsBufferPrintf("<a href=\"%s\" %s>", urlLink, target);
			if (DisplayTitle == 1)
				PrintHighlightDescription(title);
			else
				ResultsBufferPrintf("%s", url);
			ResultsBufferPrintf("</a>");				
			// end link
			if (UseCats) 
			{
				ResultsBufferPrintf("<span class=\"category\">");				
				for (int cati = 0; cati < NumCats; cati++)
				{
					if (CheckBitInByteArray(cati, catpages[ipage]) != 0)
						ResultsBufferPrintf(" [%s]", CatNames[cati]);						
				}
				ResultsBufferPrintf("</span>");
			}
			ResultsBufferPrintf("</div>\n");
		}

		if (UseMetaFields == 1 && DisplayMetaFields == 1)
		{
			BYTE* metaDataPtr = metapages[ipage];

			resultInfo.iNumMetaFields = NumMetaFields;
			resultInfo.metaList = metaInfoList;
			memset(resultInfo.metaList, 0, sizeof(zoomMetaInfo) * NumMetaFields);

			for (int fieldnum = 0; fieldnum < NumMetaFields; fieldnum++)
			{
				sprintf(cssFieldName, "result_metaname_%s", metafields[fieldnum].name);
				sprintf(cssValueName, "result_metavalue_%s", metafields[fieldnum].name);
				
				if (metafields[fieldnum].type == METAFIELD_TYPE_DROPDOWN)
				{
					BYTE metaNumValue = 0;					
					metaNumValue = metaDataPtr[0];					
					if (metaNumValue != METAFIELD_DROPDOWN_NOVALUE)
					{
						char** ddvalues = metafields[fieldnum].DropdownValues;

						strcpy(resultInfo.metaList[fieldnum].szName, metafields[fieldnum].name);
						strcpy(resultInfo.metaList[fieldnum].szValue, ddvalues[metaNumValue]);

						if (UseXML)
						{
							zoomprintf("<zoom:%s>%s</zoom:%s>\n", metafields[fieldnum].name, ddvalues[metaNumValue], metafields[fieldnum].name);
						}
						else
						{
							ResultsBufferPrintf("<div class=\"result_custommeta\">");
							ResultsBufferPrintf("<span class=\"%s\">%s: </span>", cssFieldName, metafields[fieldnum].showname);
							ResultsBufferPrintf("<span class=\"%s\">%s</span>", cssValueName, ddvalues[metaNumValue]);
							ResultsBufferPrintf("</div>");
						}
					}
					metaDataPtr += sizeof(BYTE);
				}
				else if (metafields[fieldnum].type == METAFIELD_TYPE_TEXT)
				{					
					unsigned short metaStrLen = 0;
					char tmpStrBuffer[METAFIELD_TEXT_LEN+1];					
					//metaStrLen = metaDataPtr[0];
					//metaDataPtr += sizeof(BYTE);
					if (UseBigPageInfoData)
					{
						memcpy(&metaStrLen, metaDataPtr, METAFIELD_TEXT_SIZE_BYTES);
						metaDataPtr += METAFIELD_TEXT_SIZE_BYTES;
					}
					else
					{
						memcpy(&metaStrLen, metaDataPtr, METAFIELD_OLD_TEXT_SIZE_BYTES);
						metaDataPtr += METAFIELD_OLD_TEXT_SIZE_BYTES;
					}
						
					if (metaStrLen != 0) // && metaStrLen < METAFIELD_TEXT_LEN
					{
						memcpy(tmpStrBuffer, metaDataPtr, sizeof(char)*metaStrLen);
						tmpStrBuffer[metaStrLen] = '\0';

						strcpy(resultInfo.metaList[fieldnum].szName, metafields[fieldnum].name);
						strcpy(resultInfo.metaList[fieldnum].szValue, tmpStrBuffer);

						if (UseXML)
						{
							zoomprintf("<zoom:%s>%s</zoom:%s>\n", metafields[fieldnum].name, tmpStrBuffer, metafields[fieldnum].name);
						}
						else
						{
							ResultsBufferPrintf("<div class=\"result_custommeta\">");
							ResultsBufferPrintf("<span class=\"%s\">%s: </span>", cssFieldName, metafields[fieldnum].showname);
							ResultsBufferPrintf("<span class=\"%s\">%s</span>", cssValueName, tmpStrBuffer);
							ResultsBufferPrintf("</div>");
						}
						metaDataPtr += sizeof(char) * metaStrLen;					
					}
				}
				else if (metafields[fieldnum].type == METAFIELD_TYPE_MULTI)
				{					
					BYTE metaValCount = 0;
					int metaNumValue = 0;
					metaValCount = metaDataPtr[0];					
					metaDataPtr += sizeof(BYTE);
					if (metaValCount > 0)
					{
						if (!UseXML)
						{
							ResultsBufferPrintf("<div class=\"result_custommeta\">");
							ResultsBufferPrintf("<span class=\"%s\">%s: </span>", cssFieldName, metafields[fieldnum].showname);					
							ResultsBufferPrintf("<span class=\"%s\">", cssValueName);
						}

						for (int mvi = 0; mvi < (int)metaValCount; mvi++)
						{
							metaNumValue = (int) metaDataPtr[mvi];							
							if (metaNumValue >= 0 && metaNumValue < metafields[fieldnum].DropdownCount)
							{
								strcpy(resultInfo.metaList[fieldnum].szName, metafields[fieldnum].name);
								strcpy(resultInfo.metaList[fieldnum].szValue, metafields[fieldnum].DropdownValues[metaNumValue]);

								if (UseXML)
								{
									zoomprintf("<zoom:%s>%s</zoom:%s>\n", metafields[fieldnum].name, metafields[fieldnum].DropdownValues[metaNumValue], metafields[fieldnum].name);
								}
								else
								{
									if (mvi > 0)
										ResultsBufferPrintf(", ");	
									ResultsBufferPrintf("%s", metafields[fieldnum].DropdownValues[metaNumValue]);
								}
							}													
						}

						if (!UseXML)
						{
							ResultsBufferPrintf("</span>");
							ResultsBufferPrintf("</div>");
						}
						metaDataPtr += sizeof(BYTE) * metaValCount;
					}					
				}
				else if (metafields[fieldnum].type == METAFIELD_TYPE_MONEY)
				{
					unsigned int metaNumValue = 0;
					endianAware_memcpy(&metaNumValue, metaDataPtr, sizeof(unsigned int));
					metaDataPtr += sizeof(unsigned int);
					if (metaNumValue != METAFIELD_NOVALUE_MARKER)
					{						
						if (!UseXML)
						{
							ResultsBufferPrintf("<div class=\"result_custommeta\">");
							ResultsBufferPrintf("<span class=\"%s\">%s: </span>", cssFieldName, metafields[fieldnum].showname);					
						}

						if (MetaMoneyShowDec)
						{
							float tmpFloatVal = (float)metaNumValue;
							tmpFloatVal = tmpFloatVal / 100.0f;

							strcpy(resultInfo.metaList[fieldnum].szName, metafields[fieldnum].name);
							sprintf(resultInfo.metaList[fieldnum].szValue, "%s%01.2f", MetaMoneyCurrency, tmpFloatVal);

							if (UseXML)
								zoomprintf("<zoom:%s>%s%01.2f</zoom:%s>\n", metafields[fieldnum].name, MetaMoneyCurrency, tmpFloatVal, metafields[fieldnum].name);
							else
								ResultsBufferPrintf("<span class=\"%s\">%s%01.2f</span>", cssValueName, MetaMoneyCurrency, tmpFloatVal);
						}
						else
						{
							strcpy(resultInfo.metaList[fieldnum].szName, metafields[fieldnum].name);
							sprintf(resultInfo.metaList[fieldnum].szValue, "%s%d", MetaMoneyCurrency, metaNumValue);

							if (UseXML)
								zoomprintf("<zoom:%s>%s%d</zoom:%s>\n", metafields[fieldnum].name, MetaMoneyCurrency, metaNumValue, metafields[fieldnum].name);
							else
								ResultsBufferPrintf("<span class=\"%s\">%s%d</span>", cssValueName, MetaMoneyCurrency, metaNumValue);
						}

						if (!UseXML)
							ResultsBufferPrintf("</div>");
					}								
				}
				else
				{
					unsigned int metaNumValue = 0;
					endianAware_memcpy(&metaNumValue, metaDataPtr, sizeof(unsigned int));
					metaDataPtr += sizeof(unsigned int);
					if (metaNumValue != METAFIELD_NOVALUE_MARKER)
					{					
						strcpy(resultInfo.metaList[fieldnum].szName, metafields[fieldnum].name);
						sprintf(resultInfo.metaList[fieldnum].szValue, "%d", metaNumValue);

						if (UseXML)
						{
							zoomprintf("<zoom:%s>%d</zoom:%s>\n", metafields[fieldnum].name, metaNumValue, metafields[fieldnum].name);
						}
						else
						{
							ResultsBufferPrintf("<div class=\"result_custommeta\">");
							ResultsBufferPrintf("<span class=\"%s\">%s: </span>", cssFieldName, metafields[fieldnum].showname);					
							ResultsBufferPrintf("<span class=\"%s\">%d</span>", cssValueName, metaNumValue);						
							ResultsBufferPrintf("</div>");
						}
					}									
				}				
			}			
		}

		if (DisplayMetaDesc == 1)
		{
			// Print meta description
			if (strlen(description) > 2)
			{
				strcpy(resultInfo.szDescription, description);

				if (UseXML)
					zoomprintf("<description>%s</description>\n", description);
				else
				{
					ResultsBufferPrintf("<div class=\"description\">");
					PrintHighlightDescription(description);
					ResultsBufferPrintf("</div>\n");
				}
			}
		}

		// Show no context if we are doing a blank search with only meta fields specified (IsNoKeywordQuery)
		// because we won't know what word to look for, and we have no txtptr (they're all set to zero)
		if (DisplayContext == 1 && IsNoKeywordQuery == false)
		{
			PrintUserDebug("Extracting contextual page content...");
						
			// Extract contextual page content
			unsigned long context_keywords = output[arrayline].terms; // # of terms matched
			if (context_keywords > MaxContextKeywords)
				context_keywords = MaxContextKeywords;
			
			context_word_count = (int) ceil((double)ContextSize / (double)context_keywords);

			contextArray = (context_elem*) malloc(sizeof(context_elem) * context_word_count);
			highlightArray = NULL;				
			if (Highlighting == 1)
				highlightArray = (BYTE*) malloc(sizeof(BYTE) * context_word_count);
			
			unsigned int goback  = context_word_count / 2; //removed floor()
			
			unsigned int gobackbytes = goback * DictIDLen;			

			__int64 last_startpos = 0;
			__int64 first_startpos = 0;
			__int64 last_endpos = 0;
			__int64 last_bytesread = 0;
			
			bool noGoBack = false;

			bool FoundContext = false;
			bool IsEndOfContent = false;

			if (UseXML)
				zoomprintf("<zoom:context>");
			else
				ResultsBufferPrintf("<div class=\"context\">\n");

			PrintUserDebug("Entering context keywords loop...");

			//for (unsigned long j = 0; j < context_keywords && (j == 0 || !_eof(fp_pagetext)); j++)
			for (unsigned long j = 0; IsEndOfContent == false && j < context_keywords && (j == 0 || !feof(fp_pagetext)); j++)
			{		
				__int64 origpos, startpos;
				
				if (j == 0)
					origpos = output[arrayline].txtptr1;
				else if (j == 1)
					origpos = output[arrayline].txtptr2;
				else
					origpos = output[arrayline].txtptr3;
				
				startpos = origpos;
				
				if (gobackbytes < startpos)
				{
					startpos = startpos - gobackbytes;
					noGoBack = false;
				}
				else
					noGoBack = true;

				//if (startpos < 0)
				//	startpos = 0;
				
				// Check that this will not overlap with previous extract (check also with first extract!)
				if ((startpos > last_startpos || startpos > first_startpos) && startpos < last_endpos)
					startpos = last_endpos;   // we will just continue last extract if so.	
				
				// find the pagetext pointed to				
				myfseek(fp_pagetext, startpos);
				//fseek(fp_pagetext, startpos, 0);	// why were we using this instead of myfseek() before 24/Nov/2008?

				GetNextDictWordBuffCount = 0;
				GetNextDictWordBuffRead = 0;
				
				// remember the last start position
				last_startpos = startpos;				
				last_bytesread = 0;
				if (j == 0)
					first_startpos = startpos;
								
				unsigned long word_id;
				unsigned int variant_index;
				variant_index = 0;
				bytesread = 0;
				bytesread += GetNextDictWord(fp_pagetext, &word_id, &variant_index);

				// Prepare context array of dict id				
				memset(contextArray, 0, sizeof(context_elem) * context_word_count);
								
				// Determine highlight array corresponding to context array
				if (Highlighting == 1)
					memset(highlightArray, 0, sizeof(BYTE) * context_word_count);

				for (unsigned int cti = 0; cti < context_word_count; cti++)
				{
					// check if end of page, or end of section, or invalid word_id
					if (word_id == 0 || word_id == 1 || word_id > NumKeywords)
					{   						
						// check if end of page						
						// if end of page occurs AFTER word pointer (ie: reached next page)						
						//if (noGoBack || (ftell(fp_pagetext)) > origpos)
						if (noGoBack || (startpos+bytesread) > origpos)
						{			
							IsEndOfContent = true;
							break;          // then we stop.
						}
						else                // if end of page occurs BEFORE word pointer (ie: reached previous page)
						{
							// then we clear the existing context buffer we've created.
							memset(contextArray, 0, sizeof(context_elem) * context_word_count);
							cti = 0;							
						}
					}
					else
					{
						if (word_id >= NumKeywords)
						{
							ResultsBufferPrintf("Critical error with pagetext file. Check that your files are from the same indexing session.");
						}
						else
						{							
							if (Highlighting == 1 && IsNoKeywordQuery == false && (startpos+last_bytesread) == origpos)
								highlightArray[cti] = 1;						

							contextArray[cti].dict_id = word_id;
							contextArray[cti].variant_index = variant_index;
						}
					}						
					last_bytesread = bytesread;
					bytesread += GetNextDictWord(fp_pagetext, &word_id, &variant_index);
				}
				if (word_id != 0)
					last_endpos = last_startpos+bytesread;
				
				// Work out highlighting based on string/wildcard/exact phrase matching		
				if (Highlighting == 1)
				{
					HighlightContextArray();
				}				

				// Create context output string based on the above arrays
				unsigned long prev_word_id = 0;

				char context_str[TEMP_STR_SIZE];
				memset(context_str, 0, sizeof(context_str)); //was = ""
				
				int noSpaceForNextChar = false;
				for (unsigned int cti = 0; cti < context_word_count; cti++)
				{
					word_id = contextArray[cti].dict_id;
					variant_index = contextArray[cti].variant_index;					

					if (noSpaceForNextChar == false)
					{						
		                if (word_id > DictReservedNoSpaces)
						{
							if (prev_word_id <= DictReservedPrefixes || prev_word_id > DictReservedNoSpaces)
								strcat(context_str, " ");
						}
						else if(word_id > DictReservedSuffixes && word_id <= DictReservedPrefixes)
						{
							// This is a prefix character
							strcat(context_str, " ");
							noSpaceForNextChar = true;
						}                	             
						else if(word_id > DictReservedPrefixes)	// this is a no-space character
							noSpaceForNextChar = true;
					}
					else
						noSpaceForNextChar = false;
	
					// check if end of page, or end of section, or invalid word_id
					if (word_id > 0)
					{
						if (word_id >= NumKeywords)
						{
							//char msg[256];
							//sprintf(msg, "WRONG_HERE (%ld)", word_id);
							//strcat(context_str, msg);
							ResultsBufferPrintf("Critical error with pagetext file. Check that your files are from the same indexing session.");
						}
						else
						{
							if (Highlighting == 1 && (highlightArray[cti] == HIGHLIGHT_SINGLE || highlightArray[cti] == HIGHLIGHT_START))
							{
								if (UseXML == 1)
								{
									if (XMLHighlight == 1)
										strcat(context_str, "<zoom:highlight>");
								}
								else
									strcat(context_str, "<span class=\"highlight\">");
							}

							//strcat(context_str, Dict[word_id].word);
							strcat(context_str, GetDictionaryWord(word_id, variant_index));

							if (Highlighting == 1 && (highlightArray[cti] == HIGHLIGHT_SINGLE || highlightArray[cti] == HIGHLIGHT_END))
							{
								if (UseXML == 1)
								{
									if (XMLHighlight == 1)
										strcat(context_str, "</zoom:highlight>");
								}
								else
									strcat(context_str, "</span>");
							}
						}
					}
					prev_word_id = word_id;
				}

				if (FoundContext == false && context_str[0] != '\0')
				{
					Trim(context_str);
					// this is the first context str we've found, we'll check if it matches the title.
					if (strcmp(context_str, title) == 0)
						strcpy(context_str, "");
				}

				if (context_str[0] != '\0')
				{
					FoundContext = true;		
					//htmlspecialchars((char*)context_str, TEMP_STR_SIZE);
					strcat(resultInfo.szContext, " ... ");
					strcat(resultInfo.szContext, context_str);

					if (UseXML)
					{
						zoomprintf(" ... ");
						zoomprintf("%s", context_str);
					}
					else
					{
						ResultsBufferPrintf(" <b>...</b> ");
						//PrintHighlightDescription(context_str);
						ResultsBufferPrintf("%s", context_str);
					}
				}
			}	
			if (highlightArray)
				free(highlightArray);
			if (contextArray)
				free(contextArray);

			if (FoundContext)
			{
				strcat(resultInfo.szContext, " ...");

				if (UseXML)
					zoomprintf(" ...");	
				else
					ResultsBufferPrintf(" <b>...</b>");
			}
			
			if (UseXML)
				zoomprintf("</zoom:context>\n");
			else
				ResultsBufferPrintf("</div>\n");	

		}	// end of print contextual description				
	

		char tmp_info_str[TEMP_STR_SIZE];
		char info_str[TEMP_STR_SIZE*2];
		strcpy(info_str, "");
	
		if (DisplayTerms == 1)
		{
			resultInfo.iTermsMatched = output[arrayline].terms;

			if (UseXML)
				zoomprintf("<zoom:termsMatched>%d</zoom:termsMatched>\n", output[arrayline].terms);
			else
			{
				sprintf(tmp_info_str, "%s %d", StrTable[STR_RESULT_TERMS_MATCHED], output[arrayline].terms);
				strcat(info_str, tmp_info_str);
			}
		}    	

		if (DisplayScore == 1)
		{
			resultInfo.iScore = score;

			if (UseXML)
				zoomprintf("<zoom:score>%u</zoom:score>\n", score);
			else
			{
				sprintf(tmp_info_str, "%s %u", StrTable[STR_RESULT_SCORE], score);
				if (strlen(info_str) > 0)
					strcat(info_str, "&nbsp; - &nbsp;");
				strcat(info_str, tmp_info_str);
			}
		}
				
		if (DisplayDate == 1 && filetime > 0)
		{
			time_t tt = filetime;
			//struct tm* tmptime = localtime(&tt);
			struct tm* tmptime = gmtime(&tt);			
			if (tmptime != NULL)
			{					
				sprintf(resultInfo.szPubDate, "%s, %02d %s %04d %02d:%02d:%02d UTC", DayOfWeekNames[tmptime->tm_wday], tmptime->tm_mday, MonthNames[tmptime->tm_mon], tmptime->tm_year + 1900, tmptime->tm_hour, tmptime->tm_min, tmptime->tm_sec);			

				if (UseXML)
				{
					sprintf(tmp_info_str, "%s, %02d %s %04d %02d:%02d:%02d UTC", DayOfWeekNames[tmptime->tm_wday], tmptime->tm_mday, MonthNames[tmptime->tm_mon], tmptime->tm_year + 1900, tmptime->tm_hour, tmptime->tm_min, tmptime->tm_sec);			
					zoomprintf("<pubDate>%s</pubDate>\n", tmp_info_str);
					if (ShowXMLSettings)
						zoomprintf("<zoom:datetime>%d</zoom:datetime>\n", filetime);
				}
				else
				{
					sprintf(tmp_info_str, "%02d %s %04d", tmptime->tm_mday, MonthNames[tmptime->tm_mon], tmptime->tm_year + 1900);			
					if (strlen(info_str) > 0)
						strcat(info_str, "&nbsp; - &nbsp;");
					strcat(info_str, tmp_info_str);				
				}
			}
		}		

		if (DisplayFilesize == 1)
		{
			if (UseXML)
				zoomprintf("<zoom:fileSize>%dk</zoom:fileSize>\n", filesize);
			else
			{
				long tmpFilesize = filesize / 1024;
				if (tmpFilesize == 0)
					tmpFilesize = 1;
				sprintf(tmp_info_str, "%dk", tmpFilesize);
				if (strlen(info_str) > 0)
					strcat(info_str, "&nbsp; - &nbsp;");
				strcat(info_str, tmp_info_str);
			}
		}
		
		if (DisplayURL == 1 && UseXML == 0)
		{			
			if (strlen(info_str) > 0)
				strcat(info_str, "&nbsp; - &nbsp;");
			if (TruncateShowURL > 0)
			{
				if (strlen(url) > (unsigned int)TruncateShowURL)
				{
					url[TruncateShowURL] = '\0';
					if (strlen(url)+3 < URLLENGTH)
						strcat(url, "...");
				}
			}
			sprintf(tmp_info_str, "%s %s", StrTable[STR_RESULT_URL], url);
			strcat(info_str, tmp_info_str);
		}

		if (UseXML == 0)
			ResultsBufferPrintf("<div class=\"infoline\">%s</div>\n", info_str);			

		if (UseXML)
			zoomprintf("</item>\n");
		else
			ResultsBufferPrintf("</div>\n");

#ifdef OSF
		OSFResultsCallback(&resultInfo);
#endif
		arrayline++;		
	}	

	if (metaInfoList) free(metaInfoList);
		
#if !defined(FASTCGI_ENABLED) && !defined(OSF)
	if (DisplayContext == 1 || AllowExactPhrase == 1)
	{		
		fclose(fp_pagetext);
	}

	if (IsPageDataPreloaded == false)
		fclose(fp_pagedata);
#endif
	

	if (UseXML == 0)
	{
		// Show links to other result pages
		PrintUserDebug("Show links to other pages...");
					
		/*
		unsigned char query_out[MAX_QUERY_LENGTH];
		strcpy((char*)query_out, (char*)Query);
		strReplace(" ", "+", (char*)query_out, true, MAX_QUERY_LENGTH);		
		strcpy((char*)test, (char*)query_out);
		htmlspecialchars((char*)test, MAX_QUERY_LENGTH);
		strcpy((char*)query_out, (char*)test);
		*/
		
		if(NumPagesResults > 1) 
		{
			// 10 results to the left of the current page
			int start_range = Page - 10;
			if (start_range < 1)
				start_range = 1;
			
			// 10 to the right
			unsigned int end_range = Page + 10;
			unsigned int max_end_range = MaxResultsPerQuery/Per_Page;

			if (end_range > NumPagesResults)
				end_range = NumPagesResults;
			if (end_range > max_end_range)
				end_range = max_end_range;// used to be max_end_range-1 but this omitted the last page?
			
			OutputBufferPrintf(OUTPUT_PAGENUMBERS, "<div class=\"result_pages\">%s ", StrTable[STR_RESULT_PAGES]);
			if (Page > 1)
			{				
#if defined(ASPX)
				if (DotNetUsePostBacks == 1)
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_sort=%d&amp;zoom_xml=%d", QueryForHTML, metaParams, Page-1,Per_Page,query_zoom_cats, dateRangeParams, And, Sort, UseXML);	
					snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
				}
				else
#endif
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_sort=%d&amp;zoom_xml=%d", QueryForURL, metaParams, Page-1,Per_Page,query_zoom_cats, dateRangeParams, And, Sort, UseXML);	
					snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
				}

				OutputBufferPrintf(OUTPUT_PAGENUMBERS, "<a href=\"%s\">&lt;&lt; %s</a> ", tmpLinkBuffer, StrTable[STR_RESULT_PAGES_PREVIOUS]);
			}
			for (i = start_range; i <= end_range; i++) 
			{
				if (i == Page) 
					OutputBufferPrintf(OUTPUT_PAGENUMBERS, "%d ", Page);
				else 
				{					
#if defined(ASPX)
					if (DotNetUsePostBacks == 1)
					{
						snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_sort=%d&amp;zoom_xml=%d", QueryForHTML, metaParams, i,Per_Page,query_zoom_cats,dateRangeParams,And, Sort, UseXML);
						snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
					}
					else
#endif
					{
						snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_sort=%d&amp;zoom_xml=%d", QueryForURL, metaParams, i,Per_Page,query_zoom_cats,dateRangeParams,And, Sort, UseXML);
						snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
					}
					OutputBufferPrintf(OUTPUT_PAGENUMBERS, "<a href=\"%s\">%d</a> ", tmpLinkBuffer, i);
				}
			}
			if (Page != end_range)
			{				
#if defined(ASPX)
				if (DotNetUsePostBacks == 1)
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_sort=%d&amp;zoom_xml=%d", QueryForHTML, metaParams, Page+1, Per_Page, query_zoom_cats, dateRangeParams, And, Sort, UseXML);
					snprintf(tmpLinkBuffer, URLLENGTH, "javascript:__doPostBack(&quot;zoom_postback&quot;, &quot;%s&quot;)", tmpGetParamBuffer);
				}
				else
#endif
				{
					snprintf(tmpGetParamBuffer, URLLENGTH, "zoom_query=%s%s&amp;zoom_page=%d&amp;zoom_per_page=%d%s%s&amp;zoom_and=%d&amp;zoom_sort=%d&amp;zoom_xml=%d", QueryForURL, metaParams, Page+1, Per_Page, query_zoom_cats, dateRangeParams, And, Sort, UseXML);
					snprintf(tmpLinkBuffer, URLLENGTH, "%s%s%s", ScriptName, LinkBackJoinChar, tmpGetParamBuffer);
				}

				OutputBufferPrintf(OUTPUT_PAGENUMBERS, "<a href=\"%s\">%s &gt;&gt;</a> ", tmpLinkBuffer, StrTable[STR_RESULT_PAGES_NEXT]);
			}
			OutputBufferPrintf(OUTPUT_PAGENUMBERS, "</div>");
		}
		//Let others know about Zoom.
		if (ZoomInfo == 1)
			OutputBufferPrintf(OUTPUT_PAGENUMBERS, "\n<center><p><small>%s <a href=\"http://www.wrensoft.com/zoom/\" target=\"_blank\"><b>Zoom Search Engine</b></a></small></p></center>\n", StrTable[STR_POWEREDBY]);
	}

	if (UseXML == 0)
		ResultsBufferPrintf("</div>\n");	// end of results style tag	
	
	//free output
	if (output != NULL)
		free(output);
	
	//free res_table
	if(res_table != NULL)
	{	
		free(res_table);
	}

	if (SearchPhrases != NULL)
	{
		for (unsigned long sw = 0; sw < NumSearchWords; sw++)
		{
			if (SearchPhrases[sw] != NULL)
			{
				for(int a=0; a<num_phrase_terms; a++)
				{
					if (SearchPhrases[sw][a] != NULL)
						free(SearchPhrases[sw][a]);
				}
				free(SearchPhrases[sw]);
			}
		}
		free(SearchPhrases);
	}

	if (SearchPhrasesForHL != NULL)
	{
		for (unsigned long sw = 0; sw < NumSearchWords; sw++)
		{
			if (SearchPhrasesForHL[sw] != NULL)
			{
				for(int a=0; a<num_phrase_terms; a++)
				{
					if (SearchPhrasesForHL[sw][a] != NULL)
						free(SearchPhrasesForHL[sw][a]);
				}
				free(SearchPhrasesForHL[sw]);
			}
		}
		free(SearchPhrasesForHL);
	}

	//calculte time difference
	double timeDifference = 0;
	if (Timing == 1 || Logging == 1) 
	{
		endtime = GetTimeNow();
		timeDifference = GetTimeDifference(starttime, endtime);
		if (Timing == 1)
		{
			if (UseXML)
				zoomprintf("<zoom:searchtime>%.3f</zoom:searchtime>\n", timeDifference);
			else
				OutputBufferPrintf(OUTPUT_SEARCHTIME, "<div class=\"searchtime\"><br /><br />%s %.3f %s</div>\n", StrTable[STR_SEARCH_TOOK], timeDifference, StrTable[STR_SECONDS]);		
		}
	}	
	
	//Log the search words, if required
	if (Logging == 1)
	{
		PrintUserDebug("Logging...");
		char LogQuery[MAX_QUERY_LENGTH];
		char LogString[MAX_QUERY_LENGTH+256]; //for extra date info etc
		strcpy(LogQuery, (char*)Query);
		if (strReplace("\"", "\"\"", LogQuery, false, MAX_QUERY_LENGTH) == true)
		{		
			//get time	
			struct tm *newtime;
			time_t aclock;	
			time( &aclock );              
			newtime = localtime( &aclock ); 
		
		
			sprintf(LogString,  "%d-%02d-%02d, %02d:%02d:%02d, %s, \"%s\", Matches = %d", newtime->tm_year+1900, newtime->tm_mon+1, newtime->tm_mday, newtime->tm_hour, newtime->tm_min, newtime->tm_sec, clientIP, LogQuery, matches);
			if (And == 1)
				strcat(LogString, ", AND");
			else
				strcat(LogString, ", OR");
		
			if (NewSearch == 1)
				Page = 0;
		
			sprintf(LogString, "%s, PerPage = %d, PageNum = %d",LogString, Per_Page, Page);
		
			if (UseCats == 0)
				strcat(LogString, ", No cats");
			else
			{
				if (Cat[0] == -1)
					strcat(LogString, ", \"Cat = All\"");
				else
				{				
					char logCatStr[MAX_QUERY_LENGTH+256];
					strcat(LogString, ", \"Cat = ");
					for (int cati = 0; cati < num_zoom_cats; cati++)
					{
						if (cati > 0)
							strcat(LogString, "; ");
						strcpy(logCatStr,CatNames[Cat[cati]]);
						strReplace("\"", "\"\"", logCatStr, true, MAX_QUERY_LENGTH+256);					
						strcat(LogString, logCatStr);
					}
					strcat(LogString, "\"");
				}
			}
			sprintf(LogString, "%s, Time = %.3f, Rec = %d", LogString, timeDifference, num_recs_found);
			// end of entry
			strcat(LogString, "\r\n");
		
			sprintf(tmpPath, "%s/%s", ExecutePath, LogFilename);
			FILE* logfp = fopen (tmpPath, "ab");
			if (logfp != false) 
			{
				fputs(LogString, logfp);
				fclose (logfp);
			}
			else
			{
				if (UseXML) zoomprintf("<zoom:error>");
				switch (errno)
				{
				case EACCES:
					zoomprintf("Permission denied. ");
					break;
				case EBADF:
					zoomprintf("Bad file number. ");
					break;
				case EINVAL:
					zoomprintf("Invalid argument. ");
					break;
				case EDEADLOCK:
					zoomprintf("Resource deadlock. ");
					break;
				case EEXIST:
					zoomprintf("File exists. ");
					break;
				case EMFILE:
					zoomprintf("Too many open files. ");
					break;
				case ENOENT:
					zoomprintf("No such file or directory. ");
					break;
				case ENOSPC:
					zoomprintf("No space left on device. ");
					break;
				case ENOMEM :
					zoomprintf("Not enough memory for operation. ");
					break;
				case EXDEV:
					zoomprintf("Cross device link. ");
					break;
				default:
					zoomprintf("Unknown file error: %d. ", errno);				
					break;
				}
				zoomprintf("Log filename and path not found (%s). Check that you specified the correct log filename in your Indexer settings.<br />", LogFilename);
				if (UseXML) zoomprintf("</zoom:error>\n");
			}            
		}
	}
	
	PrintUserDebug("Showing template...");
	
	//Print rest of result page template and close
	ShowTemplate();

	PrintUserDebug("Finished template.");
	
	//Closing Relation
	//You should close the relation before exit the program. 	
	/*------------------------------------------------------
	*                 Closing Relation.
	*------------------------------------------------------*/		
#if !defined(FASTCGI_ENABLED) && !defined(OSF)
	cleanup();
#endif
	return 0;
}


//Replacements for php function "preg_quote"
//takes a char and puts the delim char in front of it if it exists in the list
char* zoomSearch::preg_quote(char c, char delim)
{
	
	char list[] = {'.', '\\', '+', '*', '?', '[', '^', ']', '$', '(', ')', '{', '}', '=', '!', '<', '>', '|', ':', -1};   
	
	static char res[2];
	res[0] = c;
	res[1] = '\0';
	
	if(c==delim)
	{
		res[0]='\\';
		res[1]=c;
	}
	else
	{	
		for(int a=0; list[a] != -1; a++)
			if(c == list[a])
			{
				res[0]='\\';
				res[1]=c;
				break;
			}
	}
	
	return res;
}


// ----------------------------------------------------------------------------
// Translates a typical shell wildcard pattern ("zoo*" => "zoom" etc.)
// to a regular expression pattern. Supports only '*' and '?' characters.
// [UNDONE] UPDATED: Now changed to perform it in place. And added maxlength parameter
// to prevent overflow
// ----------------------------------------------------------------------------
char* zoomSearch::pattern2regexp(char* pattern)
//bool pattern2regexp(char* pattern, char* output, int maxlength)
{		
	int len;
	len = strlen(pattern);		
	
#define EXPANSION 10
	int i = 0;
	//need to increase size of string to allow for addition of extra charactera
	char* res = (char*)malloc(sizeof(char) * len * EXPANSION);			 
	memset(res, 0,len * EXPANSION); 
	
	if (strcspn(pattern, "$") != 0)
		strReplace("$", "\\$", pattern, false, len);
	if (strcspn(pattern, "#") != 0)
		strReplace("#", "\\#", pattern, false, len);
	
	while (i < len) 
	{
		char c = pattern[i];
		if (c == '*')
			strcat(res, "[\\d\\S]*");
		else if (c == '?')
			strcat(res, ".");
		else if (c == '.')
			strcat(res,  "\\.");
		else
			strcat(res, preg_quote(c, '/'));
		i++;
	}
	return res;

	/*
	// optimized alternative to the above, but this is limited to a maxlength*6 buffer.
	char c;	
	int destlen = 0;
	for (int i = 0; i < len; i++) 
	{		
		switch (pattern[i])
		{
		case '*':			
			destlen += 6;
			if (destlen > maxlength) return false;
			strcat(res, "[\\d\\S]*");			
			break;
		case '?':			
			strcat(res, ".");			
			break;
		case '.':			
			destlen += 1;
			if (destlen > maxlength) return false;
			strcat(res, "\\.");			
			break;
		case '$':			
			destlen += 1;
			if (destlen > maxlength) return false;
			strcat(res, "\\$");			
			break;
		case '#':		
			destlen += 1;
			strcat(res, "\\#");									
			break;
		default:
			if (strchr("/.\\+*?[^]$(){}=!<>|:", pattern[i]) != NULL)
			{
				destlen += 1;
				if (destlen > maxlength) return false;
				strcat(res, "\\");
			}
			strcat(res, pattern[i]);			
			break;
		}		
	}	
	return true; 			 
	*/	
}

int sw_compare(const void *arg1, const void *arg2 )
{	
    char* tmparg1 = *(char**)arg1;
	char* tmparg2 = *(char**)arg2;
	
	if(tmparg1[0] == '-')
		return 1;

	if (tmparg2[0] == '-')
		return -1;

	return 1;
}



// ----------------------------------------------------------------------------
// Compares the two values, used for sorting output results
// Results that match all search terms are put first, highest score
// The arguments being passed are effectively an array of 6 elements (the output array)
// only interested in elements [1] and [2]
// ----------------------------------------------------------------------------
int zoomSearch::SortCompare(const void *arg1, const void *arg2 )
{
	/*
	unsigned long **pstr1 = (unsigned long**)arg1, **pstr2 = (unsigned long**)arg2;
    unsigned long *tmparg1 = *pstr1, *tmparg2 = *pstr2;
	*/	
	output_table_item* tmparg1 = (output_table_item*)arg1;
	output_table_item* tmparg2 = (output_table_item*)arg2;


	/*
	if(tmparg1 == NULL)
		return 1;
	if(tmparg2 == NULL)
		return -1;
	*/
	
	/*
	// Previously sorting by terms matched, and then score.
	if(tmparg1[2] < tmparg2[2])
		return 1;
	else if(tmparg1[2] > tmparg2[2])
		return -1;
	else
	{
		if (tmparg1[1] < tmparg2[1])
			return 1;
        else if (tmparg1[1] > tmparg2[1])
			return -1;	
		return 0;
	}
	*/

	// Sort by score value only.
	if (tmparg1->score < tmparg2->score)
		return 1;
	else if (tmparg1->score > tmparg2->score)
		return -1;	
	return 0;
}

int zoomSearch::SortByDate(const void *arg1, const void *arg2 )
{
	/*
	unsigned long **pstr1 = (unsigned long**)arg1, **pstr2 = (unsigned long**)arg2;
    unsigned long *tmparg1 = *pstr1, *tmparg2 = *pstr2;
	*/
	output_table_item* tmparg1 = (output_table_item*)arg1;
	output_table_item* tmparg2 = (output_table_item*)arg2;
	
	/*
	if(tmparg1 == NULL)
		return 1;
	if(tmparg2 == NULL)
		return -1;
	*/
		
	__time32_t filetime1 = GetPageInfo_Filetime(pageinfo[tmparg1->pageindex], UseBigPageInfoData);
	__time32_t filetime2 = GetPageInfo_Filetime(pageinfo[tmparg2->pageindex], UseBigPageInfoData);
	if (filetime1 < filetime2)
		return 1;
	else if (filetime1 > filetime2)
		return -1;
	else
	{
		return SortCompare(arg1, arg2);
	}
}

int zoomSearch::SortByDateAsc(const void *arg1, const void *arg2 )
{
	output_table_item* tmparg1 = (output_table_item*)arg1;
	output_table_item* tmparg2 = (output_table_item*)arg2;

	__time32_t filetime1 = GetPageInfo_Filetime(pageinfo[tmparg1->pageindex], UseBigPageInfoData);
	__time32_t filetime2 = GetPageInfo_Filetime(pageinfo[tmparg2->pageindex], UseBigPageInfoData);
	if (filetime1 > filetime2)
		return 1;
	else if (filetime1 < filetime2)
		return -1;
	else
	{
		return SortCompare(arg1, arg2);
	}
}


//Get the next word id from the dictionary file
unsigned long zoomSearch::GetNextDictWord(FILE* fp_pagetext, unsigned long* dict_id, unsigned int* variant_index)
{	   			
	unsigned char bytes_buffer[MAXDICTIDLEN]={0};
	char variantByte = 0;	
	int bytesread;			

	int GetNextDictWordBuffLeft = GetNextDictWordBuffCount - GetNextDictWordBuffRead;

	if (GetNextDictWordBuffLeft < DictIDLen)
	{
		if (GetNextDictWordBuffLeft > 0)
		{
			// there are still some bytes left over

			// copy remaining bytes back to the front of the buffer
			memcpy(GetNextDictWordBuff, GetNextDictWordBuff+GetNextDictWordBuffRead, GetNextDictWordBuffLeft);
			// read in new bytes immediately following the old ones
			bytesread = fread(GetNextDictWordBuff+GetNextDictWordBuffLeft, 1, 128-GetNextDictWordBuffLeft, fp_pagetext);	
			GetNextDictWordBuffCount = bytesread+GetNextDictWordBuffLeft;
			GetNextDictWordBuffRead = 0;
		}
		else
		{
			// no bytes left over, we read more

			// first we check we're not at the end of the file
			if (feof(fp_pagetext))
			{
				*dict_id = 0;
				return 0;
			}

			bytesread = fread(GetNextDictWordBuff, 1, 128, fp_pagetext);
			GetNextDictWordBuffCount = bytesread;
			GetNextDictWordBuffRead = 0;		
		}
	}

	GetNextDictWordBuffLeft = GetNextDictWordBuffCount - GetNextDictWordBuffRead;
	if (GetNextDictWordBuffLeft < DictIDLen)
	{
		*dict_id = 0;
		return 0;
	}

	memcpy(bytes_buffer, GetNextDictWordBuff+GetNextDictWordBuffRead, DictIDLen);

	GetNextDictWordBuffRead += DictIDLen;

	// last byte is the variant
	*variant_index = (unsigned int) bytes_buffer[DictIDLen-1];

	// clear the last byte back
	bytes_buffer[DictIDLen - 1] = 0;

	memset(dict_id, 0, sizeof(unsigned long));

#if defined(NON_WINDOWS) && defined(SUN_SPARC_BUILT)
	//	Reverse the byte order for Sparc
	for (int di = 0; di < sizeof(dict_id); di++)
	{
		*dict_id <<= 8;
		*dict_id |= bytes_buffer[DictIDLen-1-di];		
	}
#else
	// For "normal" byte order
	memcpy(dict_id, bytes_buffer, sizeof(unsigned long));
#endif

	return DictIDLen;	
}

char* zoomSearch::FindVariant(char* variantPtr, int variant_index)
{
	char* retPtr = NULL;
	int varLen = 0;
	
	for (int i = 0; i < variant_index && i < NumVariants && variantPtr != NULL; i++)
	{
		varLen = variantPtr[0];
		variantPtr++;
		retPtr = variantPtr;
		variantPtr += varLen;		
	}	
	return retPtr;
}

char* zoomSearch::GetDictionaryWord(unsigned long dict_id, unsigned int variant_index)
{
	if (Dict[dict_id].variantPtr == NULL)
		return Dict[dict_id].wordPtr;	
	
	if (variant_index > 0 && variant_index <= Dict[dict_id].variantCount)
	{
		return FindVariant(Dict[dict_id].variantPtr, variant_index);
		//return Dict[dict_id].variantPtr[variant_index-1].variant;
	}

	return Dict[dict_id].wordPtr;
}

char* zoomSearch::GetSpellingWord(unsigned long dict_id)
{
	if (Dict[dict_id].variantCount > 0)
		return FindVariant(Dict[dict_id].variantPtr, 1);
	else
		return Dict[dict_id].wordPtr;
}

// Test if cahracter is a word char (letter, digit, '_')
// Return true if it is, false if it isnt
bool zoomSearch::isWordChar(char test)
{
	if((test>= 65 && test <= 90) //A-Z
		|| (test >= 97 && test <= 122) //a-z 
		|| (test >= 48 && test <= 57) //0-9
		|| test == 95)
		return true;
	else
		return false;
}

void zoomSearch::HighlightContextArray()
{
	bool bMatched;
	unsigned long word_id;
	int variant_index;
	for (unsigned int i = 0; i < context_word_count; i++)
	{
		bMatched = false;
		word_id = contextArray[i].dict_id;
		variant_index = contextArray[i].variant_index;
		for (unsigned long sw = 0; sw < NumSearchWords; sw++) 
		{						
			if (SearchPhrases != NULL && SearchPhrases[sw] != NULL)
			{
				unsigned long termNum = i;
				unsigned long pterm;

				// this is an exact phrase and has its phrase terms stored in SearchPhrases[sw]
				pterm = 0;
				//while (SearchPhrases[sw][pterm] != '\0' && termNum < context_word_count)
				while (phrase_terms_ids[sw][pterm] != 0 && termNum < context_word_count)
				{
					// only compare this word in the context if it is NOT a punctuation word
					// or if it is the first word we are looking at in this phrase
					if (termNum == i || contextArray[termNum].dict_id > DictReservedLimit)
					{						
						//if (mystrcmp_rightcase(SearchPhrases[sw][pterm], Dict[contextArray[termNum].dict_id].word) != 0)						
						if (phrase_terms_ids[sw][pterm] != contextArray[termNum].dict_id)
							break; // we break out of looking at each term of the phrase if we dont match
						pterm++;
					}
					termNum++;
				}

				if (pterm > 0 && phrase_terms_ids[sw][pterm] == 0)
				{
					highlightArray[i] = HIGHLIGHT_START;
					highlightArray[termNum-1] = HIGHLIGHT_END;					
				}
			}
			else
			{				
				if (UseWildCards[sw] == 1)
				{
					//if (WildCardMatch(Dict[word_id].word, SearchWords[sw], ToLowerSearchWords, SearchAsSubstring))
					if (WildCardCompare(Dict[word_id].wordPtr, SearchWords[sw], !ToLowerSearchWords))	
					{
						if (highlightArray[i] == HIGHLIGHT_NONE)
							highlightArray[i] = HIGHLIGHT_SINGLE;									
					}
				}
				else
				{
					if (search_terms_ids[sw] == word_id)
					{
						if (highlightArray[i] == HIGHLIGHT_NONE)
							highlightArray[i] = HIGHLIGHT_SINGLE;						
					}
				}
				/*
				else
				{
					if (mystrcmp_rightcase(SearchWords[sw], Dict[word_id].word) == 0)
					{
						if (highlightArray[i] == HIGHLIGHT_NONE)
							highlightArray[i] = HIGHLIGHT_SINGLE;
						bMatched = true;
					}
				}
				*/				
			}
		}					
	}
}

/*
// this is a custom version of strtok which will be aware of word join characters in Zoom
static char* mystrtokNextPtr = NULL;
char* mystrtok(char* strToken, char* strDelimit, char* WordJoinChars, char* outputBuffer=NULL, bool* IsPrefixOutput=NULL)
{
	int length, i;
	bool bIsDelimit;
	char prevchar, nextchar;
	char* strPtr;

	if (strToken == NULL && mystrtokNextPtr == NULL)
		return NULL;

	if (strToken == NULL)
	{
		// use the previously saved next pointer
		strToken = mystrtokNextPtr;
	}
	
	// if outputbuffer has been specified, we will blank it out.
	if (outputBuffer != NULL)
		strcpy(outputBuffer, "");

	if (IsPrefixOutput != NULL)
		*IsPrefixOutput = false;
	
	strPtr = strToken;

	length = strlen(strToken);
	for (i = 0; i < length; i++)
	{
		bIsDelimit = false;
		if (strchr(strDelimit, strToken[i]) != NULL)
		{
			bIsDelimit = true;
		}
		else
		{
			if (isalnum(strToken[i]) == false)				
			{
				// this is not an alphabetic or numeric character, we shoudl check if it is a wordjoin char
				if (strchr(WordJoinChars, strToken[i]) == NULL)
				{
					// this is not a word join char, we'll delimit!
					bIsDelimit = true;
				}
				else
				{
					// this is a word join char, let's now check its positioning...
					prevchar = '\0';
					nextchar = '\0';
					if (i > 0)
						prevchar = strToken[i-1];
					if (i < length-1)
						nextchar = strToken[i+1];
					
					if (nextchar != '\0' && 
						(isalnum(nextchar) == false && strchr(WordJoinChars, nextchar) == NULL))
					{
						bIsDelimit = true;
					}					
				}
			}
			if (bIsDelimit)
			{
				// save the delimited character in the output buffer to be returned	
				strncat(outputBuffer, strToken+i, 1);
			}
		}
		if (bIsDelimit)
		{
			strToken[i] = '\0';
			if (i+1 < length)
				mystrtokNextPtr = strToken + i + 1;
			else
				mystrtokNextPtr = NULL;

			if (*strPtr != '\0')
				return strPtr;
			else
			{
				if (IsPrefixOutput != NULL) 
					*IsPrefixOutput = true;	// this tells us that the outputBuffer (if any) is a prefix
				strPtr++;
			}
		}
	}
	if (i == length)
		mystrtokNextPtr = NULL;
	return strPtr;
}
*/

// this is a custom version of strtok which will be aware of word join characters in Zoom
#define GETTOKEN_NONE		0
#define GETTOKEN_VALID		1
#define GETTOKEN_DELIMITOR	2
static char* GetTokenNextPtr = NULL;
int GetToken(char* origStr, char* destStr, char* delimitStr, char* wordjoinChars, int destLen)
{
	if (destStr == NULL || delimitStr == NULL || wordjoinChars == NULL)
		return GETTOKEN_NONE;

	if (origStr == NULL && GetTokenNextPtr == NULL)
	{
		*destStr = '\0';
		return GETTOKEN_NONE;
	}

	int length, i;
	bool bIsDelimit;
	char prevchar, nextchar;
	
	if (origStr == NULL)
	{
		// use the previously saved next pointer
		origStr = GetTokenNextPtr;
	}
	
	// blank out destStr first
	strcpy(destStr, "");

	length = strlen(origStr);
	for (i = 0; i < length && i < WORDLENGTH; i++)
	{
		bIsDelimit = false;
		if (strchr(delimitStr, origStr[i]) != NULL)
		{
			bIsDelimit = true;
		}
		else
		{
			//if (isalnum((unsigned char)origStr[i]) == false)				
			if (ispunct((unsigned char)origStr[i]))				
			{
				if (origStr[i] == '&' && ((i > length) || origStr[i+1] == '#'))
				{
					// we need to leave entities alone.
					char* endEnt = strstr(origStr+i, ";");
					int entLen = endEnt - (origStr+i) + 1;
					// this if this is really an entity
					if (entLen < 8)
					{
						// if we already have an existing string, we return that first, and split here.
						if (*destStr != '\0')
						{
							GetTokenNextPtr = origStr+i;
							return GETTOKEN_VALID;
						}

						GetTokenNextPtr = origStr+i+entLen;
						// otherwise, we create a new string for this entity and return it alone
						// extract this out to destStr and return true, while moving the pointer to after this entity
						strncpy(destStr, origStr+i, entLen);
						destStr[entLen] = '\0';
						return GETTOKEN_DELIMITOR;
					}
				}

				// this is not an alphabetic or numeric character, we shoudl check if it is a wordjoin char
				if (strchr(wordjoinChars, origStr[i]) == NULL)
				{
					// this is not a word join char, we'll delimit!
					bIsDelimit = true;
				}
				else
				{
					// this is a word join char, let's now check its positioning...
					prevchar = '\0';
					nextchar = '\0';
					if (i > 0)
						prevchar = origStr[i-1];
					if (i < length-1)
						nextchar = origStr[i+1];
					
					if (nextchar != '\0' &&
						isalnum(nextchar) == false && strchr(wordjoinChars, nextchar) == NULL)
					{
						bIsDelimit = true;
					}					
				}
			}
		}

		if (bIsDelimit)
		{		
			GetTokenNextPtr = origStr+i;

			if (*destStr != '\0')
				return GETTOKEN_VALID;
			
			// return single delimit character as token
			GetTokenNextPtr++;

			*destStr = origStr[i];
			*(destStr+1) = '\0';
			return GETTOKEN_DELIMITOR;
		}
		else
			strncat(destStr, origStr+i, 1);
	}
	if (i == length || i == WORDLENGTH)
		GetTokenNextPtr = NULL;
	
	return GETTOKEN_VALID;
}


// Search through the passed string for instances of matched words, add the highlighting
// code to them and then print them out
//
// [optional input]	bufferIndex - index to OutputBuffers. If not specified, it is -1
// which will use the OutputResultsBuffer instead
void zoomSearch::PrintHighlightDescription(char* line, int bufferIndex)
{
	if (Highlighting == 0)
	{
		if (bufferIndex == -1)
			ResultsBufferPrintf("%s", line);
		else
			OutputBufferPrintf(bufferIndex, "%s", line);
		return;
	}

	int linelen = strlen(line);

	if (linelen == 0)
	{
		if (bufferIndex == -1)
			ResultsBufferPrintf("%s", line);
		else
			OutputBufferPrintf(bufferIndex, "%s", line);
		return;
	}

	// Need to make temp strings big enough to hold extra [;:] [:;] around highlighted words
	// needs to be a big buffer for searches like "**" 
	// The text "<span class=\"highlight\"></span>" is 34 characters long
	
	int MAXSIZE = linelen * 40;
	
	char* tempLine = (char*) calloc(MAXSIZE, sizeof(char));	
	char* outLine = (char*) calloc(MAXSIZE, sizeof(char));

	char* tmpPhraseLine = NULL;
	if (SearchPhrasesForHL != NULL)
		tmpPhraseLine = (char*) calloc(MAX_QUERY_LENGTH+1, sizeof(char));
		
	DecodeEntities(line, tempLine, WordJoinChars, MAXSIZE-1);

	strcpy(outLine, "");
		
	bool bMatched;
	bool bPhraseLineIsPrinted = false;
	
	unsigned long pterm;
	char* prevTokenPtr = NULL;
	
	int tokenType = 0;
	char word[WORDLENGTH+1];
	
	tokenType = GetToken(tempLine, (char*)&word, " ", WordJoinChars, WORDLENGTH);

	while (*word != '\0')
	{
		bMatched = false;

		// Don't bother to try to match a space character, always consider it not matched
		if (strcmp(word, " ") != 0)
		{
			for (unsigned long sw = 0; !bMatched && sw < NumSearchWords; sw++) 
			{
				if (SearchPhrasesForHL != NULL && SearchPhrasesForHL[sw] != NULL)
				{				
					// this is an exact phrase and has its phrase terms stored in SearchPhrasesForHL[sw]
					pterm = 0;
					strcpy(tmpPhraseLine, "");
					bPhraseLineIsPrinted = false;
					while (SearchPhrasesForHL[sw][pterm] != '\0' && *word != '\0')
					{
						// only valid compare if first term, or tokentype is not delimitor
						if ((pterm == 0 || tokenType == GETTOKEN_VALID) && mystrcmp_rightcase(SearchPhrasesForHL[sw][pterm], word) != 0)
						{
							// we need to go back to the original token at the start of this phrase match attempt 
							if (prevTokenPtr != NULL)
								tokenType = GetToken(prevTokenPtr, (char*)&word, " ", WordJoinChars, WORDLENGTH);
							bPhraseLineIsPrinted = true; // we don't want to print out this phrase 
							break; // we break out of looking at each term of the phrase if we dont match
						}
						// save the word in our buffer
						strcat(tmpPhraseLine, word);
						// only continue to the next phrase term if the current token was not a delimitor
						if (tokenType == GETTOKEN_VALID)
							pterm++;					
						// we're going to continue to the next word in the string to see if we match
						tokenType = GetToken(NULL, (char*)&word, " ", WordJoinChars, WORDLENGTH);
					}

					if (pterm > 0 && SearchPhrasesForHL[sw][pterm] == '\0')
					{
						// output the phrase from the buffer (highlighted)
						// Note that we do not flag bMatched, else it will try to highlight the individual word
						strcat(outLine, "[;!]");
						strcat(outLine, tmpPhraseLine);
						strcat(outLine, "[!;]");
						bPhraseLineIsPrinted = true;
					}					
				}
				else
				{
					if (UseWildCards[sw] == 1)
					{
						//if (WildCardMatch(word, SearchWords[sw], ToLowerSearchWords, SearchAsSubstring))
						if (WildCardCompare(word, RegExpSearchWords[sw], !ToLowerSearchWords))	
							bMatched = true;				
					}
					else
					{
						if (mystrcmp_rightcase(RegExpSearchWords[sw], word) == 0)
							bMatched = true;				
					}
				}
			}
		}
		
		if (bMatched)
		{
			strcat(outLine, "[;!]");
			strcat(outLine, word);
			strcat(outLine, "[!;]");
		}
		else
		{
			if (SearchPhrasesForHL != NULL)
			{
				if (!bPhraseLineIsPrinted)
					strcat(outLine, tmpPhraseLine);
				strcpy(tmpPhraseLine, "");
				bPhraseLineIsPrinted = false;
			}
			//if (pterm == 0)
				strcat(outLine, word);
		}		
		prevTokenPtr = GetTokenNextPtr;
		tokenType = GetToken(NULL, (char*)&word, " ", WordJoinChars, WORDLENGTH);
	}
	
	// We need to re-encode the wordjoinchars because we decoded them before doing the 
	// highlighting. This was necessary to match words with word-join-chars in them with
	// queries with wordjoinchars. 
	// However, this broke highlighting when colons were enabled for wordjoining
	// because it entitized the highlight marker! we had to change the marker from "[;:]"
	// and "[:;]" to "[;!]" and "[!;]" to fix this.
	EncodeEntities(outLine, tempLine, WordJoinChars, MAXSIZE-1);

	// replace the marker text with the html text
	// this is to avoid finding previous <span>'ed text.    
	strReplace("[;!]", "<span class=\"highlight\">", tempLine, false, MAXSIZE);			 
	strReplace("[!;]", "</span>", tempLine, false, MAXSIZE);
			 
	//print the highlighted line
	if (bufferIndex == -1)
		ResultsBufferPrintf("%s", tempLine);
	else
		OutputBufferPrintf(bufferIndex, "%s", tempLine);
	free(tempLine);
	free(outLine);

	if (tmpPhraseLine != NULL)
		free(tmpPhraseLine);
}

bool zoomSearch::RecLinkWordMatch(char* rec_word, __int64 rec_idx)
{
	char urlLink[1024+MAX_QUERY_LENGTH] = {0};
	char url[URLLENGTH+1] = {0};
	char title[TITLELEN+1] = {0};
	char description[MAXDESCLEN+1] = {0};
	char image[URLLENGTH+1] = {0};

	int result = 1;	// result is 0 when matched
	bool bRecLinkFound = false;

	if (LastRecLinkIdx == rec_idx)
		return false;

	for (unsigned long sw = 0; sw <= NumSearchWords; sw++)
	{
		// if finished with last search word, check the full search query
		if (sw == NumSearchWords)
		{
			result = mystrcmp_rightcase((char*)Query, rec_word);				
		}
		else
		{
			if (UseWildCards[sw] == 0)
			{
				if (SearchAsSubstring == 0)
					result = mystrcmp_rightcase(SearchWords[sw], rec_word);						
				else
				{  
					if(strlen(SearchWords[sw]) <= strlen(rec_word))
					{	
						char* tmp = iStrStr(rec_word, SearchWords[sw]);								
						if (tmp == NULL) 
							result = 1;
						else 
							result = 0;
					}
					else
						result = 1; //searchword is bigger than word, not found
				}
			} 
			else 
			{
				//if (WildCardMatch(rec_word, SearchWords[sw], ToLowerSearchWords, SearchAsSubstring))
				if (WildCardCompare(rec_word, SearchWords[sw], !ToLowerSearchWords))	
					result = 0;
				else 
					result = 1;
			}

			if (result != 0)
			{
				if (strchr(rec_word, '*') != NULL || strchr(rec_word, '?') != NULL)
				{
					//if (WildCardMatch(SearchWords[sw], rec_word, ToLowerSearchWords, SearchAsSubstring))
					if (WildCardCompare(SearchWords[sw], rec_word, !ToLowerSearchWords))	
						result = 0;
					else 
						result = 1;
				}
			}
		}

		if (result == 0) /// matched
		{
			bRecLinkFound = true;
			LastRecLinkIdx = rec_idx;

			if (num_recs_found == 0 && UseXML == 0)
			{
				OutputBufferPrintf(OUTPUT_RECOMMENDED, "<div class=\"recommended\">\n");
				OutputBufferPrintf(OUTPUT_RECOMMENDED, "<div class=\"recommended_heading\">%s</div>\n", StrTable[STR_RECOMMENDED]);
			}

			if (GetPageData(rec_idx, (char*)&url, (char*)&title, (char*)&description, (char*)image) == false)
			{
				if (UseXML) zoomprintf("<zoom:error>");
				zoomprintf("Error: Corrupted or invalid index files. Please re-index your and try again.");
				if (UseXML) zoomprintf("</zoom:error>\n");
				ShowTemplate();
				cleanup();
				return 0;
			}
			strcpy(urlLink, url);

			// Print the result link
			if (GotoHighlight == 1)
			{
				if (SearchAsSubstring == 1)
					RecLinkAddParamToURL((char*)&urlLink, url, "zoom_highlightsub=", (char*)QueryForURL);				
				else
					RecLinkAddParamToURL((char*)&urlLink, url, "zoom_highlight=", (char*)QueryForURL);				
			}
			else
				sprintf(urlLink, "%s", url);		
			
			if (PdfHighlight == 1)
			{
				if (iStrStr(urlLink, ".pdf") != NULL)
				{
					char tmpQuery[MAX_QUERY_LENGTH];
					strcpy((char*)tmpQuery, (char*)Query);
					strReplace("\"", "", (char*)tmpQuery, false, MAX_QUERY_LENGTH);
					sprintf(urlLink, "%s#search=%%22%s%%22", url, (char*)tmpQuery);
				}
			}

			if (UseXML)
			{
				zoomprintf("<zoom:recommended>\n");	
				if (UseZoomImage == 1)
				{
					if (strlen(image) > 0)
						zoomprintf("<zoom:imageURL>%s</zoom:imageURL>", image);
				}
				zoomprintf("<title>%s</title>\n", title);
				zoomprintf("<link>%s</link>\n", urlLink);
				zoomprintf("<description>%s</description>\n", description);
				//zoomprintf("<zoom:recommended value=\"yes\" />\n");
				zoomprintf("</zoom:recommended>\n");
			}
			else
			{
				OutputBufferPrintf(OUTPUT_RECOMMENDED, "<div class=\"recommend_block\">\n");						

				if (UseZoomImage == 1)
				{
					if (strlen(image) > 0)
					{
						OutputBufferPrintf(OUTPUT_RECOMMENDED, "<div class=\"recommend_image\">");
						OutputBufferPrintf(OUTPUT_RECOMMENDED, "<a href=\"%s\" %s>", urlLink, zoom_target);
						OutputBufferPrintf(OUTPUT_RECOMMENDED, "<img src=\"%s\" alt=\"\" class=\"recommend_image\">", image);
						OutputBufferPrintf(OUTPUT_RECOMMENDED, "</a></div>");
					}
				}
				int nRet = 0;
				nRet = OutputBufferPrintf(OUTPUT_RECOMMENDED, "<div class=\"recommend_title\">");
				if (nRet < 0) zoomprintf("recommended buffer could not be realloc'ed<br>");
				nRet = OutputBufferPrintf(OUTPUT_RECOMMENDED, "<a href=\"%s\" %s>", urlLink, zoom_target);
				if (nRet < 0) zoomprintf("recommended buffer could not be realloc'ed<br>");
				if (strlen(title) > 0)
					PrintHighlightDescription(title, OUTPUT_RECOMMENDED);
				else
					PrintHighlightDescription(url, OUTPUT_RECOMMENDED);
				nRet = OutputBufferPrintf(OUTPUT_RECOMMENDED, "</a></div>\n");
				if (nRet < 0) zoomprintf("recommended buffer could not be realloc'ed<br>");
				nRet = OutputBufferPrintf(OUTPUT_RECOMMENDED, "<div class=\"recommend_description\">");
				if (nRet < 0) zoomprintf("recommended buffer could not be realloc'ed<br>");
				PrintHighlightDescription(description, OUTPUT_RECOMMENDED);
				nRet = OutputBufferPrintf(OUTPUT_RECOMMENDED, "</div>\n");	
				if (nRet < 0) zoomprintf("recommended buffer could not be realloc'ed<br>");
				nRet = OutputBufferPrintf(OUTPUT_RECOMMENDED, "<div class=\"recommend_infoline\">%s</div>\n",url);	
				if (nRet < 0) zoomprintf("recommended buffer could not be realloc'ed<br>");
				nRet = OutputBufferPrintf(OUTPUT_RECOMMENDED, "</div>");
				if (nRet < 0) zoomprintf("recommended buffer could not be realloc'ed<br>");
			}
			num_recs_found++;

			break;					
		}
	}
	return bRecLinkFound;
}

void zoomSearch::ToLowerAndStem(char* destBuffer)
{
	if (UseUTF8 == 1)
		utf8_strlwr((unsigned char*)destBuffer);
	else
		_strlwr((char*)destBuffer); 

	// We need to stem the suggestion here before we compare them otherwise
	// they may not match even though they're the same word
	if (UseStemming == 1)
	{
		GetStem(destBuffer, strlen(destBuffer));
	}
}


// Return the dictionary ID of the passed word
int zoomSearch::GetDictID(char* word)
{
	UserDebugSaveTime(DEBUGTIMER_DICTLOOKUP);
    for (unsigned long i = 0; i < NumKeywords; i++) 
	{
    //   if (mystrcmp_rightcase(word, Dict[i].word) == 0)
		if (strcmp(word, Dict[i].wordPtr) == 0)
		{
			UserDebugAddTime(DEBUGTIMER_DICTLOOKUP);
            return i;
		}
    }
	UserDebugAddTime(DEBUGTIMER_DICTLOOKUP);
    return -1;  // not found
}

// Add the passed word (index of word in SearchWords array) to the skipword list
void zoomSearch::SkipSearchWord(int word)
{
	
	if(SearchWords[word] != "")
	{
		if(SkippedWords > 0)
			strcat(SkippedOutputStr, ", ");
		
		strcat(SkippedOutputStr, "\"<b>");
		strcat(SkippedOutputStr, SearchWords[word]);
		strcat(SkippedOutputStr, "</b>\"");
		strcpy(SearchWords[word], "");
	}
	
	SkippedWords++;
}


// Converts all accented characters in a word IN PLACE
// Returns FALSE if it is unable to fit the converted string in the word buffer (of WORDLENGTH)
// Returns TRUE if successful.
bool zoomSearch::ConvertAccentedWord(char* word, int MAXSIZE)
{	
	char* replacePtr;
	char* wordBuffer = (char*) malloc(MAXSIZE * sizeof(char));
	int len;
	int replace_len;
	int remain_len;
	int i;
	
	// replace all the accented chars. in 'word'
	
	for (i = 0; i < NumAccents ; i++)
	{
		// Do not attempt in the case of the AccentChar == NormalChar (this would cause an infinite loop)
		if (strcmp((char*)AccentChars[i], NormalChars[i]) != 0)
		{
			replacePtr = strstr(word, (char*)AccentChars[i]);			
			while (replacePtr  != NULL)		
			{
				len = replacePtr - word;
				replace_len = strlen(NormalChars[i]);
				remain_len = strlen(replacePtr + len);
							
				// Do not convert for words which are too long to fit in buffer
				if ((len + replace_len + remain_len) > MAXSIZE)			
				{
					free(wordBuffer);
					return false;	
				}
				
				// copy characters before replace
				strncpy(wordBuffer, word, len);
				wordBuffer[len] = '\0';
				// concatenate replace characters
				strcat(wordBuffer, NormalChars[i]);				
				
				// concatenate the characters after replace str.
				len = strlen((char*)AccentChars[i]);
				strcat(wordBuffer, replacePtr + len );
				
				// copy over original word
				// note that words which are now too long will be truncated
				strncpy(word, wordBuffer, MAXSIZE);
				replacePtr = strstr(word, (char*)AccentChars[i]);				
			}			
		}		
	} 
	free(wordBuffer);
	return true;
}

//Add word and pointer to the dictionary array
bool zoomSearch::addDictionaryEntry(char* line, unsigned long lineNo)
{	
	char* tok = NULL;
	int toklen = 0;

	tok = strtok(line, " ");
	if (tok != NULL)
	{		
		//strcpy(Dict[lineNo].word, tok);
		toklen = strlen(tok) + 1;	// +1 for null terminator

		// check if we have the room in baseword_buffer, if not, we need to realloc
		if (baseword_len+toklen > baseword_size)
		{			
			if (UseXML) zoomprintf("<zoom:error>");
			zoomprintf("Error: Out of memory for storing base words. Settings file have been modified or invalid.");
			if (UseXML) zoomprintf("</zoom:error>");
			ShowTemplate();
			cleanup();
			return false;
		}					
		Dict[lineNo].wordPtr = baseword_endPtr;
		strcpy(baseword_endPtr, tok);		// store str after it
		baseword_endPtr += toklen;			// forward end pointer
		
		baseword_len += toklen;

		tok = strtok(NULL, " ");
		if (tok != NULL)
		{
			Dict[lineNo].ptrID = ASCII_to_Int64(tok);
			tok = strtok(NULL, " \r\n");
			if (tok != NULL)
			{
				// read in the variant count
				Dict[lineNo].variantCount = ASCII_to_Integer(tok);
			}
			else
			{
				Dict[lineNo].variantCount = 0;			
			}

#ifdef OSF
			if (LoadNumResultsFromDict)
			{
				if (lineNo > DictReservedLimit)
				{
					int datacount = 0;
					__int64 wordmap_row = Dict[lineNo].ptrID;
					if (wordmap_row != -1)
					{
						if (myfseek(fp_wordmap, wordmap_row) != -1)
						{
							fread(&datacount, 4, 1, fp_wordmap);
						}
					}				
					Dict[lineNo].numResults = datacount;
				}
			}
#endif
		}
		return true;
	}
	return false;
}

//Add word and pointer to the variant buffer 
// Returns NULL if no variant available in given line
char* zoomSearch::addVariantEntry(char* line)
{	
	int toklen;
	char* tok = NULL;
	char* retPtr = NULL;

	tok = strtok(line, " \r\n");
	if (tok != NULL)
	{
		toklen = strlen(tok) + 1;	// +1 for null terminator
		if (toklen <= WORDLENGTH)
		{
			// check if we have the room in variant_buffer, if not, we need to realloc
			if (variant_len+toklen+1 > variant_size)
			{						
				if (UseXML) zoomprintf("<zoom:error>");
				zoomprintf("Error: Out of memory for storing word variants. Settings file have been modified or invalid.");
				if (UseXML) zoomprintf("</zoom:error>");
				ShowTemplate();
				cleanup();				
				return NULL;
			}

			retPtr = variant_endPtr;
			variant_endPtr[0] = toklen;			// store strlen on first byte
			variant_endPtr++;
			strcpy(variant_endPtr, tok);		// store str after it
			variant_endPtr += toklen;			// forward end pointer
			
			// include the one byte we spent on string length
			variant_len += toklen+1;

			/*
			retPtr = variantBufferPtr;
			strcpy(variantBufferPtr->variant, tok);
			variantBufferPtr++;
			*/
			return retPtr;
		}
	}
	return NULL;
}


#define DICT_READBUFFER_SIZE 32000 // number of chars
int zoomSearch::readDictionary()
{
	int fh_dict;	

	char tmpPath[1024];
	sprintf(tmpPath, "%s/zoom_dictionary.zdat", ExecutePath);

	fh_dict = _open(tmpPath, _O_RDONLY|_O_BINARY);
	//fh_dict = myfopen(tmpPath, "r");

	if (fh_dict == -1)
	{
		zoomprintf("Can not find one or more of the Zoom index files.<br />Please make sure the generated index files are uploaded to the same path as this search script.<br />Also check file permissions have been set to allow reading.<br />\n");		
		return FAILURE;
	}

	char* buffer = (char*) malloc((DICT_READBUFFER_SIZE+1) * sizeof(char));
	Dict = (dictstruct*) calloc((NumKeywords + 1), sizeof(dictstruct));		  
	if(buffer == NULL || Dict == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Error: out of memory");		
		if (UseXML) zoomprintf("</zoom:error>");
		ShowTemplate();
		cleanup();
		return FAILURE;
	}

	baseword_size = sizeof(char)*OutputBasewordBufferSize;		
	baseword_buffer = (char*) malloc(baseword_size);
	if (baseword_buffer == NULL || baseword_size == 0)
	{
		zoomprintf("Error: Could not allocate memory for storing base words.<br />");
		return FAILURE;
	}
	baseword_len = 0;
	baseword_endPtr = baseword_buffer;
	
	if (OutputVariantBufferSize == 0)
	{
		// it can be zero if there are no variants! commented out warning below
		//zoomprintf("Warning: Variant buffer size invalid. Please use the unmodified settings file generated with the index files.");
		variant_size = (sizeof(char)*VARIANT_BUFFER_AVGWORDLEN) * NumVariants;		
	}
	else
		variant_size = sizeof(char)*OutputVariantBufferSize;	
	variant_buffer = (char*) malloc(variant_size);	
	if (variant_buffer == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Error: Out of memory for storing word variants.");
		if (UseXML) zoomprintf("</zoom:error>");
		return FAILURE;
	}	
	variant_len = 0;
	variant_endPtr = variant_buffer;

	char* buffReadPtr = buffer;
	int buffReadSize = DICT_READBUFFER_SIZE;
	int buffBytesRead = 0;

	char* buffEndPtr = buffer + DICT_READBUFFER_SIZE;

	char* tok = NULL;
	char* nlpos = NULL;
	unsigned long dict_id = 0;	
	int bytesread = 1;
	int IsVariant = 0;
	char* retVarPtr = NULL;
	
	while (bytesread != 0)
	{
#ifdef NON_WINDOWS
		bytesread = read(fh_dict, buffReadPtr, buffReadSize-1);
#else
		bytesread = _read(fh_dict, buffReadPtr, buffReadSize-1);
#endif
		if (bytesread > 0)
		{
			buffReadPtr[bytesread] = '\0';		

			buffBytesRead += bytesread;

			buffEndPtr = buffReadPtr + buffBytesRead;
			
			tok = buffer;
			nlpos = strchr(tok, '\n');
			while (nlpos != NULL)
			{				
				*nlpos = '\0';

				if (IsVariant > 0 && dict_id > 0)
				{
					// handle variant row
					retVarPtr = addVariantEntry(tok);						
					// only remember a pointer to the first one
					if (IsVariant == Dict[dict_id-1].variantCount)
						Dict[dict_id-1].variantPtr = retVarPtr;
					IsVariant--;					
				}
				else
				{
					IsVariant = 0;
					
					if (strchr(tok, ' ') == 0)
					{
						if (UseXML) zoomprintf("<zoom:error>");
						zoomprintf("Error: Corrupted dictionary data. Invalid dictionary entry at record #%d<br />", dict_id);
						if (UseXML) zoomprintf("</zoom:error>");
						free(buffer);
						ShowTemplate();
						cleanup();
						return FAILURE;
						// OR we could tolerate the error and populate a blank dictionary entry??
						// need to make sure dict_id matches up still and we increment
						// otherwise we exit(); right here and discontinue.
					}

					// handle dictionary row
					if (addDictionaryEntry(tok, dict_id))
					{
						if (Dict[dict_id].variantCount > 0)
						{
							IsVariant = Dict[dict_id].variantCount;
						}
						dict_id++;
						if (dict_id > NumKeywords)
						{
							if (UseXML) zoomprintf("<zoom:error>");
							zoomprintf("Error: Dictionary data (entry %d) does not match NumKeywords (%d) specified in settings.zdat<br />", dict_id, NumKeywords);
							if (UseXML) zoomprintf("</zoom:error>");
							free(buffer);
							ShowTemplate();
							cleanup();
							return FAILURE;
						}
					}
				}

				if ((nlpos+1) > buffEndPtr)
					break;
				
				tok = nlpos+1;				
				nlpos = strchr(tok, '\n');							
			}
			if (nlpos == NULL && tok < buffEndPtr)
			{
				int toklen = strlen(tok);
				memcpy(buffer, tok, toklen);
				buffReadPtr = buffer + toklen;
				buffBytesRead = toklen;
				buffReadSize = DICT_READBUFFER_SIZE - toklen;				
			}
		}		
	}

#ifdef NON_WINDOWS
	close(fh_dict);
#else
	_close(fh_dict);
#endif
	
	free(buffer);

	return SUCCESS;
}


int zoomSearch::GetBytes(FILE* fp, int numbytes)
{
	int retvalue = 0;

	unsigned char* readbytes;
	readbytes = (unsigned char*) malloc(sizeof(unsigned char) * numbytes);
	memset(readbytes, 0, sizeof(unsigned char) * numbytes);

	fread(readbytes, 1, numbytes, fp);
	for (int k = 0; k < numbytes; k++)
	{
		retvalue = retvalue | (readbytes[k] << (8*k));
	}	

	return retvalue;
}

// Read the zoom_pageinfo file into memory
int zoomSearch::readPageinfo()
{	
	char tmpPath[1024];

	sprintf(tmpPath, "%s/%s", ExecutePath, ZoomPageinfoFilename);
	FILE* fp_pageinfo = myfopen(tmpPath, "rb"); 	
	if (fp_pageinfo == NULL)
		return FAILURE;
	
	if (pageinfobin != NULL)
		free(pageinfobin);

	pageinfobin = (BYTE*) malloc(PageInfoSize);
	if (pageinfobin == NULL)
	{
		fclose(fp_pageinfo);
		return FAILURE;
	}

	int bytesread = 0;	

#if defined(NON_WINDOWS) && defined(SUN_SPARC_BUILT)
	//	Reverse the byte order for Sparc
	BYTE* tmpPageInfoBin = (BYTE*) malloc(PageInfoSize);
	if (tmpPageInfoBin == NULL)
	{
		fclose(fp_pageinfo);
		return FAILURE;
	}
	
	bytesread = fread(tmpPageInfoBin, sizeof(BYTE), PageInfoSize, fp_pageinfo);

	// Copy over everything - we'll fiddle the bytes as we need
	memcpy(pageinfobin, tmpPageInfoBin, PageInfoSize);

	unsigned long tmpPtr = 0;
	unsigned long nextSrcPtr;

	BYTE* nextDestPtr;	
	PageInfoData* binPtr = NULL;	
	int sizeof_recsize;
	
	nextDestPtr = pageinfobin;

	for (int pagenum = 0; pagenum < NumPages; pagenum++)
	{						
		binPtr = (PageInfoData*) nextDestPtr;	
		// unsigned short recsize;
		binPtr->recsize = 0;
		
		//binPtr->recsize = tmpPageInfoBin[tmpPtr++] | tmpPageInfoBin[tmpPtr++]<<8;
		// the post increment seem to not work and might be risky with different compilers... so we'll
		// hard code the tmpPtr increments.		
		if (UseBigPageInfoData)
		{
			// this was when recsize was 2 bytes:
			//binPtr->recsize = tmpPageInfoBin[tmpPtr] | tmpPageInfoBin[tmpPtr+1]<<8 ;		
			// now it's 4 bytes:
			binPtr->recsize = tmpPageInfoBin[tmpPtr] | tmpPageInfoBin[tmpPtr+1]<<8 | tmpPageInfoBin[tmpPtr+2]<<16 | tmpPageInfoBin[tmpPtr+3]<<24;		
			sizeof_recsize = sizeof(PageInfoData.recsize);
		}
		else
		{
			binPtr->recsize = tmpPageInfoBin[tmpPtr] | tmpPageInfoBin[tmpPtr+1]<<8 ;
			sizeof_recsize = sizeof(OldPageInfoData.recsize);
		}

		// we work out the pointer to the next record in advanced, 
		// while we have the pointer to the start of this record, and recsize here
		nextDestPtr += binPtr->recsize;
		nextSrcPtr = tmpPtr + binPtr->recsize;		
		tmpPtr += sizeof_recsize;		

		
		// unsigned long dataoffset;
		//binPtr->dataoffset = 0;
		//endianAware_memcpy(binPtr->dataoffset, tmpPageInfoBin, PAGEINFODATA_DATAOFFSET_SIZE);
		//binPtr->dataoffset = tmpPageInfoBin[tmpPtr] | tmpPageInfoBin[tmpPtr+1]<<8 | tmpPageInfoBin[tmpPtr+2]<<16 | tmpPageInfoBin[tmpPtr+3]<<24 | tmpPageInfoBin[tmpPtr+4]<<32;		
		tmpPtr += PAGEINFODATA_DATAOFFSET_SIZE;
		
		// unsigned long filesize;
		//binPtr->filesize = 0;
		binPtr->filesize = tmpPageInfoBin[tmpPtr] | tmpPageInfoBin[tmpPtr+1]<<8 | tmpPageInfoBin[tmpPtr+2]<<16 | tmpPageInfoBin[tmpPtr+3]<<24;		
		tmpPtr += 4;		

		// __time32_t filetime;	// this is 4 bytes		
		//binPtr->filetime = 0;
		binPtr->filetime = tmpPageInfoBin[tmpPtr] | tmpPageInfoBin[tmpPtr+1]<<8 | tmpPageInfoBin[tmpPtr+2]<<16 | tmpPageInfoBin[tmpPtr+3]<<24;		
		tmpPtr += 4;		

		// single chars don't need to change order

		// unsigned char boost;
		//binPtr->boost = 0;
		//binPtr->boost = tmpPageInfoBin[tmpPtr];		
		tmpPtr += 1;

		// unsigned char linkaction;
		//binPtr->linkaction = 0;
		//binPtr->linkaction = tmpPageInfoBin[tmpPtr++];		
		tmpPtr += 1;

		// data is left as is

		// BYTE metadata;
		// leave this as it is... move the recPtr forward
		tmpPtr = nextSrcPtr;
	}

#else
	bytesread = fread(pageinfobin, sizeof(BYTE), PageInfoSize, fp_pageinfo);
	if (bytesread != PageInfoSize)
	{
		zoomprintf("Pageinfo file does not match PageInfoSize specified.");
		fclose(fp_pageinfo);
		return FAILURE;
	}
#endif

	fclose(fp_pageinfo);
	
	pageinfo = NULL;
	
	// build index of offsets to pageinfobin in memory
	pageinfo = (PageInfoData**) malloc(sizeof(PageInfoData*) * NumPages);

	if (UseCats == 1)
	{
		// build index of offsets to the cat number
		catpages = (BYTE**) malloc(sizeof(BYTE*) * NumPages);
	}

	if (UseMetaFields == 1)
	{
		// build index of offsets to the meta data
		metapages = (BYTE**) malloc(sizeof(BYTE*) * NumPages);
	}
	
	BYTE* pageinfoPtr;
	unsigned int pageinfoSizeCheck = 0;
	
	pageinfoPtr = pageinfobin;
	for (int pagenum = 0; pagenum < NumPages; pagenum++)
	{		
		if (UseBigPageInfoData)
		{
			pageinfo[pagenum] = (PageInfoData*) pageinfoPtr;	

			if (UseCats == 1)
			{
				catpages[pagenum] = (BYTE*) &(pageinfo[pagenum]->data);
			}

			if (UseMetaFields == 1)
			{
				metapages[pagenum] = (BYTE*) &(pageinfo[pagenum]->data);
				if (UseCats == 1 && NumCatBytes > 0)
					metapages[pagenum] = metapages[pagenum] + NumCatBytes;
			}

			if (pageinfoSizeCheck + pageinfo[pagenum]->recsize > PageInfoSize)
			{
				zoomprintf("Pageinfo file and PageInfoSize does not match NumPages specified.");
				return FAILURE;
			}

			pageinfoPtr += pageinfo[pagenum]->recsize;
			pageinfoSizeCheck += pageinfo[pagenum]->recsize;
		}
		else
		{
			// This is gonna be tricky to document. 
			// So our pageinfoPtr will actually contain the full pageinfo file in OldPageInfoData format (despite typed as PageInfoData**)
			OldPageInfoData* opid = (OldPageInfoData*) pageinfoPtr;

			// pageinfo[] array is just a bunch of pointers for each page to jump quickly to the appropriate record. So we'll
			// point them to the right place for each OldPageInfoData record, but note that they must be casted to a OldPageInfoData struct 
			// to be used.
			pageinfo[pagenum] = (PageInfoData*)opid;	// this will contain references to each record. but they are OldPageInfoData structs 

			// The rest of these arrays are OK and just have to point to the right place.
			if (UseCats == 1)
			{
				catpages[pagenum] = (BYTE*) &(opid->data);
			}

			if (UseMetaFields == 1)
			{
				metapages[pagenum] = (BYTE*) &(opid->data);
				if (UseCats == 1 && NumCatBytes > 0)
					metapages[pagenum] = metapages[pagenum] + NumCatBytes;
			}

			if (pageinfoSizeCheck + opid->recsize > PageInfoSize)
			{
				zoomprintf("Pageinfo file and PageInfoSize does not match NumPages specified.");
				return FAILURE;
			}

			pageinfoPtr += opid->recsize;
			pageinfoSizeCheck += opid->recsize;
		}
	}

	if (pageinfoSizeCheck < PageInfoSize)
	{
		zoomprintf("Warning: Pageinfo file and PageInfoSize does not match NumPages specified.");
		return FAILURE;
	}
	
	return SUCCESS;
}

int zoomSearch::readRecommended()
{
	char tmpPath[1024];
	sprintf(tmpPath, "%s/%s", ExecutePath, ZoomRecommendedFilename);
	FILE* fp_rec = myfopen(tmpPath, "r"); 
	char buffer[TEMP_STR_SIZE]="";

	if (fp_rec == NULL)
	{
		zoomprintf("Can not find one or more of the Zoom index files.<br />Please make sure the generated index files are uploaded to the same path as this search script.<br />Also check file permissions have been set to allow reading.<br />\n");
		return 0;
	}
		
	unsigned int i = 0;
	int sep = 0;
	char* sepPtr = NULL;
	char* tempLine = (char*) malloc(MaxKeyWordLineLen * sizeof(char));
	
	if (tempLine == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Error: out of memory");
		if (UseXML) zoomprintf("</zoom:error>");
		ShowTemplate();
		cleanup();
		return 0;
	}			  
	
	Rec = (recstruct*) malloc((NumRecommended + 1) * sizeof(recstruct));
	if (Rec == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Error: out of memory");
		if (UseXML) zoomprintf("</zoom:error>");
		ShowTemplate();
		cleanup();
		return 0;
	}			  
		  
	while (!feof(fp_rec) && i < NumRecommended)
	{
		if (fgets(tempLine, MaxKeyWordLineLen, fp_rec) == NULL)
			break;
		
		sepPtr = strrchr(tempLine, ' ');
		if (sepPtr != NULL)
		{
			sep = sepPtr - tempLine;
			if (sep > 0)
			{
				strncpy(Rec[i].word, tempLine, sep);
				Rec[i].word[sep] = '\0';
				// get the index
				Rec[i].index = ASCII_to_Int64(sepPtr);
				i++;		
			}
		}		
	}
	rec_count = i;

	fclose(fp_rec);
	free(tempLine);	
	return 1;
}

void zoomSearch::ShowDefaultForm()
{
	if (OutputBuffers[OUTPUT_JAVASCRIPTS]) PrintStr(OutputBuffers[OUTPUT_JAVASCRIPTS]);
	if (OutputBuffers[OUTPUT_FORM_START]) PrintStr(OutputBuffers[OUTPUT_FORM_START]);
	if (OutputBuffers[OUTPUT_FORM_SEARCHBOX]) PrintStr(OutputBuffers[OUTPUT_FORM_SEARCHBOX]);
	if (OutputBuffers[OUTPUT_FORM_SEARCHBUTTON]) PrintStr(OutputBuffers[OUTPUT_FORM_SEARCHBUTTON]);
	if (OutputBuffers[OUTPUT_FORM_RESULTSPERPAGE]) PrintStr(OutputBuffers[OUTPUT_FORM_RESULTSPERPAGE]);
	if (OutputBuffers[OUTPUT_FORM_MATCH]) PrintStr(OutputBuffers[OUTPUT_FORM_MATCH]);
	if (OutputBuffers[OUTPUT_FORM_CATEGORIES]) PrintStr(OutputBuffers[OUTPUT_FORM_CATEGORIES]);
	if (OutputBuffers[OUTPUT_FORM_CUSTOMMETA]) PrintStr(OutputBuffers[OUTPUT_FORM_CUSTOMMETA]);
	if (OutputBuffers[OUTPUT_FORM_DATERANGE]) PrintStr(OutputBuffers[OUTPUT_FORM_DATERANGE]);
	if (OutputBuffers[OUTPUT_FORM_END]) PrintStr(OutputBuffers[OUTPUT_FORM_END]);
}

bool zoomSearch::IncludeFile(char* filepath)
{
	char buffer[1024+1]="";
	FILE* include_fp;
	char tmpPath[1024+1]="";

	sprintf(tmpPath, "%s/%s", ExecutePath, filepath);
	include_fp = fopen(tmpPath, "r");
	if (include_fp == NULL)
	{
		printf("Error: Failed to include file: %s\n<br \\>", filepath);
		return false;
	}

	while (!feof(include_fp))
	{
		if (fgets(buffer, 1024, include_fp) == NULL)
			break;

		PrintStr(buffer);
	}
	fclose(include_fp);
	return true;
}

void zoomSearch::IncludeAdditionalJS()
{
	if (UseAutoComplete == 1)	
	{		
		char functionStr[1024];		
		PrintStr("<script language=\"javascript\" type=\"text/javascript\">\n");
		//zoomprintf("var ZoomAutoComplete_URL = \"%s%szoom_ac=1\";\n", ExeName, LinkBackJoinChar);		
		zoomprintf("var ZoomAutoComplete_URL = \"%s?zoom_ac=1\";\n", ExeName);		
		IncludeFile(ZOOM_AUTOCOMPLETE_JS);
		PrintStr("if (window.addEventListener) window.addEventListener('load', ");
		sprintf(functionStr, "function() { ZoomAutoComplete_OnLoad('zoom_searchbox');}");
		PrintStr(functionStr);
		PrintStr(", false);");
		PrintStr(" else if (window.attachEvent) window.attachEvent('onload', ");
		PrintStr(functionStr);
		PrintStr(");");
		PrintStr(" else window['onload'] = ");
		PrintStr(functionStr);
		PrintStr("</script>\n");
		PrintStr("<style>\n");
		IncludeFile(ZOOM_AUTOCOMPLETE_CSS);
		PrintStr("</style>\n");
	}
	if (DateRangeSearch == 1)
	{
		PrintStr("<style>\n");
		IncludeFile(ZOOM_DATEPICKER_CSS);
		PrintStr("</style>\n");
		PrintStr("<script language=\"javascript\" type=\"text/javascript\">\n");
		IncludeFile(ZOOM_DATEPICKER_JS);
		PrintStr("</script>\n");
	}

}

//Read and print the first bit of the template file (up to "<!--ZOOMSEARCH-->") 
void zoomSearch::ShowTemplate()
{		
	if (UseXML)
	{
		// this is only called at the end, and for XML output, we just need to close our opening tags.
		zoomprintf("</channel>\n");
		zoomprintf("</rss>");
		return;
	}

	char buffer[1024]="";
	char tmpPath[1024];
	char* tagPtr;
	char* templatePtr;

	bool bIncludedJS = false;
	bool bNeedJS = false;

	if (UseAutoComplete == 1 || DateRangeSearch == 1)
		bNeedJS = true;

	sprintf(tmpPath, "%s/%s", ExecutePath, TemplateFilename);
	templateFp = fopen(tmpPath, "r");
	if(templateFp == NULL)
	{
		zoomprintf("Error: Could not open template file.");
		return;
	}
	
	tagPtr = NULL;
	templatePtr = NULL;

	while(!feof(templateFp))
	{				
		if (templatePtr == NULL)
		{			
			if (fgets(buffer, 1024, templateFp) == NULL)
				break;
			templatePtr = buffer;
		}
		
		if (bNeedJS == true && bIncludedJS == false && strstr(templatePtr, "</head>") != NULL)
		{
			IncludeAdditionalJS();
			bIncludedJS = true;
		}		

		tagPtr = strstr(templatePtr, "<!--ZOOM");
		if (tagPtr != NULL)
		{			
			if ((tagPtr-templatePtr) > 0)
			{
				// print out the bit before the tag			
				*tagPtr = '\0';	//temporary terminate string
				PrintStr(templatePtr);	// prints out content up to tag
				*tagPtr = '<';	// restore the saved char. (should be '<')
			}
		}

		if (tagPtr == NULL)
		{
			PrintStr(templatePtr);		
			templatePtr = NULL;
		}
		else if (strncmp(tagPtr, TEMPLATE_DEFAULT_TAG, strlen(TEMPLATE_DEFAULT_TAG)) == 0)
		{
			ShowDefaultForm();
			if (OutputBuffers[OUTPUT_HEADING]) PrintStr(OutputBuffers[OUTPUT_HEADING]);
			if (OutputBuffers[OUTPUT_SUMMARY]) PrintStr(OutputBuffers[OUTPUT_SUMMARY]);
			if (OutputBuffers[OUTPUT_CATSUMMARY]) PrintStr(OutputBuffers[OUTPUT_CATSUMMARY]);
			if (OutputBuffers[OUTPUT_SUGGESTION]) PrintStr(OutputBuffers[OUTPUT_SUGGESTION]);
			if (OutputBuffers[OUTPUT_PAGESCOUNT]) PrintStr(OutputBuffers[OUTPUT_PAGESCOUNT]);
			if (OutputBuffers[OUTPUT_RECOMMENDED]) PrintStr(OutputBuffers[OUTPUT_RECOMMENDED]);
			if (OutputBuffers[OUTPUT_SORTING]) PrintStr(OutputBuffers[OUTPUT_SORTING]);
			PrintStr(OutputResultsBuffer);
			if (OutputBuffers[OUTPUT_PAGENUMBERS]) PrintStr(OutputBuffers[OUTPUT_PAGENUMBERS]);
			if (OutputBuffers[OUTPUT_SEARCHTIME]) PrintStr(OutputBuffers[OUTPUT_SEARCHTIME]);
			templatePtr = tagPtr + strlen(TEMPLATE_DEFAULT_TAG);
		}
		else if (strncmp(tagPtr, TEMPLATE_SEARCHFORM_TAG, TEMPLATE_SEARCHFORM_TAG_LEN) == 0)
		{
			ShowDefaultForm();
			templatePtr = tagPtr + TEMPLATE_SEARCHFORM_TAG_LEN;
		}
		else if (strncmp(tagPtr, TEMPLATE_RESULTS_TAG, TEMPLATE_RESULTS_TAG_LEN) == 0)
		{
			PrintStr(OutputResultsBuffer);
			templatePtr = tagPtr + TEMPLATE_RESULTS_TAG_LEN;
		}
		else if (strncmp(tagPtr, TEMPLATE_QUERY_TAG, TEMPLATE_QUERY_TAG_LEN) == 0)
		{
			if (strlen((char*)&QueryForHTML) > 0)
				PrintStr((char*)&QueryForHTML);
			templatePtr = tagPtr + TEMPLATE_QUERY_TAG_LEN;
		}
		else
		{
			int tagLen = 0;
			int tagnum = 0;
			for (tagnum = 0; tagnum < OUTPUT_TAG_COUNT; tagnum++)
			{
				tagLen = strlen(TemplateShowTags[tagnum]);
				if (strncmp(tagPtr, TemplateShowTags[tagnum], tagLen) == 0)
				{
					PrintStr(OutputBuffers[tagnum]);
					templatePtr = tagPtr + tagLen;
					break;
				}
			}
			if (tagnum == OUTPUT_TAG_COUNT)
			{
				PrintStr(tagPtr);
				templatePtr = NULL;
			}

			/*
			// Changed from iStrStr to strstr() by RL (8/5/2007)
			// Should be safe to assume it needs to be in all caps? Want to avoid calling
			// iStrStr which is UTF8 aware and really does too many lowercase conversions anyway.		
			if (strstr(buffer, "<!--ZOOMSEARCH-->") == NULL) 
			{
				zoomprintf("%s", buffer);
				//fputs(buffer, stdout);
			}
			else
			{
				break;
			}
			*/
		}
	}

	if (templateFp)
		fclose(templateFp);

	if (bNeedJS == true && bIncludedJS == false)
	{
		IncludeAdditionalJS();
		bIncludedJS = true;
	}	
}


//Read the settings file and set the globals accordingly
//If it cant read the settings file the default initialisations values at start are used
int zoomSearch::readSettings()
{
	char buffer[TEMP_STR_SIZE]="";
	char tmpPath[1024];
	sprintf(tmpPath, "%s/%s", ExecutePath, SettingsFilename);
	FILE* fp = fopen(tmpPath, "rb");
	
	int i = 0;
	int len;

	int tmpStrTableCount = 0;
	
	
	if(fp == NULL)
		return FILE_NOT_FOUND; //failed to open file
	
	int res=0;
		
		
	//try getting a line at a time and matching to a keyword
	//so can change the vars (and order of vars) in the settings file more easily
	char setting[TEMP_STR_SIZE];

	char tmpStrValue[STR_TABLE_LINELEN];
	
	while(!feof(fp))
	{
		if(fgets(buffer, TEMP_STR_SIZE, fp) == NULL)
			break;
		
		sscanf(buffer, "%s", setting);
		
		if(strcmp(setting, "UseUTF8")==0)
			sscanf(buffer,  "UseUTF8 = %d", &UseUTF8);
		else if(strcmp(setting, "Charset")==0)
			sscanf(buffer,  "Charset = \"%[^\"]\"", &Charset);
		else if(strcmp(setting, "Codepage")==0)
			sscanf(buffer,  "Codepage = %d", &Codepage);
		else if(strcmp(setting, "UseStemming")==0)
			sscanf(buffer,  "UseStemming = %d", &UseStemming);
		else if(strcmp(setting, "StemmingAlgo")==0)
			sscanf(buffer,  "StemmingAlgo = %d", &StemmingAlgo);
		else if(strcmp(setting, "NoCharset")==0)
			sscanf(buffer,  "NoCharset = %d", &NoCharset);
		else if(strcmp(setting, "MinWordLen")==0)
			sscanf(buffer,  "MinWordLen = %d", &MinWordLen);
		else if(strcmp(setting, "Highlighting")==0)
			sscanf(buffer,  "Highlighting = %d", &Highlighting);
		else if(strcmp(setting, "HighlightLimit")==0)
			sscanf(buffer,  "HighlightLimit = %d", &HighlightLimit);
		else if(strcmp(setting, "GotoHighlight")==0)
			sscanf(buffer,  "GotoHighlight = %d", &GotoHighlight);		
		else if(strcmp(setting, "PdfHighlight")==0)
			sscanf(buffer,  "PdfHighlight = %d", &PdfHighlight);
		//else if(strcmp(setting, "TemplateFilename")==0)
		//	sscanf(buffer, "TemplateFilename = \"%[^\"]\"", TemplateFilename);
		else if(strcmp(setting, "FormFormat")==0)
			sscanf(buffer, "FormFormat = %d", &FormFormat);
		else if(strcmp(setting, "DisplayMetaDesc")==0)
			sscanf(buffer, "DisplayMetaDesc = %d", &DisplayMetaDesc);
		else if(strcmp(setting, "DisplayMetaFields")==0)
			sscanf(buffer, "DisplayMetaFields = %d", &DisplayMetaFields);
		else if(strcmp(setting, "DotNetUseFormTags")==0)
			sscanf(buffer, "DotNetUseFormTags = %d", &DotNetUseFormTags);
		else if(strcmp(setting, "DotNetUsePostBacks")==0)
			sscanf(buffer, "DotNetUsePostBacks = %d", &DotNetUsePostBacks);
		else if(strcmp(setting, "UseDateTime")==0)
			sscanf(buffer, "UseDateTime = %d", &UseDateTime);
		else if(strcmp(setting, "DisplayDate")==0)
			sscanf(buffer, "DisplayDate = %d", &DisplayDate);
		else if(strcmp(setting, "DefaultSort")==0)
			sscanf(buffer, "DefaultSort = %d", &DefaultSort);		
		else if(strcmp(setting, "DateRangeSearch")==0)
			sscanf(buffer, "DateRangeSearch = %d", &DateRangeSearch);
		else if(strcmp(setting, "DateRangeFormat")==0)
			sscanf(buffer, "DateRangeFormat = %d", &DateRangeFormat);
		else if(strcmp(setting, "UseAutoComplete")==0)
			sscanf(buffer, "UseAutoComplete = %d", &UseAutoComplete);
		else if(strcmp(setting, "DisplayContext")==0)
			sscanf(buffer, "DisplayContext = %d", &DisplayContext);
		else if(strcmp(setting, "DatePickerPath")==0)
		{
			strcpy(ScriptName, "");
			sscanf(buffer,  "DatePickerPath = \"%[^\"]\"", &DatePickerPath);
		}
		else if(strcmp(setting, "DisplayNumber")==0)
			sscanf(buffer, "DisplayNumber = %d", &DisplayNumber);
		else if(strcmp(setting, "DisplayTitle")==0)
			sscanf(buffer, "DisplayTitle = %d", &DisplayTitle);
		else if(strcmp(setting, "DisplayTerms")==0)
			sscanf(buffer, "DisplayTerms = %d", &DisplayTerms);
		else if(strcmp(setting, "DisplayScore")==0)
			sscanf(buffer, "DisplayScore = %d", &DisplayScore);
		else if(strcmp(setting, "DisplayURL")==0)
			sscanf(buffer, "DisplayURL = %d", &DisplayURL);
		else if(strcmp(setting, "TruncateShowURL")==0)
			sscanf(buffer, "TruncateShowURL = %d", &TruncateShowURL);
		else if(strcmp(setting, "DisplayFilesize")==0)
			sscanf(buffer, "DisplayFilesize = %d", &DisplayFilesize);
		else if(strcmp(setting, "UseZoomImage")==0)
			sscanf(buffer, "UseZoomImage = %d", &UseZoomImage);
		else if(strcmp(setting, "DisplayCatSummary")==0)
			sscanf(buffer, "DisplayCatSummary = %d", &DisplayCatSummary);
		else if(strcmp(setting, "Version")==0)
			sscanf(buffer, "Version = \"%[^\"]\"", &VersionInfo);		
		else if(strcmp(setting, "Logging")==0)
			sscanf(buffer, "Logging = %d", &Logging);
		else if(strcmp(setting, "LogFileName")==0)
			sscanf(buffer, "LogFileName = \"%[^\"]\"", LogFilename); //remove the " " surrounding filename
		else if(strcmp(setting, "MaxKeyWordLineLen")==0)
			sscanf(buffer, "MaxKeyWordLineLen = %d", &MaxKeyWordLineLen);
		else if(strcmp(setting, "OutputVariantBufferSize")==0)
			sscanf(buffer, "OutputVariantBufferSize = %d", &OutputVariantBufferSize);		
		else if(strcmp(setting, "OutputBasewordBufferSize")==0)
			sscanf(buffer, "OutputBasewordBufferSize = %d", &OutputBasewordBufferSize);				
		else if(strcmp(setting, "DictIDLen")==0)
		{
			sscanf(buffer,  "DictIDLen = %d", &DictIDLen);
			if (DictIDLen > MAXDICTIDLEN)
				return UNSUPPORTED_FORMAT;
		}
		else if(strcmp(setting, "DictReservedLimit")==0)
			sscanf(buffer,  "DictReservedLimit = %d", &DictReservedLimit);
		else if(strcmp(setting, "DictReservedPrefixes")==0)
			sscanf(buffer,  "DictReservedPrefixes = %d",&DictReservedPrefixes);
		else if(strcmp(setting, "DictReservedSuffixes")==0)
			sscanf(buffer,  "DictReservedSuffixes = %d", &DictReservedSuffixes);
		else if(strcmp(setting, "DictReservedNoSpaces")==0)
			sscanf(buffer,  "DictReservedNoSpaces = %d", &DictReservedNoSpaces);
		else if(strcmp(setting, "UseBigPageInfoData")==0)
			sscanf(buffer,  "UseBigPageInfoData = %d", &UseBigPageInfoData);
		else if(strcmp(setting, "PluginOpenNewWindow")==0)
			sscanf(buffer,  "PluginOpenNewWindow = %d", &PluginOpenNewWindow);
		else if(strcmp(setting, "FileTypePluginStart")==0)
			sscanf(buffer,  "FileTypePluginStart = %d", &FileTypePluginStart);
		else if(strcmp(setting, "WordSplit")==0)
			sscanf(buffer,  "WordSplit = %d", &WordSplit);
		else if(strcmp(setting, "ZoomInfo")==0)
			sscanf(buffer,  "ZoomInfo = %d",&ZoomInfo );
		else if(strcmp(setting, "Timing")==0)
			sscanf(buffer, "Timing = %d", &Timing);
		else if(strcmp(setting, "SearchAsSubstring")==0)
			sscanf(buffer,  "SearchAsSubstring = %d", &SearchAsSubstring);
		else if(strcmp(setting, "ToLowerSearchWords")==0)
			sscanf(buffer,  "ToLowerSearchWords = %d", &ToLowerSearchWords);
		else if(strcmp(setting, "ContextSize")==0)
			sscanf(buffer,  "ContextSize = %d", &ContextSize);
		else if(strcmp(setting, "MaxContextKeywords")==0)
			sscanf(buffer, "MaxContextKeywords = %d", &MaxContextKeywords);
		else if(strcmp(setting, "MaxResultsPerQuery")==0)
			sscanf(buffer, "MaxResultsPerQuery = %d", &MaxResultsPerQuery);
		else if(strcmp(setting, "WeightProximity")==0)
			sscanf(buffer, "WeightProximity = %d", &WeightProximity);		
		else if(strcmp(setting, "AllowExactPhrase")==0)
			sscanf(buffer, "AllowExactPhrase = %d", &AllowExactPhrase);
		else if(strcmp(setting, "MaxContextSeeks")==0)
			sscanf(buffer,  "MaxContextSeeks = %d", &MaxContextSeeks);
		else if(strcmp(setting, "MaxMatches")==0)
			sscanf(buffer,  "MaxMatches = %d", &MaxMatches);
		else if(strcmp(setting, "MaxSearchTime")==0)
			sscanf(buffer,  "MaxSearchTime = %d", &MaxSearchTime);
		else if(strcmp(setting, "UseLinkTarget")==0)
			sscanf(buffer, "UseLinkTarget = %d",&UseLinkTarget);
		else if(strcmp(setting, "LinkTarget")==0)
			sscanf(buffer, "LinkTarget = \"%[^\"]\"",LinkTarget);
		else if(strcmp(setting, "WordJoinChars")==0)
			sscanf(buffer,  "WordJoinChars = \"%[^\"]\"", WordJoinChars);
		else if(strcmp(setting, "Spelling")==0)
			sscanf(buffer,  "Spelling = %d", &Spelling);
		else if(strcmp(setting, "SpellingWhenLessThan")==0)
			sscanf(buffer,  "SpellingWhenLessThan = %d", &SpellingWhenLessThan);
		else if(strcmp(setting, "NumSpellings")==0)
			sscanf(buffer,  "NumSpellings = %d", &NumSpellings);
		else if(strcmp(setting, "Recommended")==0)
			sscanf(buffer,  "Recommended = %d", &Recommended);
		else if(strcmp(setting, "NumRecommended")==0)
			sscanf(buffer,  "NumRecommended = %d", &NumRecommended);
		else if(strcmp(setting, "RecommendedMax")==0)
			sscanf(buffer,  "RecommendedMax = %d", &RecommendedMax);
		else if(strcmp(setting, "UseCats")==0)
			sscanf(buffer,  "UseCats = %d", &UseCats);
		else if(strcmp(setting, "SearchMultiCats")==0)
			sscanf(buffer,  "SearchMultiCats = %d", &SearchMultiCats);
		else if(strcmp(setting, "NumKeywords")==0)
			sscanf(buffer,  "NumKeywords = %ld", &NumKeywords);
		else if(strcmp(setting, "NumPages")==0)
			sscanf(buffer,  "NumPages = %ld", &NumPages);				
		else if(strcmp(setting, "NumVariants")==0)
			sscanf(buffer,  "NumVariants = %ld", &NumVariants);		
		else if(strcmp(setting, "PageInfoSize")==0)
			sscanf(buffer,  "PageInfoSize = %ld", &PageInfoSize);		
		else if(strcmp(setting, "LinkBackURL")==0)
		{
			strcpy(ScriptName, "");
			sscanf(buffer,  "LinkBackURL = \"%[^\"]\"", ScriptName);
			if (strchr(ScriptName, '?') != NULL)
				strcpy(LinkBackJoinChar, "&amp;");
		}
		else if(strcmp(setting, "DefaultToAnd")==0)
		{
			sscanf(buffer,  "DefaultToAnd = %ld", &DefaultToAnd);			
		}
		else if(strcmp(setting, "UseXML")==0)
			sscanf(buffer,  "UseXML = %d", &DefaultToXML);
		else if(strcmp(setting, "XMLTitle")==0)
		{
			strcpy(XMLTitle, "");
			sscanf(buffer,  "XMLTitle = \"%[^\"]\"", XMLTitle);
		}
		else if(strcmp(setting, "XMLDesc")==0)
		{
			strcpy(XMLDescription, "");
			sscanf(buffer,  "XMLDesc = \"%[^\"]\"", XMLDescription);
		}
		else if(strcmp(setting, "XMLURL")==0)
		{
			strcpy(XMLLink, "");
			sscanf(buffer,  "XMLURL = \"%[^\"]\"", XMLLink);
		}
		else if(strcmp(setting, "XMLXSLTURL")==0)
		{
			strcpy(XMLStyleSheetURL, "");
			sscanf(buffer,  "XMLXSLTURL = \"%[^\"]\"", XMLStyleSheetURL);
		}
		else if(strcmp(setting, "XMLOpenSearchDescURL")==0)
		{
			strcpy(XMLOpenSearchDescURL, "");
			sscanf(buffer,  "XMLOpenSearchDescURL = \"%[^\"]\"", XMLOpenSearchDescURL);
		}
		else if(strcmp(setting, "XMLHighlight")==0)
		{
			sscanf(buffer,  "XMLHighlight = %d", &XMLHighlight);
		}		
		else if(strncmp(setting, "STR_", 4) == 0)
		{						
			UserDebugSaveTime(DEBUGTIMER_LOADSTRTABLE);
			for (int i = 0; i < STR_TABLE_COUNT; i++)
			{
				if (strcmp(setting, StrTableVarNames[i]) == 0)
				{
					strcpy(tmpStrValue, "");
					sscanf(buffer, "%s = \"%[^\"]\"", setting, tmpStrValue);
					len = strlen(tmpStrValue);
					StrTable[i] = (char*) malloc((len+1) * sizeof(char));					
					if (StrTable[i] == NULL)
						return OUT_OF_MEMORY;
					memset(StrTable[i], 0, (len+1) * sizeof(char));
					strncpy(StrTable[i], tmpStrValue, len);
					StrTable[i][len] = '\0';
					tmpStrTableCount++;
					break;
				}
			}	
			UserDebugAddTime(DEBUGTIMER_LOADSTRTABLE);
		}
		else if(strcmp(setting, "NumCats")==0)
			sscanf(buffer,  "NumCats = %d", &NumCats);
		else if(strcmp(setting, "NumCatBytes")==0)
			sscanf(buffer,  "NumCatBytes = %d", &NumCatBytes);
		else if(strcmp(setting, "catnames") == 0)
		{			
			UserDebugSaveTime(DEBUGTIMER_LOADCATS);
			
			if (NumCats > 0)
			{
				//allocate 2d array
				CatNames = (char**) malloc(NumCats * sizeof(char*));
				if(CatNames == NULL)
					return OUT_OF_MEMORY;				
				
				char* ret;
				i = 0;
				while (!feof(fp) && i < NumCats)
				{
					ret = fgets(buffer, TEMP_STR_SIZE, fp); 									
					if (ret != NULL)
					{
						if (strlen(buffer) < 1)
							break;

						CatNames[i] = (char*) malloc((strlen(buffer)+1)  * sizeof(char)); 
						if(CatNames[i] == NULL)
							return OUT_OF_MEMORY;

						strcpy(CatNames[i], buffer);						
						//CatNames[i][strlen(buffer)-2] = '\0'; // trim CRLF						
						Trim(CatNames[i]);
						i++;
					}
				}
			}			
			UserDebugAddTime(DEBUGTIMER_LOADCATS);
		}
		else if(strcmp(setting, "UseMetaFields")==0)
			sscanf(buffer,  "UseMetaFields = %ld", &UseMetaFields);
		else if(strcmp(setting, "NumMetaFields")==0)
			sscanf(buffer,  "NumMetaFields = %ld", &NumMetaFields);
		else if(strcmp(setting, "MetaMoneyShowDec")==0)
			sscanf(buffer,  "MetaMoneyShowDec = %ld", &MetaMoneyShowDec);
		else if(strcmp(setting, "MetaMoneyCurrency")==0)
			sscanf(buffer,  "MetaMoneyCurrency = \"%[^\"]\"", MetaMoneyCurrency);
		else if(strcmp(setting, "metafields") == 0)
		{						
			UserDebugSaveTime(DEBUGTIMER_LOADMETA);

			if (NumMetaFields > 0)
			{
				metafields = (metafield_elem*) malloc(sizeof(metafield_elem)*NumMetaFields);
				if (metafields == NULL)
					return OUT_OF_MEMORY;
				memset(metafields, 0, sizeof(metafield_elem)*NumMetaFields);

				// read in the metafields settings
				char* ret;
				i = 0;
				while (!feof(fp) && i < NumMetaFields)
				{
					ret = fgets(buffer, TEMP_STR_SIZE, fp); 									
					if (ret != NULL)
					{
						if (strlen(buffer) < 1)
							break;

						// get type
						metafields[i].type = atoi(buffer);					

						// get name
						ret = fgets(buffer, TEMP_STR_SIZE, fp); 	
						if (ret == NULL)
							return FAILURE;						
						strcpy(metafields[i].name, buffer);
						//metafields[i].name[strlen(metafields[i].name)-2] = '\0';	// trim CRLF
						Trim(metafields[i].name);

						// get showname
						ret = fgets(buffer, TEMP_STR_SIZE, fp); 	
						if (ret == NULL)
							return FAILURE;
						strcpy(metafields[i].showname, buffer);
						//metafields[i].showname[strlen(metafields[i].showname)-2] = '\0';	// trim CRLF
						Trim(metafields[i].showname);

						// get formname
						ret = fgets(buffer, TEMP_STR_SIZE, fp); 	
						if (ret == NULL)
							return FAILURE;
						strcpy(metafields[i].formname, buffer);
						//metafields[i].formname[strlen(metafields[i].formname)-2] = '\0';	// trim CRLF
						Trim(metafields[i].formname);

						// get method
						ret = fgets(buffer, TEMP_STR_SIZE, fp); 	
						if (ret == NULL)
							return FAILURE;
						metafields[i].method = atoi(buffer);						

						// get dropdown values
						ret = fgets(buffer, TEMP_STR_SIZE, fp); 	
						if (ret == NULL)
							return FAILURE;
						metafields[i].DropdownCount = atoi(buffer);
						if (metafields[i].DropdownCount > 0 && 
							(metafields[i].type == METAFIELD_TYPE_DROPDOWN || metafields[i].type == METAFIELD_TYPE_MULTI))
						{
							metafields[i].DropdownValues = (char**) malloc(metafields[i].DropdownCount * sizeof(char*));
							if (metafields[i].DropdownValues == NULL)
								return OUT_OF_MEMORY;

							int di = 0;
							for (di = 0; di < metafields[i].DropdownCount; di++)
							{
								ret = fgets(buffer, TEMP_STR_SIZE, fp);
								if (ret == NULL)
									return FAILURE;

								metafields[i].DropdownValues[di] = (char*) malloc((METAFIELD_DROPDOWN_LEN+1)*sizeof(char));
								if (metafields[i].DropdownValues[di] == NULL)
									return OUT_OF_MEMORY;
								strcpy(metafields[i].DropdownValues[di], buffer);
								//metafields[i].DropdownValues[di][strlen(metafields[i].DropdownValues[di])-2] = '\0';	// trim CRLF
								Trim(metafields[i].DropdownValues[di]);
							}							
						}

						i++;
					}
				}
			}
			UserDebugAddTime(DEBUGTIMER_LOADMETA);
		}
		else if(strcmp(setting, "MapAccents")==0)
		{
			sscanf(buffer,  "MapAccents = %d", &MapAccents);
			if(MapAccents == 1) //only read if map accents are on and included
			{
				char accentLine[TEMP_STR_SIZE];
				int numAccents=0;
				fgets(buffer, TEMP_STR_SIZE, fp); //get accented chars
				
				//copy accented chars into buffer
				//count number of accents chars, create buffer, copy
				
				sscanf(buffer, "AccentChars =  %[^\n]", accentLine);
				
				char* first; 
				char* next;
				int startPos=0, endPos=0;
				
				//count num accents
				while(1)
				{
					first = strchr(accentLine+startPos, '"');
					if(first == NULL)
						break;
					startPos = first - accentLine+ 1;
					next = strchr(accentLine+startPos, '"');
					endPos = next - accentLine +1;
					
					startPos = endPos;
					numAccents++;
				}
				
				NumAccents = numAccents;		
				
				//allocate memory
				AccentChars = (unsigned char**)malloc(numAccents * sizeof(unsigned char*));
				
				if(	AccentChars == NULL)
					return OUT_OF_MEMORY;
				int a=0;
				for( a=0; a<numAccents; a++)
				{
					AccentChars[a] = (unsigned char*) malloc(10 * sizeof(unsigned char));
					if(	AccentChars[a] == NULL)
						return OUT_OF_MEMORY;
					
					memset(AccentChars[a], 0, 10);
				}
				
				//find first ", get data till next ", 
				int i=0;
				startPos=0, endPos=0;
				memset(buffer, 0, sizeof(buffer));
				while(i < NumAccents)
				{
					first = strchr(accentLine+startPos, '"');
					if(first == NULL)
						break;
					startPos = first - accentLine+ 1;
					next = strchr(accentLine+startPos, '"');
					endPos = next - accentLine +1;
					
					strncpy(buffer, accentLine+startPos, endPos-startPos-1);
					buffer[endPos-startPos-1] = '\0';
					strcpy((char*)AccentChars[i], buffer);
					
					startPos = endPos;
					i++;
				}
				
				fgets(buffer, TEMP_STR_SIZE, fp); //get normal chars
				//copy Normal chars into buffer
				
				NormalChars = (char**)malloc(numAccents * sizeof(char*));
				
				if(	NormalChars == NULL)
					return OUT_OF_MEMORY;
				
				for(a=0; a<numAccents; a++)
				{
					NormalChars[a] = (char*) malloc(5 * sizeof(char));
					if(	NormalChars[a] == NULL)
						return OUT_OF_MEMORY;
					memset(NormalChars[a], 0, 4);
				}
				
				
				sscanf(buffer, "NormalChars =  %[^\n]", accentLine);
				
				startPos=0;
				endPos=0;
				//find first ", get data till next ", 
				i=0;
				while(i < NumAccents)
				{
					first = strchr(accentLine+startPos, '"');
					if(first == NULL)
						break;
					startPos = first - accentLine+ 1;
					next = strchr(accentLine+startPos, '"');
					endPos = next - accentLine +1;
					
					strncpy(buffer, accentLine+startPos, endPos-startPos);
					buffer[endPos-startPos-1] = '\0';			
					
					strcpy(NormalChars[i], buffer);
					startPos = endPos;
					i++;
				}
			}
		}

	}	
	fclose(fp);

	// Check string table is fully populated.
	if (tmpStrTableCount < STR_TABLE_COUNT)
	{
		// error: string table not fully populated - missing strings.		
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("Error: Missing text in the settings.zdat file. Check that your ZLANG file is valid if you have modified it.<br />\n");
		if (UseXML) zoomprintf("</zoom:error>");
		return FAILURE;
	}

	return SUCCESS;
}

void zoomSearch::DisplayXMLSettings()
{
	zoomprintf("<zoom:setting name=\"UseUTF8\" value=\"%d\" />\n", UseUTF8);
	zoomprintf("<zoom:setting name=\"MinWordLen\" value=\"%d\" />\n", MinWordLen);
	zoomprintf("<zoom:setting name=\"Highlighting\" value=\"%d\" />\n", Highlighting);
	zoomprintf("<zoom:setting name=\"HighlightLimit\" value=\"%d\" />\n", HighlightLimit);
	zoomprintf("<zoom:setting name=\"GotoHighlight\" value=\"%d\" />\n", GotoHighlight);
	zoomprintf("<zoom:setting name=\"PdfHighlight\" value=\"%d\" />\n", PdfHighlight);
	zoomprintf("<zoom:setting name=\"PluginOpenNewWindow\" value=\"%d\" />\n", PluginOpenNewWindow);
	zoomprintf("<zoom:setting name=\"DisplayMetaDesc\" value=\"%d\" />\n", DisplayMetaDesc);
	zoomprintf("<zoom:setting name=\"DisplayCatSummary\" value=\"%d\" />\n", DisplayCatSummary);
	zoomprintf("<zoom:setting name=\"UseDateTime\" value=\"%d\" />\n", UseDateTime);
	zoomprintf("<zoom:setting name=\"DisplayDate\" value=\"%d\" />\n", DisplayDate);
	zoomprintf("<zoom:setting name=\"DisplayContext\" value=\"%d\" />\n", DisplayContext);
	zoomprintf("<zoom:setting name=\"DisplayTitle\" value=\"%d\" />\n", DisplayTitle);
	zoomprintf("<zoom:setting name=\"DisplayTerms\" value=\"%d\" />\n", DisplayTerms);
	zoomprintf("<zoom:setting name=\"DisplayScore\" value=\"%d\" />\n", DisplayScore);
	zoomprintf("<zoom:setting name=\"DisplayURL\" value=\"%d\" />\n", DisplayURL);
	zoomprintf("<zoom:setting name=\"DisplayFilesize\" value=\"%d\" />\n", DisplayFilesize);
	zoomprintf("<zoom:setting name=\"LinkTarget\" value=\"%s\" />\n", LinkTarget);
	zoomprintf("<zoom:setting name=\"UseZoomImage\" value=\"%d\" />\n", UseZoomImage);
	zoomprintf("<zoom:setting name=\"Version\" value=\"%s\" />\n", VersionInfo);
	zoomprintf("<zoom:setting name=\"WordSplit\" value=\"%d\" />\n", WordSplit);
	zoomprintf("<zoom:setting name=\"SearchAsSubstring\" value=\"%d\" />\n", SearchAsSubstring);
	zoomprintf("<zoom:setting name=\"ToLowerSearchWords\" value=\"%d\" />\n", ToLowerSearchWords);
	zoomprintf("<zoom:setting name=\"ContextSize\" value=\"%d\" />\n", ContextSize);
	zoomprintf("<zoom:setting name=\"MaxContextKeywords\" value=\"%d\" />\n", MaxContextKeywords);
	zoomprintf("<zoom:setting name=\"AllowExactPhrase\" value=\"%d\" />\n", AllowExactPhrase);
	zoomprintf("<zoom:setting name=\"MaxContextSeeks\" value=\"%d\" />\n", MaxContextSeeks);
	zoomprintf("<zoom:setting name=\"MaxSearchTime\" value=\"%d\" />\n", MaxSearchTime);
	zoomprintf("<zoom:setting name=\"WordJoinChars\" value=\"%s\" />\n", WordJoinChars);
	zoomprintf("<zoom:setting name=\"Spelling\" value=\"%d\" />\n", Spelling);
	zoomprintf("<zoom:setting name=\"SpellingWhenLessThan\" value=\"%d\" />\n", SpellingWhenLessThan);	
	zoomprintf("<zoom:setting name=\"Recommended\" value=\"%d\" />\n", Recommended);
	zoomprintf("<zoom:setting name=\"RecommendedMax\" value=\"%d\" />\n", RecommendedMax);
	zoomprintf("<zoom:setting name=\"UseCats\" value=\"%d\" />\n", UseCats);
	zoomprintf("<zoom:setting name=\"SearchMultiCats\" value=\"%d\" />\n", SearchMultiCats);
	zoomprintf("<zoom:setting name=\"XMLHighlight\" value=\"%d\" />\n", XMLHighlight);	
	if (UseCats == 1)
	{
		zoomprintf("<zoom:setting name=\"NumCats\" value=\"%d\" />\n", NumCats);
		for (int a = 0; a < NumCats; a++) 
			zoomprintf("<zoom:category>%s</zoom:category>\n", CatNames[a]);
	}
	zoomprintf("<zoom:setting name=\"MapAccents\" value=\"%d\" />\n", MapAccents);
	if (AccentChars != NULL)
	{
		zoomprintf("<zoom:setting name=\"AccentChars\">");
		for (int a=0; a<NumAccents; a++)
			zoomprintf("%s", AccentChars[a]);
		zoomprintf("</zoom:setting>\n");
	}
	if (NormalChars != NULL)
	{
		zoomprintf("<zoom:setting name=\"NormalChars\">");
		for (int a=0; a < NumAccents; a++)
			zoomprintf("%s", NormalChars[a]);
		zoomprintf("</zoom:setting>\n");
	}	
}

FILE* zoomSearch::myfopen(const char* filename, const char* mode)
{
#ifdef NON_WINDOWS
	// for linux, etc. we need O_LARGEFILE to access files > 2GB in size
	int flags = 0;
	if (strchr(mode, 'r') != 0)
		flags |= O_RDONLY;
	if (strchr(mode, 'b') != 0)
		flags |= O_BINARY;
	if (strchr(mode, 'a') != 0)
		flags |= O_APPEND;

	int fd;
	// low level open so we can use O_LARGEFILE
	fd = open(filename, flags|O_LARGEFILE);
	if (fd == -1)
		return NULL;
	
	// associate the pre-established stream with a new file handle
	return fdopen(fd, mode);

#else
	// for windows:
	return fopen(filename, mode);
#endif
}

//unsigned long myfseek(int stream, unsigned long pos)
long zoomSearch::myfseek(FILE* stream, __int64 pos)
{
	long ret;
#ifndef NON_WINDOWS
	ret = _fseeki64(stream, (__int64)pos, SEEK_SET);	
	//ret = (unsigned long) _lseeki64(stream, (__int64)pos, 0);
	// Windows here:	
	/*	
	fpos_t tmppos;
	tmppos = (fpos_t) pos;
	ret = fsetpos(stream, &tmppos);
	*/
#else
	// Linux, FreeBSD
	// ==============
	//

#ifdef LINUX_LARGEFILE_USEFSEEKO
	// Uncomment the following to support >2GB on Ubuntu VM (the one used for Zoom For Linux builds)
	ret = fseeko(stream, (off_t)pos, 0);					// requires LARGEFILE support, off_t can vary from platform to platform
#else
	// Below was an alternative method, but requires changing fopen/fclose calls to _open(), etc.
	//ret = (unsigned long) lseek(stream, (off_t)pos, 0);	// requires _open()/_read() and LARGEFILE support

	// This is the widely supported call for glibc earlier than 2.1 (1999). Our old Debian VM has glibc 1.9.
	// But it only supports 2 GB fseeks.
	ret = fseek(stream, pos, SEEK_SET);		// less than 2 GB only
#endif

#endif
	if (ret != 0)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		if (pos > 2000000000)
			zoomprintf("Error seeking in file. This CGI build does not support files larger than 2GB in size. Please contact Wrensoft for help.");
		else
			zoomprintf("Error seeking in file. This may be due to an incompatible CGI build for your operating system. Please contact Wrensoft for help.");
		//exit(1); //Now that this is a class we exit after 01 is returned to search function		
	}
	return ret;
}

//remove leading and trailing spaces from buffer
int zoomSearch::Trim(char* buffer)
{
	if(buffer == NULL)
		return SUCCESS;
	
	int size = strlen(buffer);
	if(size < 1)
		return SUCCESS;

	int startindex=0, stopindex=size-1;
	//find leading spaces
	int a=0;
	for(a=0; a<size; a++)
	{
		if(buffer[a] != 0x20 && buffer[a] != 0x09 && buffer[a] != 0x0A && buffer[a] != 0x0D)//remove space & tab & CR & LF
		{
			startindex=a;
			break;
		}
	}	
	//find trailing spaces
	for(a=stopindex; a>-1; a--)
	{
		if(buffer[a] != 0x20 && buffer[a] != 0x09 && buffer[a] != 0x0A && buffer[a] != 0x0D) //remove space & tab & CR & LF
		{
			stopindex=a;
			break;
		}	
	}
	if (startindex == 0 && stopindex == size-1)
		return SUCCESS;

	char* tmpBuffer = (char*) calloc(size+1, sizeof(char*));
	
	if(tmpBuffer == NULL)
		return OUT_OF_MEMORY;
	
	
	stopindex = stopindex-startindex+1;
	strcpy(tmpBuffer, buffer);
	//memset(buffer, 0, size+1);
	strncpy(buffer, tmpBuffer+startindex, stopindex);
	buffer[stopindex] = '\0';
	
	free(tmpBuffer);
	return SUCCESS;
	
}


void zoomSearch::cleanupQuery()
{
	if (SearchWords != NULL)
	{
		for(unsigned long a=0; a<NumSearchWords; a++)
			free(SearchWords[a]);
		free(SearchWords);
		SearchWords = NULL;
	}

	if (UseWildCards != NULL)
	{
		free(UseWildCards);
		UseWildCards = NULL;
	}

	if (search_terms_ids != NULL)
	{
		free(search_terms_ids);
		search_terms_ids = NULL;
	}

	if (phrase_terms_ids != NULL)
	{	
		for(unsigned int sw=0; sw < NumSearchWords; sw++)
		{
			if (phrase_terms_ids[sw] != NULL)
				free(phrase_terms_ids[sw]);
		}
		free(phrase_terms_ids);
		phrase_terms_ids = NULL;
	}
	
	if (RegExpSearchWords != NULL)
	{
		for(unsigned int a=0; a < NumSearchWords; a++)
			free(RegExpSearchWords[a]);
		free(RegExpSearchWords);
		RegExpSearchWords = NULL;
	}	

	if (GetNextDictWordBuff != NULL)
	{
		free(GetNextDictWordBuff);
		GetNextDictWordBuff = NULL;
	}

	if (OutputBuffers != NULL)
	{
		for (int bufnum = 0; bufnum < OUTPUT_TAG_COUNT; bufnum++)
		{
			if (OutputBuffers[bufnum]) 
				free(OutputBuffers[bufnum]);
		}
		free(OutputBuffers);
		OutputBuffers = NULL;
	}
	if (OutputBufferSizes) 
	{
		free(OutputBufferSizes);
		OutputBufferSizes = NULL;
	}
	if (OutputBufferLens)
	{
		free(OutputBufferLens);
		OutputBufferLens = NULL;
	}
	if (OutputResultsBuffer)
	{
		free(OutputResultsBuffer);
		OutputResultsBuffer = NULL;
	}

	if (stemmer)
	{
		sb_stemmer_delete(stemmer);
		stemmer = NULL;
	}

#ifdef DEBUG_TIMER
	memset(&UserDebugStats, 0, sizeof(USERDEBUGSTATS));
#endif

}

//de-allocate memory etc
void zoomSearch::cleanup()
{	
	PrintUserDebug("Cleaning up...");
	cleanupQuery();

#if defined(FASTCGI_ENABLED) || defined(OSF)
	if (fp_wordmap != NULL)
	{
		// Close the files that we didn't close because we're in FCGI mode
		fclose(fp_wordmap);
		fp_wordmap = NULL;
	}

	if (DisplayContext == 1 || AllowExactPhrase == 1)
	{		
		if (fp_pagetext != NULL)
		{
			fclose(fp_pagetext);
			fp_pagetext = NULL;
		}
	}
#endif

	if(CatNames != NULL)
	{
		for(int a=0; a<NumCats; a++)
			free(CatNames[a]);
		free(CatNames);
		CatNames = NULL;
	}

	if (pageinfo != NULL)
	{
		free(pageinfo);
		pageinfo = NULL;
	}

	if (catpages != NULL)
	{
		free(catpages);
		catpages = NULL;
	}

	if (metapages != NULL)
	{
		free(metapages);
		metapages = NULL;
	}

	if (pageinfobin != NULL)
	{
		free(pageinfobin);
		pageinfobin = NULL;
	}

	if (metafields != NULL)
	{
		for (int a = 0; a < NumMetaFields; a++)
		{
			if (metafields[a].DropdownValues != NULL && metafields[a].DropdownCount > 0)
			{				
				for (int ddi = 0; ddi < metafields[a].DropdownCount; ddi++)
					free(metafields[a].DropdownValues[ddi]);
				free(metafields[a].DropdownValues);
			}
		}
		free(metafields);
		metafields = NULL;
	}
		
	if(AccentChars != NULL)
	{
		for(int a=0; a<NumAccents; a++)
			free(AccentChars[a]);
		free(AccentChars);
		AccentChars = NULL;
	}
	
	if(NormalChars != NULL)
	{
		for(int a=0; a<NumAccents; a++)
			free(NormalChars[a]);
		free(NormalChars);
		NormalChars = NULL;
	}
	
	if(Dict != NULL)
	{
		free(Dict);
		Dict = NULL;
	}

	if (baseword_buffer)
	{
		free(baseword_buffer);
		baseword_buffer = NULL;
	}

	if (variant_buffer != NULL)
	{
		free(variant_buffer);
		variant_buffer = NULL;
	}

	if (Rec != NULL)
	{
		free(Rec);
		Rec = NULL;
	}

	if (pagedatabin)
	{
		free(pagedatabin);
		pagedatabin = NULL;
	}
	
			
	// delete string table
	if (StrTable != NULL)
	{
		for (int a = 0; a < STR_TABLE_COUNT; a++)
		{
			if (StrTable[a] != NULL)
				free(StrTable[a]);
		}
		free(StrTable);
		StrTable = NULL;
	}

	if (UseUTF8 == 1)
	{
		ucdata_cleanup();
	}



	PrintUserDebug("Cleanup finished.");
}

void zoomSearch::AddParamToURL(char* destURL, char* srcURL, char* paramStr, char* paramValue)
{
	// determine whether we should add parameters with "?" or "&amp;" depending on
	// whether there are any existing parameters in the URL
	char addParamStr[10] = {0};				
	if (strchr(srcURL, '?') != NULL)
		strcpy(addParamStr, "&amp;");
	else
		strcpy(addParamStr, "?");

	sprintf(destURL, "%s%s%s%s", srcURL, addParamStr, paramStr, paramValue);
}

void zoomSearch::RecLinkAddParamToURL(char* destURL, char* srcURL, char* paramStr, char* paramValue)
{
	// determine whether we should add parameters with "?" or "&amp;" depending on
	// whether there are any existing parameters in the URL
	char addParamStr[10] = {0};				
	if (strchr(srcURL, '?') != NULL)
		strcpy(addParamStr, "&amp;");
	else
		strcpy(addParamStr, "?");

	char* hashPtr = strchr(srcURL, '#');
	if (hashPtr != NULL)
	{
		*hashPtr = '\0';
		sprintf(destURL, "%s%s%s%s#%s", srcURL, addParamStr, paramStr, paramValue, hashPtr+1);		
	}
	else
		sprintf(destURL, "%s%s%s%s", srcURL, addParamStr, paramStr, paramValue);
}



//char4char == true, replace each character found in seach with corresponding char in replace
//char4char == false, replace each occurance of string search with string replace
bool zoomSearch::strReplace(char* search, char* replace, char* subject, bool char4char, int MAXSIZE)
{
	char* replacePtr;
	int i=0;

	bool ret = true;

	// nothing to do if search string is same as replace string.
	if (strcmp(search, replace) == 0)
		return true;
				
	// replace all the "search" chars in "subject" with "replace"
	if(char4char == true)
	{
		while(search[i] !='\0')
		{
			replacePtr = strchr(subject, search[i]);
			while (replacePtr  != NULL)		
			{			
				subject[replacePtr-subject] = replace[i];
				replacePtr = strchr(subject, search[i]);
			}
			i++;			
		}
	}
	else
	{	
		char* buf = (char*) calloc(MAXSIZE*2, sizeof(char));
		
		replacePtr = strstr(subject, search);
		int last=0;
		int result=0;
		
		while(replacePtr != NULL)
		{
			result = replacePtr - subject;
			strncat(buf, subject+last, result-last);
			strcat(buf, replace);
			replacePtr = strstr(subject+result+1,search);
			last = result+strlen(search);
			
		}
		strcat(buf, subject+last);	
		
		if(strlen(buf) < (unsigned int) MAXSIZE)
		{
			strcpy(subject, buf);
		}
		else
			ret = false;
				
		free(buf);		
	}

	return ret;
}

#define NUMSPECIALCHARS	5
static char* specialChars[] = { "&", "<", ">", "\"", "'" }; //VERY important: do amp first or it will match all other amps, eg will match &lt
static char* charReplacements[] = {"&#38;", "&#60;", "&#62;", "&#34;", "&#39;" };

//Replaces special characters with their html entities
void zoomSearch::htmlspecialchars(char* str, int MAXSIZE)
{	
	char* res = (char*) calloc(MAXSIZE, sizeof(char));
	strcpy(res, str);	
	for(int b=0; b < NUMSPECIALCHARS; b++)
	{		
		strReplace(specialChars[b], charReplacements[b], res, false, MAXSIZE);
	}	
	strcpy(str, res);
	free(res);
}

#define NUMQUERYENTCHARS	3
static char* queryEntityChars[] = { "&", "<", ">" }; //VERY important: do amp first or it will match all other amps, eg will match &lt
static char* queryEntityCharReplacements[] = {"&#38;", "&#60;", "&#62;" };
void zoomSearch::QueryEntities(char* str, int MAXSIZE)
{		
	char* res = (char*) calloc(MAXSIZE, sizeof(char));
	strcpy(res, str);	
	for(int b=0; b < NUMQUERYENTCHARS; b++)
	{		
		strReplace(queryEntityChars[b], queryEntityCharReplacements[b], res, false, MAXSIZE);
	}	
	strcpy(str, res);
	free(res);
}

//Replaces special characters with their html entities
void zoomSearch::urlencode(char* src, int MAXSIZE)
{		
	int srcLen = strlen(src);

	char* res = (char*) calloc(MAXSIZE, sizeof(char));
	int resPos = 0;

	unsigned char ch;	// needs to be unsigned for the %X hex value to be correct

	for (int srcPos = 0; srcPos < srcLen && resPos < MAXSIZE; srcPos++)
	{
		ch = src[srcPos];
		if (!isalnum(ch) && ch != '-' && ch != '_')
		{
			if (ch == ' ')
				res[resPos++] = '+';
			else
			{
				// not alphanumeric and should be encoded in hex				
				sprintf(res+resPos, "%%%X", ch);
				resPos += 3;
			}			
		}
		else
			res[resPos++] = ch;
	}

	strcpy(src, res);
	free(res);

	/*
	char* res = (char*) calloc(MAXSIZE, sizeof(char));
	strcpy(res, str);	
	for(int b=0; b < NUMSPECIALCHARS; b++)
	{		
		strReplace(specialChars[b], charReplacements[b], res, false, MAXSIZE);
	}	
	strcpy(str, res);
	free(res);
	*/
}


//Case insentive strstr function
char* zoomSearch::iStrStr(const char *string, const char *strCharSet, int nStrLen)
{
	
	char *res;
	int result=0;
	
	//duplicate stings
	char* tmpString = _strdup(string);	
	if(tmpString == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("<br />Out of memory.<br />");
		if (UseXML) zoomprintf("</zoom:error>");
		ShowTemplate();
		return NULL;
	}

	// if length of string is specified, we truncate it so we only match
	// against given string length
	if (nStrLen > 0)
		tmpString[nStrLen] = '\0';
				
	char* tmpStrCharSet = _strdup(strCharSet);
	
	if(tmpStrCharSet == NULL)
	{
		if (UseXML) zoomprintf("<zoom:error>");
		zoomprintf("<br />Out of memory.<br />");
		if (UseXML) zoomprintf("</zoom:error>");
		free(tmpString);
		ShowTemplate();
		cleanup();
		return NULL;
	}
	
	//convert to lowercase
	if (UseUTF8 == 1)
	{
		utf8_strlwr((unsigned char*)tmpString);
		utf8_strlwr((unsigned char*)tmpStrCharSet);
	}
	else
	{
		_strlwr(tmpString);
		_strlwr(tmpStrCharSet);
	}
		
	res = strstr(tmpString, tmpStrCharSet); //call strstr
	
	if(res == NULL)
	{
		free(tmpString);
		free(tmpStrCharSet);
		return NULL;
	}
	
	result = res - tmpString;
	free(tmpString);
	free(tmpStrCharSet);
	
	return (char*)(string+result);  //convert pointer to position in oringal string 
	
}



bool zoomSearch::GetSPCode(char* word, char* soundex)
{
	char tmpword[WORDLENGTH] = {0};

	int wordlen;
	int i, j;

	wordlen = strlen(word);
	
	if (wordlen > WORDLENGTH)
		return false;
		
	// copy to tmpword (stripping out spaces and non-alphabetic characters)
	// and change everything to lowercase
	for (i = 0, j = 0; i < wordlen; i++)
	{	
		//if (isalpha(word[i]))
		if ((word[i] >= 'a' && word[i] <= 'z') ||
			(word[i] >= 'A' && word[i] <= 'Z') )		
		{
			tmpword[j] = tolower(word[i]);
			j++;
		}
	}
	// recalc length of word
	wordlen = j;

	// quit if all stripped out
	if (wordlen < 1 || wordlen > WORDLENGTH)
		return false;
	
	// note that we skip the first letter and leave it intact
	for (i = 1; i < wordlen; i++)
	{
		switch (tmpword[i])
		{
		case 'a':
		case 'e':
		case 'i':
		case 'o': 
		case 'u': 
		case 'h': 
		case 'w': 
		case 'y':
			tmpword[i] = '0';
			break;
		case 'b':
		case 'p': 
		case 'f': 
		case 'v':
			tmpword[i] = '1';
			break;
		case 'c':
		case 'g':
		case 'j': 
		case 'k': 
		case 'q': 
		case 's':
		case 'x':
		case 'z':
			tmpword[i] = '2';
			break;
		case 'd':
		case 't':
			tmpword[i] = '3';
			break;
		case 'l':
			tmpword[i] = '4';
			break;
		case 'm':
		case 'n':
			tmpword[i] = '5';
			break;
		case 'r':
			tmpword[i] = '6';
			break;	
		default:
			// Contains a non-alphabetic character, so we replace it with a space
			tmpword[i] = ' ';
			break;
		}		
	}
		
	// squeeze out equal adjacent digits
	for (j = i = 1; i < wordlen; i++)
	{
		tmpword[j] = tmpword[i];
		if (tmpword[j] != tmpword[i+1])
			j++;	// only move on to next char if different.
	}	
	tmpword[j] = '\0';

	// squeeze out spaces and zeroes
	for (j = i = 1; i < wordlen; i++)
	{
		tmpword[j] = tmpword[i];
		if (tmpword[j] != ' ' && tmpword[j] != '0')
			j++;	// only move on to next char if not the above
	}	
	tmpword[j] = '\0';

	// check the new length
	int soundex_len = strlen(tmpword);
	
	if (soundex_len < 1)
		return false;	// something's gone wrong, the first letter should always be present
	
	// zero-pad it out to 4 characters
	for (i = soundex_len; i < 4; i++)
	{
		tmpword[i] = '0';
	}

	// null-terminate 
	tmpword[4] = '\0';

	// copy to soundex buffer including null terminator
	strncpy(soundex, tmpword, 5);


	return true;
}


//--------------------------------------------------------------------------
// Metaphone code below:
//--------------------------------------------------------------------------

#define NULLCHAR (char *) 0

#ifdef NON_WINDOWS
#define TRUE	1
#define FALSE	0
#endif

char *VOWELS="AEIOU",
     *FRONTV="EIY",   // special cases for letters in FRONT of these 
     *VARSON="CSPTG", // variable sound--those modified by adding an "h"    
     *DBL=".";     // let these double letters through 

char *excpPAIR="AGKPW", // exceptions "ae-", "gn-", "kn-", "pn-", "wr-" 
     *nextLTR ="ENNNR";
char *chrptr, *chrptr1;

bool zoomSearch::GetMetaphone(char* name, char* metaph, int metalen)
{	
	int  ii, jj, silent, hard, Lng, lastChr;
	
	char curLtr, prevLtr, nextLtr, nextLtr2, nextLtr3;
	
	int vowelAfter, vowelBefore, frontvAfter;
	
	char wname[60];
	char *ename = wname;
	
	jj = 0;
	for (ii=0; name[ii] != '\0'; ii++) 
	{
		if (isalpha(name[ii]) ) 
		{
			ename[jj] = toupper(name[ii]);
			jj++;
		}
	}
	ename[jj] = '\0';
	
	if (strlen(ename) == 0) 
		return FALSE;

	memset(metaph, 0, sizeof(char)*(metalen+1));
	
	/* if ae, gn, kn, pn, wr then drop the first letter */
	if ( (chrptr=strchr(excpPAIR,ename[0]) ) != NULL ) 
	{
		chrptr1 = nextLTR + (chrptr-excpPAIR);
		if ( *chrptr1 == ename[1] )  
			strcpy(ename,&ename[1]);
	}
	/* change x to s */
	if (ename[0] == 'X') 
		ename[0] = 'S';

	/* get rid of the "h" in "wh" */
	if (strncmp(ename, "WH",2) == 0 ) 
		strcpy(&ename[1], &ename[2]);
	
	Lng = strlen(ename);
	lastChr = Lng -1;   /* index to last character in string makes code easier*/
	
	/* Remove an S from the end of the string */
	if ( ename[lastChr] == 'S' ) {
		ename[lastChr] = '\0';
		Lng = strlen(ename);
		lastChr = Lng -1;
	}
	
	for (ii=0; ( ((int)strlen(metaph) < metalen) && (ii < Lng) ); ii++) {
		
		curLtr = ename[ii];
		
		vowelBefore = FALSE;  prevLtr = ' ';
		if (ii > 0) {
			prevLtr = ename[ii-1];
			if ( strchr(VOWELS,prevLtr) != NULL ) 
				vowelBefore = TRUE;
		}
		/* if first letter is a vowel KEEP it */
		if (ii == 0 && (strchr(VOWELS,curLtr) != NULL) ) {
			strncat(metaph,&curLtr,1);
			continue;
		}
		
		vowelAfter = FALSE;   frontvAfter = FALSE;   nextLtr = ' ';
		if ( ii < lastChr ) {
			nextLtr = ename[ii+1];
			if ( strchr(VOWELS,nextLtr) != NULL ) vowelAfter = TRUE;
			if ( strchr(FRONTV,nextLtr) != NULL ) frontvAfter = TRUE;
		}
		/* skip double letters except ones in list */
		if (curLtr == nextLtr && (strchr(DBL,nextLtr) == NULL) ) continue;
		
		nextLtr2 = ' ';
		if (ii < (lastChr-1) ) nextLtr2 = ename[ii+2];
		
		nextLtr3 = ' ';
		if (ii < (lastChr-2) ) nextLtr3 = ename[ii+3];
		
		switch (curLtr) 
		{
			
		case 'B': 
			silent = FALSE;
			if (ii == lastChr && prevLtr == 'M') silent = TRUE;
			if (! silent) strncat(metaph,&curLtr,1);
			break;
			
			/*silent -sci-,-sce-,-scy-;  sci-, etc OK*/
		case 'C': 
			if (! (ii > 1 && prevLtr == 'S' && frontvAfter) )
			{
				if ( ii > 0 && nextLtr == 'I' && nextLtr2 == 'A' )
					strncat(metaph,"X",1);
				else if (frontvAfter)
					strncat(metaph,"S",1);
				else if (ii > 1 && prevLtr == 'S' && nextLtr == 'H')
					strncat(metaph,"K",1);
				else if (nextLtr == 'H')
				{
					if (ii == 0 && (strchr(VOWELS,nextLtr2) == NULL) )
						strncat(metaph,"K",1);
					else
						strncat(metaph,"X",1);
				}
				else if (prevLtr == 'C')
					strncat(metaph,"C",1);
				else
					strncat(metaph,"K",1);				
			}
			break;
										  
		case 'D': 
			if (nextLtr == 'G' && (strchr(FRONTV,nextLtr2) != NULL))
				strncat(metaph,"J",1);
			else
				strncat(metaph,"T",1);
			break;
			
		case 'G': silent=FALSE;
			/* SILENT -gh- except for -gh and no vowel after h */
			if ( (ii < (lastChr-1) && nextLtr == 'H')
				&& (strchr(VOWELS,nextLtr2) == NULL) )
				silent=TRUE;
			
			if ( (ii == (lastChr-3) )
				&& nextLtr == 'N' && nextLtr2 == 'E' && nextLtr3 == 'D')
                silent=TRUE;
			else if ( (ii == (lastChr-1)) && nextLtr == 'N') 
				silent=TRUE;
				
			if (prevLtr == 'D' && frontvAfter) 
				silent=TRUE;
			
			if (prevLtr == 'G')
				hard=TRUE;
			else
				hard=FALSE;
			
			if (!silent)
				if (frontvAfter && (! hard) )
					strncat(metaph,"J",1);
				else
					strncat(metaph,"K",1);
				break;
					
		case 'H': silent = FALSE;
			if ( strchr(VARSON,prevLtr) != NULL ) silent = TRUE;
			
			if ( vowelBefore && !vowelAfter) silent = TRUE;
			
			if (!silent) strncat(metaph,&curLtr,1);
			break;
			
		case 'F':
		case 'J':
		case 'L':
		case 'M':
		case 'N':
		case 'R': 
			strncat(metaph,&curLtr,1);
			break;
			
		case 'K': 
			if (prevLtr != 'C') 
				strncat(metaph,&curLtr,1);
			break;
			
		case 'P': 
			if (nextLtr == 'H')
				strncat(metaph,"F",1);
			else
				strncat(metaph,"P",1);
			break;
			
		case 'Q': strncat(metaph,"K",1);
			break;
			
		case 'S': 
			if (ii > 1 && nextLtr == 'I' && ( nextLtr2 == 'O' || nextLtr2 == 'A') )
				strncat(metaph,"X",1);
			else
			{
                if (nextLtr == 'H')
					strncat(metaph,"X",1);
                else
					strncat(metaph,"S",1);
			}
			break;
				
		case 'T': 
			if (ii > 1 && nextLtr == 'I' && ( nextLtr2 == 'O' || nextLtr2 == 'A') )
				strncat(metaph,"X",1);
			else if (nextLtr == 'H')         /* The=0, Tho=T, Withrow=0 */
			{
				if (ii > 0 || (strchr(VOWELS,nextLtr2) != NULL) )
					strncat(metaph,"0",1);
				else
					strncat(metaph,"T",1);
			}
			else if (! (ii < (lastChr-2) && nextLtr == 'C' && nextLtr2 == 'H'))
				strncat(metaph,"T",1);
			break;
						
		case 'V': 
			strncat(metaph,"F",1);
			break;
			
		case 'W':
		case 'Y': 
			if (ii < lastChr && vowelAfter) 
				strncat(metaph,&curLtr,1);
			break;
			
		case 'X': 
			// be careful adding the extra character in for "KS"
			// check before hand, truncate otherwise.
			if (strlen(metaph)+2 < SPCODE_LEN)
				strncat(metaph,"KS",2);
			else
				strncat(metaph,"K",1);
			break;
			
		case 'Z': 
			strncat(metaph,"S",1);
			break;
    }
	
  }
  
  /*  DON'T DO THIS NOW, REMOVING "S" IN BEGINNING HAS the same effect
  with plurals, in addition imbedded S's in the Metaphone are included
  Lng = strlen(metaph);
  lastChr = Lng -1;
  if ( metaph[lastChr] == 'S' && Lng >= 3 ) metaph[lastChr] = '\0';
  */
  
  if (strlen(metaph) == 0)
	  return FALSE;

  // null terminate and truncate (in case) based on metalen
  //metaph[SOUNDEX_LEN-1] = _T('\0');

  return TRUE;
}


bool zoomSearch::GetPageData(__int64 index, char* url, char* title, char* desc, char* image)
{
	char buffer[PAGEDATA_MAX_LINELEN+1] = "";

	if (url != NULL)
		strcpy(url, "");
	if (title != NULL)
		strcpy(title, "");
	if (desc != NULL)
		strcpy(desc, "");
	if (image != NULL)
		strcpy(image, "");

	//int result = fseek(fp_pagedata, pgdataoffset[index], 0);
	//int result = fseek(fp_pagedata, pageinfo[index]->dataoffset, 0);	 // changed to myfseek for 64bit support (24/nov/08)
	//int result = myfseek(fp_pagedata, pageinfo[index]->dataoffset);

	__int64 dataoffset = ExpandBytesToInt64(GetPageInfo_DataOffset(pageinfo[index], UseBigPageInfoData), PAGEINFODATA_DATAOFFSET_SIZE);

	if (IsPageDataPreloaded)
	{
		// Copy the line of data from memory to "buffer" 
		char ch;
		int i = 0;
		for (i = 0; i < PAGEDATA_MAX_LINELEN; i++)
		{
			ch = (pagedatabin+dataoffset)[i];
			if (ch == '\r' || ch == '\n' || ch == '\0')
				break;
			buffer[i] = ch;
		}
		buffer[i] = '\0';
	}
	else
	{			
		int result = myfseek(fp_pagedata, dataoffset);
		fgets(buffer, PAGEDATA_MAX_LINELEN, fp_pagedata);
	}

	int tmplen;

	char* urlPtr = NULL;
	char* titlePtr = NULL;
	char* descPtr = NULL;
	char* imgPtr = NULL;
	char* startPtr =NULL;
	char* token;
	int toklen;

	startPtr = buffer;
	token = strchr(startPtr, '|');
	if (token == NULL)
		return false;
	toklen = token - startPtr;
	if (toklen > 0)
	{
		urlPtr = startPtr;
		*token = '\0';
	}
	else			
		urlPtr = NULL;

	startPtr = ++token;
	token = strchr(startPtr, '|');
	if (token == NULL)
		return false;
	toklen = token - startPtr;
	if (toklen > 0)
	{
		titlePtr = startPtr;
		*token = '\0';
	}
	else			
		titlePtr = NULL;

	startPtr = ++token;
	token = strchr(startPtr, '|');
	if (token == NULL)
		return false;
	toklen = token - startPtr;
	if (toklen > 0)
	{
		descPtr = startPtr;
		*token = '\0';
	}
	else			
		descPtr = NULL;

	startPtr = ++token;
	token = strchr(startPtr, '|');
	if (token == NULL)
		return false;
	toklen = token - startPtr;
	if (toklen > 0)
	{
		imgPtr = startPtr;
		*token = '\0';
	}
	else			
		imgPtr = NULL;
	
	if (urlPtr != NULL && url != NULL)
	{
		tmplen = strlen(urlPtr);
		if (tmplen > 0)
		{			
			strcpy(url, urlPtr);
		}
	}

	if (titlePtr != NULL && title != NULL)
	{
		tmplen = strlen(titlePtr);
		if (tmplen > 0)
		{			
			strcpy(title, titlePtr);
		}
	}

	if (descPtr != NULL && desc != NULL)
	{
		tmplen = strlen(descPtr);
		if (tmplen > 0)
		{			
			strcpy(desc, descPtr);
		}
	}

	if (imgPtr != NULL && image != NULL)
	{
		tmplen = strlen(imgPtr);
		if (tmplen > 0)
		{			
			strcpy(image, imgPtr);
		}
	}
	
	return true;
}

// This is a specialized version of GetPageData() which only returns the URL,
// and does not allocate extra memory to do. This is purposely designed for
// the "site:blah" search for speed reasons.
// [in]	index		the page index
// [in] urlBuffer	a pre-allocated buffer for storing the resultant url. It must be PAGEDATA_MAX_LINELEN+1 in size.
bool zoomSearch::GetUrlFromPageData(__int64 index, char* urlBuffer)
{
	if (urlBuffer == NULL)
		return false;
	
	__int64 dataoffset = ExpandBytesToInt64(GetPageInfo_DataOffset(pageinfo[index], UseBigPageInfoData), PAGEINFODATA_DATAOFFSET_SIZE);

	if (IsPageDataPreloaded)
	{
		// copy only up to the first '|' character to urlBuffer and return
		char ch;
		int i = 0;
		for (i = 0; i < PAGEDATA_MAX_LINELEN; i++)
		{
			ch = (pagedatabin+dataoffset)[i];
			if (ch == '|' || ch == '\r' || ch == '\n' || ch == '\0')
				break;
			urlBuffer[i] = ch;
		}
		urlBuffer[i] = '\0';
		return true;
	}
	else
	{
		int result = myfseek(fp_pagedata, dataoffset);
		fgets(urlBuffer, PAGEDATA_MAX_LINELEN, fp_pagedata);
	}

	char* token;
	int toklen;

	token = strchr(urlBuffer, '|');
	if (token == NULL)
		return false;
	toklen = token - urlBuffer;
	if (toklen > 0)
	{
		*token = '\0';
		return true;
	}
	return false;
}




//ASCII_to_Integer
// Fast ASCII to Integer conversion.
// Unlike the standard C library we
//	- Only handle base 10 (not base 16,32, or anything else)
//	- Assume there is no overflow of an integer
//	- Don't check the location to do international character conversions
//	- Does only limited checking for leading white space
//As the result there is a dramatic speed increase, but less error handling
//INPUT: The input string must be ASCII number between 0 and 9 and be NULL terminated.
//OUTPUT: The converted number as an Int.
inline int zoomSearch::ASCII_to_Integer(char *ASCIIString) {

	char	CurrentChar;
	int		DigitVal;
	unsigned int Result = 0;
	int		Negative = 1;
	char	*StrPtr;

	StrPtr = ASCIIString;
	CurrentChar = *StrPtr++;

	//Skip leading white space
	while (CurrentChar == ' ')
		CurrentChar = *StrPtr++;

	//Check if string negative
	if (CurrentChar == '-') {
		Negative = -1;
		CurrentChar = *StrPtr++;
	}

	while (CurrentChar >= '0' && CurrentChar <= '9') {
		
		// Convert CurrentChar to a value and add to total
		DigitVal = CurrentChar - '0';
		Result = Result * 10 + DigitVal;

		//read next digit
		CurrentChar = *StrPtr++; 
	}
	return (Result * Negative);
}

inline __int64 zoomSearch::ASCII_to_Int64(char *ASCIIString) {

	char	CurrentChar;
	int		DigitVal;
	__int64 Result = 0;
	int		Negative = 1;
	char	*StrPtr;

	StrPtr = ASCIIString;
	CurrentChar = *StrPtr++;

	//Skip leading white space
	while (CurrentChar == ' ')
		CurrentChar = *StrPtr++;

	//Check if string negative
	if (CurrentChar == '-') {
		Negative = -1;
		CurrentChar = *StrPtr++;
	}

	while (CurrentChar >= '0' && CurrentChar <= '9') {
		
		// Convert CurrentChar to a value and add to total
		DigitVal = CurrentChar - '0';
		Result = Result * 10 + DigitVal;

		//read next digit
		CurrentChar = *StrPtr++; 
	}
	return (Result * Negative);
}

// This function compares text strings, one of which can have wildcards ('*').
// By Kirk J. Krauss
// From "Matching Wildcards: An Algorithm" (Dr Dobb's, 26/Aug/2008)
//
// Modified by RL@Wrensoft to add support for '?' operator (30/Sep/2008)
bool zoomSearch::WildCardCompare(
						char * pTameText,             // A string without wildcards
						char * pWildText,             // A (potentially) corresponding string with wildcards
						bool bCaseSensitive,  // By default, match on 'X' vs 'x'
						char cAltTerminator    // For function names, for example, you can stop at the first '('
						)
{
	bool bMatch = true;
	char * pAfterLastWild = NULL; // The location after the last '*', if we've encountered one
	char * pAfterLastTame = NULL; // The location in the tame string, from which we started after last wildcard
	char t, w;

	// Walk the text strings one character at a time.
	while (1)
	{
		t = *pTameText;
		w = *pWildText;

		// How do you match a unique text string?
		if (!t || t == cAltTerminator)
		{
			// Easy: unique up on it!
			if (!w || w == cAltTerminator)
			{
				break;                                   // "x" matches "x"
			}
			else if (w == '*')
			{
				pWildText++;
				continue;                           // "x*" matches "x" or "xy"
			}
			else if (pAfterLastTame)
			{
				if (!(*pAfterLastTame) || *pAfterLastTame == cAltTerminator)
				{
					bMatch = false;
					break;
				}
				pTameText = pAfterLastTame++;
				pWildText = pAfterLastWild;
				continue;
			}

			bMatch = false;
			break;                                           // "x" doesn't match "xy"
		}
		else
		{
			if (!bCaseSensitive)
			{
				// Lowercase the characters to be compared.
				if (t >= 'A' && t <= 'Z')
				{
					t += ('a' - 'A');
				}

				if (w >= 'A' && w <= 'Z')
				{
					w += ('a' - 'A');
				}
			}

			// How do you match a tame text string?
			if (t != w && w != '?')	// added support for "?" by RL@Wrensoft (30/sep/08)
			{
				// The tame way: unique up on it!
				if (w == '*')
				{
					pAfterLastWild = ++pWildText;
					pAfterLastTame = pTameText;
					w = *pWildText;

					if (!w || w == cAltTerminator)
					{
						break;                           // "*" matches "x"
					}
					continue;                           // "*y" matches "xy"
				}
				else if (pAfterLastWild)
				{
					if (pAfterLastWild != pWildText)
					{
						pWildText = pAfterLastWild;
						w = *pWildText;

						if (!bCaseSensitive && w >= 'A' && w <= 'Z')
						{
							w += ('a' - 'A');
						}

						if (t == w)
						{
							pWildText++;
						}
					}
					pTameText++;
					continue;                           // "*sip*" matches "mississippi"
				}
				else
				{
					bMatch = false;
					break;                                   // "x" doesn't match "y"
				}
			}
		}
		pTameText++;
		pWildText++;
	}
	return bMatch;
}

/*
//WildCardMatch 
//	Perform a simple fast wild card match.
//	Only handles * and ?
//	Input:	haystack (string to search in),
//			needle	 (string to search for)
//			CaseInsensitive flag
//			SubstringMatch	flag
//	Output: true if match
//			false if no match found
//	Warnings: Doesn't handle case UTF-8 very well. (might need to convert to wide chars to support UFT-8 !!, but will be slow)
#define WCMATCH_NORMAL		1
#define WCMATCH_WILDCARD	2
int zoomSearch::WildCardMatch (char* haystack, char* needle, int CaseInsensitive, int SubstringMatch) 
{
	int		State = WCMATCH_NORMAL;	
	unsigned char	needleChar, haystackChar;	
	
	if (haystack == NULL || needle == NULL)
		return false;

	int		IsStartWildcard = 0;
	char*	needleStart = needle;

	needleChar = *needle++;
	haystackChar = *haystack++;

	if (SubstringMatch)
	{
		State = WCMATCH_WILDCARD;	// start in wildcard match mode since substring matching
		IsStartWildcard = 1;
	}
	
	//Loop until we reach the end of the word(s)
	while (haystackChar && needleChar) 
	{
		if (CaseInsensitive) 
		{
			needleChar = toupper(needleChar); 
			haystackChar = toupper(haystackChar);
		}
		switch (needleChar) {
		case '*':
			IsStartWildcard = 1;
			needleStart = needle;
			needleChar = *needle++;			//Wild card matches any char so advance the pointer
			if (needleChar == '\0')
				return true;				//Reached end of word without any mis-matches
			else
				State = WCMATCH_WILDCARD;
			break;
		case '?':
			needleChar = *needle++;			//Question mark matches any char so advance the pointers
			haystackChar = *haystack++;
			break;
		default:							//Normal character 
			if (State == WCMATCH_WILDCARD) {
				if (needleChar != haystackChar)	{	//In wild card mode so skip characters until a match or end of string
					haystackChar = *haystack++;
				}
				else {
					State = WCMATCH_NORMAL;	//Characters after a * matched so advance the pointers and go back to normal match mode
					needleChar = *needle++;
					haystackChar = *haystack++;
				}
			}
			else {
				if (needleChar != haystackChar)
				{
					if (IsStartWildcard)
					{
						//need to restart the needle string and keep looking
						needle = needleStart;
						needleChar = *needle++;					
						if (SubstringMatch)
							State = WCMATCH_WILDCARD;
						else
							IsStartWildcard = false;
					}
					else
						return false;			//No match
				}
				else
				{
					haystackChar = *haystack++;
					needleChar = *needle++;		//Character matched so advance the pointers
				}								
			}
		}
	}
	if (haystackChar == '\0' && needleChar == '\0')
		return true;						//Reached end of word without any mis-matches
	else if (needleChar == '\0' && (State == WCMATCH_WILDCARD || SubstringMatch))
		return true;
	else if (needleChar == '*' && haystackChar == '\0')
		return true;			// needle ending with a "*" matches no last character as well
	else
		return false;
}
*/

// converts a UTF8 string to its lowercase form INPLACE in memory.
void zoomSearch::utf8_strlwr(unsigned char* string)
{	
	unsigned char* strPtr = NULL;
	unsigned char* loweredStr = NULL;
	int llen = 0;
	int newbytes = 0;

	strPtr = (unsigned char*)string;
	while (*strPtr != '\0' && (llen = ucgetutf8lower(strPtr, &loweredStr)) > 0)
	{
		//printf("strPtr: %d, strPtr+1: %d<br>", *strPtr, *(strPtr+1));
		if ((llen == 2 && *strPtr == 207 && *(strPtr+1) == 128) ||
			(llen == 2 && *strPtr == 209 && *(strPtr+1) == 128))
		{				
			// Hack workaround to avoid lowercasing the small greek letter pi to a further
			// lowercase character (which for some reason ucgetutf8lower returns as
			// "cyrillic small letter ie with grave": e'). Not sure if this is a bug in ucgetutf8lower() 
			// or an incompatibility with the latest UnicodeData.txt file retrieved from 
			// the http://www.unicode.org website.			
			// Update (7/Mar/08): Same workaround for "cyrillic small letter er"
			strPtr += llen;
			newbytes += llen;
		}
		/*
		// This doesn't work atm! Why?
		else if ((llen == 1 && *strPtr == 196 && *(strPtr+1) == 176))	// this is a capitol dotted I character
		{
			// Note that ucgetutf8lower() fails to recognize this UTF-8 char and returns llen==1.

			// Hack workaround for the Turkish dotless I debacle. This is explained
			// in detail here:
			// http://www.i18nguy.com/unicode/turkish-i18n.html
			// Basically, Turkish has a dotted i character (with upper and lower case forms)
			// and a dotless i character (with upper and lower case forms)
			// but they do not roundtrip when you switch cases on them. e.g.
			// (dotless little i) -uc-> (dotless capital i) -lc-> (dotted little i)
			
			// We can only provide a partial solution to fix the fact that UnicodeData.txt
			// seem to break the dotted capital I. 
			string[newbytes] = 'i';			
			strPtr += 2;
			newbytes++;
		}
		*/
		else
		{
			memcpy(string+newbytes, loweredStr, llen);
			strPtr += llen;
			newbytes += llen;
		}
	}
	//memcpy(string, newStr, newbytes);
	string[newbytes] = '\0';
}

int zoomSearch::mystricmp(const char *string1, const char *string2)
{
	if (ToLowerSearchWords == 0)
		return strcmp(string1, string2);

	if (UseUTF8 == 1)
	{		
		int ret;
		unsigned char* tmpstr1;
		unsigned char* tmpstr2;
		tmpstr1 = (unsigned char*) _strdup(string1);
		tmpstr2 = (unsigned char*) _strdup(string2);
		utf8_strlwr(tmpstr1);
		utf8_strlwr(tmpstr2);
		ret = strcmp((char*)&tmpstr1, (char*)&tmpstr2);
		free(tmpstr1);
		free(tmpstr2);
		return ret;
	}
	else
		return _stricmp(string1, string2);
}

int zoomSearch::mystrnicmp(const char *string1, const char *string2, int len)
{
	if (ToLowerSearchWords == 0)
		return strncmp(string1, string2, len);

	if (UseUTF8 == 1)
	{		
		int ret;
		unsigned char* tmpstr1;
		unsigned char* tmpstr2;
		tmpstr1 = (unsigned char*) malloc(sizeof(char) * (len+1));
		if (tmpstr1 == NULL)
			return -1;
		tmpstr2 = (unsigned char*) malloc(sizeof(char) * (len+1));
		if (tmpstr2 == NULL)
			return -1;
		memcpy(tmpstr1, string1, sizeof(char)*len);
		tmpstr1[len] = '\0';
		memcpy(tmpstr2, string2, sizeof(char)*len);
		tmpstr2[len] = '\0';
		utf8_strlwr(tmpstr1);
		utf8_strlwr(tmpstr2);
		ret = strncmp((char*)tmpstr1, (char*)tmpstr2, len);
		free(tmpstr1);
		free(tmpstr2);
		return ret;
	}
	else
		return _strnicmp(string1, string2, len);
}

// This is a custom strcmp function which assumes that only the right-hand side (second) parameter
// requires converting to lowercase. This is used whenever we need to make frequent 
// stricmp's where we know one of the parameters is already converted to lowercase.
int zoomSearch::mystrcmp_rightcase(const char *string1, const char *string2)
{
	if (ToLowerSearchWords == 0)
		return strcmp(string1, string2);

	if (UseUTF8 == 1)
	{
		int ret;
		unsigned char* tmpstr;
		tmpstr = (unsigned char*) _strdup(string2);		
		utf8_strlwr(tmpstr);
		ret = strcmp((char*)string1, (char*)tmpstr);
		free(tmpstr);
		return ret;
	}
	else
		return _stricmp(string1, string2);
}

#ifndef NON_WINDOWS 
//Return the last error as a string, returned error message should be freed with LocalFree()
char* zoomSearch::GetLastErrorString()
{
	{
		LPVOID lpMsgBuf;
		DWORD error = GetLastError();
		
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			error,
			0,
			(LPTSTR) &lpMsgBuf, 
			0, 
			NULL);	
		return (char*) lpMsgBuf;
	}	
}
#endif

/*
void utf8_strlwr(unsigned char* string)
{
#ifdef NON_WINDOWS
	
	char* oldlocale;
	oldlocale = setlocale(LC_ALL, ".65001");
	if (oldlocale == NULL)
		oldlocale = setlocale(LC_ALL, "en_US.UTF-8");
	zoomprintf("oldlocale: %s<br>", oldlocale);
	zoomprintf("pre lowered: %s<br>", Query);
	mbslwr((unsigned char*)Query);
	zoomprintf("to lowered: %ls<br>", Query);
	setlocale(LC_ALL, oldlocale);			

#else
	char* oldlocale;
	oldlocale = setlocale(LC_ALL, ".65001");
	if (oldlocale == NULL)
		oldlocale = setlocale(LC_ALL, "en_US.UTF-8");
	zoomprintf("oldlocale: %s<br>", oldlocale);
	zoomprintf("pre lowered: %s<br>", Query);
	_mbslwr((unsigned char*)Query);
	zoomprintf("to lowered: %ls<br>", Query);
	setlocale(LC_ALL, oldlocale);			
#endif
}
*/

//---------------------------------------------------------------------------------
// Replacement functions for windows api calls for non windows compiles
//---------------------------------------------------------------------------------

#ifdef NON_WINDOWS

void zoomSearch::_strlwr(char* string)
{	
	for(int a=0; a<strlen(string); a++)
	{
		if((string[a] >= 'A' && string[a] <= 'Z')) 
			string[a] += 32;
	}
	
}

void zoomSearch::_mbslwr(unsigned char* string)
{
	//currently have no replacement
	//try setting locale to utf
	//then calling tolower() on each char 	
}


int zoomSearch::_stricmp(const char *string1, const char *string2)
{	
	int ret=0;
	char* tmpString1 = strdup(string1);
	
	if(tmpString1 == NULL)
	{
		zoomprintf("<br />Out of memory<br />");
		return 0;
	}
	
	char* tmpString2 = strdup(string2);	
	
	if(tmpString2 == NULL)
	{
		zoomprintf("<br />Out of memory<br />");
		free(tmpString1);
		return 0;
	}
	
	if (UseUTF8 == 1)
	{
		utf8_strlwr((unsigned char*)tmpString1);
		utf8_strlwr((unsigned char*)tmpString2);
	}
	else
	{
		_strlwr(tmpString1);
		_strlwr(tmpString2);
	}
	
	ret = strcmp(tmpString1, tmpString2);
	
	free(tmpString1);
	free(tmpString2);
	
	return ret;
}

int zoomSearch::_strnicmp(const char *string1, const char *string2, int len)
{	
	int ret=0;	
	char* tmpString1 = (char*) malloc(sizeof(char) * (len+1));		
	if(tmpString1 == NULL)
	{
		zoomprintf("<br />Out of memory<br />");
		return 0;
	}
	strncpy(tmpString1, string1, len);
	tmpString1[len] = '\0';
	
	char* tmpString2 = (char*) malloc(sizeof(char) * (len+1));	
	if(tmpString2 == NULL)
	{
		zoomprintf("<br />Out of memory<br />");
		free(tmpString1);
		return 0;
	}
	strncpy(tmpString2, string2, len);
	tmpString2[len] = '\0';
	
	if (UseUTF8 == 1)
	{
		utf8_strlwr((unsigned char*)tmpString1);
		utf8_strlwr((unsigned char*)tmpString2);
	}
	else
	{
		_strlwr(tmpString1);
		_strlwr(tmpString2);
	}
	
	ret = strncmp(tmpString1, tmpString2, len);
	
	free(tmpString1);
	free(tmpString2);
	
	return ret;
}


char* zoomSearch::_strdup(const char *strSource)
{
	char* newString = NULL;
	newString = (char*)calloc(strlen(strSource)+1, sizeof(char));
	
	if(newString != NULL)
	{
		strcpy(newString, strSource);
	}				
	return newString;
	
}


int zoomSearch::_eof(int filedes)
{	
	off_t here;
	off_t end;
	int retval;

	if ( ((here = lseek(filedes, 0, SEEK_CUR)) == -1) ||
		((end = lseek(filedes, 0, SEEK_END)) == -1) )
		retval = -1;
	else if ( here == end )
		retval = 1;
	else {
		lseek(filedes, here, SEEK_SET);
		retval = 0;
	}

	return retval;
}
#endif


// ----------------------------------------------------------------
// Some entity handling functions
// ----------------------------------------------------------------

// This function encodes the characters in entitychars found in origStr as HTML entities
// and write out the new string in the destStr buffer.
// It will also recognize HTML entities of the form &#yyyy; in origStr and NOT re-encode 
// the ampersand character.
bool EncodeEntities(char* origStr, char* destStr, char* entitychars, int destLen)
{
	if (origStr == NULL || destStr == NULL || entitychars == NULL)
		return false;

	int origLen;
	char* NewStartPt;
	char* EndPt;
	
	char entityStr[10];
	int entityLen;
	bool bIsEntity;
	
	origLen = strlen(origStr);	

	int destStrCount = 0;

	int i = 0;

	while (i < origLen)
	{
		NewStartPt = origStr+i;
		bIsEntity = false;

		if (strchr(entitychars, *NewStartPt) != NULL)
		{
			// We need to check for any of these word join chars which are potentially entities!
			// We do so by checking if they're in the form of "&#..;" and don't encode them
			// if so. Otherwise, we presume they are words (e.g. "bed&breakfast" or "C#") and
			// we DO encode them.
			if (*NewStartPt == '&' && i+3 < origLen)
			{
				// special handling for ampersand since it could be a HTML entity				
				if (origStr[i+1] == '#')
				{
					EndPt = strstr(NewStartPt, ";");
					if ((EndPt - NewStartPt) < 8)
						bIsEntity = true;						
				}
			}			
			else if (i > 0 && *NewStartPt == '#' && i+2 < origLen)
			{
				// special handling for hash character since it could be a HTML entity				
				if (origStr[i-1] == '&')
				{
					EndPt = strstr(NewStartPt, ";");
					if ((EndPt - NewStartPt) < 8)
						bIsEntity = true;						
				}
			}			
			
			if (bIsEntity)
			{
				// this is most likely an entity, we should leave it alone
				// and copy it as is.
				destStr[destStrCount] = origStr[i];
				destStrCount++;
			}
			else
			{
				sprintf(entityStr, "&#%d;", *NewStartPt);
				entityLen = strlen(entityStr);
				for (int j = 0; j < entityLen; j++)
				{
					destStr[destStrCount] = entityStr[j];
					destStrCount++;
				}
			}
		}
		else
		{
			destStr[destStrCount] = origStr[i];
			destStrCount++;
		}
		i++;
	}	
	return true;
}

// This function will decode only the characters defined in entitychars in the origStr
// and write the output in the destStr buffer. It will only decode entities of the format
// &#yyyy; where yyyy is the decimal or hex value of the character entity.
bool DecodeEntities(char* origStr, char* destStr, char* entitychars, int destLen)
{
	if (origStr == NULL || destStr == NULL || entitychars == NULL)
		return false;

	int origLen;
	bool bIsDecoded;
	int Val;
	
	char* EndPt;
	char* NewStartPt;
	
	origLen = strlen(origStr);

	int destStrCount = 0;

	int i = 0;

	while (i < origLen)
	{
		bIsDecoded = false;

		NewStartPt = origStr+i;
		if (*NewStartPt == '&' && i+3 < origLen)
		{			
			if (origStr[i+1] == '#')
			{
				EndPt = strstr(NewStartPt, ";");
				if ((EndPt - NewStartPt) < 8)
				{
					Val = atoi(NewStartPt+2);
					if (Val == 0)
					{
						// not a number, check if 'x' which means hex number
						if (NewStartPt[2] == 'x')
						{
							char* endofhex;
							Val = strtol(NewStartPt+3, &endofhex, 16);
							// Check hex value parsed was stopped at the ';'
							if (endofhex != EndPt)
								Val = 0;	// if not, we won't provide a value.
						}
					}

					// we only convert the characters in entitychars string									
					if (strchr(entitychars, Val) != NULL)
						bIsDecoded = true;
				}
			}
		}

		if (bIsDecoded)
		{
			// write the decoded entity character value
			destStr[destStrCount] = (char)Val;
			i += EndPt - NewStartPt;	// move the pointer after the entity				
		}
		else
		{
			// copy the character as it is
			destStr[destStrCount] = origStr[i];
		}

		i++;
		destStrCount++;
	}
	return true;
}

//#pragma managed(push, off)
int zoomSearch::zoomprintf(const char *fmt, ...)
{			
	int nRet;
	int nLen;

	char TempPrintfBuffer[MAX_PRINTF_BUFFERLEN+1];
	memset(TempPrintfBuffer, 0, sizeof(char)*(MAX_PRINTF_BUFFERLEN+1));
	//TempPrintfBuffer[0] = 0;
	
	va_list	ap;		// variable list
	if (fmt == NULL) 
	{
		return -1;
	}

	va_start(ap, fmt);	
		nLen = vsnprintf(TempPrintfBuffer, MAX_PRINTF_BUFFERLEN, fmt, ap);
	va_end(ap);											
	
	// vsnprintf returns -1 if it can't fit it into the allocated buffer
	if (nLen == -1)
		return -1;

#ifdef ASPX
	// NOTES:
	// Compiling with /clr:pure will allow this function to compile managed (without the pragma
	// commands to disable managed). But it has to be compiled with Whole Program Optimizations
	// disabled (which is normally enabled for Release builds, and disabled for Debug).
	//
	// Alternative if we are unable to call managed function from unmanaged,
	// is to write into global buffer, bounds check, and realloc when necessary.

	nRet = strlen(TempPrintfBuffer);

	/*
	wchar_t* tempBufferW = (wchar_t*) malloc(sizeof(wchar_t)*(nRet+1));
	if (tempBufferW == NULL)
		return -1;

	DWORD dwSize = nRet;
	MultiByteToWideChar(codePage, 0, TempPrintfBuffer, -1, tempBufferW, *dwSize);		
	aspxoutput->Write(&tempBufferW, 0, nRet);
	free(tempBufferW);
	*/

	/*
	String^ tmpStr = gcnew String(TempPrintfBuffer, 0, nRet, aspxoutput->Encoding);
	aspxoutput->Write(tmpStr);
	*/
	
	aspxoutput->Write(gcnew String(TempPrintfBuffer, 0, nRet, aspxoutput->Encoding));
	
	//aspxoutput->Write(marshal_as<String^>(TempPrintfBuffer));
	
	
#elif defined(FASTCGI_ENABLED)
	nRet = FastCGIPrintf((char*)TempPrintfBuffer);
#elif defined(OSF)
	nRet = OSFPrintf(TempPrintfBuffer);
#else
	// A possible improvement on this and above, is to write the output to a buffer in
	// memory and only write to stdout (aspxoutput) when the buffer is full. However
	// the measurable differences seem small at this point.
	nRet = printf("%s", TempPrintfBuffer);
#endif

	return nRet;	
}
//#pragma managed(pop)


// This is a simpler version of zoomprintf which does not take printf formating
// and does not need to use a temporary string for storage (thus no need to 
// know size of string in advance.
int zoomSearch::PrintStr(const char* str)
{
	int nRet;

#ifdef ASPX
	// NOTES:
	// Compiling with /clr:pure will allow this function to compile managed (without the pragma
	// commands to disable managed). But it has to be compiled with Whole Program Optimizations
	// disabled (which is normally enabled for Release builds, and disabled for Debug).
	//
	// Alternative if we are unable to call managed function from unmanaged,
	// is to write into global buffer, bounds check, and realloc when necessary.

	nRet = strlen(str);
	aspxoutput->Write(gcnew String(str, 0, nRet, aspxoutput->Encoding));
	//aspxoutput->Write(gcnew String(str));
	/*
	String^ tmpStr = gcnew String(str, 0, nRet, aspxoutput->Encoding);
	aspxoutput->Write(tmpStr);
	*/

	//aspxoutput->Write(marshal_as<String^>(str));
	
#else
#ifdef FASTCGI_ENABLED
	nRet = FastCGIPrintf((char*)str);
#else
	// A possible improvement on this and above, is to write the output to a buffer in
	// memory and only write to stdout (aspxoutput) when the buffer is full. However
	// the measurable differences seem small at this point.
	nRet = printf("%s", str);
#endif
#endif

	return nRet;
}

int zoomSearch::ResultsBufferPrintf(const char *fmt, ...)
{			
	int nLen;	

	char TempPrintfBuffer[MAX_PRINTF_BUFFERLEN+1];
	memset(TempPrintfBuffer, 0, sizeof(char)*(MAX_PRINTF_BUFFERLEN+1));
	
	va_list	ap;		// variable list
	if (fmt == NULL) 
	{
		return -1;
	}

	va_start(ap, fmt);									
		nLen = vsnprintf(TempPrintfBuffer, MAX_PRINTF_BUFFERLEN, fmt, ap);					
	va_end(ap);	

	// vsnprintf returns -1 if it can't fit it into the allocated buffer
	if (nLen == -1)
		return -1;

	int bufferlen;
	bufferlen = OutputResultsLen;
	if (bufferlen+nLen >= OutputResultsSize)
	{
		// realloc is necessary
		int newsize;
		newsize = (int) (OutputResultsSize * 1.5f);
		if (newsize < (bufferlen+nLen))
			newsize = bufferlen+nLen+1;
		
		char* reallocRet = NULL;
		reallocRet = (char*) realloc(OutputResultsBuffer, newsize);
		if (reallocRet == NULL)
			return -1;
				
		OutputResultsBuffer = reallocRet;
		// update buffersize
		OutputResultsSize = newsize;
	}

	strncat(OutputResultsBuffer, TempPrintfBuffer, nLen);

	OutputResultsLen += nLen;

	return nLen;	
}

int zoomSearch::OutputBufferPrintf(int bufferIndex, const char *fmt, ...)
{
	int nLen;	

	char TempPrintfBuffer[MAX_PRINTF_BUFFERLEN+1];
	memset(TempPrintfBuffer, 0, sizeof(char)*(MAX_PRINTF_BUFFERLEN+1));
	
	va_list	ap;		// variable list
	if (fmt == NULL) 
	{
		return -1;
	}

	va_start(ap, fmt);	
		nLen = vsnprintf(TempPrintfBuffer, MAX_PRINTF_BUFFERLEN, fmt, ap);		
	va_end(ap);											

	// vsnprintf returns -1 if it can't fit it into the allocated buffer
	if (nLen == -1)
		return -1;

	int bufferlen;
	bufferlen = OutputBufferLens[bufferIndex];
	if (bufferlen+nLen >= OutputBufferSizes[bufferIndex])
	{
		// realloc is necessary
		int newsize;
		newsize = (int) (OutputBufferSizes[bufferIndex] * 1.5f);
		if (newsize < (bufferlen+nLen))
			newsize = bufferlen+nLen+1;

		char* reallocRet = NULL;
		reallocRet = (char*) realloc(OutputBuffers[bufferIndex], newsize);
		if (reallocRet == NULL)
			return -1;
		
		OutputBuffers[bufferIndex] = reallocRet;
		// update buffersize
		OutputBufferSizes[bufferIndex] = newsize;
	}

	strncat(OutputBuffers[bufferIndex], TempPrintfBuffer, nLen);

	OutputBufferLens[bufferIndex] += nLen;

	return nLen;	
}

void zoomSearch::OutputBufferClear(int bufferIndex)
{
	strcpy(OutputBuffers[bufferIndex], "");
	OutputBufferLens[bufferIndex] = 0;
	return;
}


void zoomSearch::ResetStartTime()
{	
	memset(&starttime, 0, sizeof(TIMER_VALUE));
}

static const char StemStopChars[50] = "`1234567890-=[]\\;',./~!@#$%^&*_+|:\"<>?";

#ifdef _WINDOWS
wchar_t StemmerWordBufferW[WORDLENGTH+1];
#endif


bool zoomSearch::GetStem(char* word, unsigned int wordlen)
{	
	UserDebugSaveTime(DEBUGTIMER_STEMMING);

	if (strcspn(word, StemStopChars) < wordlen)
	{
		UserDebugAddTime(DEBUGTIMER_STEMMING);
		return false;
	}

	int stemmedlen = 0;

#ifdef _WINDOWS
	int wordBufferLen = 0;
	StemmerWordBufferW[0] = L'\0';
	wordBufferLen = MultiByteToWideChar(CP_UTF8, 0, word, wordlen, StemmerWordBufferW, wordlen+1);
	if (wordBufferLen < 1)
		return false;
	StemmerWordBufferW[wordBufferLen] = L'\0';
	stemmed = sb_stemmer_stem(stemmer, StemmerWordBufferW, wordBufferLen);
	if (stemmed != NULL)
		stemmedlen = wcslen((wchar_t*)stemmed);
#else
	/*
	char* oldlocale = setlocale(LC_CTYPE, "en_US.UTF-8");
	wordBufferLen = mbstowcs(StemmerWordBufferW, word, wordlen);
	if (wordBufferLen == -1)
		return false;
	StemmerWordBufferW[wordBufferLen] = L'\0';
	setlocale(LC_CTYPE, oldlocale);
	*/
	stemmed = sb_stemmer_stem(stemmer, (unsigned char*)word, wordlen);
	if (stemmed != NULL)
		stemmedlen = strlen((char*)stemmed);
#endif
	
	if (stemmedlen > 0 && stemmedlen < MAX_QUERY_LENGTH)
	{
#ifdef _WINDOWS
		wordlen = WideCharToMultiByte(CP_UTF8, 0, stemmed, stemmedlen, word, wordlen, NULL, NULL);
		if (wordlen < 1)
			return false;
		word[wordlen] = '\0';
#else
		/*
		char* oldlocale = setlocale(LC_CTYPE, "en_US.UTF-8");
		wordlen = wcstombs(word, StemmerWordBufferW, wordBufferLen);
		if (wordlen == -1)
			return false;
		word[wordlen] = '\0';
		setlocale(LC_CTYPE, oldlocale);
		*/
		strcpy(word, (char*)stemmed);	
#endif			
		UserDebugAddTime(DEBUGTIMER_STEMMING);			
		return true;
	}
	UserDebugAddTime(DEBUGTIMER_STEMMING);	
	return false;
}

// Given a byte array containing a truncated int64 value, 
// we return the full int64 value
//__int64 PageInfoDataGetDataOffset(BYTE* dataoffset)
__int64 zoomSearch::ExpandBytesToInt64(BYTE* byteArray, int numBytes)
{	
	__int64 retValue;
	memset(&retValue, 0, sizeof(__int64));
	endianAware_memcpy(&retValue, byteArray, sizeof(BYTE)*numBytes, sizeof(__int64));
	return retValue;
}

	
// We need to store the __int64 (8 bytes) pagedataOffset value in a 5-byte array
// in PageInfoData.dataoffset. This saves us space and also avoids having to switch
// between two different struct types.
//void PageInfoDataSaveDataOffset(__int64 srcOffset, BYTE* destOffset)
void zoomSearch::TruncateInt64ToBytes(__int64 srcValue, BYTE* destBytesArray, int numBytes)
{
	BYTE tmpBytes[8];
	memcpy(&tmpBytes, &srcValue, sizeof(BYTE) * 8);
	memset(destBytesArray, 0, sizeof(BYTE) * numBytes);
	for (int x = 0; x < numBytes; x++)
	{
		destBytesArray[x] = tmpBytes[x];
	}
}

int zoomSearch::CheckBitInByteArray(int bitnum, BYTE* byteArray)
{
	int bytenum;
	int newBitnum;
	bytenum = (int)ceil((float)(bitnum+1) / 8.0f);

	if (bytenum > 1)
	{
		newBitnum = bitnum - ((bytenum-1)*8);
		bytenum = bytenum - 1;
	}
	else
	{
		newBitnum = bitnum;
		bytenum = 0;
	}

	if (bytenum >= NumCatBytes)
	{
		if (UseXML) zoomprintf("<zoom:error>");		
		zoomprintf("Error: Category number is invalid. Incorrect settings file used?");
		if (UseXML) zoomprintf("</zoom:error>\n");							
		ShowTemplate();
		cleanup();
		exit(1);
	}

	return (byteArray[bytenum] & (1 << newBitnum));
}

//Performs a memcpy, however in the case of system with switched endianess (SUN_SPARC_BUILT) to windows
//it reverses the order of the bytes
//size_dst only needs to be defined when the variable we are putting the data into is a different
//size to the data set we are copying out of
void zoomSearch::endianAware_memcpy(void *dst, void *src, size_t size, size_t size_dst)
{
	
	
#ifndef SUN_SPARC_BUILT
	memcpy(dst, src, size);
#else
	//Typecast input to BYTE*
	BYTE* bPtrDst = (BYTE*)dst;
	BYTE* bPtrSrc = (BYTE*)src;
	
	//the default of 0 means that size_dst is the same as size;
	if (size_dst == 0)
		size_dst = size;
	else
		memset(dst, 0, size_dst);	//Only if sizes don't match, otherwise pointless
		
	if (size > size_dst)
		return;
	size_t szDif = size_dst - size;
	
	for (int i = 0; i < size; i++)
		bPtrDst[szDif+i] = bPtrSrc[(size-1)-i];

#endif
}
